#!/usr/libexec/platform-python
"""
python -m grpc_tools.protoc -I../../protos --python_out=. --grpc_python_out=. ../../protos/ctpd.proto
- see pbcomp.sh
"""

from concurrent import futures
import os,time,random,threading,string,sys,traceback
import auxfuns,ilogger,tri_grpcs
#!/usr/bin/python3
import atbwrap
import atb.partdef as partdef
import atb.ctp as ctp
#!/usr/bin/python3
import grpc
import ctpd_pb2
import ctpd_pb2_grpc
import ltud_pb2
import ltud_pb2_grpc
from grpc_reflection.v1alpha import reflection
#!/usr/bin/python3
import ltu_grpcs

_ONE_DAY_IN_SECONDS = 60 * 60 * 24

server= None

def checkport(port):
  #cmd= "ss -tap |grep ':%d'"%port
  cmd= "ss -tap |grep ':%d' |grep '^LISTEN'"%port
  # not reliable for afs. 'fuser port/tcp -fine for trigger, not for afs'
  rc= os.popen(cmd).read()
  #print("action",rc,":",len(rc))
  return rc

class CTPd(ctpd_pb2_grpc.CTPd):
  RUNACTIVE= 0
  RUNPAUSED= 1
  RUNLOADED= 2
  RUNVALID= 3
  # Errors
  RUNUNKNOWN= 5
  RUNLOADERR= 10
  RUNSTARTERR= 11
  def __init__(self, netname, init=None, dummy=False):
    self.myN= 0   # number of Sleep() requests
    self.finish= None     # or string giving a reason finishing
    self.montimestamp= 0  # the time of last monitoring 
    self.monsleep= 10.0   # monitoring interval [s]
    self.sruns= {}        # stdalone+global runs, key:runn v: detname 
    self.gruns= {}        # v: ref. to Partition instance
    self.lockedfor= None  # userid if locked for somebody
    self.portn= tri_grpcs.grpc_channels[netname]
    self.logger=  ilogger.ilogger("ctpd_%s.log"%netname)
    self.pid= os.getpid()
    self.mutex= threading.Lock()
    #von let's ctpd is ttcpon agnostic
    #if dummy:
    #  self.tpnet= ttcpon.ttcponnet.TtcPonNetDummy(self.logger, netname,
    #  server=True, init=init)
    #else:
    #  #todo: maybe should be replaced by ttcpon.netfile.NetFile(netname)
    #  # or TtcPonNet should deal only with OLTs (ONU addrs should be set
    #  # in the ltu.net or if set here, it is the only action for ltu.net)
    #  self.tpnet= ttcpon.ttcponnet.TtcPonNet(self.logger, netname,
    #  server=True, init=init)
    ##print("ctpd waserror:", self.tpnet.waserror)
    #if self.tpnet.waserror:
    #  # sys.exit(8)
    #  self.finish= "Cannot join ttcpon network. init:"+str(init)
    #  return
    #self.tpnet.measureConfig()
    ##
    # instead open board
    ctpbname= partdef.LTUS.ctpbname
    print("PID:{} bname:{}".format(self.pid,ctpbname))
    self.logger.info("PID:{} bname:{}".format(self.pid,ctpbname))
    atbwrap.Init()
    self.board= atbwrap.OpenATBoard(ctpbname)
    #self.fpga= self.tpnet.oltboard.findDevice("fpga")
    self.fpga= self.board.findDevice("fpga")
    self.fpga.setDport(0xffffffff)   # i.e. ctpd active
    print("pardef.LTUS:",partdef.LTUS.ltus.keys())
    #self.ctp= self.tpnet.oltboard.findDevice("ctp")
    self.ctp= self.board.findDevice("ctp")
    self.ctp.initCtp()   # i.e. ctpd restarts destroys all global runs. todo:
    # check all ltud putting them in stdalone if they are not.
    # ltuds: active daemons, i.e.:
    #        - present in $CTP3_DBCTP/ctpboards.cfg
    #        - present in $CTP3_DBCTP/ctp.net
    # todo: (to be checked)
    # 1. when ltud later restarted, connection become ON automatically
    # 2. New ltud (i.e. the one not active at the ctpd start) should
    #    send a 'i am here' request to the ctpd to establish connection
    # von:
    #ltus= partdef.LTUS.ltus
    #self.ltuds= {}
    #for detn in tri_grpcs.grpc_channels.keys():   # all ltu + ctp daemons
    #  if detn=="ctp": continue
    #  if not (detn in ltus):
    #    print("{} not in LTUS".format(detn))
    #    continue
    #  ltu= ltus[detn]
    #  if ltu == None: continue    # known, but not in ctpboards.cfg
    #  if ltu.emun == None: continue # not part of ctp.net
    #  self.makegrpc(detn)
    #  
    self.ltugrpcs= ltu_grpcs.LTU_grpcs()
    self.ltuds= self.ltugrpcs.ltuds
  #def makegrpc(self, detn):
  #  self.ltuds[detn]= tri_grpcs.TRG_grpc(detn)
  #  if self.ltuds[detn].channel:
  #    print("ltud:{} active.".format(detn))
  #  else:
  #    print("ltud:{} connected (not active).".format(detn))
  #  return self.ltuds[detn].channel
  def printN(self, label=None):
    print("CTPd myN:%d"%self.myN, label)
  def SayHello(self, request, context):
    return ctpd_pb2.HelloReply(message='Hello %s %f.' %(request.name,
      request.floatn))
  def killme(self):
    #sys.exit(8)
    os.system("kill %d"%self.pid)
  def LTUHello(self, request, context):
    ltus= partdef.LTUS.ltus
    detn=request.name
    r= ctpd_pb2.SReply()
    if detn=="ctp": # Cautious check: Don't know if it's needed
      r.rc=8
      r.msg="LTUHello request from ctpd, internal error"
      return r  
    print("LTUHello request from {}".format(detn))
    stubrc= self.ltugrpcs.getltustub(detn)   #
    if detn in self.ltuds:  # check in active ltuds
      #if detn==ctpd_pb2.SRequest.name: #requesting LTU is already in dictionary
      stubrc= self.ltuds[detn].stub   #
      if stubrc!=None:
        r.rc=4
        r.msg="ltud:{} already active.".format(detn)
      else:
        # detn was not alive in time of ctpd start, detn ltud started now.
        #rcmsg= self.ltuds[detn].openchannel()
        #r.rc= rcmsg[0]
        #r.msg= rcmsg[1]
        #print("Hello dbg: channel:{}, rcmsg:{}".format(self.ltuds[detn].channel, rcmsg))
        # above: 2 attempts always, if first one done after ctpd start. 
        if self.ltugrpcs.makegrpc(detn)==None:
          r.rc= 8 ; r.msg= "cannot connect"
        else:
          r.rc= 0 ; r.msg= "ok"
    else:
      r.rc= 12
      r.msg="Unknown detector:{} (e.g. not wired in ctp.net)".format(detn)
    print("LTUHello rc:",r)
    return r
  def Stop(self, request, context):
    global server
    self.finish= "Stop cmd received."
    #self.tpnet.Stop()
    self.fpga.setDport(0)
    #orc= self.tpnet.awmc.ws_close()   # is in tpnet.Stop()
    threading.Timer(2.0, self.killme).start()
    server.stop(1)   # 1s grace period, 
    #server does not respond to any client from now (but it does not finish)
    #raise ValueError("stopping ctpd...")
    print("stopping...")
    return ctpd_pb2.HelloReply(message='stop.')
  def getthreadid(self):
    #return str(threading.current_thread().ident)
    return threading.current_thread().getName()
    #getName(), .ident unique number?
  def findrun(self,detn):
    for runn in self.sruns:
      if self.sruns[runn]==detn: return runn
      return None
  def Sleep(self, request, context):
    print("Sleep: %s %s"%(request.name, self.getthreadid()))
    #newn= self.myN= self.myN+1
    # deliberately put sleep time between read and update of self.myN: 
    #self._lock()
    newn= self.myN
    if request.floatn>0.0: time.sleep(request.floatn)
    newn= newn+1 ; self.myN= newn
    #self._unlock()
    rsecs= request.floatn; msg= "ok"
    rr= ctpd_pb2.HelloReply()
    rr.message='Sleep newn:%d %ds, %s!'%(newn, request.floatn,msg)
    rr.secs= rsecs
    rr.dyna.append(888)
    n= random.randint(0,4)
    for ix in range(n):
      rr.dyna.append(random.randint(1,100))
    print(rr)
    return rr
  def Status(self, req, context):
    print("%s %s"%("Status:", self.getthreadid()))
    lst= "{:3} {:6} {:2} {:6} {}\n".format("det", "inrun", "sg", "lturun","ltumsg")
    rc= 0
    # "detn run ESG"
    for detn in self.ltuds:
      #grpc= self.ltugrpcs.ltuds[detn]
      #ltustub= grpc.stub
      ltustub,msg= self.ltugrpcs.getltustub(detn)
      if ltustub==None:
        #if grpc.channel==None:
        lst= lst + "{:3} not connected {}\n".format(detn, msg)
        #elif tri_grpcs.grpc_server_on(grpc.channel):
      else:
        connected= 'Y'
        inrun= self.findrun(detn)
        if inrun==None:
          inrun= ""
        emptymsg= ltud_pb2.Empty()
        rcltu= ltustub.LtuStatus(emptymsg)
        print("Status: {}: rcltu.mode:{}. ".format(detn, rcltu.mode))
        imode= int(rcltu.mode)  # 1,2 (ok) 9: error in rcltu.msg
        if imode>2: imode= 0
        sg= ("E","S","G")[imode]
        # seems grpc returns in rcltu just 0 in case of (0,"")
        #if hasattr(rcltu,"runn"): lturunn= rcltu.runn
        lturunn= getattr(rcltu,"runn", 0)
        ltumsg= getattr(rcltu,"msg", "")
        lst= lst+"{:3} {:6} {:2} {:6} {}\n".format(detn, inrun, sg, lturunn,ltumsg)
    return ctpd_pb2.SReply(rc=0, msg=lst)
  def RunStatus(self, req, context):
    runn= req.runn
    self.logger.info("%s of run %d"%("RunStatus:", runn))
    # status..
    #print(req, dir(context))
    rr= ctpd_pb2.RunReply()
    if runn in self.sruns:
      # check ltud
      rr.rc= self.RUNACTIVE
    elif runn in self.gruns:
      st= self.gruns[runn].getStatus()   # R L(also in hw) ? l(not in hw)
      if st=="L":
        rr.rc= self.RUNLOADED
      if st=="l":
        rr.rc= self.RUNVALID
      elif st=="R":
        rr.rc= self.RUNACTIVE
      else:
        rr.rc= self.RUNUNKNOWN
    else:
      rr.rc= self.RUNUNKNOWN
      rr.msg="Run: {} unknown".format(runn)
    return rr
  def RunList(self, req, context):
    self.logger.info("%s %s"%("RunList:", self.getthreadid()))
    # status..
    #print(req, dir(context))
    rr= ctpd_pb2.RunReply()
    lst= "" ; rc= 0
    for runn in self.sruns:
      lst= lst+"{:6} R {:3}\n".format(runn, self.sruns[runn])
      rc= rc+1
    for runn in self.gruns:
      st= self.gruns[runn].getStatus()
      lst= lst+"{:6} {:1}     {}\n".format(runn, st, self.gruns[runn].pname)
      rc= rc+1
    rr.rc= rc
    rr.msg= lst
    return rr
  def setModeLtus(self, detlist, mode, runn):
    """ put all detlist detectors in mode. runn -passed to ltud
rc:0,""         -all required detectors set to requested mode GLOBAL
   0,eor_report -ok, all ltud put in STDALONE, EOR report from each ltud
                 assembled in eor_report
   8,errmsg     -some detectors were not put into required mode
"""
    ltudr= ltud_pb2.SetModeRequest(mode=mode, runn=runn)
    detsing= 0   # numb. of ltus which were changed
    deterrmsg= None
    eor_report= ""
    for ixdet in range(len(detlist)):
      det= detlist[ixdet]
      #ltustub,msg= self.ltuds.getltustub(det)
      ltustub,msg= self.ltugrpcs.getltustub(det)
      if ltustub!=None:
        rc= ltustub.SetMode(ltudr)
        print("ltu {} rc:{} {}".format(det, rc.rc, rc.msg))
        if rc.rc==0:   # ltud_pb2.Ltumode.GLOBAL
          detsing= detsing+1
          if mode==ltud_pb2.Ltumode.STDALONE:   # global run stopped
            eor_report= eor_report + det + ' ' + rc.msg + '\n'
        else:
          deterrmsg= "{}: {}".format(det, rc.msg)
          break
      else:
        deterrmsg= msg
        break
    #if detsing<len(dets):
    if deterrmsg!=None: 
      print("setModeLtus:", deterrmsg)
      return 8,deterrmsg
    else:
      if mode==ltud_pb2.Ltumode.STDALONE:   # global run stopped
        return 0,eor_report
      else:
        return 0, ""
  def RunLoad(self, req, context):
    """
1. check if started already (gruns/sruns)
2. partLoad
"""
    runn= req.runn
    # load ctp config in ctp:
    detectors= req.detectors   # available detectors
    pname= req.config
    #if lines[0].find("consul")>=0:
    #  print("RunLoad consul line[0]:{}:".format(lines[0]))
    # only .pd files in CTP3_DBCTP/pd directory for now:
    # todo: lock
    self.logger.info("{} {} runn:{} dets:{} pname:{}".format("RunLoad:", self.getthreadid(),runn,detectors,pname))
    rr= ctpd_pb2.RunReply()
    # 1. check if started already
    if runn in self.gruns:
      rr.rc= self.RUNLOADERR
      rr.msg= "Global run {0} already active".format(runn)
      return rr 
    if runn in self.sruns:
      detactive= self.sruns[runn]
      rr.rc= self.RUNLOADERR
      rr.msg= "Stdalone run {0} already active for {}".format(runn, detactive)
      return rr 
    # 2. partLoad
    part= self.ctp.partLoad(pname, runn, detectors)
    if part.status==part.LOADEDHW:
      if len(self.gruns)==0:
        # this should be removed/improved when we have fills (see OrbitReset service)
        # e.g. reset orbit in case we are not in the fill and not done long time.
        self.fpga.orbitset(None, 3564)   
        time.sleep(0.025)    # give 1TF to orbit reset. todo: not clear if sync needed now
        self.logger.info("Orbit reset to 0")
      dets= part.getDets().split(" ")
      moderc,msg= self.setModeLtus(dets, ltud_pb2.GLOBAL, runn)
      if moderc!=0:
        #todo: does setModeLtus change state for all (i.e. ok:changed all, err: no ltu was changed)?
        rr.rc= 8
        rr.msg= msg
      else:
        self.gruns[runn]= part
        rr.rc= 0
        rr.msg="%s loaded."%pname
    else:
      rr.rc= self.RUNLOADERR
      rr.msg= part.allerrs
    # unlock
    return rr
  def RunStart(self, req, context):
    """
1. if runn in sruns: err
2. if runn in gruns:
     if not LOADEDHW: err: running already or interr?
   else:
     if detector==config=="":        ->det not provided, i.e. global run
       err: global not loaded
     else:                           -> starting stdalone run
       check detector,config -both must be defined 
3. if detector=="":
     -start global
   else:
     -start stdalone
"""
    runn= req.runn
    detector= req.detector
    allpars= {}
    print("%s %s detector:%s runn:%d"%("RunStart:", self.getthreadid(), detector, runn))
    self.logger.info("%s detector:%s runn:%d"%("RunStart:", detector, runn))
    # apply default config:
    #
    #pdb.set_trace()  
    rr= ctpd_pb2.RunReply()
    rr.rc= 0
    if runn in self.sruns:   # 1.
      detactive= self.sruns[runn]
      # check ltud
      rr.rc= self.RUNSTARTERR
      rr.msg= "Stdalone run {0} already active in ltu {1}.".format(runn, detactive)
      return rr 
    if runn in self.gruns:   # 2.
      # check if LOADED or RUNNING
      #pname= self.gruns[runn]
      #part= self.ctp.parts.getPart(pname)
      part= self.gruns[runn]
      if part==None:
        rr.rc= self.RUNSTARTERR
        rr.msg= "Run {0} pname:{1} not found (global) int. error?."\
          .format(runn,part.pname)
        return rr
      if part.status != part.LOADEDHW:
        rr.rc= self.RUNSTARTERR
        rr.msg= "Run {0} (global) not LOADED in ctp hw.".format(runn)
        return rr 
    else:
      if req.detector == "" and req.config == "":
        rr.rc= self.RUNSTARTERR
        rr.msg= "Run {0} (global) not loaded.".format(runn)
        return rr 
      elif req.detector == "":
        rr.rc= self.RUNSTARTERR
        rr.msg= "Run {0} (stdalone) detector missing.".format(runn)
        return rr 
      #elif req.config == "":
      #  rr.rc= self.RUNSTARTERR
      #  rr.msg= "Run {0} (stdalone) config missing.".format(runn)
      #  return rr 
      else:
        pass
    allpars["config"]= req.config
    allpars["detector"]= detector
    allpars["runn"]= runn
    if detector == "":     # start global
      #if lines[0].find("consul")>=0:
      #  print("RunStart consul line[0]:{}:".format(lines[0]))
      # todo: lock
      # run is LOADEDHW (checked above).
      rc= self.ctp.partStart(int(runn))
      if rc!="":
        rr.rc= 8
        rr.msg= rc
      else:
        rr.rc= 0
        rr.msg= ""
    else:                  # start stdalone
      # check if ltu in stdalone: done in ltud
      ltustub,msg= self.ltugrpcs.getltustub(detector, reconnect=True)
      if ltustub==None:
        rr.rc= 12
        rr.msg= msg
      else:
        #rc= ltustub.RunStart(ltud_pb2.RunStartRequest(runn=req.runn, config=req.config))
        # both ok
        rc= ltustub.RunStart(ltud_pb2.RunStartRequest(**allpars))
        print("rc:{}\n{}".format(rc.rc, rc.msg))
        if rc.rc<=4:   # started in ltu
          self.sruns[runn]= detector
        rr.rc= rc.rc
        rr.msg= rc.msg
    # todo: unlock
    return rr
  def RunStop(self, req, context):
    runn= req.runn
    detn= req.detector
    print("{} runn:{} det:{}".format("RunStop:", runn,detn))
    self.logger.info("{} {} runn:{} det:{}".format("RunStop:", self.getthreadid(),runn,detn))
    # stop run ...
    rr= ctpd_pb2.RunReply()
    if detn!="":                  # forced stdalone run stop
      #ltustub,msg= self.ltuds.getltustub(detn)
      ltustub,msg= self.ltugrpcs.getltustub(detn)
      if ltustub==None:
        rr.rc= 12
        rr.msg= msg
      else:
        if (runn>0) and (runn in self.sruns):
          # request to stop runn for detn
          #todo: check if detn agrees with self.runs[runn]
          pass
        else:    
          for rix in self.sruns:
            detnix=self.sruns[rix]
            if detn==detnix:
              print("self.sruns[{}]: {}".format(rix, detnix))
              runn= rix   # let's correct it in case not given by client
        rcltu= ltustub.RunStop(ltud_pb2.RunStartRequest(runn=runn))
        print("rc:{}\n{}".format(rcltu.rc, rcltu.msg))
        rr.rc= rcltu.rc
        rr.msg= rcltu.msg
        if runn in self.sruns:
          del(self.sruns[runn])
    elif runn>0:
      if runn in self.sruns:      # stdalone run
        detector= self.sruns[runn]
        #ltustub,msg= self.ltuds.getltustub(detector)
        ltustub,msg= self.ltugrpcs.getltustub(detector)
        if ltustub==None:
          rr.rc= 12
          rr.msg= msg
        else:
          rc= ltustub.RunStop(ltud_pb2.RunStartRequest(runn=runn))
          print("rc:{}\n{}".format(rc.rc, rc.msg))
          del(self.sruns[runn])
          rr.rc= rc.rc
          rr.msg= rc.msg
      elif runn in self.gruns:    # global run
        # todo: lock
        part= self.gruns[runn]
        #if part==None:
        #  rr.rc= self.RUNSTARTERR
        pname= part.pname
        self.ctp.partStop(int(runn))
        dets= part.getDets().split(" ")
        if runn==88888888:
          rr.msg= "{} left in global mode (run 88888888).".format(dets)
          rr.rc= 4
        else:
          moderc,msg= self.setModeLtus(dets, ltud_pb2.STDALONE, 0)
          if moderc!=0:
            #todo: put back changed ltus to stdalone...
            rr.msg= msg
          else:
            rr.msg="global run {} stopped.\n{}".format(runn, msg)
          rr.rc= 0
        del(self.gruns[runn])
        # todo: unlock
      else:
        rr.rc= 8
        rr.msg="run {} not started.".format(runn)
    else:
      rr.rc= 8
      rr.msg="Both run number and detector name missing, at least one expected."
    return rr
  def RunConfig(self, req, context):
    runn= req.runn
    print("%s %s %d"%("RunConfig:", self.getthreadid(), runn))
    #
    rr= ctpd_pb2.RunReply()
    if runn>0:
      if runn in self.sruns:      # stdalone
        detector= self.sruns[runn]
        #ltustub,msg= self.ltuds.getltustub(detector)
        ltustub,msg= self.ltugrpcs.getltustub(detector)
        if ltustub==None:
          rr.rc= 12
          rr.msg= msg
        else:
          rcltu= ltustub.RunConfig(ltud_pb2.RunStopRequest(runn=runn))
          print("rc:{}\n{}".format(rcltu.rc, rcltu.msg))
          rr.rc= rcltu.rc
          rr.msg= rcltu.msg
      elif runn in self.gruns:    # global
        # todo: lock?
        part= self.gruns[runn]
        #if part==None:
        #  rr.rc= self.RUNSTARTERR
        pname= part.pname
        # todo: unlock?
        rr.rc= 0
        rr.msg="rconfig not available (yet) for global runs."
      else:
        rr.rc= 8
        rr.msg="run {0} not started.".format(runn)
    else:
      rr.rc= 8
      rr.msg="Run number not given, no action"
    return rr
  def OrbitReset(self, request, context):
    name=request.name
    r= ctpd_pb2.SReply()
    self.fpga.orbitset(None, 3564)
    r.rc= 0
    r.msg=""
    print("OrbitReset done")
    return r
  def MonCycle(self, req, context):
    rc= self.tpnet.measureSlow()
    self.montimestamp= rc[0]
    #print("MonCycle measureSlow rc:", rc)
    return ctpd_pb2.Empty()
  def Lock(self, req, context):
    """ From: awmtors onl
"""
    if (self.lockedfor!= None):   # and (req.user=="None"):
      #self.logger.warning("Lock from %s, but was locked for %s"%(req.user,self.lockedfor))
      print("Lock from %s, but was locked for %s"%(req.user,self.lockedfor))
      action= "lockother"
    else:
      self.lockedfor= req.user
      print("Lock by "+req.user)
      action= "lockme"
    return ctpd_pb2.LockReply(lockaction=action)
  def Unlock(self, req, context):
    """ From: awmtors or ctpd_client ("None" request)
"""
    if (req.user!= self.lockedfor) and (req.user!="None"):
      #self.logger.error("Unlock from %s, but was locked for %s"%(req.user,self.lockedfor))
      print("Unlock from %s, but was locked for %s"%(req.user,self.lockedfor))
      action= "lockother"
    else:
      if req.user=="None":
        # forced unlock request from ctpd_client.
        # wsup-msg to awm from where sent further to all ws clients:
        orc= self.tpnet.awmc.forceunlock()
      print("Unlock by {} (was {}).".format(req.user, self.lockedfor))
      self.lockedfor= None
      action= "unlockme"
    return ctpd_pb2.LockReply(lockaction=action)
  def _lock(self):
    #rc= fcntlflock.acquire("CTPd", -1)  does not work for threads
    #print("_lock:", rc)
    self.mutex.acquire()   # ok for threads
  def _unlock(self):
    #rc= fcntlflock.release("CTPd")
    #print("_unlock:", rc)
    self.mutex.release()
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
def monStub(server):
  threadid= threading.current_thread().getName()
  iloop= 0
  channel= grpc.insecure_channel('localhost:%d'%server.portn)
  if channel== None:
    print("monStub error.")
    return
  ctpstub = ctpd_pb2_grpc.CTPd(channel)
  while True:
    if server.finish: break
    iloop= iloop+1
    rc = ctpstub.MonCycle(ctpd_pb2.Empty())
    if (iloop<3) or ((iloop%100)==1):
      server.printN("monStub %d thread:%s rc:%s"%(iloop, str(threadid), str(rc)))
    time.sleep(server.monsleep)
  print("monStub finishing...")
def serve():
  global server
  netname= auxfuns.getarg1()
  if netname=="":
    print("""
ctpd ctp [start | kill ] [init] [mon] [dummy]
where:
netname  $CTP3_CONFIG_DIRECTORY/nets/netname.net
         must exist. One of: {0} 
start
kill     kill running daemon (to be used only in case 'ctpdc netname qs'
         does not work)

init     initialise the net (master board(PLL)+INIT_OLTS)
         Without this, ttcpon net is expected to be active
mon      start monitoring thread in addition to the ctpd thread
dummy    no ipbus operation (only .net file read + monitor operation).
         requires:   
         /opt/cactus/bin/uhal/tests/DummyHardwareTcp.exe -p 50003 -v 2 &
         and
         a master board in .net file must be defined in alice_cf.xml as:
         <connection id="dummyhw" uri="ipbustcp-2.0://localhost:50003" address_table="file://minimal.xml"/>

""".format(str(tri_grpcs.grpc_channels.keys())))
    return
  # CTPd() instantiated here once when server starts
  startserver= False
  init= None
  dummy= False
  for iarg in range(2, len(sys.argv)):
    arg= sys.argv[iarg]
    if arg == "start":
      if iarg!=2: 
        print("start allowed only after netname")
      else:
        startserver= True
    elif arg == "init":
      init= "init"
    elif arg == "dummy":
      dummy= True
    elif arg == "mon":
      pass
    else:
      print("Bad arg:", arg)
      return
  #if auxfuns.isinargs("init"): init= "init"
  #if auxfuns.isinargs("dummy"): dummy= True
  if netname in tri_grpcs.grpc_channels:
    portn= tri_grpcs.grpc_channels[netname]
  else:
    print("Unknown netname:%s"%netname)
    return
  # check if portn already taken:
  rc= tri_grpcs.checkport(portn)
  if len(rc)>0:
    print("ctpd for %s already active.\n%s"%(netname,rc))
    return
  ctpserver= CTPd(netname, init, dummy)
  if ctpserver.finish:
    print("ctpd not started.", ctpserver.finish)
    return
  portn= ctpserver.portn
  max_workers= 1  # if 1, _lock/_unlock of ATB resources not involved
  #if max_workers>1:
  print("grpc.server,max.workers:%d."%max_workers)
  server= grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
  #print("grpc.server add...")
  ctpd_pb2_grpc.add_CTPdServicer_to_server(ctpserver, server)
  SERVICE_NAMES = (
    ctpd_pb2.DESCRIPTOR.services_by_name['CTPd'].full_name,
    reflection.SERVICE_NAME,
  )
  reflection.enable_server_reflection(SERVICE_NAMES, server)
  try:
    server.add_insecure_port('[::]:%d'%portn)
    #print("rc:",rc)
  except:
    print("Can't bind to %d"%portn)
    return
  #server.start()
  mon_t= None
  if startserver==False: return
  try:
    server.start()
    print("server [::]:%d started."%portn)
    if auxfuns.isinargs("mon"):
      # optionA: using max_workers>=1 with _lock/_unlock 
      #mon_t= threading.Thread(target= monThread, args=(ctpserver,), name= "mon_t")
      # optionB: max_workers:1, no _lock/_unlock
      mon_t= threading.Thread(target= monStub, args=(ctpserver,), name= "mon_t")
      print("Starting mon_t...")
      ctpserver.logger.info("Starting mon_t...")
      # Note: measureConfig already done (when server instantiated).
      mon_t.start()
    ixmon= 0
    while True:
      #time.sleep(_ONE_DAY_IN_SECONDS)
      time.sleep(ctpserver.monsleep*3)
      ixmon= ixmon+1
      if (mon_t!= None) and (not mon_t.isAlive()):
        print("mon_t isAlive():{}".format(mon_t.isAlive()))
        if ixmon>3:
          ctpserver.finish= "monitoring test failed too many times."
          break
  #except KeyboardInterrupt:
  except:
    print("Error:"+str(sys.exc_info()))
    traceback.print_exc(file=sys.stdout)
    server.stop(0)
if __name__ == '__main__':
  auxfuns.cdworkdir()
  #logging.basicConfig()
  serve()
