#!/usr/bin/python3
from sys import argv
from datetime import datetime
from time import sleep
from ltudobject import *

from pcp.pmda import PMDA

from logging import info

class Broker:
    cfg = []
    heartbeat = 0
    interval = 1
    path = ""
    host = ""
    options = {
        "config": "/opt/cockpit/alice-ctp/input.txt",
        "debug": False,
        "host": "localhost"
    }

    def __init__(self, path, heartbeat = 5, interval = 1, options = None):
        self.options = options if options else self.options
        if options.names:
            self.options.debug = False
        if options.debug:
            PMDA.dbg("Broker Initialisation")
        self.path = path
        self.heartbeat = heartbeat * interval
        self.interval = interval
        self.host = options.host if options else "localhost"
        self.load_config(path)

    def load_config(self, path):
        if self.options.debug:
            PMDA.dbg("Broker - Loading config from %s" % path)
        with open(path, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if line.startswith('#'):
                    continue
                data = line[:-1]
                description = ""
                if '#' in line:
                    [data, description] = line[:-1].split('#')
                    data = data.strip()
                    description = description.strip()
                [name, num, port] = data.split(" ")
                obj = LTUDObject(self.host, port, name, description, self.options, self.heartbeat)
                self.cfg.append(obj)

    def run(self):
        print("[INFO] Path: " + self.path)
        print("[INFO] Heartbeat: " + str(self.heartbeat))
        print("[INFO] Interval: " + str(self.interval))
        cnt = 0
        while True:
            print("----")
            for obj in self.cfg:
                obj.update()
            if cnt >= self.heartbeat:
                cnt = 0
                for obj in self.cfg:
                    if obj.disabled == True:
                        print("[WARN] Trying to reconnect to " + obj.host + ":" + obj.port)
                        obj.disabled = False
            cnt += 1
            sleep(self.interval)

    def pcp(self, cluster, item, inst):
        if self.options.debug:
            PMDA.log("Broker pcp() called")
        [counters, changed] = self.cfg[cluster].update()
        if self.options.debug:
            PMDA.log("Changed: %d" % changed)
        return [counters[item], changed]

    def init(self):
        for obj in self.cfg:
            obj.update_grpc()


if __name__ == "__main__":
    if len(argv) != 2:
        print("[ERR] Path not specified")
        exit(1)
    path = argv[1]
    broker = Broker(path, 3, 1)
    broker.run()
