#!/usr/bin/python3
from datetime import datetime
from ltud_pb2 import Empty
from ltud_pb2_grpc import LTUdStub
import grpc
from pcp.pmda import PMDA

class LTUDObject:
    counters = []
    names = []
    timestamp = 0.0
    updated = 0
    host = ""
    port = ""
    timeout = -1
    heartbeat = 0
    disabled = False
    name = ""
    description = ""
    channel = None
    stub = None
    options = None

    def __init__(self, host, port, name, description, options, timeout = 2):
        if options.debug:
            PMDA.dbg("[%s] LTUdObject Initialisation" % name)
        self.options = options
        self.host = host
        self.port = port
        self.timeout = timeout
        self.updated = datetime.now()
        self.name = name
        self.description = description
        self.channel = grpc.insecure_channel(host + ":" + port)
        self.stub = LTUdStub(self.channel)
        if options.names:
            self.update_grpc()
            for n in self.names:
                if len(n) > 0:
                    print("cts." + name + "." + n)
            return
    
    def update(self):
        if self.options.debug:
            PMDA.dbg("[%s] LTUdObject update" % self.name)
        if (datetime.now() - self.updated).total_seconds() > self.timeout:
            if self.options.debug:
                PMDA.dbg("[%s] LTUdObject update - Cache invalid, need to contact gRPC server" % self.name)
            self.update_grpc()
            return [self.counters, 0]
        return [self.counters, 1]

    def update_grpc(self):
        if self.disabled:
            if self.options.debug:
                PMDA.dbg("[%s] LTUdObject update_grpc - this object is disabled" % self.name)
            return
        try:
            if len(self.names) == 0:
                if self.options.debug:
                    PMDA.dbg("[%s] LTUdObject update_grpc - this object has no names yet" % self.name)
                self.names = self.stub.OutCntsGetNames(Empty()).names
                if self.options.debug:
                    for idx, n in enumerate(self.names):
                        PMDA.dbg("[%d] cts.%s.%s" % (idx, self.name, n))
            raw_counters = self.stub.OutCntsGet(Empty())
            self.timestamp = raw_counters.timestamp
            self.counters = raw_counters.outcnts
            self.updated = datetime.now()
            return 0
        except:
            self.updated = datetime.now()
            self.disabled = True
            return 0
