#!/usr/libexec/platform-python
from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
import ipbuswrap,i2c,fcntlflock
import time,traceback,json
import rpyc, threading

MEASUREINT= 5
Boards= {}   # list of ATBoard references (None: dead) 
class AtbsmonService(rpyc.Service):
  def __init__(self):
    #self.conns= 0
    self.quit= False   # True: stop atbmon
    self.mutex_lastabs= threading.Lock()   # self.lastabs exclusive access
    self.reopen= False # ->True  if main should update
    self.lastabs= {}   # recently received from 'checkboards.py'
    # status:
    self.started= str(time.time())
    self.loops= 0 ; self.atbupdates= 0
  def incloops(self):
    self.loops= self.loops+1
  def exposed_stopmon(self):
    print("stop cmd received.")
    self.quit= True
  def exposed_getstatus(self):
    return ("started, loops, atbupdates, quit:",self.started,self.loops,self.atbupdates, self.quit)
  def exposed_update(self,freshatbsstr):
    ''' see: saxipbus.py ATBmonclient()
! if exception here, it is seen only in client side!
    '''
    self.atbupdates= self.atbupdates + 1
    # check if changed:
    #if len(self.lastatbs) != len(self.freshatbs):
    freshatbs= eval(freshatbsstr)
    if freshatbs== self.lastabs:
      print("update atb: List of active boards equal, no need to update.")
      return
    print("update atb...", str(time.time()), freshatbs)
    print("   lastabs:",self.lastabs)
    #print "AtbsmonService:", freshatbs,type(freshatbs)
    #print "AtbsmonService:", freshatbs.keys()   # nebavi
    self.mutex_lastabs.acquire()
    self.badded= []
    for fkey in list(freshatbs.keys()):  # add new
      if fkey in self.lastabs: 
        continue
      else:
        self.lastabs[fkey]= freshatbs[fkey]
        self.badded.append(fkey)
    print("added:", self.badded)
    self.bdeleted= []
    for lkey in list(self.lastabs.keys()):   # remove dead
      if lkey in freshatbs: 
        continue
      else:
        del self.lastabs[lkey]
        self.bdeleted.append(lkey)
    print("dels:", self.bdeleted)
    self.reopen= True   # will be taken into account in approppriate time
    self.mutex_lastabs.release()
class ATBoard(object):
  def __init__(self, name, ftl=None):
    self.bname= name
    self.ftl= ftl   # db  to write to (open seq. file)
    self.dbtimeout= None
    if name=="atb_29":
      self.dbtimeout= 60*MEASUREINT
    #print "OpenBoard:",name
    self.board= ipbuswrap.OpenBoard(name)
    # i2c:
    #self.i2cs= [] 
    i2cmuxs= (0x74, 0x76)
    if self.board.nodExists("i2cmainttc", True):
      self.i2cg= i2c.I2Ctpcore(self.board,"i2cmainttc",i2cmuxs)
      #self.board.Write("i2cmainttc.mux", 11)   # 1..9: olt1..9   10:onu   11:ipbus
      #print "i2cmainttc bus."
      #self.i2cs.append(self.i2cg)
    elif self.board.nodExists("i2cmain", True):   # used ONLY for PLLs and FMC card
      self.i2cg= i2c.I2Ccore(self.board,"i2cmain",i2cmuxs)
      #self.i2cs.append(self.i2cg)
    else:
      print("NO i2cmain[ttc] present!")
    # self.dev_...  -reference to a device on the board
    #self.dev_tc74= self.i2cg.addDevice(i2c.TC74(self.i2cg, "tc74", mux=(0x76,0, 0x74,1<<0)))
    self.dev_tc74= i2c.TC74(self.i2cg, "tc74", mux=(0x76,0, 0x74,1<<0))
    #self.dev_sfp= self.i2cg.addDevice(i2c.SFP(self.i2cg, "SFP", "ipbus", mux=(0x76,0, 0x74,1<<1)))   # ipbus connection
    self.dev_sfp= i2c.SFP(self.i2cg, "SFP", "ipbus", mux=(0x76,0, 0x74,1<<1))   # ipbus connection
    if self.board.nodExists("i2cpll", True):
      self.i2cpll= i2c.I2Ccore(self.board,"i2cpll")
      #self.i2cs.append(i2cpll)
      self.dev_pll= i2c.Si534x(self.i2cpll, "si5345")
      #i2cpll.addDevice(dev_pll)
    # gathered monitored values:
    self.timestamp= None
    self.timestamp_db= None   # when last time writtten in db
    self.tc74_db= None        # last value written in db
    self.nlocked= 0           # number of consecutive loops when board was locked
    self.tc74= 0
  def temp_tc74(self):
    try:
      temp= self.dev_tc74.gettemp()
    #except:
    except Exception as e:
      temp= None 
      print("temp_tc74 exception...")
      #print("temp_tc74 exception:" + traceback.format_exc())
    return temp
  def tss(self):
    return str(self.timestamp)
  def measure(self,dead=False):
    """ dead: True if we want to remove board from monitoring
    rc: None:
        - board not available
        - bad measurement (e.g. tc74 is 0)
        - board locked (self.nlocked>0), i.e. not clear if ON/connected
    """
    self.timestamp= time.time()   
    if dead:
      self.tc74= None   # seems board dead/disconnected
    else:
      if fcntlflock.acquire(self.bname,1):
        if self.nlocked > 0:   # was just released
          print("released: %s %s after lockedloops:%d"%\
           (self.bname, self.tss(),self.nlocked))
          self.nlocked= 0
        try:
          self.board.Write("i2cmainttc.mux", 11)
          #cca 0.1s difference when writing to seq file also
          # without the write to seq file:
          self.tc74= self.temp_tc74()
        except:
          self.tc74= None   # seems board dead/disconnected
        finally:
          fcntlflock.release(self.bname)
      else:
        self.tc74= None   # cannot measure (locked by other process)
        self.nlocked= self.nlocked+1
      if (self.tc74== None) or (self.tc74==0):
        print(("%s %s None, lockedloops:%d)"%(self.bname,\
          self.tss(),self.nlocked)), self.tc74)
        if self.tc74==0: self.tc74= 1  # can't divide by 0 
      else:
        pass # print("%s %s %d"%(self.bname, self.tss(), self.tc74))
    return self.tc74
  def dbupdate(self):
    # consider  smoothing  (_db)
    if (self.timestamp_db!=None) and (self.dbtimeout!=None):
      if abs(self.timestamp - self.timestamp_db)>self.dbtimeout:
        forcedupdate= True
      else:
        forcedupdate= False
    else:
      forcedupdate= False
    if self.tc74==None:        # not available now (dead or locked)
      if self.tc74_db!=None:   # was available, update the time
        #self.ftl.write("%s %s %s -\n"%(self.bname, "tc74", self.tss()))
        msg= {"name":self.bname, "time":self.timestamp, "dead":"1"}
        self.ftl.write(json.dumps(msg)+"\n")
        self.timestamp_db= self.timestamp   # when last time writtten in db
        self.tc74_db= self.tc74        # last value written in db
    #elif (self.tc74_db==None) or (abs(self.tc74-self.tc74_db)/self.tc74db >0.01):
    elif (self.tc74_db==None) or (abs(self.tc74-self.tc74_db) > 1.1) or forcedupdate:
      # just appeared back or significant difference w.r.t. previous measurement
      #self.ftl.write("%s %s %s %d\n"%(self.bname, "tc74", self.tss(), self.tc74))
      msg= {"name":self.bname, "time":self.timestamp, "tc74":(self.tc74,0)}
      self.ftl.write(json.dumps(msg)+"\n")
      self.timestamp_db= self.timestamp   # when last time writtten in db
      self.tc74_db= self.tc74        # last value written in db
    else:
      pass #print "no change %s tc74:"%self.bname, self.tss()
def main(asinst, thrs):
  ignoremeas= 0   # ignore measuring when >0
  #Boards= {}
  print("opening /tmp/atbmon.log...")
  ftl= open("/tmp/atbmon.log", "w")
  #with open("/tmp/atbmon.log", "w") as ftl:
  while True:
    if asinst.quit: 
      print("thrs.close...")
      thrs.close()
      break
    if ignoremeas>0:   # monitoring paused for a while, DB dead?
      ignoremeas= ignoremeas-1
      if ignoremeas>0: continue
      print("Measure again, loop:%d..."%asinst.loops)
    #print "loop:",asinst.loops
    if asinst.reopen: 
      # throw out not available:
      for dkey in asinst.bdeleted:
        print("closing %s..."%dkey)
        Boards[dkey].measure(dead=True)
        Boards[dkey].dbupdate()
        del Boards[dkey]
      print("removed boards:",asinst.bdeleted)
      #print "or closing ipbus manager (todo ...)"
      #m.close() ???
      m= ipbuswrap.Init()
      if m==None: break
      #ipbuswrap.LoggingOff()
      #
      # add new boards:
      for dkey in asinst.badded:
        print("adding %s..."%dkey)
        Boards[dkey]= ATBoard(dkey, ftl)
      print("added boards:",asinst.badded)
      #for atbname in asinst.lastabs:
      asinst.reopen= False
    # measure:
    #timest= str(time.time())   # ctime:Sun Oct  7 17:49:33 2018
    notavail= []   # list of 'not responding boards'
    for ix in list(Boards.keys()):
      b= Boards[ix]
      if b.measure()==None: 
        if b.nlocked == 0 : 
          notavail.append(ix)
        else:   # board is locked, not availabe for monitor, regard it as dead
          if b.nlocked > 3 : 
            print("Board %s locked too long, will be removed..."%b.bname)
            notavail.append(ix)
    # update database:
    for ix in list(Boards.keys()):
      try:
        Boards[ix].dbupdate()   # i.e.also when board went dead or up
      except:
      #except Exception, e:
        #print "e:",str(e)   # is in format_exc output
        print("dbupdate exception:" + traceback.format_exc())
        print("Cannot write to db at loop:%d:%s, breaking recording to db for 10 loops"%\
          (asinst.loops,str(time.time())))
        ignoremeas=10
        break
    # remove 'dead/unavailable' boards:
    asinst.mutex_lastabs.acquire()
    for ix in range(len(notavail)-1,-1,-1):
      bname= Boards[notavail[ix]].bname
      print("Removing board:",bname, notavail[ix])
      del asinst.lastabs[bname]
      del Boards[notavail[ix]]
    asinst.mutex_lastabs.release()
    asinst.incloops()
    #if (asinst.loops%10) == 0: ftl.flush()
    ftl.flush()
    if asinst.quit: 
      print("thrs.close2...")
      thrs.close()
      break
    time.sleep(MEASUREINT)  
  ftl.close()

def run():
  from rpyc.utils.server import ThreadedServer
  asinst = AtbsmonService()
  thrs = ThreadedServer(asinst, port=18861)
  # start counters readingthread:
  mon_t= threading.Thread(target= main, args=(asinst, thrs), name= "mon_t")
  print("starting monitoring thread (main)...")
  mon_t.start()
  print("starting Atbsmon service (listening to commands)...")
  # start rpyc service thread:
  thrs.start()
  print("quitting...")
  asinst.exposed_stopmon()  

if __name__ == "__main__":
    run()
