#!/usr/libexec/platform-python
"""
Translate .slm (mnemonic SLM instructions) to .seq (binary-like) file.
1 line ->  to be saved as 4 32bit words in emulator memory
0x... [01]+112/4 hexa digits =  113bit
0x code L/R Dmask Imask Ttype
   3    2   16    74    16   bits
See slmdefs.Slmdefs.bps...

"""
from __future__ import print_function
from builtins import range
from builtins import object
import os.path, os, slmdefs

sdf= slmdefs.Slmdefs
slmseqpath='slmseq.seq'
errtext=''
wartext=''
def Err(text=None, warn=None, line=None):
  global errtext,wartext
  if text==None: 
    errtext=""; wartext=""
  else:
    if warn:
      if line: wartext= wartext+line+'\n'
      wartext= wartext+"Warning:"+text+'\n'
    else:
      if line: errtext= errtext+line+'\n'
      errtext= errtext+"Error:"+text+'\n'
def PrintErrors():
  if wartext:
    print(wartext)
  if errtext:
    print(errtext)

class Sequence(object):
  def __init__(self, line):
    #print line
    self.sx= 0
    # RAn instruction format:
    # code dmask imask flags (flags: see Slmdefs.bps in slmdefs.py)
    self.noswc=0               # NOSWC flag not present
    self.line=line
    atms= line.split()
    self.code=0
    for ix in range(len(slmdefs.Slmdefs.scodes)):
      if ix==0: continue
      cix = slmdefs.Slmdefs.scodes[ix]
      if atms[0] == cix:
        self.code= ix
    self.placebits(self.code, slmdefs.Slmdefs.bps["scode"])
    if self.code==0:   
      #  Err("Unknown sequence code:"+atms[0])
      pass # Let's allow ZERO
    if len(atms)<3:
      Err("Imask missing, set to 0","Warning",line)
      atms.append("0x0")
    if len(atms)<2:
      Err("Dmask missing, set to 0","Warning",line)
      atms.append("0x0")
    dmask= eval(atms[1])
    imask= eval(atms[2])
    #if not dmask:
    #  Err("Bad definition of dmask:"+atms[1])
    #  return
    self.placebits(dmask, slmdefs.Slmdefs.bps["dmask"])
    self.placebits(imask, slmdefs.Slmdefs.bps["imask"])
    for flag in atms[3:]:
      self.procflag(flag)
  def placebits(self, bpat, poslen):
    pshift= poslen[0]
    self.sx= self.sx | (bpat<< pshift)
  def procflag(self, flag):
    pshift= slmdefs.Slmdefs.bps[flag]
    self.sx= self.sx | (1<< pshift)
  def savebin(self, of):
    of.write("%x\n"%self.sx)
class Cmpslm(object):
  def __init__(self, slmfile):
    global sdf
    sdf= slmdefs.Slmdefs
    Err()
    self.slm=[]   # max. 51 items (sequences) 113 bits long
    #self.allowederrs=[0,0,0,0,0,0,0]   # allowed errors
    if os.access(slmfile, os.R_OK) == 0:
      Err(slmfile+" doesn't exist")
      return
    self.bname= os.path.basename(slmfile)
    sf= open(slmfile, 'r')
    for line in sf.readlines():
      #print "Cmpslm:",line
      #if line[:7]== 'Errors:':
      #  for er in string.split(line[7: -1]):
      #    if er=='PP': self.allowederrs[0]=1
      #    elif er=='L0': self.allowederrs[1]=1
      #    elif er=='L1': self.allowederrs[2]=1
      #    elif er=='L1M': self.allowederrs[3]=1
      #    elif er=='L1&L1M': self.allowederrs[4]=1
      #    elif er=='L2aM': self.allowederrs[5]=1
      #    elif er=='L2rWord': self.allowederrs[6]=1
      #    else: Err("Unknown allowed errors flag:"+er)
      #  continue
      if line[0]=='#': 
        #print line
        continue
      if line[0]=='\n': continue
      if line[-1]=='\n': line= line[0:-1]
      self.slm.append(Sequence(line))
    #if len(self.slm)>0 and (self.slm[-1].flglast()==0):
    #  self.slm[-1].flglast(1)
    #  Err("LAST flag forced to 1 in the last sequence", "Warning")
    sf.close()
  def save2file(self, fn):
    of= open(fn, "w")
    #of.write( string.split(os.path.basename(fn),'.')[0]+'\n')
    #of.write( str(len(self.slm))+'\n')
    for ix in range(len(self.slm)):
      self.slm[ix].savebin(of)
    of.close()
  def ememBlock(self):
    emumem=[]
    for ix in range(len(self.slm)):
      v113= self.slm[ix].sx
      for ix2 in range(3,-1,-1):
        v32= (v113 >> (ix2*32)) & 0xffffffff
        emumem.append(v32)
    return emumem
def main():
  global errtext, wartext
  import sys
  if len(sys.argv) < 2:
    print("""
Convert .slm file to slmseq.seq file.

Usage:
       assemu.py [-emublock] name.slm

name.slm:   abs. path or relative path. name
-emublock:  show result as an emulator memory block (to be loaded)

Operation:
File slmseq.seq is created, which can be than:
 -loaded to LTU seq. memory (see Cmpslm.ememBlock() )
 -reverse-compiled by disemu.py ->should return original .slm file
""")
    #a= disslm("all.slm")
    #print a.getlist()
  else:
    #os.chdir(os.environ['VMEWORKDIR'])
    emublock= 0
    for bn in sys.argv[1:]:
      if bn=="-emublock": emublock=1 ; continue
      a= Cmpslm(bn)
      if wartext:
        print(wartext)
      if errtext:
        print(errtext)
      else:
        a.save2file(slmseqpath)
        print("Saved: %s"%(slmseqpath))
        if emublock==1:
          emem= a.ememBlock()
          for ix in range(0, len(emem), 4):
            line= "%4d:"%ix
            for ix2 in range(4):
              if (ix+ix2)>= len(emem): break
              line= line+ " 0x%x"%emem[ix+ix2]
            print(line)
if __name__ == "__main__":
  main()

