#!/usr/libexec/platform-python
"""slm disassembler
Usage: 
disemu.py file.seq
  -> converts 'binary' form to human readable form
Input file (file.seq):
"""
from __future__ import print_function
#!/usr/bin/python3
from builtins import hex
from builtins import range
from builtins import object
import os.path, os, types, slmdefs
sdf= slmdefs.Slmdefs
class Seq(object):
  SNAME=['ZERO', 'L0  ', 'L2A ', 'L2R ', 'CPP ', 'CL0 ', 'CL2A', 'CL2R']
  def __init__(self, hexaline):
    #self.sbin= eval("0x"+hexaline)
    self.sbin= hexaline
    self.errs=self.warns=''
  def fieldpos(self, fieldname):
    pl= slmdefs.Slmdefs.bps[fieldname]
    return pl
  def fieldv(self, fieldname):
    pl= self.fieldpos(fieldname)
    #if type(pl) is types.IntType:
    if type(pl) is tuple:
      return (self.sbin >> pl[0]) & (2**(pl[1])-1)   # field
    else:
      return (self.sbin >> pl) & 1   # flag
  def printcheck(self):
    seqok=1
    #print Seq.SNAME[self.s[0]&0x7],':',self.s[0]
    vbin= self.fieldv("scode")
    if vbin == 0: seqok= 0
    scode= slmdefs.Slmdefs.scodes[vbin]
    vbin= self.fieldv("dmask") ; dmask= hex(vbin)
    if dmask[-1]=='L': dmask= dmask[0:-1]
    vbin= self.fieldv("imask") ; imask= hex(vbin)
    if imask[-1]=='L': imask= imask[0:-1]
    instrline= "%s %s %s" % (scode, dmask, imask)
    flags=""
    for flagn in list(slmdefs.Slmdefs.bps.keys()):
      if type(self.fieldpos(flagn)) is tuple: continue
      flbin= self.fieldv(flagn)
      #flag
      if flbin == 1:
        if flags!="":
          flags= flags + " " + flagn
        else:
          flags= flagn
    instrline= instrline + " " + flags
    if self.warns:
      instrline= instrline+"Warnings: "+ self.warns+"\n"
    return instrline,seqok
class Disslm(object):
  def __init__(self, seqinp, number=0):
    global sdf
    self.loglist= ''
    self.number= number
    self.fileok= 1   # 0 if load not recommended
    self.slm= []   #max. 51 items (Seq objects)
    slmbin= []   #max. 51 items (Seq objects)
    #print "Disslm.init: seqinp type:", type(seqinp)
    #if type(seqinp) == types.ListType:   not, seqinp type: <class 'uhal._core.ValVector_uint32'>
    if type(seqinp) != bytes:
      # emu memory (4 words per 1 instruction):
      for ix in range(0, len(seqinp), 4):
        seqbin= 0 ; lastlen= 0
        for ix2 in range(4):
          if ix+ix2 >= len(seqinp): break
          seqbin= (seqbin<<32) | seqinp[ix+ix2]
          lastlen= lastlen +1
        if lastlen != 4:
          self.error("Error:short mem (%d), multiple of 4 expected"%len(seqinp));
          self.fileok=0
          break
        #print "seqbin:", ix, hex(seqbin)
        slmbin.append(seqbin)
    else:
      if os.access(seqinp, os.R_OK) == 0:
        self.error(seqinp+" doesn't exist")
        return
      sf= open(seqinp, 'r')
      for line in sf.readlines():
        if line[-1]=='\n': line= line[0:-1]
        lenline= len(line)
        # at least 111 bits (code:1) + 0x -> 108/4+2 = 29 chars
        if line=='' or line==None or lenline<28: 
          self.error("Error:too short line (%d) at least 28 hex. chars expected"%lenline);
          self.fileok=0
          continue
        slmbin.append(eval("0x"+line))
      sf.close()
    # disassembly slmbin[]:
    for line in slmbin:
      self.slm.append(Seq(line))
      #if self.slm[-1].word(ln%sdf.wordseq, line):
      #  self.error("Error: incorrect format of input file")
      #  del(self.slm[-1])
      #  self.fileok=0
      #  break
  def a2bin(self,line,bits):
    """Convert binary string 10111 into 0x17"""
    binword=bit=0
    #print "a2bin:",line
    #for i in range(bits-1,-1,-1):
    if len(line)<bits+1:
      self.error("Incorrect ascii binary(short line):"+line[:-1])
      return binword
    for i in range(bits):
      if line[i]=='1':
        binword= binword | 1<<bit
      elif line[i]!='0':
        self.error("Incorrect ascii binary:"+line[:-1])
        break
      bit= bit+1
    return binword
  def add2list(self,text):
    self.loglist= self.loglist+text
  def error(self,text):
    self.add2list(text+"\n")
  def getaerrs(self):   # called form ltu_u.py
    return self.AllowedErrors
  def getlist(self):
    n=0     # sequence number (printed from 1:)
    for sq in self.slm:
      lst,sqok= sq.printcheck(); n=n+1
      if self.number==1:
        self.add2list("%2d: %s\n"%(n, lst))
      else:
        self.add2list("%s\n"%(lst))
      if sqok==0:
        self.fileok=0   # load not recommended
    if len(self.slm)>0 and self.slm[-1].fieldv("last")==0:
      self.error("Warning: LAST flag not set in the last sequence")
      self.fileok=0
    return self.loglist

def main():
  import sys
  if len(sys.argv) < 2:
    print("""
Reverse compilation of .seq file (text file 
representing bits in LTU-emulator memory). The format of .seq
file is described in file assemu.py.

Usage: disemu.py [-test4] name.seq

Expected abs. path or relative path. name
Operation: .slm file printed to stdout
-test4:   test input from memory block (not from text file)
""")
  else:
    #os.chdir(os.environ['VMECFDIR'])
    number= 0
    for bn in sys.argv[1:]:
      if bn=="-number":
        number= 1
        print("-number : with line numbers")
        continue
      if bn=="-test4":
        mem=[0x5fff, 0xe0000000, 0x00000000, 0x00010001,\
0x5fff, 0xc0000000, 0x00000000, 0x00020001,\
0x5fff, 0x80000000, 0x00000000, 0x00040001]
        a= Disslm(mem, number)
        print(a.getlist())
        break
      a= Disslm(bn, number)
      print(a.getlist())

if __name__ == "__main__":
  main()

