#!/usr/libexec/platform-python
from __future__ import absolute_import
from __future__ import print_function
import sys,random,time
class CDC:
  """ CDC from fi 11_09.
1.
Measure CDC with ONU inserted and connected to the CTP
2.
Set CDC phase in a board + save in the file
3.
In stdalone, set CDC phase using a value saved in a file (measured before in global mode)
4.
Always use value from a file (if present) when checking PLL
"""
  def __init__(self, board, tdomux):
    self.board= board
    self.tdomux= tdomux
    self.phase= self.getphase()
    print("cdc phase now: {0}".format(self.phase))
  def measure(self, setnew= False, n=5):   # n: number of measurements
    if self.tdomux.mux_name != "TDOmux":
      print("Cannot measure, not LTU.")
      return(None)
    omux= self.tdomux.getmux()
    if omux==8:
      print("tdomux:{0:d}({1}) ok".format(omux, self.tdomux.mux_docs[omux]))
    else:
      print(" Changing tdomux:{0:d}({1}) -> 8({2})...".format(omux, self.tdomux.mux_docs[omux],
        self.tdomux.mux_docs[8]))
      self.tdomux.setmux(8)
    oldphase= self.getphase()
    cdcsleep= 50
    self.reset()
    print("CDC reset done, sleeping randomly 1-{}ms between measurements...".format(cdcsleep))
    alls= {}   # res1:repetitions, res2:rep2,...
    for ix in range(n):
      # start phase scan
      self.board.Write("cdc_phase_scan.cdc_ctrl",1)   # start scan
      # wait time sufficient even shorter (50) when PLL lock sticky checked before start:
      # done: in readConfig, cclosl is checked when cdcscan done for ltus connected to ctp
      rn= random.randint(1,cdcsleep) ; time.sleep(0.001*rn)
      # read result of the scan
      cdcres= self.board.Read("cdc_phase_scan.cdc_status") & 0xfff
      self.board.Write("cdc_phase_scan.cdc_ctrl",0)   # 0: stop scan
      if alls.has_key(cdcres):
        alls[cdcres]= alls[cdcres]+1
      else:
        alls[cdcres]= 1
    # let's stop inside the loop - we should get 2 patterns with roughly the same
    # occurence
    #self.board.Write("cdc_phase_scan.cdc_ctrl",0)   # 0: stop scan
    self.allsorted= sorted(alls.iteritems(),key = lambda x : x[1], reverse=True)
    print("cdc scan results: {}".format(self.allsorted))
    phase1= self.cdcfind(self.allsorted[0][0])
    if len(self.allsorted)>1:
      phase2= self.cdcfind(self.allsorted[1][0], grid=False)
      if phase1==phase2:
        pass
      else:
        # only check if different phase, e.g. the following one is ok:
        # cdc scan results: [(3087, 4), (1008, 1)]
        # 012345012345
        # 110000001111   3087 0xc0f
        #           ^---suggested phase:4, length of 1s:6
        # 001111110000   1008 0x3f0
        #     ^---suggested phase:4, length of 1s:6
        print("Warning: {0}(phase1) != {1}((phase2)".format(phase1,phase2))
        phase1= None   # todo: consider if phases close
        if 1.0*self.allsorted[0][1]/self.allsorted[1][1] > 1.2:
          print("Too big difference in pattern occurences")
          phase1= None
      if len(self.allsorted)>2:
        print("Warning: too many patterns (>2)")
        phase1= None
    if (phase1!=None) and setnew:
      self.setphase(phase1)  
      print("New phase {0} set.".format(phase1))
    else:
      self.setphase(oldphase)  
      print("Original phase {0} restored.".format(oldphase))
    return(phase1)
  def getphase(self):
    return (self.board.Read("cdc_phase_scan.cdc_ctrl")>>3) & 0x7
  def setphase(self, phase):
    """ phase: 0..5  -result of cdcfind() """
    oldp= self.getphase()
    self.board.Write("cdc_phase_scan.cdc_ctrl",((phase&0x7)<<3)|0x2)
    self.board.Write("cdc_phase_scan.cdc_ctrl",((phase&0x7)<<3))
    self.phase= phase
    #self.reset()
    #print("cdc phase set: {0} -> {1}, cdc reset done after. ".format(oldp, self.phase))
    #print("cdc phase set: {0} -> {1}.".format(oldp, self.phase))
  def reset(self):
    cr= self.board.Read("cdc_phase_scan.cdc_ctrl")
    self.board.Write("cdc_phase_scan.cdc_ctrl",cr|0x4)
    self.board.Write("cdc_phase_scan.cdc_ctrl",cr)
  def cdcfind(self, sr, grid=True):
    """ sr: scan result: 12 bits
ideal case: 6*0 + 6*1 e.g.: 110000001111
- sometimes one '1' get lost -> leads to 7*0
- legal patterns: '6*0 6*1'  '7*0 5*1'

Operation: find 1st '1' in group of consecutive 6*1 (or 5*1)
1. find 10
2. find 01, not done:check consecutive N*0 (N=6 or 7)
3. check number of '1's
rc: 4 for this case:
012345012345  -grid number used for rc (phase)
001111110000
    ^
or 0 for this case:
012345012345
111100000011
^    
or 0 for this case:
012345012345
000011111100
^    

rc: phase ='grid number' of 3rd bit in the 1-train
    None + error message on stdout

todo: we can get 2 diffrent patterns (roughly in 50% of cases) with the
      same calculated phase:
A. stop scan inside the loop:
cdc scan results: {3843: 23, 252: 27} maxres: [252, 27]

B. stop scan after the loop:
atb_13>cdc m 550
phase now:  0
tdomux:8(ONU data) ok
cdc scan results: {252: 550} maxres: [252, 550]
012345012345
000011111100
      ^---phase:0 length of 1s:6
atb_13>cdc m 550
phase now:  0
tdomux:8(ONU data) ok
cdc scan results: {3843: 550} maxres: [3843, 550]
012345012345
111100000011
^---phase:0 length of 1s:6

"""
    #pos1s= [ 1<<ix for ix in range(12-1,-1,-1)] 
    pos1s= [2048, 1024, 512, 256, 128, 64, 32, 16, 8, 4, 2, 1]
    #
    # 1. find 10
    #ones= [one for one in pos1s if sr&one]
    #print(ones)
    #oneto0= [ix for ix in range(11) if (sr&pos1s[ix]) and (not (sr&pos1s[ix+1]))]
    #if (sr&pos1s[11]) and not (sr&pos1s[0]): oneto0.append(11) # check last bit
    # oneto0: indexes of last '1' in train of '1's. e.g. 5 for pattern 0xfc0
    #
    oneto0= [ix for ix in range(1,12) if ((sr&pos1s[ix])==0) and (sr&pos1s[ix-1])]
    if ((sr&pos1s[0])==0) and (sr&pos1s[11]): oneto0= [0]+oneto0 # check 1st bit
    # oneto0: indexes of 1st '0' after a train of '1's. e.g. 6 for pattern 0xfc0
    #print("index of 1st 0 after 1, oneto0:", oneto0)
    if len(oneto0) != 1:
      print("Bad pattern: {0:0>12b}".format(sr))
      return(None)
    #print("Pattern: {0:0>12b}".format(sr))
    #
    #2. find 01, start from oneto0[0]
    r13= [ix for ix in range(oneto0[0], 12)+range(oneto0[0])+[oneto0[0]]]
    #print("r13:",r13) 
    zeroto1= [ r13[ix] for ix in range(len(r13)-1) if (sr&pos1s[r13[ix]]==0) and (sr&pos1s[r13[ix+1]])]
    #print("indexes of '0's before '1's zeroto1:", zeroto1)
    if len(zeroto1) != 1:
      print("Bad pattern: {0:0>12b}".format(sr))
      return(None)
    #
    # 3. check number of '1's starting from zeroto1[0]
    ix1= zeroto1[0]+1            # index of 1st '1' in the train of 1s
    ixphase= (zeroto1[0]+3)%12   # index of 'phase is 2 BCs further
    ones= 0
    for ix in range(12):
      if ix1>11: ix1= 0
      if sr&pos1s[ix1]: ones= ones+1
      ix1= ix1+1
    if ones<5:
      print("Bad pattern: {0:0>12b}".format(sr))
      return(None)
    rc= ixphase%6
    pointer= "^".rjust(ixphase+1)
    if grid: print("012345012345")
    print("{0:0>12b}   {0:4d} 0x{0:x}".format(sr))   # padding '0' align right (^ center align)
    print("{0}---suggested phase:{1:d}, length of 1s:{2:d}".format(pointer,rc,ones))
    return(rc)
  def prtHelp(self):
    print("""
cdc m Nrep    -measure phase: read result Nrep times and set it
cdc g         -get current phase, should display a number 0..5
cdc set PHASE -set phase, where PHASE is a number 0..5
cdc r         -reset
""")
if __name__ == "__main__":
  cdc= CDC()
  cdc.cdcfind(eval(sys.argv[1]))
