#!/usr/libexec/platform-python
from __future__ import print_function
import sys
TFL=256
def checksynctf(tfa, allpairs=False):
  """ check if TF in sync.
tfa: a dictionary, N=len(tfa),  
     tfa[detname]=orbit_number acquired from tdg snapshot

allpairs: True -check all pairs, i.e. 2-combinations of set N
          False-quick check for pairs A vs B where
                A: max(tfa)= tfa[ixmax]   -the highest orbit item
                B: tfa[ix] where ix!=xmax -the orbit of remaining items

rc: number of good pairs.
    N-1: success, i.e. (A-B)/256 is an integer for any A,B pair
    <(N-1): at least 1 pair failed

for len(tfa)==3, allpairs=True, stdout:
./checksynctf.py 23*256 67*256 67*255 all
{'1': 8960, '3': 26265, '2': 26368} ['2', '3', '1']
____   2   3
  3:   ?
  1:   .   ?
rc: 0

stdout: table of [all] pairs, where x is: 
. -ok, orbit in sync   or
? -orbit not in sync

for len(tfa)==3, allpairs== False: print:
./checksynctf.py 23*256 67*256 67*256
{'1': 8960, '3': 26368, '2': 26368} ['3', '2', '1']
____   3
  2:   .
  1:   .

"""
  rc= 0
  #tfa.sort(reverse=True)   # tfa: list of orbits
  tfas= sorted(tfa, key=tfa.get, reverse=True)
  # tfas: the list of sorted keys by descending values in tfa
  print(tfa,tfas)
  # compare all possible pairs:
  head= "____"
  if allpairs:
    for name in tfas[:-1]:
      head= head+" "+"{:>3s}".format(name)
  else:
      head= head+" "+"{:>3s}".format(tfas[0])
  print(head)
  for irow in range(1,len(tfa)):
    maxc= irow-1
    #n1= tfa[irow]   # tfa: list
    n1= tfa[tfas[irow]]
    #line= "{:2d}:".format(irow)   # tfa: list
    line= "{:>3s}:".format(tfas[irow])
    rowok= True
    for icol in range(maxc+1):
      #n2= tfa[icol]   # tfa: list
      n2= tfa[tfas[icol]]
      dif= n1-n2
      ratio= dif//TFL
      if ratio*TFL == dif:
        ok= '.'
      else:
        ok= '?' ; rowok= False
      #line= line+"  {}".format(ok)   # tfa: list
      line= line+"   {}".format(ok)
      if allpairs==False: break
    if rowok: rc= rc+1
    print(line)
  return(rc)
def main():
  if len(sys.argv) < 2 or (len(sys.argv)==2 and sys.argv[1]=="help"):
    #print("""   python sys .hexversion .version .version_info: {}
    #{}
    #{}
    #""".format(hex(sys.hexversion), sys.version, sys.version_info))
    print(""" argv:{}

checksynctf.py hexa1 hexa2 hexa3...[all]
hexa*  -haxadecimal numbers (e.g. 0xfab123)
all    -check all pairs (default: minimal check)

example:
./checksynctf.py 10 110 210 310 all
____   4   3   2
  3:   .
  2:   .   .
  1:   .   .   .

{}
""".format(sys.argv, checksynctf.func_doc))
    sys.exit(8)
  #tfa= []
  #for tf1 in sys.argv[1:]:
  #  tfa.append(eval("0x"+tf1))
  tfa= {}
  allpairs= False
  for ix in range(1, len(sys.argv[1:])+1):
    tf1= sys.argv[ix]
    if tf1=="all": allpairs= True ; continue
    #tfa[str(ix)]= (eval("0x"+tf1))
    tfa[str(ix)]= (eval(tf1))
  rc= checksynctf(tfa, allpairs)
  print("rc:",rc)

if __name__== "__main__":
  main()
