#!/usr/libexec/platform-python
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

from __future__ import division
from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
from past.utils import old_div
from tabulate import tabulate
import time #,pdb
import hexdump
def prtErr(msg):
  print("ERROR:%s"%msg)
def bit_not(n, numbits=32):
  """set 0s in low numbits bits of n"""
  return (1 << numbits) - 1 - n

class Counters(object):
  """
parent of CountersCTPEMU, CountersTTC[_13], CountersLECO, CountersGBT,...TTCIT
                used_by     ipsl/ipslblock devname
CountersCTPEMU  atb/ltu     "outcnts.TTcnts"  outcnts
                atb/ctp     "outcnts1/2/3/...18.TTcnts" outcntsX (common outcnts)
CountersLTU         ltu     "cnts.cnts"       icnts
CountersCTP     atb/ctp     "gcnts.data"      gcnts
CountersTTC         ltu     "ttccnts.cnts"    ttccnts
CountersTTC_13      ltu     "ttccnts.REGNAME"     -obsolete
CountersLECO    atb/ltu     "leco1.data"          -obsolete
                            "oltrx_lec"       oltrx_lec -from ltu_v15, own prtall()
CountersGBT     atb/ltu     "gbtX.REGNAME"  X:1,2,... not array of regs!
CountersTTCIT   atb/ttcit   "ttcitcnts.ttcitcnts" (len:135)
"""
  #SIZE= 0    # virtual
  PERLINE_HLP= 2
  PERLINE= 2 # number of counters in 1 line when displayed
  PERLINE_RELPOS= 5
  mymsg= "Output counters."
  def __init__(self, board, ipsl,devname=None, ttccs=None):
    """
ipsl: top node name
devname:  optional (default: ipsl)
ttccs: CountersTTC instance (only for ipsl="outcnts")
"""
    self.ipsl= ipsl
    #if not "ipslblock" in vars(self):   -moved to CountersCTPEMU.__init__
    #  self.ipslblock= self.ipsl+".TTcnts"   # default for Counters[CTPEMU]
    #self.addrname= ipsl+".addr"
    #self.csa_slave= ipsl+"ctrl"
    if devname==None:
      self.devname= ipsl
    else:
      self.devname= devname
    self.board= board
    self.ttccs= ttccs
    self.cgroup= None    # group_print not active
    #self.fromad= 0; self.nws= 10; 
    self.dbgblock= [0]*self.SIZE
    self.lasttime= time.time()
    self.lastrr= self.readRaw()   # last readings for readInc()
    self.lastinctime= self.lasttime
    #print "%s counters read in (SIZE:%d)."%(self.devname, self.SIZE)
    self.incs= []               # (last - previous) readings
    for ix in range(self.SIZE):
      self.incs.append(0)
    # todo: check ctrl.pid_1 (fpga.getDport) if active run (g or s):
    # let's assume run is not active, i.e. recent xox was EOX
    self.xox= 'E' ; self.xoxraw=[]
    self.xoxtime= self.lasttime
    for ix in range(self.SIZE):
      self.xoxraw.append(self.lastrr[ix])
    if self.ipsl=="ttcitcnts":
      self.fmtnvals= "%18s:%10d    "
      self.fmthead= "%+18s "
      self.fmtval= "%18d "
    else:
      self.fmtnvals= "%8s:%10d    "
      self.fmthead= "%+12s "
      self.fmtval= "%12d "
  def xoxReport(self, prevxoxraw):
    """
rc: [[name, count], [name, count],...]
"""
    reptable= []
    for name in self.reportnames:
      relpos= self.cnames[name][1]
      cntdif= self.dodif32(prevxoxraw[relpos], self.xoxraw[relpos])
      reptable.append((name, cntdif))
    return reptable
  def readRawxox(self, xox):
    """ must be called in time of AfterEOR and BeforeSOR.
rc: report in multiline string"""
    prevxoxraw= [] ; prevxoxtime= self.xoxtime ; prevxox= self.xox
    for ix in range(self.SIZE):
      prevxoxraw.append(self.xoxraw[ix])
    if self.ttccs:
      ttcprevxoxraw= [] ; ttcprevxoxtime= self.ttccs.xoxtime ; ttcprevxox= self.ttccs.xox
      for ix in range(self.ttccs.SIZE):
        ttcprevxoxraw.append(self.ttccs.xoxraw[ix])
    self.readRaw(xox=xox)
    #if report: 
    # later todo: ttcnames=("tPHIn", tPH")
    # all (we do not need all for report):
    #self.calcupdate(prevxoxraw,self.xoxraw,xoxincs):
    counts= self.xoxReport(prevxoxraw)
    secs= self.xoxtime-prevxoxtime
    if secs>60.:
      mors= "{:.2f} minutes".format(secs/60.)
    else:
      mors= "{:.2f} secs".format(secs)
    rc= "{}OR. {}OR at:{} Duration: {}\n".format(\
      self.xox, prevxox, time.ctime(prevxoxtime), mors)
    if self.ttccs:
      #print("ttc (should be the same):\n{}OR. Previous {}OR at:{} Duration: {:.2f} minutes".format(\
      #  self.ttccs.xox, ttcprevxox, time.ctime(ttcprevxoxtime), (self.ttccs.xoxtime-ttcprevxoxtime)/60.))
      counts= counts + self.ttccs.xoxReport(ttcprevxoxraw)
    rc= rc + tabulate(counts, headers=("name","count"))
    return rc
  def readRaw(self, N=None, fromad=0, grpcreply=None, xox=None):   #Counters
    """op:
operation:
  -read counters
  -set self.lasttime
grpcreply: + fill grpcreply message (involves also ttccs read if necessary)
  -used in ltud_server.py
  -grpcreply present only for CountersCTPEMU
xox: E(EOX) S(SOX) None(reading inside or outside of the run)
     E,S: exclusive with grpc (i.e. grpcreply has to be None)
     E,S: update only self[.ttccs].xoxraw/xoxtime), 
     E,S: return None
     E,S overwrites self.xox* when read, i.e. should be called like:
       - store self.xox* to have them available for the calculation after read:
       - ...readRaw(xox='E')
       - calculate+print difference of necessary counters/time
rc: block of raw (32) counters
"""
    #print("readRaw dbg ipslblock:{} xox:{}".format(self.ipslblock, xox))
    size1= self.SIZE
    if N==None: N= size1
    if fromad>=size1: fromad= 0
    if (fromad+N)>size1: N= size1-fromad
    #
    #print("readRaw: %d counters from:%d dbgmode:%d"%(N, fromad, self.board.dbgmode))
    now= time.time()
    if self.board.dbgmode>0:
      msecs= int((now-self.lasttime)*1000)
      for i in range(len(self.dbgblock)):
        self.dbgblock[i]= self.dbgblock[i] + msecs
      block= self.dbgblock
    else:
      block= self.board.ReadBlock(self.ipslblock,N, fromad)
    if self.ttccs and ((grpcreply!=None) | (xox!=None)):
      #ttcblock= self.ttccs.readRaw(xox=xox)
      if self.board.dbgmode>0:
        for i in range(len(self.ttccs.dbgblock)):
          self.ttccs.dbgblock[i]= self.ttccs.dbgblock[i] + msecs
        ttcblock= self.ttccs.dbgblock
      else:
        ttcblock= self.board.ReadBlock(self.ttccs.ipslblock, self.ttccs.SIZE)
      if xox:
        for ix in range(len(ttcblock)):
          self.ttccs.xoxraw[ix]= ttcblock[ix]
        self.ttccs.xoxtime= now
        self.ttccs.xox= xox
        self.ttccs.lasttime= now
    self.lasttime= now
    if xox:
      for ix in range(len(block)):
        self.xoxraw[ix]= block[ix]
      self.xoxtime= now
      self.xox= xox
      return None
    #1.
    #blocko= block
    # ArgumentError: Python argument types in
    #     ValVector_uint32.__getitem__(ValVector_uint32, slice)
    # did not match C++ signature:
    #     __getitem__(uhal::ValVector<unsigned int>, int)
    # 
    # 2. anyhow, seems the block of all cnts have to be readout 
    # to get proper values
    blocko= []
    if grpcreply!=None:
      #print "cblock:",block
      grpcreply.timestamp= self.lasttime
      grpcreply.ClearField("outcnts")   # set default value
      #grpcreply.outcnts= []   #  Assignment not allowed to repeated field
      #del grpcreply.outcnts[:]    # should be ok too
      for ix in range(len(block)):
        blocko.append(block[ix]); grpcreply.outcnts.append(block[ix])
        # note: python float is actually double, i.e. use double (not float)
        # in .proto to get the same precision
        #blocko.apppresentend(ix) ; grpcreply.outcnts.append(ix)   # dbg
      if self.ttccs:
        for ix in range(len(ttcblock)):
          #blocko.append(ttcblock[ix]); 
          grpcreply.outcnts.append(ttcblock[ix])
    else:
      for ix in range(len(block)):
        blocko.append(block[ix])
    #print("readRaw dbg len(rc):",len(blocko))
    return blocko
  def dodif32(self, before, now):
    if now >= before: 
      dif= now-before
    else: 
      dif= now+ (0xffffffff-before) +1
    #if(DBGcnts) printf("dodif32:%d\n", dif)
    return dif
  def incsupdate(self, block):
    for ix in range(self.SIZE):
      cv= block[ix]
      self.incs[ix]= self.dodif32(self.lastrr[ix], cv)
      self.lastrr[ix]= cv
  def calcupdate(self, before,now,now_before):
    """ 
1. now_before= now-before
2. before= now
"""
    for ix in range(self.SIZE):
      cv= now[ix]
      now_before[ix]= self.dodif32(before[ix], cv)
      before[ix]= cv
  def readInc(self, grpcreply=None):
    """ op:
- blok= readRaw
- calculate increments:
    self.incs= block-self.lastrr
    self.inctime= corresponding diff. in time
- update: self.lastrr= block

rc: time difference between last readings (=self.inctime)
    self.incs -calculated increments

todo: 
1. ttccs should not be treated here, but in excmd, after readInc() call
2. self.lastrr: not updated in readRaw but it should be (together with
   self.lasttime to get the same time for both values):
   solution for now:
   readRaw time is kept in self.lasttime
   readInc uses its own time of last reading in self.lastinctime
"""
    prevtime= self.lastinctime
    if self.ttccs: ttcblock= self.ttccs.readRaw()
    block= self.readRaw()
    self.lastinctime= self.lasttime
    self.inctime= self.lasttime-prevtime
    if self.ttccs: self.ttccs.incsupdate(ttcblock)
    self.incsupdate(block)
    if grpcreply!=None:
      grpcreply.timestamp= self.inctime
      for ix in range(len(self.incs)):
        grpcreply.outcnts.append(self.incs[ix])
        #grpcreply.outcnts.append(ix*ix)   # dbg
      if self.ttccs:
        for ix in range(len(self.ttccs.incs)):
          grpcreply.outcnts.append(self.ttccs.incs[ix])
    return self.inctime
  #def getlast1(self, ix):
  #  return self.lastrr[ix] 
  def getinc1(self, ix, raw=None):
    if raw=="raw":
      return self.lastrr[ix]   # self.lastinctime
    else:
      return self.incs[ix]   # self.lastinctime/inctime
  #def getraw1(self, ix):
  # return self.lastrr[ix]   # self.lastinctime
  def SortByIndex(self, cname):
    return self.cnames[cname][1]
  def checktick(self):   # by: CountersCTP, CountersLTU
    timedif= self.readInc()
    time.sleep(1)
    timedif= self.readInc()
    ticks= self.getinc1(0)
    c240= self.getinc1(1)
    tick_freq= old_div(ticks,timedif)
    ticks_vs_240= abs(ticks*6 - c240)
    #print "abs(ticks*6:%d - clk240:%d)= %d <---max.5 expected."%(ticks*6, c240, ticks_vs_240)
    print("abs(ticks*6:%d - clk240:%d)= %d <---max.5 expected.%f hz"%\
      (ticks*6, c240, ticks_vs_240, tick_freq))
    if ticks_vs_240 < 6:
      return 0   # ok
    else:
      return 1   # bad tick
  def dumpdata(self, fn, blok):
    os.chdir(os.path.join(os.getenv('HOME'),'WORK'))
    #fp= os.path.join(TRGDBDIR, "ctpinputs_built.cfg")
    f= open(fn, "w")
    for ix in range(len(blok)):
      f.write("%x\n"%blok[ix])
    f.close()
    print("dumpdata: ~/WORK/%s written"%fn)
  def group_print(self, cgroup=None):
    if cgroup==None or len(cgroup)==0:
      self.cgroup= None   # deactivate
    else:  
      self.cgroup= cgroup   # activate todo: deep copy?
      #print "group_print active:", self.cgroup
  def prtall(self, names=False, cgroup=None, raw=None):
    """ names==True  print help text (counter names)
             ==False print counter values

self.group: ordered list of addresses to be printed.
When group_print is ON i.e. self.cgroup != None:
  prtall()      -print line with group counters values only
  prtall(True)  -print header line with group counters names only

raw: "raw" -print raw counters (self.lastrr)
     !=    -print increments (self.incs)
"""
    if names==True:
      if self.cgroup== None:
        print("%s, %d counters:"%(self.devname, self.SIZE))
      else:
        print("%s, %d (out of %d) counters:"%(self.devname, len(self.cgroup), self.SIZE))
    ix= 0; 
    if self.cgroup==None:   # help text
      line= ""
    else:
      if names==True:   # print help text or header
        line= "   "
      else:
        line= "   "
        if "outcnts"==self.ipsl[:7] and len(self.ipsl)>7:
          line= "%+2s:"%self.ipsl[7:]
    if self.cnames!=None:
      for cn in self.cnames_sorted:
        cnta= self.cnames[cn][1]    # address
        #print "cn:",cn, cnta
        if names==True:   # print help text or header
          lname= self.cnames[cn][0]   # longer name
          if self.cgroup==None:   # help text
            #if lname=="":   # (ttcit -all counters are "")
            #  line= line+ "%20.20s: %3d"%(cn, cnta)
            #else:
            #  #line= line+ "%16.16s: %3d,%25.25s "%(cn, cnta, lname)
            line= line+ "%20.20s: %3d,%-12s "%(cn, cnta, lname)
            perline= self.PERLINE_HLP
          else:   # 1 line header
            if cnta in self.cgroup:
              #line= line + "%+12s "%cn
              line= line + self.fmthead%cn
            perline=1000
        else:   # print values
          val= self.getinc1(cnta, raw)
          if self.cgroup==None:
            line= line+ self.fmtnvals%(cn, val)
            perline= self.PERLINE
          else:
            if cnta in self.cgroup:
              #line= line+ "%12d "%val
              line= line+ self.fmtval%val
            perline=1000
        ix= ix+1
        if ((ix%perline)==0) or (ix==len(self.cnames)): 
          if (names!=True) and (self.cgroup!=None):
            # add time
            if raw=="raw":
              secs= " raw"
            else:
              secs= " +%.2fs"%self.inctime
            line= line + secs
          print(line)
          line=""
    else:
      if names:
        print("what's it?",names)
      else:
        for cnta in range(self.SIZE):
          cn= str(cnta)
          val= self.getinc1(cnta)
          line= line+ "%3s:%10d    "%(cn, val)
          perline= self.PERLINE_RELPOS
          ix= ix+1
          if ((ix%perline)==0) or (ix==self.SIZE): 
            print(line)
            line=""
    if self.cgroup== None:
      if self.ttccs:
        print("\noldTTC counters:")
        self.ttccs.prtall(names, raw=raw)
  def excmd(self, cmd, stdprint="stdout"):
    """ stdprint: checked only for r rr r2 cmds
    """
    if len(cmd) < 2:
      self.prtHelp()
      return
    N= self.SIZE
    fromad= 0
    if cmd[1]=='rr' or cmd[1]=='r' or cmd[1]=='r2':
      allcs= True
      if len(cmd)>2: N= int(cmd[2]) ; allcs= False
      if len(cmd)>3: fromad= int(cmd[3]) ; allcs= False
      if cmd[1]=='rr':
        #cnts= self.readRaw(N, fromad)
        #if stdprint=="stdout":
        #  hexdump.hd_dump(cnts, 0, len(cnts), fromad, dumptyp='d')
        #else:
        #  return(hexdump.hd_dump(cnts, 0, N-1, fromad, dumptyp='d', stdprint="rc")+" raw values")
        timedif= self.readInc()
        self.prtall(raw="raw")
      elif cmd[1]=='r' or cmd[1]=='r2':
        if cmd[1]=='r2':
          timedif= self.readInc()
          time.sleep(1)
        timedif= self.readInc()
        #if (self.ipsl=="leco1") or (self.ipsl=="ttc") or (self.ipsl=="gbtX"):
        #if self.ipsl[:7]!="outcnts" and self.ipsl!="ttccnts":
        if self.ipsl[:3]=="gbt":
          tds= "%.2fs"%(timedif)
          if stdprint=="stdout":
            #print "Time delta (system): %.2fs"%(timedif)
            print("Time delta (system): %s"%(tds))
            hexdump.hd_dump(self.incs[fromad:], 0, N-1, fromad, dumptyp='d')
          else:
            return(hexdump.hd_dump(self.incs[fromad:], 0, N-1, fromad, dumptyp='d', stdprint="rc")+" time:"+tds)
        else:
          ms= old_div(self.getinc1(0)*3.564E6,40078500)
          print("Time delta (system): %.2fs, orbits:%.3fms (+-1 orbit)"%(timedif, ms))
          if allcs:
            self.prtall()
          else:
            hexdump.hd_dump(self.incs[fromad:], 0, N-1, fromad, dumptyp='d')
      else:
        self.prtHelp()
    elif cmd[1]=='rS' or cmd[1]=='rE':
      print(self.readRawxox(cmd[1][1]))
      #if self.ttccs:
      #  self.ttccs.readRawxox(cmd[1][1])
    elif "group".find(cmd[1])==0:
      if len(cmd)== 2:   # deactivate
        self.group_print()
      else:
        gl= []
        for gr in cmd[2:]:
          gl.append(int(gr))
        self.group_print(gl)
    else:
      if self.ipsl=="leco1":
        if cmd[1]=="clear":
          # this is not valid for fi>= v15_00
          print("1->ctrl.ctrl.bit2 + 0->ctrl.ctrl.bit2 ")
          ctrlctrl= self.board.Read("ctrl.ctrl")
          ctrlctrl= ctrlctrl | 0x4
          self.board.Write("ctrl.ctrl", ctrlctrl)
          ctrlctrl= ctrlctrl & 0xfffffffb
          self.board.Write("ctrl.ctrl", ctrlctrl)
        else:        
          self.prtHelp()
      #elif (self.ipsl[:8]=="outcnts") or (self.ipsl[:6]=="gcnts"):
      elif (self.ipsl=="cnts") or (self.ipsl=="gcnts"):
        if cmd[1]=="checktick":
          rc= self.checktick()
          print(rc,"  (0:ok, 1:bad tick, 2: cannot be checked (e.g. bad firmware))")
        elif cmd[1]=="clock":
          secs=10
          if len(cmd)>2: secs= int(cmd[2])
          self.clock(secs)
        else:
          self.prtHelp()
      else:
        self.prtHelp()
  def prtHelp(self):
    print(""" %s

r [N] [fromad]    -show difference from the last counters reading
                   N      -number of counters (def: all)
                   fromad -start from fromad counter (def: 0)
r2 [N] [fromad]   -the difference between 2 readings 1s apart,i.e. show rate
rr [N] [fromad]   -read + show raw counters values
group a1 a2 ...   -addresses of counters, activate group_print
group             -deactivate group_print
"""%self.mymsg)
    #if (self.board.isfiwa(self.board.FW_LTU) and self.board.isfiverhe(0x12)):
    if (self.ipsl=="cnts") or (self.ipsl=="gcnts"):
      print("""
checktick         -check tick (only for icnts or gcnts devices)
clock secs        -check clockfrekv (only LTU), wait secs between readings, 10 secs default.
""")
    if (self.ipsl=="cnts"):
      print("""
Note for PLSR, FASTLM: counting 3x less due to synchr. circuit
""")
    if self.ipsl!="leco1":   # obsolete
      self.prtall(True)
    elif self.ipsl!="oltrx_lec":
      print("""
clear             -clear all 64 link error counters
""")
    else:
      print("prtHelp: ", self.ipsl)
class CountersLECO(Counters):
  """
Link Error COunters (for olt1 only with fi above d_00)
each counter is 32 bits counter
for fi >= 15_01: 9*64 counters (16 bits each!) for all possible CRUs
"""
  mymsg= "Link error (upstream) output counters."
  def __init__(self, board, ipsl,devname=None):
    if ipsl=="leco1":
      self.SIZE= 64
      # rel_position_in_outcnts or bit number in TriggerType word
      self.cnames={   # "ShortName":("LongName", rel_position_in_outcnts)
      "le0":("link error 0",0),
      "le1":("link error 1",1)}
      self.cnames_sorted= sorted(self.cnames.keys())
      #self.cnames_sorted= sorted(self.cnames.items(), key=lambda kv: kv[1][0])
    else:   # oltrx_lec (for LTU fi >15)
      self.SIZE= 9*64
      self.cnames= None   # use rel. address instead of name
    self.ipslblock= ipsl+".data"
    Counters.__init__(self, board, ipsl,devname)
  def readRaw(self, N=None, fromad=0):   # CountersLECO
    if N==None: N= self.SIZE
    if fromad>=self.SIZE: fromad= 0
    if (fromad+N)>self.SIZE: N= self.SIZE-fromad
    print("countersLECO.readRaw: N:%d fromad:%d"%(N,fromad))
    block= self.board.ReadBlock(self.ipslblock,N, fromad)
    self.lasttime= time.time()
    #1.
    #blocko= block
    # ArgumentError: Python argument types in
    #     ValVector_uint32.__getitem__(ValVector_uint32, slice)
    # did not match C++ signature:
    #     __getitem__(uhal::ValVector<unsigned int>, int)
    # 
    # 2. anyhow, seems the block of all cnts have to be readout 
    # to get proper values
    blocko= []
    for ix in range(len(block)):
      blocko.append(block[ix])
    return blocko
  def prtall(self, names= False):
    if names:
      print("9*64 Link counters")
    else:
      fromad= 0 ; N= self.SIZE
      hexdump.hd_dump(self.incs[fromad:], 0, N-1, fromad, dumptyp='d')
class CountersCTPEMU(Counters):
  """
ipbus_ctrs_ported in fi
"""
  SIZE= 34   # 32 counters  from fi 12.0: 2 countrs added: bc40 (tick) and clk240
  mymsg= "CTPEMU output counters."
  def __init__(self, board, ipsl,devname=None, ttccs=None):
    """ ipsl: outcnts (ltu)
outcntsN (N: 1..18) counters of ctpemuN
"""
    # rel_position_in_outcnts or bit number in TriggerType word
    self.cnames={   # "ShortName":("LongName", rel_position_in_outcnts)
    "ORB":("Orbit",0),
    "HB":("Heart Beat",1),
    "HBr":("HB Reject", 2),
    "HC":("Health Check",3),
    "GAP1":("LHCgap1",27),
    "GAP2":("LHCgap2",28),
    "TPC_sync":("TPC sync",29),
    "TPC_rst":("TPC rst",30),
    "TOF":("TOF rst",31),
    "PH":("Physics",4),
    "PP":("PrePulse",5),
    "CAL":("Calibration",6),
    "SOT":("SOT", 7),
    "EOT":("EOT", 8),
    "SOC":("SOC", 9),
    "EOC":("SOC",10),
    "TF": ("Time Frame",11),
    "FERST": ("FE reset",12),
    "RT": ("Run Type (1:cont)",13),
    "RS": ("R.Stat(1:ON)",14),
    "bc40": ("BC (tick)",32),
    "clk240": ("240Mhz clock",33)}
    self.reportnames=("PH","RS","HB","HBr")
    if board.isfiwa(board.FW_LTU) and board.isfiverhe(0x1b):
      self.SIZE= 32   # bc40+clk240 moved to cnts.cnts
      del(self.cnames["bc40"])
      del(self.cnames["clk240"])
    elif board.isfiwa(board.FW_LTU) and board.isfiverhe(0x12):
      self.SIZE= 34
    else:
      self.SIZE= 32   # for ctp
      #print self.cnames
      del(self.cnames["bc40"])
      del(self.cnames["clk240"])
    #if board.dbgmode>0:
    #self.dbgblock= [0]*self.SIZE
    #self.cnames_sorted= sorted(self.cnames.keys())
    self.cnames_sorted= list(self.cnames.keys())
    self.cnames_sorted.sort(key=self.SortByIndex)
    #print "CountersCTPEMU selfSIZE:",self.SIZE, "dbgmode:", board.dbgmode
    self.ipslblock= ipsl+".TTcnts"
    Counters.__init__(self, board, ipsl,devname, ttccs)
  def SortByIndex(self, cname):
    return self.cnames[cname][1]
  def getnames(self, grpcreply):
    size1= self.SIZE ; size2= 0
    if self.ttccs:
      size2= self.ttccs.SIZE
    sizeall= size1+size2
    cn= [""]*sizeall
    for name in self.cnames:
      ix= self.cnames[name][1]
      cn[ix]= name
    if self.ttccs:
      for name in self.ttccs.cnames:
        ix= self.ttccs.cnames[name][1] + size1
        cn[ix]= name
    for ix in range(sizeall):
      grpcreply.names.append(cn[ix])
  #def readRaw(self, N=None, fromad=0, grpcreply=None):
  #  if N==None: N= self.SIZE
  #  if fromad>=self.SIZE: fromad= 0
  #  if (fromad+N)>self.SIZE: N= self.SIZE-fromad
  #  #print "readRaw: %d counters from:%d"%(N, fromad)
  #  if self.board.dbgmode>0:
  #    for i in range(len(self.dbgblock)):
  #      self.dbgblock[i]= self.dbgblock[i] + 1
  #    block= self.dbgblock
  #  else:
  #    block= self.board.ReadBlock(self.ipsl+".TTcnts",N, fromad)
  #  self.lasttime= time.time()
  #  #self.lasttime= 1604164415.17
  #  #print "readRaw:%.2f %d counters from:%d"%(self.lasttime,N, fromad)
  #  blocko= []
  #  if grpcreply!=None:
  #    #print "cblock:",block
  #    grpcreply.timestamp= self.lasttime
  #    #print "readRaw ts:",type(grpcreply.timestamp),grpcreply.timestamp, self.lasttime
  #    grpcreply.ClearField("outcnts")   # set default value
  #    #del grpcreply.outcnts[:]    # should be ok too
  #    for ix in range(len(block)):
  #      blocko.append(block[ix]); grpcreply.outcnts.append(block[ix])
  #      #blocko.append(ix) ; grpcreply.outcnts.append(ix)   # dbg
  #  else:
  #    for ix in range(len(block)):
  #      blocko.append(block[ix])
  #  return blocko
  def clock(self, tint=10):
    if self.SIZE<=32:
      return 2   # cannot be checked
    t1= time.time()
    timedif= self.readInc()
    time.sleep(tint)
    timedif= self.readInc()
    t2= time.time()
    ticks= self.getinc1(32)
    timedif= t2-t1
    tick_freq= old_div(ticks,timedif)
    print("ticks:%d time:%f = %f hz"%(ticks, timedif, tick_freq))
  def checktick(self):   # not valid for ltu >= v1b
    if self.SIZE<=32:
      return 2   # cannot be checked
    timedif= self.readInc()
    time.sleep(1)
    timedif= self.readInc()
    ticks= self.getinc1(32)
    c240= self.getinc1(33)
    tick_freq= old_div(ticks,timedif)
    ticks_vs_240= abs(ticks*6 - c240)
    print("abs(ticks*6:%d - clk240:%d)= %d <---max.5 expected.%f hz"%\
      (ticks*6, c240, ticks_vs_240, tick_freq))
    if ticks_vs_240 < 6:
      return 0   # ok
    else:
      return 1   # bad tick
class CountersCTP(Counters):
  """ CTP global counters
"""
  SIZE= 7   #3  4 added 21.4.2021
  mymsg= "CTP global counters."
  def __init__(self, board, ipsl="gcnts"):
    """ ipsl: only gcnts possible (fi 1.2)
"""
    self.cnames={   # "ShortName":("LongName", rel_position_in_outcnts)
      "bc40": ("BC (tick)",0),
      "clk240": ("240Mhz clock",1),
      "extorb": ("External orbit",2),   # after shortening 1us width to 25ns pulse
      "PLSRin": ("PLSRin",3),           # these are available from ctp_v9_00
      "FastLMin": ("FastLMin",4),
      "BUSYin": ("BUSYin",5),
      "SPAREin": ("SPAREin",6) }
    #self.cnames_sorted= sorted(self.cnames.keys())
    print("selfSIZE CountersCTP:",self.SIZE)
    self.cnames_sorted= list(self.cnames.keys())
    #self.cnames_sorted.sort(key=self.SortByIndex)
    self.ipslblock= ipsl+".data"
    Counters.__init__(self, board, ipsl)
    #def getnames(self, grpcreply):
    #  """ not used yet (as of fi 1.2) """
    #  cn= [""]*self.SIZE
    #  for name in self.cnames:
    #    ix= self.cnames[name][1]
    #    cn[ix]= name
    #  for ix in range(self.SIZE):
    #    grpcreply.names.append(cn[ix])
class CountersLTU(Counters):
  """ LTU input counters cnts.cnts
"""
  SIZE= 7
  mymsg= "LTU ph. input counters."
  def __init__(self, board, ipsl="cnts", devname="icnts"):
    """ ipsl: cnts.cnts (from ltu_v1b_00)
"""
    self.cnames={   # "ShortName":("LongName", rel_position_in_outcnts)
    "bc40": ("BC (tick)",0),
    "clk240": ("240Mhz clock",1),
    "extorb": ("External orbit",2),
    "PLSR": ("Pulser in/3",3),
    "FASTLM": ("Fast LM in/3",4),
    "BSYin":("Busy time (BCs), FP LVDS", 5 ),
    "SPARE": ("",6) }
    #self.cnames_sorted= sorted(self.cnames.keys())
    #print "selfSIZE:",self.SIZE
    self.cnames_sorted= list(self.cnames.keys())
    #self.cnames_sorted.sort(key=self.SortByIndex)
    self.ipslblock= ipsl+".cnts"
    Counters.__init__(self, board, ipsl, devname=devname)
class CountersTTC_13(Counters):
  """
TTC counters (+busy timers when ready). Valid till v13.xx See CountersTTC from v14.xx
"""
  SIZE= 7   # 7 counters
  mymsg= "TTC output counters (ltu fi <14.xx)."
  ttcclist= ["L0pre","L0","L1","Orbit","PP","L1r","CAL"]
  cnames={   # "ShortName":("LongName", rel_pos)
  # ttc.ShortName_count is the name of the counter reg in .xml
  # rel_pos: not used in fi 8_00
  "L0pre":("CTPEMU output",0),
  "L0":("L0/A bsy", 1),
  "L1":("L1/A",2),
  "Orbit":("Orbit",3),
  "PP":("Prepulse",4),
  "L1r":("L1 reject",5),
  "CAL":("TRD calib",6)}
  cnames_sorted= sorted(cnames.keys())
  def __init__(self, board, ipsl,devname=None):
    Counters.__init__(self, board, ipsl,devname)
  def readRaw(self, N=SIZE, fromad=0):   # CountersTTC_13, obsolete
    if fromad>=self.SIZE: fromad= 0
    if (fromad+N)>self.SIZE: N= self.SIZE-fromad
    #print "readRaw: %d counters from:%d"%(N, fromad)
    blocko= []
    for ix in range(fromad, fromad+N):
      regname= self.ttcclist[ix]+"_count"
      blocko.append(self.board.Read(self.ipsl+"."+regname))
    self.lasttime= time.time()
    return blocko
class CountersTTC(Counters):
  """
TTC counters ltu fi>=14
each counter is 32 bits counter
"""
  SIZE= 32
  PERLINE_HLP= 1
  PERLINE= 3
  mymsg= "TTC output counters."
  # rel_position_in_outcnts or bit number in TriggerType word
  cnames={   # "ShortName":("LongName", rel_position_in_outcnts)
"tPHIn":("PHpre, generated by emu",0),
"tOrbitIn":("ORBIT",2),
"tPPIn":("PPs received",3),
"tL1r":("L1 rejected triggers",4),
"tPH":("PH sent to detector",5),
"tSOT":("SOT received",6),
"tEOT":("EOT received",7),
"tL0tra":("L0 transmission start for PH/CAL/SOT/EOT",8),
"tL1tra":("L1 transmission start to TTC-A",9),
"tCALtra":("CAL transmission start to TTC-A",10),
"tL1FIwr":("# of words written to L1 FIFO, must be 7xPH",12),
"tDaFIwr":("# of words written to Data FIFO,",13),
"tL1FIrd":("# of words read from L1 FIFO",14),
"tDaFIrd":("# of words read from Data FIFO",15),
"tOrbitO":("trasmission of Orbit to TTC-B",16),
"tPPOut":   ("transmission of PP to TTC-B",17),
"tL1mTS":  ("transmission of L1 to TTC-B",18),
"tDataTS":   ("transmission of Data word to TTC-B",19),
"tTTCend": ("End of transmission to TTC-B",20),
"tTTCBSY":("Time (in BC) of active transmission in TTC-B",21),
"tL1Fmax":("Max word in FIFO",22),
"tL1Ffull":("L1_FIFO_FULL   -number of BCs when L1 FIFO full",23),
"tDaFfull":("DATA_FIFO_FULL -number of BCs",24),
"tBSYin":("Busy time (in BC) in Front panel LVDS or TTC-PON upstream", 26),
"t3exBUSY":("TTC3exBUSY -additional BSY when FIFOs are full",27),
"tBSYout":("BUSY_OUT   -DET busy or oldTTC internal BSY itransmitted out over FP or TTCPON downstream",28)}
  # "spare":("SPARE",1), 11 25 29 30 31
  #cnames_sorted= sorted(self.cnames.items(), key=lambda kv: kv[1][0])
  cnames_sorted= sorted(cnames.keys())
  def __init__(self, board, ipsl,devname=None):
    self.ipslblock= ipsl+".cnts"
    self.reportnames=("tPHIn","tPH","tBSYin","tBSYout")
    #self.dbgblock= [0]*self.SIZE
    Counters.__init__(self, board, ipsl,devname)
  #von def readRaw(self, N=SIZE, fromad=0):
  #  if fromad>=self.SIZE: fromad= 0
  #  if (fromad+N)>self.SIZE: N= self.SIZE-fromad
  #  #print "countersTTC.readRaw: N:%d fromad:%d"%(N,fromad)
  #  block= self.board.ReadBlock(self.ipslblock, N, fromad)
  #  self.lasttime= time.time()
  #  #1.
  #  #blocko= block
  #  # ArgumentError: Python argument types in
  #  #     ValVector_uint32.__getitem__(ValVector_uint32, slice)
  #  # did not match C++ signature:
  #  #     __getitem__(uhal::ValVector<unsigned int>, int)
  #  # 
  #  # 2. anyhow, seems the block of all cnts have to be readout 
  #  # to get proper values
  #  blocko= []
  #  for ix in range(len(block)):
  #    blocko.append(block[ix])
  #  return blocko
class CountersGBT(Counters):
  """
GBT counters
"""
  SIZE= 3   # Word_count Error_count(errors) Correct_count(corrections)
  ttcclist= ["Word","Error","Correct"]
  cnames={   # "ShortName":("LongName", rel_pos)
  # ShortName is the name of the counter reg in .xml
  # rel_pos: not used in fi 8_00
  "Word_count":("# of GBT packets received",0),
  "Error_count":("# of errors detected (incorrect data received)", 1),
  "Correct_count":("# of corrections",2)}
  PERLINE= 1
  cnames_sorted= sorted(cnames.keys())
  def __init__(self, board, ipsl,devname=None):
    Counters.__init__(self, board, ipsl,devname)
  def readRaw(self, N=SIZE, fromad=0):   # CountersGBT
    if fromad>=self.SIZE: fromad= 0
    if (fromad+N)>self.SIZE: N= self.SIZE-fromad
    #print "readRaw: %d counters from:%d"%(N, fromad)
    blocko= []
    for ix in range(fromad, fromad+N):
      regname= self.ttcclist[ix]+"_count"
      blocko.append(self.board.Read(self.ipsl+"."+regname))
    self.lasttime= time.time()
    return blocko
class CountersTTCIT(Counters):
  """
ipbus_ctrs_ported in fi
"""
  SIZE= 135
  mymsg= "TTCIT input counters."
  # rel_position_in_outcnts 
  cnames={   # "ShortName":("LongName", rel_position_in_cnts)
  "data_i.clk40_gated":("",0),
  "data_i.adr_d1..8":("",1),
  "data_i.adr_s1..8":("",9),
  "data_i.adr_e":("L1m words(L1*7)",17),
  "data_i.adr_a1..14":("",18),
  "data_i.adr_strobe":("",32),
  "data_i.brc_b":("",33),
  "data_i.brc_e":("",34),
  "data_i.brc_d1..4":("",35),
  "data_i.brc_t1":("",39),
  "data_i.brc_t2":("",40),
  "data_i.brc_strobe":("",41),
  "data_i.l1accept":("L0+2*L1",42),
  "status_i.div_nrst":("",43),
  "status_i.err_comm":("",44),
  "status_i.err_dbl":("",45),
  "status_i.err_sng":("",46),
  "status_i.ready":("",47),
  "HB0":("",48),
  "HBr_vector":("",49),
  "TF0":("",50),
  "PP":("",51),
  "ErrSingle":("",52),
  "ErrDouble":("",53),
  "ErrSingle2":("",54),
  "L0":("",55),
  "L1":("",56),
  "CAL":("",57),
  "ssm_flag":("",58),
  "TTorbit":("",59),
  "TThb":("",60),
  "TThbr":("",61),
  "TThealth":("",62),
  "TTph":("",63),
  "TTpp":("",64),
  "TTcal":("",65),
  "TTsot":("",66),
  "TTeot":("",67),
  "TTsoc":("",68),
  "TTeoc":("",69),
  "TTtf":("",70),
  "TTtpcsync":("",88),
  "TTtpcrst":("",89),
  "TTtof":("",90),
  "Orbit0..31":("",91),
  "BC_ID0..11":("",123) }
  #cnames_sorted= sorted(cnames.keys())
  def __init__(self, board, ipsl,devname=None, ttccs=None):
    """ ipsl: ttcitcnts (ttcit_logic_v4.xml)
"""
    self.ipslblock= ipsl+".TTCitcnts"
    Counters.__init__(self, board, ipsl,devname, ttccs)
    self.cnames_sorted= list(self.cnames.keys())
    self.cnames_sorted.sort(key=self.SortByIndex)
  #von def readRaw(self, N=SIZE, fromad=0):
  #  if fromad>=self.SIZE: fromad= 0
  #  if (fromad+N)>self.SIZE: N= self.SIZE-fromad
  #  #print "readRaw: %d counters from:%d"%(N, fromad)
  #  block= self.board.ReadBlock(self.ipslblock,N, fromad)
  #  self.lasttime= time.time()
  #  #1.
  #  #blocko= block
  #  # ArgumentError: Python argument types in
  #  #     ValVector_uint32.__getitem__(ValVector_uint32, slice)
  #  # did not match C++ signature:
  #  #     __getitem__(uhal::ValVector<unsigned int>, int)
  #  # 
  #  # 2. anyhow, seems the block of all cnts have to be readout 
  #  # to get proper values
  #  blocko= []
  #  for ix in range(len(block)):
  #    blocko.append(block[ix])
  #  return blocko
class CtpEmuCounters(object):
  """ self.ocnts: list of 18 CountersCTPEMU instances
"""
  def __init__(self, board):
    #if board.isfiwa((board.FW_CTP_LOGIC,)):
    self.board= board
    self.ocnts= []
    self.cgroup= [1,2,4,5,6]   # HB HBr PH PP CAL
    for did in range(18):
      #if dets & (1<<did):
      ocnname= "outcnts%d"%(did+1)
      cnt= self.board.atbdevs[ocnname]
      self.ocnts.append(cnt)
      cnt.group_print(self.cgroup)
  def group_print(self, cgroup=None):
    if cgroup==None or len(cgroup)==0:
      # should never happen
      self.cgroup= None   # deactivate
    else:  
      self.cgroup= cgroup   # activate todo: deep copy?
      print("group_print active:", self.cgroup)
    for did in range(len(self.ocnts)):
      self.ocnts[did].group_print(self.cgroup)
  def prt(self, dets=0x3ffff):
    nactive= 0
    for did in range(18):
      if dets & (1<<did):
        #cntname= "outcnts%d"%(did+1)
        #cnt= self.board.atbdevs[cntname]
        cnt= self.ocnts[did]
        if nactive==0:
          print("%3s"%"emu",cnt.prtall(True))
        nactive= nactive+1
        print("%2d:"%(did+1),cnt.prtall())
  def excmd(self, cmd):
    """ 
    """
    if len(cmd) < 2:
      self.prtHelp()
      return
    #N= self.SIZE
    #fromad= 0
    if "group".find(cmd[1])==0:
      if len(cmd)== 2:   # deactivate
        self.group_print()
      else:
        gl= []
        for gr in cmd[2:]:
          gl.append(int(gr))
        self.group_print(gl)
    elif cmd[1]=='rr' or cmd[1]=='r' or cmd[1]=='r2':
      if len(cmd)>2 and "all".find(cmd[2])==0:
        dets=0x3ffff
      else:
        gcgr= self.board.Read("gcore.global_run")
        dets= (gcgr>>1) & 0x3ffff
      nactive= 0
      for did in range(18):
        if dets & (1<<did):
          cnt= self.ocnts[did]
          if nactive==0:
            cnt.prtall(names=True)
          nactive= nactive+1
          #print "%2d:"%(did+1),cnt.prtall()
          cnt.excmd(("outcntsX", cmd[1]))
    else:
      self.prtHelp()
  def prtHelp(self):
    print(""" EmuX (outcnsX) for X: 1..18
group a1 a2 ...   -addresses of counters, activate group_print mode
group             -deactivate group_print
r  [all]          -show difference from the last counters reading
r2 [all]          -the difference between 2 readings 1s apart,i.e. show rates
rr [all]          -read + show raw counters values
Active group: """, self.cgroup,":")
    self.ocnts[0].prtall(names=True)
