#!/usr/libexec/platform-python
from __future__ import print_function
from __future__ import absolute_import
from builtins import range
from builtins import object
import os, string, time
from . import bcmask,partdef
from auxfuns import printError
class Ctp(object):
  def __init__(self, board):
    #global LTUS
    self.devname= "ctp"
    self.board= board
    self.runn= 0   # if not given by ECS, use self.getCtpRunn()
    #LTUS= Ltus()
    self.parts= partdef.Partitions()
  def initCtp(self):
    printError("clearing gcore.global_run[0..18]")
    gcgr= self.board.Read("gcore.global_run")
    gcgr= gcgr& 0xfff80000
    self.board.Write("gcore.global_run", gcgr)
  def getCtpRunn(self):
    self.runn= self.runn+1
    return self.runn
  def partShow(self, runn=None):
    if runn==None:
      self.parts.listParts()
    else:
      if runn in self.parts.parts:
        part= self.parts.parts[runn]
        part.printPart()
      else:
        printError("Run %d not loaded"%runn)
  def partLoad(self, pname, runn=None, detectors=""):
    """ todo:
1. take into account active+offline ltus: done in ctpd_server
2. check (load if not available) bcms
3. global generators
4. check/load classes

rc: Partition instance
    rc.status: rc.LOADEDHW -ok loaded in hw
               rc.errs: >0 if error(s) encountered
               rc.allerrs: "" if no errors
"""
    if runn==None: runn= self.getCtpRunn()  # not given by ECS, assign the own one
    # von
    #part= self.parts.getPart(pname)  -not applicable (runn needed)
    #if part != None:
    #  if part.status==part.RUNNING:
    #    printError("Run {} already active, {} cannot be loaded".format(runn,pname))
    #    return None
    #  else:
    #    print("Partition %s already loaded, replacing..."%pname)
    FillNotActive= True   # todo: check if fill is active
    if (self.parts.getN()==0) and (FillNotActive):
      # reset to 0 if no other run loaded/active
      self.board.atbdevs["fpga"].orbitset(None,3564)
      # todo:do the orbit clear also when fill starts
    part= self.parts.addPart(pname, runn, detectors)
    if part.status!=part.LOADED:
      printError("Partition {} (run:{}) not loaded.".format(pname,runn))
      return part
    if part.errs>0:
      print(part.allerrs)
      #return part
    # part. definition loaded in part (part.status==LOADED). Load hw now:
    #dets= part.getDets()
    # check globals for all dets:
    bcm= self.board.atbdevs["bcm"]   # available masks
    for mskname in part.masks:
      newmsk= part.masks[mskname]
      if not mskname in bcmask.symname:
        part.regError("Unknown mask :"+mskname)
        return part
      mskid= bcmask.symname.index(mskname)
      # todo: check if mskid already loaded, if yes check if the same
      bcm.pmsk[mskid].update(newmsk.maskdef)   # default: also hw updated
    #
    # load all LTGs:
    for detname in part.ltgs:
      ltg= part.ltgs[detname]
      print("LTG %s cmds:"%detname, ltg.loadcmds)
      emuname= "emu%d"%partdef.findEmuN(detname)
      emu= self.board.atbdevs[emuname]
      for ixc in range(len(ltg.loadcmds)):
        #print "  ",ltg.loadcmds[ixc]      
        emu.excmd(ltg.loadcmds[ixc])      
    part.setStatus(part.LOADEDHW)
    return part
  def syncTF(self, newdets, force8=False):
    #force8= True
    fwi= (self.board.fwinfo>>8)&0xffff
    if (fwi>=0x9) and (not force8):
      print("syncTF using ctrl.rst bit 7 (fwi:0x%x)"%fwi)
      emus_bit= 7    # see fpga.Sysmon.ctprst dict.
      self.board.Set10("ctrl.rst", emus_bit)
    else:
      print("syncTF for fi ctp_v<9_0x(sw patch) or forced (fi:0x%x)"%fwi)
      wargs= []
      for did in range(18):
        if newdets & (1<<did):
          nodname= "ctpemu%d.ctrl"%(did+1)
          nodv= self.board.Read(nodname) | 0x1
          wargs.append([nodname,nodv])   
      for ix in range(len(wargs)):
        self.board.Write_nd(wargs[ix][0], wargs[ix][1])
      self.board.mydispatch()
      for ix in range(len(wargs)):
        self.board.Write_nd(wargs[ix][0], wargs[ix][1]&0xfffffffe)
      self.board.mydispatch()
  def reset_global_run(self, newdets):
    """ reset (i.e. set to 0) for newdets, i.e. with 1 in their bit
Check newdets (if active) before clearing bits in global_run[18..1]
global_run[0]: set to 1 if any det. left active (0 if no act. dets)
"""
    gcgr= self.board.Read("gcore.global_run")
    dets= (gcgr>>1) & 0x3ffff
    # emu.stop for all choosen newdets
    for did in range(18):
      if newdets & (1<<did):
        if (dets & (1<<did)) == 0:   # OFF?
          emuname= "emu%d"%(did+1)
          printError("%s OFF, dets:0x%x hw:0x%x"%(emuname,newdets,dets))
          return False 
    invdets= ~newdets & 0x3ffff
    alldets= dets & invdets
    if alldets==0:   # clear completely if not active dets left
      new18= ((invdets<<1) | 0) & gcgr
      print("gcore.global_run[18:1] set to 0x%x, [0] set to 0."%alldets)
    else:
      new18= ((invdets<<1) | 1) & gcgr
      print("gcore.global_run[18:1] set to 0x%x, [0] set to 1."%alldets)
    self.board.Write("gcore.global_run", new18)
    time.sleep(0.05)   # the same sleep in ctpemu.stop (stdalone runs stop)
    # emu.stop for all choosen newdets
    #for did in range(18):
    #  if newdets & (1<<did):
    #    emuname= "emu%d"%(did+1)
    #    emu= self.board.atbdevs[emuname]
    #    emu.stop()
    return True
  def set_global_run(self, newdets, setrunbit=False,
    synctf= False, tdg=False, stopms=0):
    """ 
newdets: 0xffffffff -set only global_run[0] bit
         Check newdets before setting bits in global_run[18..1]
         if det. already in, run won't be started
setrunbit: by default, global_run[0] bit not set (i.e. run not started 
           if it is 0 before)
tdg:       start tdg recording before start
stopms: stop partition after stopms ms (0: do not stop partition)
rc: True  -ok, started (and stopped if requested)
    False -error
"""
    if newdets!=0xffffffff:
      if synctf: self.syncTF(newdets) 
      gcgr= self.board.Read("gcore.global_run")
      dets= (gcgr>>1) & 0x3ffff
      # emu.start for all choosen newdets
      for did in range(18):
        if newdets & (1<<did):
          emuname= "emu%d"%(did+1)
          if dets & (1<<did):   # already ON?
            printError("%s already ON, global run not started."%emuname)
            return False 
          #if dets & (1<<did):
          #  print "already in:", emuname
          emu= self.board.atbdevs[emuname]
          # todo: check if running already?
          emu.start()
      #new18= (newdets<<1) | (gcgr&(~(0x3ffff<<1)))
      new18= (newdets<<1) | (gcgr&(0x3ffff<<1))   # add dets already in
      if setrunbit:
        new18= new18 | 1
        print("gcore.global_run[1] set to 1. [18:1] set to 0x%x."%newdets)
      else:
        print("gcore.global_run[0] not touched. [18:1] set to 0x%x."%newdets)
      if tdg:
        trc= self.board.atbdevs["tdg"].start_recording()
      self.board.Write("gcore.global_run", new18)
      if stopms>0:
        time.sleep(stopms/1000.)
        self.reset_global_run(newdets)
      if tdg:
        if stopms>0:
          time.sleep(0.023)   # to get an EOX
        else:
          time.sleep(1)       # if tdg started without stopms take 1s
        print("TDG recording started %.6fs"%trc)
        self.board.atbdevs["tdg"].stop_recording()
    else:
      if tdg:
        trc= self.board.atbdevs["tdg"].start_recording()
      self.board.SetBit("gcore.global_run", 0)
      print("1->gcore.global_run[0].")
      if stopms>0:
        time.sleep(stopms/1000.)
        #self.reset_global_run(newdets)
      self.board.ResetBit("gcore.global_run", 0)
      if tdg:
        if stopms>0:
          time.sleep(0.023)   # to get an EOX
        else:
          time.sleep(1)       # if tdg started without stopms take 1s
        print("TDG recording started %.6fs"%trc)
        self.board.atbdevs["tdg"].stop_recording()
    return True
  def partStop(self, runn):
    if runn in self.parts.parts:
      part= self.parts.parts[runn]
    else:
      return("Run %d cannot be stopped (not loaded)"%runn)
    if part.status!=part.RUNNING:
      printError("Run %d not RUNNING, stopping anyhow..."%runn)
      part.printPart()
    allclus= part.getDets18()
    self.reset_global_run(allclus)
    part.setStatus(part.STOPPED)
    self.parts.delPart(runn)
    return ""
  def partStart(self,runn, synctf= True, tdg=False, stopms=0):
    print("partStart:",list(self.parts.parts.keys()))
    if runn in self.parts.parts:
      part= self.parts.parts[runn]
    else:
      return "Run %d cannot be started (not loaded)"%runn
    allclus= part.getDets18()
    rc= self.set_global_run(allclus, setrunbit=True,
      synctf=synctf, tdg=tdg, stopms=stopms)
    if rc and (stopms==0):
      part.status= part.RUNNING
      return ""
    else:
      return "global run not started. set_global_run rc:%d"%rc
  def excmd(self, cmd):
    if len(cmd) < 2:
      self.prtHelp()
      return
    elif 'synctf'==cmd[1]:
      forceopt= False
      if len(cmd)>3:
        forceopt= True
      self.syncTF(eval(cmd[2]), forceopt)
    elif 'show'==cmd[1]:
      runn=None
      if len(cmd)>2:
        runn= int(cmd[2])
      self.partShow(runn)
    elif 'load'==cmd[1]:
      # ctp load N pname [detectors]
      if len(cmd)<4:
        printError("expected: ctp load N pname")
      else:
        part= self.partLoad(cmd[3], int(cmd[2]), "")
        part.printPart()
    elif 'start'==cmd[1]:
      synctf= False ; tdg= False ; stopms=0 ; optserr= False
      if len(cmd) > 2:
        # ctp start PNAME tdg synctf 500
        pstp= "Run won't be stopped"
        for ix in range(len(cmd)-3):
          opt= cmd[ix+3]   # emu start f3 opts
          if opt=="synctf": synctf= True
          elif opt=="tdg": tdg= True
          else:
            try:
              stopms= int(opt)
            except ValueError as e:
              print("Bad stopms option:%s. Expected: 0..1600). Using %dms"%\
                (opt,stopms))
              optserr= True
            if stopms==0:
              pstp= "Run won't be stopped"
            else:
              pstp= "Run will be stopped after %d ms"%stopms
        if not optserr:
          print("Starting %s with opts:"%cmd[2],"synctf:",synctf, "tdg:",tdg,'.',pstp)
          rc= self.partStart(int(cmd[2]), synctf, tdg, stopms)
          print(rc)
    elif 'stop'==cmd[1]:
      rc= self.partStop(int(cmd[2]))
      print(rc)
    elif 'startrun'.find(cmd[1])==0:
      #todo: use set_global_run instead...
      gcgr= self.board.Read("gcore.global_run")
      dets= (gcgr>>1) & 0x3ffff
      #print "TFsync for ctp_v<9_0x..."
      #ec1= self.board.Read("ctpemu1.ctrl") | 0x1
      #ec2= self.board.Read("ctpemu2.ctrl") | 0x1
      #self.board.Write_nd("ctpemu1.ctrl", ec1)
      #self.board.Write_nd("ctpemu2.ctrl", ec2)
      #self.board.mydispatch()
      #ec1= ec1&0xfffffffe
      #ec2= ec2&0xfffffffe
      #self.board.Write_nd("ctpemu1.ctrl", ec1)
      #self.board.Write_nd("ctpemu2.ctrl", ec2)
      #self.board.mydispatch()
      if len(cmd)>2:
        newdets= eval(cmd[2])
        dets= (self.board.Read("gcore.global_run")>>1) & 0x3ffff
        #self.syncTF(newdets)
        # emu.start for all choosen newdets
        for did in range(18):
          if newdets & (1<<did):
            emuname= "emu%d"%(did+1)
            if dets & (1<<did):
              print("already in:", emuname)
            emu= self.board.atbdevs[emuname]
            # todo: check if running already?
            emu.start()   # emu.rnc() + set ctpemuX.TTena
        self.board.Write("gcore.global_run", (newdets<<1) | (gcgr&0x1))
        print("gcore.global_run[0] not touched. [18:1] set to 0x%x."%newdets)
      else:
        self.board.SetBit("gcore.global_run", 0)
        print("1->gcore.global_run[0].")
    elif 'stoprun'.find(cmd[1])==0:
      # following test not needed (Ctp object exists ONLY with ctpboard)
      if self.board.isfiwa((self.board.FW_CTP_LOGIC,)):
        if len(cmd)>2:
          rmdets= eval(cmd[2])
          gcgr= self.board.Read("gcore.global_run")
          origdets=dets= (gcgr>>1) & 0x3ffff
          intersec= rmdets & dets
          if not intersec==rmdets:
            print("Cannot remove all, removing only 0x%x dets..."%intersec)
          for did in range(18):
            bit= 1<<did
            if intersec & bit:
              dets= dets & (~bit)
              emuname= "emu%d"%(did+1)
              emu= self.board.atbdevs[emuname]
              emu.stop()
          gcgrn= (gcgr&0xfff80001) | (dets<<1)
          self.board.Write("gcore.global_run", gcgrn)
          print("gcore.global_run[0] not touched. [18:1] set to 0x%x (was: 0x%x)."%\
            (dets, origdets))
        else:
          self.board.ResetBit("gcore.global_run", 0)
          print("Only 0->gcore.global_run[0].")
      else:
        print("Not CTP, no action.")
    elif 'run'.find(cmd[1])==0:
      if self.board.isfiwa((self.board.FW_CTP_LOGIC,)):
        gcgr= self.board.Read("gcore.global_run")
        dets= (gcgr>>1) & 0x3ffff
        gr= gcgr & 0x1
        print("enabled dets: 0x%x globalrun:%d"%(dets, gr))
        nactive=0
        for did in range(18):
          if dets & (1<<did):
            emuname= "emu%d"%(did+1)
            emu= self.board.atbdevs[emuname]
            if nactive==0:
              print("%5s"%"emu",emu.bstatus(fmt='h'))
            nactive= nactive+1
            print("%5s"%emuname,emu.bstatus())
    #elif 'outcnts'.find(cmd[1])==0:
    elif 'irsel'.find(cmd[1])==0:
      if self.board.isfiwa((self.board.FW_CTP_LOGIC,)):
        if len(cmd) > 2:
          if 'normal'.find(cmd[2])==0:
            self.board.ResetBit("ctrl.ctrl", 3)
          elif 'counter'.find(cmd[2])==0:
            self.board.SetBit("ctrl.ctrl", 3)
          else:
            pass
        cc= self.board.Read("ctrl.ctrl")
        print("ctrl.ctrl: %x & 0x8 = %x. 0: IR: normal(i.e. prbs) 0x8: counter"%(cc, cc&0x8))
    else:
      self.prtHelp()
  def prtHelp(self):
    print("""
init            -initialise ctp (startrun 0x0 + stoprun + ...)
startrun        -set gcore.global_run[0] bit (the only action)
startrun 0x..   -0x..: hex. pattern of enabled/disabled detectors (18 bits)
                 e.g.: 0x3: enable emu1+2 AND disable emu3-18
                 Note: for all given detectors 'emuX start' invoked
                       (run preparation)
stoprun         -clear gcore.global_run[0] bit
stoprun 0x..    -clear only given bits in gcore.global_run[18:1]
                 e.g.: 0x3: disable emu1+2 in case it was enabled,
                       emu3-18 not touched
run             -show status. In the list of emulators rates of signals
                 are given in hz/khz:
                 0: repetition is 0 (i.e. no signal generated)
                 -: trigger signal not choosen in TTena
                 HBacc: the ratio of accepted TF in %

irsel normal|counter   - set data pattern in IR stream (see also 'fpga reset')
                         normal: prbs pattern now (for debugging)
                         counter: counter pattern
synctf detpat    manual sync of TF for detpat detectors

Partitions:
show [N]
load  N parname    -load (always needed before start) CTP3_DBCTP/pd/parname.pd
start N [opts]     -opts: synctf, tdg , stopms(number)
                     defaults: False   False 0
                     synctf: needed once for ctp_v8_0a
                     tdg:    record tdg (0.023ms if stopms==0)
                     stopms: 0: do not stop partition
                            >0: stop after 'stopms' ms
stop  N            -stop+unload, 'parname' had to be started before
""")


