#!/usr/libexec/platform-python
#!/usr/bin/python3
from __future__ import division
from __future__ import print_function
from builtins import map
from builtins import str
from builtins import range
from builtins import object
from past.utils import old_div
import time
import colors, symbreg, auxfuns
import atb.bcmask as bcmask

xf8= 0xffffffff   # ~m&0xf8 or use ctypes.c_uint32(~m).value
emubatchcmds= {"trig":None, 
"cont":None, 
"nosox":None, 
"rnc":None, 
"pulserin":None,    # options pulser fastlm
"pulserdet":None,   # options strobe edge
"ferst":None,       # 0..15, off
"hbaccept":None,    #find
"hbacc":None,
"bcm":None,
"cal":None,
"ttc":None,
}   # + CTPemu.signames
def check_rn(cmd):
  """
rate,nn,immediate,Pp= check_rn(cmd)
rate: None -no change requested (apart from Pp) or error in rate argument
"""
  signam= cmd[1]
  rate= None; nn= "inf"; immediate= False ; Pp= False
  if len(cmd)<3:
    return rate,nn,immediate,Pp
  # 'emu PH_bc 1hz ... ', or  'emu PH_bc P'
  if (cmd[2]=='P') or (cmd[2]=='p'): # emu PH_bc P just connect to pulser
    Pp= True
    return rate,nn,immediate,Pp
  # emu PH_bc 1hz ...
  if len(signam)>3 and signam[-3:]=="_bc":
    rate=auxfuns.fromms(cmd[2])
  elif signam=="TPC_SYNC_orbit":
    rate=auxfuns.fromms(cmd[2], old_div(auxfuns.BCFREQ,auxfuns.ORBITL))
  elif signam=="PH_rnd":
    rate=auxfuns.frommsRandom(cmd[2])
  else:
    try:
      if (len(cmd[2])>2) and (cmd[2][:2]=='0x'):  # direct hexa
        intrep=eval(cmd[2])
        rate= str(intrep)
      else:     # n given directly as int
        intrep= int(cmd[2])
        rate= cmd[2]
    except:
      rate=None
  #print "excmdBatch rate:", rate
  if rate==None:   # bad rate argument
    return rate,nn,immediate,Pp
  if len(cmd)>=4: nn=cmd[3]
  if (len(cmd)>=5) and (cmd[4]=='i'): immediate= True
  return rate,nn,immediate,Pp
def check_emubatchcmd(cmd):
  emsg= ""   # ok
  #print("check_emubatchcmd:",cmd)
  if len(cmd)<2: 
    emsg= "parameter(s) missing in line:"
  elif cmd[1] in CTPemu.signames:   # emu PH_bc rate
    if len(cmd)==2:
      emsg="rate parameter missing in emu line:"
    else:
      rate,nn,immediate,Pp= check_rn(cmd)
      if rate==None:
        emsg= "bad rate in 'emu SIGNAME rate ...' in line:"
  elif cmd[1] in emubatchcmds:
    pass   # todo: better check (esp. for emu ttc)
  else:
    emsg= "Unknown emu cmd in line:"
  if emsg!="":
    emsg= emsg+" "+" ".join(cmd)
  return emsg
class CTPemu(object):
  # NAME_rate:("..._n", TTenabit, outcnts_index)
  outcnts="""
  x -> counter index in signames[][2]
  .    not present in signames

  .  0 Orbit
  .  1 Heart Beat
  x  2 HB Reject
  x  3 Health Check
  x 29 TPC sync
  x 30 TPC rst
  x 31 TOF rst
  x  4 Physics
  x  5 PrePulse
  x  6 Calibration
  x  7 SOT
  x  8 EOT
  x  9 SOC
  x 10 EOC
  x 11 Time Frame

signames[signal]: (Nname,TTenabit,countaddr,PLSRbit) 
Nname: "" -ok exists and is defined in .xml with: 
             ctpemu.signal_n    -the limit on number of trigger signals
             ctpemu.signal_rate -the rate (BCs between triggers)
       None: does not exists
TTenabit: see above or .xml
countaddr: rel. address of the counter. None: counter does not exist
PLSRbit: None: cannot be connected to pulser
         N:    ctpemu.ctrl bit enabling pulser for this signal
following from v7:
"""
  #signames= {"HBr_orbit":("",2,2),"HBr_rnd":(None,2,None), 
  signames= {"HBr_orbit":("",2,2,None),
    "HC":("",3,3,None), 
    "PH_bc":("",4,4,8), "PH_rnd":("",4,None,8),   # + PP mask
    "PP_bc":("",5,5,9), "CAL_bc":("",6,6,10),     # + CAL mask
    # _tf: Only SOX/EOX are synced with TF signal (ltu_logic_v12)
    "SOT_tf":(None,7,7,None), "EOT_tf":(None,8,8,None),   # just TTena and counter positions
    "SOC_tf":(None,9,9,None), "EOC_tf":(None,10,10,None), # _n/_rate regs do not exist
    # _orbit: synced with orbit signal
    "TF_orbit":("",11, 11,None),   # TTena.0x800 bit valid from d02
    "TOF_bc":("",31,31,None),      # + TOF trigger can be driven also by TOF mask
    "TPC_SYNC_orbit":("",29,29,11),
    "TPC_RST_bc":("",30,30,12) }
  # CAL_sel bits 16..17 not valid from 15_01
  ctrlbits=("reset","EMUstarted","spare2","spare3","rst_out_cnts","?5","?6","FASTLM",
    "P_PH","P_PP","P_CAL", "P_TPC_SYNC","P_TPC_RST","P_edge","?14","?15","?16","?17",
    "PH_reject_ena","?19","?20","?21","?22","?23","?24","?25","?26","?27","?28","?29","?30","?31")
  ttenabits= ("TOF","TPC_RST","TPC_SYNC","","","","","",
    "", "", "", "", "", "", "", "",
    "", "", "", "FEErst", "", "EOC", "SOC", "EOT",
    "SOT", "CAL", "PP", "PH", "HC", "HBr", "HBr_HBM", "CALafterPP")
  DETMODETX= ("CRU", "TTC", "TRD", "CPV")   # 1 2 3 4 in .cfg/self.DETMODE
  calmodes= ["pp", "ppcal", "cal", "off"]
  expTTenabits= [2,3,4,0]   # pp ppcal cal noppcal
  ferbit=0x1000
  delchans= ("A","B","1","2")   # position in TTCLVDS_delay in ltuIP.cfg
  # bit position in ttc.masks ix*8 i.e. 21BA
  delbc2val= {"d":0, "0":0x10, "1":0x20, "2":1, "3":2, "4":4, "5":8}
  delval2bc= {0:"d", 0x10:"0", 0x20:"1", 1:"2", 2:"3", 4:"4", 8:"5"}
  def __init__(self, board, ltudevs, emuname):
    #if emuname=="ctpemu":
    #  self.tn= emuname    # top node
    self.board= board
    self.tn= emuname    # top node
    self.options={}
    oname= "pulserin"
    self.options[oname]= board.muxOption(emuname+".ctrl", (7,1),
      oname, pulser=0,fastlm=1)
    oname= "pulserdet"
    self.options[oname]= board.muxOption(emuname+".ctrl", (13,1),
      oname, strobe=0,edge=1)
    if emuname=="ctpemu":  # ltu logic  (ctpemu1,2, for ctp)
      self.version1501= self.board.isfiverhe(0x15)
      # emudefaults items: name:("ltuipcfgname",value)
      #   ltuipcfgname: "" -the same name used in ltuIP.cfg file
      # TTCLVDS_delay and DETMODE specially set in readDefaults
      self.emudefaults= {"enableBUSY":("enableBUSY", 0),
        "sourceBUSY":("",0),
        "XOX_ttc":("",0),
        "LLprotection":("",280),
        "PP_BC_mask":("",""),        # "" should be 3564L
        "CAL_BC_mask":("",""),       # should be 3564L
        "PPCALmode":("",3),          # cal. off
        "PPCALfrekv":("",0xdeadbeaf),   # 0: means off
        "ttcdelays":("", [ "1", "2", "3", "4"]),  # [A B 1 2]
        "DETMODE":("", 1),           # new TTCPON dets by default
        "TForbits":("",256),
        "CAL_protection_time":("",63)}   # was 62 before ltu_v1d_0c
    else:   # $DBCTP/ctp.cfg
      self.emudefaults= {"TForbits":("", 256),
        "DETMODE":("", 1)}
      self.version1501= True
    self.emuname= emuname[3:]   # "emu,emu1,2,...
    #print "ctpemu init version>=1501:", self.version1501, emuname
    self.tnttc= None    # top node for old ttc not used
    self.ctrln= self.tn+".ctrl"
    self.ltudevs= ltudevs   # ltudevs["tdg"], ltudevs["outcnts"],...
    # Defaults available in ltuIP.cfg:
    self.ppcalprotl= None   # None: no PP_CAL_Protection line in ltuIP.cfg
    self.ttena= self.board.Read(self.tn+".TTena")
    #delme:
    #self.TForbitsActual= self.board.Read(self.tn+".TF_orbit_rate")
    #if (self.TForbitsActual != 256) and (self.TForbitsActual != 128):
    #  auxfuns.printError("TF length in orbits not 256 neither 128: %d"%\
    #    self.TForbitsActual)
    self.readDefaults()
    self.checksetTF()
    if self.TForbitsActual==0:
      auxfuns.printError("TForbitsActual:0")
    #
    # Other defaults (not available in ltuIP.cfg):
    self.L1reject=  0  # 0: no reject
    self.CALrate_trd= 0   #numb. of PH triggers between cal. triggers
    #
    #if (not self.emuactive()) and\
    #  (self.board.pllwasloaded or auxfuns.isinargs("emuinit")):
    #  print colors.color("Emulation not active, loading emu+BCmasks defaults from ltuIP.cfg file...","fg_red")
    if self.board.nodExists("ttc", topnode=True):
      self.tnttc= "ttc"   # top node for old ttc
      self.ttcctrl= self.board.Read(self.tnttc+".ctrl")
      # following moved to emuinit:
      #if  (self.board.pllwasloaded or auxfuns.isinargs("emuinit")):
      #  self.ttc_setdefaultsHW()
      #
      self.ttcstatusbits= symbreg.Register("ttc.statusbits", L1Fempty=0, L1Ffull=1,
        DATAFempty=2, DATAFfull=3, L1Renabled=4, TRDCALenabled=5)
      self.ttcerrors= symbreg.Register("ttc.errorbits", OrbitEarly=0, OrbitLate=1,
        OrbitInhibit=2, PPInhibit=3, TransBusyOverflow=4, OrbitPPOverlay=5,
        BMsgCorrupted=6, WrongMode=7, WrongAMask=8, WrongBMask=9,
        WrongLvds1Mask=10, WrongLvds2Mask=11, LLProtection=12,TTCAtick=13,
        TTCBtick=14,TTC1tick=15,TTC2tick=16)
      self.ttcctrlsym= symbreg.Register("ttc.ctrl",Reset=0, ResetSticky=1,
        BusyEnabled=4, BusySRC_PON=5, DisableXOX=8, Mode=(12,4))
      #  BusyEnabled=4, BusySRC_PON=5, Global=8, Mode=(12,4))  Global thrown out from ltu_v1c_05 
      self.Sttcdelays= symbreg.Register("ttc.masks", TTC_A=(0,6), TTC_B=(8,6),
        LVDS1=(16,6), LVDS2=(24,6))
    self.sox= self.getsox(self.ttena)
    #if (self.board.pllwasloaded or auxfuns.isinargs("emuinit")): self.emuinit()
    # moved to atbwrap.py fpgainit
  def emuinit(self):
    print(colors.color("loading emu+BCmasks defaults from {ltuIP|ctp}.cfg file...","fg_red"))
    if self.emuactive():
      self.stop(True)
    self.checksetTFHBr() # also set TF HBr in TTena
    self.sox= "SOCEOC"   # SOTEOT SOCEOC NOSOX
    self.setsoxbits(self.sox)
    self.clearemu()
    #
    bcmsdev= self.board.atbdevs["bcm"]   # all bcmasks
    bcmsdev.allSetDefaults()   # all bcm
    #
    # todo: set in hw the ltuIP.cfg definitions
    if self.ppcalprotl!=None:   # PP_CAL_Protection line in ltuIP.cfg, set hw:
      # PP_CAL_Protection PP_BC CAL_BC ProtectionTime Frequency
      # rewriting: PP+CAL masks, CAL_protection_time (PP_del_oldTTC 62 less),
      #            PP_CAL_rate (because CALIB mode is PPCAL mode)
      #
      mskid= bcmask.symname.index("CAL")
      bcmsdev.pmsk[mskid].update(self.CAL_BC_mask, alsohw=False)   # -> bcm -> hardware
      mskid= bcmask.symname.index("PP")
      bcmsdev.pmsk[mskid].update(self.PP_BC_mask, alsohw=False)
      bcmsdev.updatehw()
      self.cal_setmode(self.PPCALmode) 
      self.cal_phreject(1)
      self.cal_rate(self.PPCALfrekv, "inf")   
      self.setCAL_protection_time(self.CAL_protection_time)
    nbits= 0
    for key in list(self.signames.keys()):
      rname,nname,enabit,pulser= self.getnames(key)
      if enabit==None or nname==None: continue
      if self.signames[key][0]!=None:
        n= self.board.Read(nname)
        if n>0: nbits= nbits | (1<< enabit)
    if (self.ttena & 0xfffff07b) != (nbits & 0xfffff07b):  # do not check TF XOX HBr
      #print colors.color("CTPemu: TTena:0x%x does not correspond to choosen triggers:0x%x (N>0)"%\
      #  (self.ttena&0xfffff07b,nbits&0xfffff07b), "fg_red")
      print("CTPemu: TTena:0x%x does not correspond to choosen triggers:0x%x (N>0)"%\
        (self.ttena&0xfffff07b,nbits&0xfffff07b))
      print("TTena:",self.prtena(self.ttena),",N>0 for:",self.prtena(nbits))
      #self.ttena= self.ttena | nbits
    if self.board.nodExists("ttc", topnode=True):
      self.ttc_setdefaultsHW()
  def readDefaults(self):
    """ operation: do not touch hw, only:
set self.XXX default values used later in allSetDefaults
"""
    # 1. set defaults (i.e. parameters NOT PRESENT in ltuIP|ctp.cfg file):
    for dk in self.emudefaults:
      varname= self.emudefaults[dk][0]
      if varname=="":
        varname= dk
      vars(self)[dk]= self.emudefaults[dk][1]
    #self.ttcdelays= [ "1", "2", "3", "4"]   # [A B 1 2]
    #self.DETMODE= 1   # new TTCPON dets by default
    if self.tn=="ctpemu":
      fp= self.board.openltucfg()
      cfgname= "CFG/ltuIP.cfg"
    else:
      fp= self.board.openctpcfg()
      cfgname= "DBCTP/ctp.cfg"
    # 2. read values from .cfg file
    if fp==None: return
    print("readDefaults from %s."%cfgname)
    for line in fp.readlines():
      ls= line.split()
      if len(ls)==0: continue
      if line[0]=='#': continue
      parn= ls[0]
      #print parn ,"=", ls[1:]
      if parn =="DETMODE":
        detmode= ls[1]
        if (detmode<"1") or (detmode>"4"):
        # DETMODE: cru=1 ttc=2 trd=3 cpv=4
          auxfuns.printError("Bad DETMODE in ltuIP.cfg file:%s"%detmode)
        else:
          self.DETMODE= int(detmode)
        print("ctpemu: DETMODE:",self.DETMODE, self.DETMODETX[self.DETMODE-1])
      elif parn == "TTCLVDS_delay":
        ds= ls[1:]
        # ltuIP.cfg line:
        # TTCLVDS_delay A B 1 2 (i.e. invert the order):
        for ix in range(4):
          self.ttcdelays[ix]= ds[ix]   #ds[3-ix]
        print("TTCLVDS_delay [TTCA, TTCB, lvds1, lvds2]:", self.ttcdelays)
      elif parn == "PP_CAL_Protection":
        if len(ls)!= 5:
          auxfuns.printError("Wrong PP_CAL_Protection line in ltuIP.cfg file.")
          continue
        self.ppcalprotl= ls[1:]
      elif parn in self.emudefaults:
        varname= self.emudefaults[parn][0]
        if varname=="":
          varname= parn
        vars(self)[varname]= int(ls[1])
        print("readDefaults %s:"%parn, vars(self)[varname])
    # 3. ppcalproc: special line rewriting parameters available in ltuIP.cfg
    #    by pcalprot calculated values
    if self.ppcalprotl!=None:   # calculate + overwrite 
      # pplcalprotl: [PP_BC_mask CAL_BC_mask ProtTime Frekv]
      #               0          1           2        3
      print("readDefaults  PP_CAL_Protection:", self.ppcalprotl)
      # calculate+overwrite defaults only:
      pci= list(map(int, self.ppcalprotl))
      if pci[2]>(pci[1]-pci[0]):
        self.PP_BC_mask= str((pci[1]+bcmask.ORBITLENGTH - pci[2])%bcmask.ORBITLENGTH)
        self.CAL_BC_mask= self.ppcalprotl[1]
        self.CAL_protection_time= (pci[0] + pci[2] - pci[1])%bcmask.ORBITLENGTH
      else:
        self.PP_BC_mask= self.ppcalprotl[0]
        self.CAL_BC_mask= self.ppcalprotl[1]
        self.CAL_protection_time= 0
      print("         HW  PP_CAL_Protection:", self.PP_BC_mask,\
        self.CAL_BC_mask,self.CAL_protection_time)
      self.PPCALfrekv= self.ppcalprotl[3]
      # todo for Frekv :"0" -do not set any, otherwise BCs between caltriggs
      # ppcal mode IS SET WHEN PP_CAL_Protection LINE PRESENT!
      self.PPCALmode= 1         # maybe to 3 if frekv=0 ?
      #
      # hw not set here (see CTPemu.__init__):
    fp.close()
  def getDETMODE(self):
    return self.DETMODE   # 1 2 3 4  i.e. what is set in ltuIP.cfg
  def rnc(self, bits=0xffffffff):
    self.board.Write(self.tn+".rst_N_cnts",bits);   # reset cnts
    self.board.Write(self.tn+".rst_N_cnts",0);
  def start(self, msg=""):
    """
return: (rc,emsg)
LTU: prepare+start stdalone run
     (0,message) -run started
     (4,message) -run started
     (>4,message) -run not started
CTP: prepare only -self.start() needed also before setting bits in gcore.global_run
"""
    ctrl= self.board.Read(self.ctrln)
    if self.emuactive(ctrl):   # ctp or ltu
      return 8,"%s is already started, no action."%self.emuname
    if self.board.isfiwa(self.board.FW_LTU):
      #ctrlctrl= self.board.Read(self.ctrl.ctrl)   # better: use fpga.getMode
      #if (ctrlctrl & 0x10)==0x10:
      rc,msg= self.ltudevs["fpga"].getMode() 
      if msg!="":
        return 8,msg+", no action."
      if rc==1:
        return 8,"LTU in global mode, no action."
    rc= 0 ; emsg=""
    self.rnc()
    if self.board.isfiwa(self.board.FW_LTU):
      # following action done in ctp.py separately (TF sync)
      self.emureset(0x1)   # has to be done before startemu (orbit counter not cleared)
      self.ltudevs["fpga"].orbitset(None,3564)   # clear orbit counter with each stdalone run
      #rw= self.board.Read(self.tnttc+".ctrl", rw)
      DETMODE= (self.ttcctrl >>12 )&0xf + 1
      if self.DETMODE==3:   # TRD
        bsy_se= (self.ttcctrl >>4 )&3
        self.ltudevs["hbmb"].hbmbinit()   # let's init always (even if PON disbaled)
        if bsy_se==3:   # this is required only by TRD
          bsys= self.ltudevs["hbmb"].getBusy()   # check if busy
          if bsys!=0:
            emsg= "TRD busy CRUs: 0x%x, . Valid CRUs:0x%x"%(bsys, self.ltudevs["hbmb"].validmap)
            auxfuns.printError(emsg)
            rc= 4
        else:
          emsg= "DETMODE TRD, but BUSY:%d (3 expected, i.e. sourcePON+enabled"%bsy_se
          auxfuns.printError(emsg)
          return 8,emsg
    self.board.Write(self.tn+".TTena", self.ttena);
    if msg!="": print(msg)
    if self.board.isfiwa(self.board.FW_LTU):
      # read counters BeforeSOX (similarly in fpga.setGlobalMode for global run)
      self.board.Write(self.ctrln, ctrl|0x2)   # EMU_run bit: 1
      time_rs= time.time()
      if emsg!="":
        emsg= emsg+"\n"+"ctpemu started"
      else:
        emsg= "ctpemu started"
      print("ctpemu started. Time: %.6fs, TTena set to 0x%x, EMU_run:1."%\
        (time_rs,self.ttena))
    else:
      emsg= "%s: LTG TTena set to 0x%x."%(self.emuname,self.ttena)
      print(emsg)
    return rc,emsg
  def emuactive(self, ctrlreg=None):
    if ctrlreg==None: ctrlreg= self.board.Read(self.ctrln)
    if self.emuname=="emu":
      if (ctrlreg & 2)==0:
        return False # emulation off
      else:
        return True
    else:   # ctp
      gcgr= self.board.Read("gcore.global_run")
      detbit= int(self.emuname[3:])-1   # 0..17
      dets= (gcgr>>1) & 0x3ffff
      gr= gcgr & 0x1
      #if (gr==0) and (((dets>>detbit)&1)==0): # grun stopped AND det. not choosen!
      if ((dets>>detbit)&1)==0: # det. not choosen!
        return False
      else:
        return True
  def clearemu(self):
    for signam in list(self.signames.keys()):
      if signam=="TF_orbit": continue
      if signam=="HBr_orbit": continue
      self.set_rn(signam, "0", "0")
  def stop(self,clear=False):
    """ only for LTU
"""
    ctrl= self.board.Read(self.ctrln)
    if (ctrl & 0x2)==0:
      print("CTP emulator was already stopped")
    if self.board.isfiwa((self.board.FW_CTP_LOGIC,)):
      # LTG in ctp board:
      #print "CTP, i.e. use fpga stoprun, no action."
      #return
      print("CTP %s stop."%self.tn)
    else:
      if (ctrl & 0x10)==0x10:
        print("LTU in global mode, no action.")
        return 
    # clear TTena (do not touch XOX TF_orbit and HBr_orbit):
    ttena= self.board.Read(self.tn+".TTena")
    #ttena= ttena & 0xf84   # TF EOC SOC EOT SOT HBr
    ttena= ttena & 0xfe4   # TF EOC SOC EOT SOT CAL PP HBr
    self.board.Write(self.tn+".TTena", ttena);
    self.board.Write(self.ctrln, ctrl&0xfd)   # EMU_run bit: 1
    print("emu stop. TTena set to 0 (but TF_orbit XOX HBr_orbit bits kept), EMU_run:0")
    time.sleep(0.050)   # wait 2 TF after stopped
    # read counters AfterEOX (similarly in fpga.setStdaloneMode for global run)
    if self.board.isfiwa(self.board.FW_LTU):
      if "hbmb" in self.ltudevs:
        self.ltudevs["hbmb"].trdstop()
    if clear:
      self.ttena= ttena
      self.clearemu()
  def getnames(self, rnreg):
    """ rc: rate_regname, N_regname, enabit_position, plsr_bit_position
"""
    nname= self.signames[rnreg][0]
    enablebit= self.signames[rnreg][1]
    if nname==None:
      n= None
      rate= None
    elif nname=="":
      n= self.tn+"."+rnreg+"_n"
      rate= self.tn+"."+rnreg+"_rate"
      if rnreg=="TF_orbit": n= None 
    else:
      # should not happen:
      n= self.tn+"."+nname+"_n"
      rate= self.tn+"."+nname+"_n"
    pulser= self.signames[rnreg][3]
    #print "getnames:",rate,n,enablebit,pulser
    return rate,n,enablebit,pulser
  def n2str(self, n):
    if n==0xffffffff:
      strn= "inf"
    else:
      strn= "0x%x"%n
    return strn
  def hbr_perc2n(self, fperc):
    if fperc>100: fperc= 100
    if fperc<0: fperc= 0
    #rateint= int((257/100.)*fperc)   # v 15_03
    rateint= int((self.TForbitsActual/100.)*fperc)
    return rateint
  def hbr_n2perc(self, hbrn):
    acc_perc= hbrn*100./self.TForbitsActual
    #acc_perc= hbrn*100/257.   # v15_03
    return acc_perc
  def prtena(self, ebits):
    ib= 31 ; rc= ""
    while ib>=0:
      if ebits & (1<<ib):
        tx= self.ttenabits[31-ib]
        if tx!="":
          rc= rc+tx+" "
      ib= ib-1
    return rc
  def prt_rn(self, rnreg, fmt="prtcolor"):
    """ fmt:
prtcolor: print line to stdout (colored)
plaintext: return line (plain text)
                  None if not choosen in TTena
"rate": return string:
"0"      -trigger repetition is 0
None     -not choosen in TTena
"xx.xxhz -rate in hz khz mhz
"""
    rname,nname,enabit,pulserbit= self.getnames(rnreg)
    if rname==None:       # trigger rate not defined
      ratestrn="    -"
      rate= None
    else:
      rate= self.board.Read(rname)
      ratestrn= "%5d"%rate
    if pulserbit!=None:   # pulser connected?
      ctrl= self.board.Read(self.ctrln) 
      pulser= ctrl & (1<<pulserbit)
    else:
      pulser= 0
    if nname==None:       # number of triggers limit exist?
      strN="-"
      N= 0
      if rate==None:
        # XOX_tf     -      -     (SOT EOT SOC EOC+_tf not printed)
        return
    else:
      N= self.board.Read(nname)
      strN= self.n2str(N)
    pulsernote= ""
    choosen= ' '
    if (enabit!=None) and (self.ttena & (1<<enabit)):   # choosen signal?
      choosen= '*'
    if pulser==0:
      ppt= ""
    else:
      ppt= "P"
    if rnreg.find("_bc")>0:
      if N>0:
        pulsernote= auxfuns.bc2hz(rate)
    elif rnreg.find("_rnd")>0:
        pulsernote= auxfuns.rnd2hz(rate)
    else:
      if fmt=="prtcolor":
        if rnreg=="HBr_orbit":
          #if rate<=1:   # v15_03
          if rate==0:
            pulsernote="All rejected. 'emu hbaccept 100%' for min. rejection"
          else:
            rateF= self.hbr_n2perc(rate)
            pulsernote= "%.2f%% TFs accepted"%(rateF)
      else:
        if rnreg=="HBr_orbit":
          if rate==0:
            pulsernote="0"
          else:
            rateF= self.hbr_n2perc(rate)
            pulsernote= "%.2f"%(rateF)
    if fmt=="plaintext":
      if choosen==' ': return None
      if (rnreg[:3]=="PH_") and (N==0): return None
      line= "%s %s %6s %s"%(rnreg, ratestrn,strN, ppt)
      return line
    elif fmt=="prtcolor":
      if ppt=="P":
        pulsernote= "PLSR "+pulsernote
      line= "%14s%c%s %6s %s"%(rnreg, choosen, ratestrn,strN, pulsernote)
      if enabit==None:
        print(line)
      else:
        edbits= self.board.Read(self.tn+".TTena")
        if (edbits & (1<<enabit)) or (ppt=="P"):
          enadiscol= "fg_green"
        else:
          enadiscol= "fg_red"
        #print "prt_rn:", enadiscol, hex(edbits)
        print(colors.color(line, enadiscol)) 
    else:   # fmt=="rate"
      if choosen==' ': return None
      if N==0: return "0"
      return  pulsernote
  def emureset(self, cbit):
    ctrl= self.board.Read(self.ctrln)
    self.board.Write(self.ctrln,ctrl | cbit)
    self.board.Write(self.ctrln,ctrl & (~cbit))
  def enadis(self, enabit, dis0):
    """ set/clear self.ttena bit 
enadis(enabit, 1): set bit
enadis(enabit, 0): clear bit
"""
    if dis0==0:
      self.ttena= self.ttena & (~(1<<enabit))
    else:
      self.ttena= self.ttena | (1<<enabit)
    #self.board.Write(self.tn+".TTena", TTena)
  def invertpulser(self, rnreg):
    ctrl= self.board.Read(self.ctrln)
    enabit= self.signames[rnreg][3]
    if enabit==None:
      print(colors.color("%s cannot be connected/disconnected to/from pulser"%\
        rnreg, "fg_red"))
      return
    dis0= ctrl & (1<<enabit)
    if dis0!=0:
      ctrl= ctrl & (~(1<<enabit))
    else:
      ctrl= ctrl | (1<<enabit)
    self.board.Write(self.ctrln, ctrl)
  def set_rn(self, rnreg, rate, nn, immediate=False):
    # rate: int  nn:string
    rname,nname,enabit,pulser= self.getnames(rnreg)
    #print "set_rn:", rnreg,rate,nn,rname,nname,enabit
    #
    # set_rn: PH_bc 34 inf ctpemu.PH_bc_rate ctpemu.PH_bc_n 4
    # ltu_logic_v12: problem with writing BC_rate/BC_n registers
    # seen in boards: 16,9,10
    # atb_10 (tof 10.4.2020:): -without an errmsg seen from WriteCheck!
    # PH_bc*40078 4294966783    <- hex: 0xfffffdff
    # after rewriting again it is ok:
    # emu PH_bc 1khz inf i
    # PH_bc*40078    inf
    #
    # not obsereved: with ctp_logic_v1 neither ltu_logic_itsmft_v12
    # (boards 61,13)
    if (rate!=None) and (rname!=None) and (rname!=""):
      self.board.WriteCheck(rname, int(rate))
    if nname!=None:
      if nn:
        if nn=='inf': intn=0xffffffff
        else:
          intn= int(nn)
        if intn==0:
          self.enadis(enabit, 0)   # disable
        else:
          self.enadis(enabit, 1)   # enable
        self.board.WriteCheck(nname, intn)
        #newintn= self.board.Read(nname)
        #print "set_rn intn:", intn, newintn
      if immediate:  
        self.board.Write(self.tn+".TTena", self.ttena)
    self.prt_rn(rnreg)
  def checksetTF(self):
    tfname= self.tn+".TF_orbit_rate"
    tfrate= self.board.Read(tfname)
    self.TForbitsActual= tfrate
    #if (tfrate != 256) and (tfrate != 128):
    if tfrate != self.TForbits:
      self.board.Write(tfname, self.TForbits)
      self.TForbitsActual= self.TForbits
      print(colors.color("%s: was %d, rewritten to %d"%(tfname, tfrate, self.TForbits), "fg_red"))
  def checksetTFHBr(self):
    """ set default values 
    """
    if self.board.isfiverhe(0x19):
      # is in ctrl.ctrl[31:20] ORBIT_OFFSET_OLDTTC, also for ctp_v8
      #self.board.SetBits("ctrl.ctrl", 60, 0xfff, 20)
      #self.board.SetBits("ctrl.ctrl", 3564-60, 0xfff, 20) before ltu_v1d_0c
      # todo: remove following line, anyhow set it fpga.setStdalone()
      self.board.SetBits("ctrl.ctrl", 3564-61, 0xfff, 20) #0xdaf from ltu_v1d_0c
      # stdalone: 0xdaf   global: 0xdb5
    elif self.board.isfiverhe(0x18):
      self.board.Write(self.tn+".ORBIT_offset_oldTTC", 60)
    else:
      pass
    self.checksetTF()
    #self.ttena= self.board.Read(self.tn+".TTena")
    if (self.ttena & 0x004) != 0x004:   # TF HBr
      # 8.10.2020: 0x4 is enough (Marian: 0x800 is not used for TF). changed 12.5.2021
      print(colors.color("TTena was: 0x%x, setting TF and HBr bits (0x004)..."%self.ttena, "fg_red"))
      self.ttena= self.ttena | 0x004
      self.board.Write(self.tn+".TTena", self.ttena)
  def setCAL_protection_time(self, delay=None):
    if delay==None: delay= self.CAL_protection_time
    CPTD= self.emudefaults["CAL_protection_time"][1]
    #cal_62= delay-CPTD
    cal_62= delay-CPTD   # 63 from ltu_v1d_0c (62 before)
    if cal_62<0:
      #auxfuns.printError("CAL_protection_time (%d) must be >=62. Set to 62"%delay)
      auxfuns.printError("CAL_protection_time (%d) must be >=%d. Set to %d"%(delay, CPTD,CPTD))
      delay= CPTD ; cal_62= 0
    else:
      self.board.Write(self.tn+".CAL_protection_time", delay)
    # keep it -62:
    if self.board.isfiverhe(0x18):
      self.board.Write(self.tn+".PP_del_oldTTC", cal_62)
  def bstatus(self, fmt=None):
    """ brief status, return the list of:
fmt: None
active X Mode HBa% PH_bc PH_rnd PP_bc CAL_bc TOF_bc TPCbc TPCorb
active: 0: stopped 1: started
X: Continuous Triggered - (nosox)
Mode:
HBa%: .xx in % (string)
PH_bc,...  string ...hz ...khz or 0: N=0 -: not choosen in TTena

fmt: "h"
"""
    rc= []
    if fmt=='h':
      h="%-2s %-1s %-3s %-6s"%("Ac", "X", "MOD", "HBacc%")
      for ke in ("PH_bc","PH_rnd","PP_bc","CAL_bc","TOF_bc",\
        "TPC_RST_bc","TPC_SYNC_orbit"):
        h= h+ " %-12s"%ke
      return h
    ctrl= self.board.Read(self.ctrln)
    active= str((ctrl>>1)&1)
    rc.append(active) ; rlin= "%-2s"%active
    TTena= self.board.Read(self.tn+".TTena")
    sox= self.getsox(TTena, fmt=2)
    rc.append(sox) ; rlin= rlin + " %-1s"%sox
    if self.tnttc==None:
      modet= self.DETMODETX[0]
    else:
      rw= self.board.Read(self.tnttc+".ctrl") ; self.ttcctrl= rw
      mode= (rw>>12)&0xf
      modet= self.DETMODETX[mode]
      #print "DETMODE: %s, see also 'emu ttc status'"%modet
    rc.append(modet)
    rlin= rlin + " %-3s"%modet
    trgs= {}
    for key in list(self.signames.keys()):
      trgs[key]= self.prt_rn(key, fmt="rate")
      #print "bstatus:", key, trgs[key]
    rc.append(trgs["HBr_orbit"])
    rlin= rlin + " %-6s"%trgs["HBr_orbit"]
    for ke in ("PH_bc","PH_rnd","PP_bc","CAL_bc","TOF_bc",\
      "TPC_RST_bc","TPC_SYNC_orbit"):
      #print "bst:", ke, trgs[ke]
      if ke in trgs:
        v= trgs[ke]
        if v==None: v= "-"
      else:
        v= "?"
      rc.append(v)
      rlin= rlin + " %-12s"%v
    #return rc
    return rlin
  def status(self):
    ctrl= self.board.Read(self.ctrln)
    TTena= self.board.Read(self.tn+".TTena")
    self.sox= self.getsox(TTena)
    sctrl="ctrl:0x%x ("%ctrl
    for ixb in range(32):
      if not self.version1501:
        if ixb==17: continue   # already checked with 16
      if ctrl & (1<<ixb):   # 1
        if not self.version1501:
          if ixb==16:
            ppcal= (ctrl>>16)&0x3
            sctrl= sctrl+self.calmodes[ppcal]+" "
          else:
            sctrl= sctrl+ self.ctrlbits[ixb]+" "
        else:
          sctrl= sctrl+ self.ctrlbits[ixb]+" "
      else:   # 0
        if (ixb==1):
          sctrl= sctrl+ "EMUstopped "
    sctrl= sctrl+")"
    if TTena & self.ferbit:
      fertxt= "FErst:%d,enabled"%self.board.Read(self.tn+".FEErst_tf_delay")
    else:
      fertxt= "FErst:%d,disabled"%self.board.Read(self.tn+".FEErst_tf_delay")
    print("%s SOX/EOX:%s %s"%(sctrl, self.sox, fertxt))
    print("TTena(in hw):0x%x %s"%(TTena, self.prtena(TTena)))
    print("USRena(used with next emu start):0x%x %s"%\
      (self.ttena,self.prtena(self.ttena)))
    #print "pllmode:",self.board.pllmode
    for key in list(self.signames.keys()):
      self.prt_rn(key)
    # oldttc part:
    if self.tnttc==None: return
    # show only in case it is relevant
    rw= self.board.Read(self.tnttc+".ctrl") ; self.ttcctrl= rw
    mode= (rw>>12)&0xf
    if mode>0:
      self.ttcstatus()
      #modet= self.DETMODETX[mode]
      #print("DETMODE: %s, see also 'emu ttc status'"%modet)
  def showMasks(self):
    mskena= self.board.Read(self.tn+".BCmask_ena")
    enabled= ""
    for ib in range(len(bcmask.symname)):
      if mskena & (1<<ib):
        enabled= enabled+bcmask.symname[ib]+" "
    return "%s enabled BCmasks: %s ('emu bcm TOF e/d' to enable/disable mask)"%(self.tn, enabled)
  def ttcstatus(self):
    rw= self.board.Read(self.tnttc+".ctrl") ; self.ttcctrl= rw
    mode= (rw>>12)&0xf
    print("Mode:", self.DETMODETX[mode])
    LLprot= self.board.Read(self.tnttc+".LLprot")
    L1reject= self.board.Read(self.tnttc+".L1reject")
    if L1reject!=0:
      L1rejectp= "(%.2f%%)"%(100./L1reject)
    else:
      L1rejectp= ""
    CALrate= self.board.Read(self.tnttc+".CAL_rate")
    print("LLprot:%d   L1reject:%d%s   CALrate_trd:%d"%(LLprot, \
      L1reject, L1rejectp, CALrate))
    self.ttcctrlsym.prt(rw)
    #if rw& 0x2:
    #  detbusy=1
    #else:
    #  detbusy=0
    #print "oldttc ctrl:0x%x = mode:0x%x(%s) DETbusy enabled:%d"%(rw, mode, modet, detbusy)
    ttcstatusbits= self.board.Read(self.tnttc+".statusbits")
    self.ttcstatusbits.prt(ttcstatusbits)
    errs= self.board.Read(self.tnttc+".errorbits")
    self.ttcerrors.prt(errs)
    ctrl= self.board.Read(self.tnttc+".masks")
    self.Sttcdelays.prt(ctrl)
  def trigs_config(self):
    """ rc: "" if stopped
"""
    txconfig= ""
    ctrl= self.board.Read(self.ctrln)
    TTena= self.board.Read(self.tn+".TTena")
    self.sox= self.getsox(TTena, fmt=1)
    if not self.emuactive(ctrl):
      return txconfig # emu stopped
    txconfig= self.sox
    for key in list(self.signames.keys()):
      if (key[-3:]=="_tf") or (key in ("TF_orbit",)): continue
      line= self.prt_rn(key,fmt="plaintext")
      if line!=None: txconfig= txconfig+"\n"+line
    # oldttc part:
    return txconfig
  def delme_hbrorbit(self, cmd):
    if len(cmd)==2:
      rate= self.board.Read(self.tn+".HBr_orbit_rate") # period in BCs
      raten= self.board.Read(self.tn+".HBr_orbit_n")   # 0:disbled, 0xffffffff:infinite
      print("%s: %d %d"%(cmd[1], rate, raten))
    elif len(cmd)>=3:
      rate=int(cmd[2])
      self.board.Write(self.tn+".HBr_orbit_rate", rate)
      if len(cmd)==4:
        raten=int(cmd[3])
        self.board.Write(self.tn+".HBr_orbit_n", rate)
  def setttcreg(self,regname, cmd):
    if len(cmd)>3:
      value= eval(cmd[3])
      self.board.Write(regname, value)
      print("%s: 0x%x written"%(regname, value))
      return value
    else:
      #print "%s content missing"%regname
      value= self.board.Read(regname)
      print("%s: 0x%x"%(regname,value))
      return None
  def printDelay(self,chan, masks=None):
    if masks==None:
      masks= self.board.Read(self.tnttc+".masks")
    if not (chan in self.delchans):
      print("Error: delay %s ? Only A B 1 2 allowed."%chan)
      return
    ix= self.delchans.index(chan)*8
    val= (masks>>ix)& 0x3f
    if val in self.delval2bc:
      bc= self.delval2bc[val]
      if bc=="d":
        print("Channel %s disabled."%chan)
      else:
        print("Delay %s:%s/6 BCs"%(chan, bc))  
    else:
      print("Int. error: forbidden delay 0x%x for channel %s"%\
        (val, chan))
  def ttc_setdel(self, ab12, dvaltx):   # i.e. ("2", "d" or "0".."5")
    if not (dvaltx in self.delbc2val):
      print("Only d(disable) or 0..5(delay in 1/6BC) allowed for TTCLVDS delay")
      return False
    dval= self.delbc2val[dvaltx]
    if not (ab12 in self.delchans):
      print("One of the following allowed for TTCLVDS delay:", self.delchans)
      return False
    dval= dval & 0x3f
    regname= self.tnttc+".masks"      # in hw: 2 1 B A
    ix= self.delchans.index(ab12)*8   # delchans= ("A","B","1","2") ix: 0 8 16 24
    value= self.board.Read(regname)
    value= value & (xf8 & ~(0x3f<<ix))
    nvalue= value | (dval<<ix)
    self.board.Write(regname, nvalue)
    print("%s set to 0x%x (%s: 0x%x)."%(ab12, dval, regname, nvalue))
    return True
  def ttc_setdelall(self, ab12):
    """ ab12: list of string values of 4 delays: (a, b, 1, 2) -> 0x21ba
values: "d" or "0".."5"
"""
    dvals= 0 ; bshift= 0
    #idels= [ int(dl) for dl in ab12 ] 
    #for ix in (0,1,2,3):
    for ix in range(4):
      dltx= ab12[ix]
      if not (dltx in self.delbc2val):
        print("Only d(disable) or 0..5(delay in 1/6BC) allowed for TTCLVDS delay")
        return False
      dl= self.delbc2val[dltx]
      dvals= dvals | ((dl&0x3f)<<(ix*8))
      #print "ttc_setdelall:", hex(dvals)
    self.board.Write(self.tnttc+".masks", dvals)
    print("TTCLVDS delays (A,B,1,2) set to", ab12)  
  #def ttc_setgs(self, gs):   # gs: 0:stdalone 1: global
  # obsolete -set by ctrl.ctrl g/s bit
  #  ctrl= self.board.Read(self.tnttc+".ctrl")  
  #  gsbits= (ctrl & 0xffff0ff) | (gs<<8)
  #  self.board.Write(self.tnttc+".ctrl", gsbits)
  def getsox(self, ttenabits=None, fmt=0):   # rc: SOTEOT SOCEOC NOSOX ERRSOX
    if ttenabits==None:
      ttenabits= self.board.Read(self.tn+".TTena")
    b1trig= (1<<self.signames["SOT_tf"][1]) | (1<<self.signames["EOT_tf"][1])
    b1cont=  (1<<self.signames["SOC_tf"][1]) | (1<<self.signames["EOC_tf"][1])
    if (ttenabits & b1trig) == b1trig:
      rc= ("SOTEOT","trig","T")
    elif (ttenabits & b1cont) == b1cont:
      rc= ("SOCEOC","cont","C")
    elif ((ttenabits & b1cont) == 0) and ((ttenabits & b1trig) == 0):
      rc= ("NOSOX","nomode","-")
    else:
      rc=("ERRSOX","errmode","E")
    return rc[fmt]
  def setsoxbits(self, sox):   # sox: SOTEOT SOCEOC NOSOX
    """set self.ttena bits (i.e. hw will be set when run starts)."""
    if sox=="SOTEOT":
      bits_one= (1<<self.signames["SOT_tf"][1]) | (1<<self.signames["EOT_tf"][1])
      bits_zero= ~( (1<<self.signames["SOC_tf"][1]) | (1<<self.signames["EOC_tf"][1]) )
    elif sox=="SOCEOC":
      # ->0: all 1s but 0 in bits which will be set to 0
      bits_zero= ~( (1<<self.signames["SOT_tf"][1]) | (1<<self.signames["EOT_tf"][1]) )
      # ->1: all 0s, but 1 in bits set to 1
      bits_one=  (1<<self.signames["SOC_tf"][1]) | (1<<self.signames["EOC_tf"][1])
    elif sox=='NOSOX':
      bits_zero= ~( (1<<self.signames["SOT_tf"][1]) | (1<<self.signames["EOT_tf"][1]) |\
        (1<<self.signames["SOC_tf"][1]) | (1<<self.signames["EOC_tf"][1]) )
      bits_one= 0
    else:
      print("Bad sox: %s (SOTEOT, SOCEOC or NOSOX expected) "%sox)
      return
    #ttenabits1= self.board.Read(self.tn+".TTena")
    ttenabits1= self.ttena
    ttenabits= ttenabits1 & bits_zero
    ttenabits= ttenabits | bits_one
    self.ttena= ttenabits
    print("setsoxbits(%s): ttena:0x%x -> 0x%x 0s:%x 1s:%x"%(sox, ttenabits1, ttenabits, 
      bits_zero&xf8, bits_one&xf8))
    self.sox= sox
    #self.board.Write(self.tn+".TTena", ttenabits);
  def setferst(self, tfs):
    #ttenabits= self.board.Read(self.tn+".TTena")
    ttenabits= self.ttena
    if tfs==None:
      tfs= self.board.Read(self.tn+".FEErst_tf_delay")
      ttenabits= ttenabits & (~self.ferbit)
      print("FErst: %d, disabled (ttena bit will be set to 0 at next SOX)"%tfs)
    else:
      ttenabits= ttenabits | (self.ferbit)
      print("FErst: %d before SOX, ttena bit will be enabled at next SOX"%tfs)
      self.board.Write(self.tn+".FEErst_tf_delay", tfs);
    self.ttena= ttenabits
    #self.board.Write(self.tn+".TTena", ttenabits);
  def ttc_setdefaultsHW(self):
    self.ttc_reset()
    self.board.Write(self.tnttc+".LLprot", self.LLprotection)
    self.board.Write(self.tnttc+".L1reject", self.L1reject)
    self.board.Write(self.tnttc+".CAL_rate", self.CALrate_trd)
    #rw= self.board.Read(self.tnttc+".ctrl")
    #rw= (rw & 0xff0f) | (self.defttcmode<<4)
    #if self.board.pllmode=="pllin1o0lvpecl":
    #  #rw= 0x10100   # for TTCex fed by LEMO00B cable from ATB
    rw= 0x0020   # MSP0: M=0:CRU, S=0:XOX over LVDS/TTC enabled, 
    #            SP= 2:   XOX+ PON+disabled   -default
    #            SP=13: noXOX+ PON+enabled    -should be for TRD
    bsy_se= (self.sourceBUSY<<1) | (self.enableBUSY&1) | (self.XOX_ttc<<4)
    rw= (rw&0xfffffecf) | (bsy_se<<4)
    #rw= rw | ((self.DETMODE-1)<<4)  <v16_00
    rw= rw | ((self.DETMODE-1)<<12)
    self.board.Write(self.tnttc+".ctrl", rw) ; self.ttcctrl= rw
    print("ttc_setdefaultsHW: LLprot:%d L1reject:%d CAL_rate:%d ctrl:0x%x"%\
      (self.LLprotection, self.L1reject, self.CALrate_trd, rw))
    self.ttc_setdelall(self.ttcdelays)   # a b 1 2
    if self.DETMODE==3:    # default required by ltuIP.cfg is TRD
      if bsy_se==0x13:   # this is required only by TRD
        #self.ltudevs["hbmb"].trdinit("trd.net")  done in hbmb
        pass
      else:
        auxfuns.printError("DETMODE: %d, but 0x13 (TRD) expected (= noXOX + busy over PON + busy enabled"%\
          self.DETMODE)
    self.ttc_resetsticky()
  def ttc_reset(self):
    self.board.Set10(self.tnttc+".ctrl",0)
  def ttc_resetsticky(self):
    self.board.Set10(self.tnttc+".ctrl",1)
  def cal_status(self):
    ppcal= self.cal_getmode()
    #ttenabits= (self.board.Read(self.tn+".TTena")>>5)&3
    ttenabits= (self.ttena>>4)&6 | self.ttena&1
    modeerr= ""
    #if ttenabits!=self.expTTenabits[ppcal]:
    if not (ttenabits in self.expTTenabits):
      modeerr= " TTena err: %d (expected:%d)"%(ttenabits,self.expTTenabits[ppcal])
    ppcaltx= self.calmodes[ppcal]
    ctrl= self.board.Read(self.ctrln)
    phrej= (ctrl>>18)&0x1
    phrejtx= ["off", "on"][phrej]
    delay= self.board.Read(self.tn+".CAL_protection_time")
    print("Mode:%s%s PHreject:%s CAL_protection_time:%d"%\
      (ppcaltx,modeerr, phrejtx,delay))
    self.cal_rate()
    if ppcal==0:
      #bcm= bcmask.BCmaskhw(self.board, 1) ; bcm.prt()
      self.bcmprt(1)
    elif ppcal==1 or ppcal==2:
      #bcm= bcmask.BCmaskhw(self.board, 1) ; bcm.prt()
      #bcm= bcmask.BCmaskhw(self.board, 2) ; bcm.prt()
      self.bcmprt(1)
      self.bcmprt(2)
    else:
      pass
  def bcmprt(self,msknumber):
    enabit= self.board.Read(self.tn+".BCmask_ena") & (1<<msknumber)
    if msknumber==1 or msknumber==2:
      enadis="(%d) always enabled"%enabit
    else:
      enadis= "enabled"
      if enabit==0: enadis= "disabled"
    print("%s: %s, BCmask:"%(bcmask.symname[msknumber], enadis),\
      self.ltudevs["bcm"].getbcmstr(msknumber))
  def cal_getmode(self):
    """ rc: 0:pp 1:ppcal 2:cal 3:noppcal
"""
    if self.version1501:
      b3= ((self.ttena>>4)&6) | (self.ttena&1 )
      #print "cal_getmode:",hex(self.ttena), b3
      if b3 == 2:
        rc= 0   # pp
      elif b3 == 3:
        rc= 1   # ppcal
      elif b3 == 4:
        rc= 2   # cal
      else:
        rc= 3   # off
    else:
      ctrl= self.board.Read(self.ctrln)
      rc= (ctrl>>16)&0x3
    print("cal_getmode:%s (rc:%d)"%(self.calmodes[rc],rc))
    return rc
  def cal_setmode(self, calsel):
    """ calsel: 0,1,2 -> ctpemu.ctrl[CAL_sel bits]
<15_01
Mode   ctrl.    TTena.  TTena.
       CAL_sel  CAL_ena PP_ena
       17.16    6       5
pp     0        0       1  =1
ppcal  1        0       1  =3
cal    2        1       0  =2
                0       0   CAL_sel does not matter (disabled both)

from v15_01:
Mode            TTena.  TTena. TTena.
       calsel   CAL_ena PP_ena CAL_after_PP_ena
bit in ctrl:    6       5      0
pp     0        0       1      0
ppcal  1        0       1      1
cal    2        1       0      0
off    3        0       0      0 no pp neither cal
"""
    #print "cal_setmode:", self.version1501, calsel
    if self.version1501:
      if calsel==1:
        self.enadis(0, 1)   # set bit CAL_after_PP_ena
      else:
        self.enadis(0, 0)   # clear bit CAL_after_PP_ena
    else:
      if calsel==3: calpp= 0
      elif calsel==0: calpp= 1
      else: calpp= calsel
      self.ttena= self.board.SetBits(self.ctrln, calpp, 0x3, 16)
      #  it is wiser to change only self.ttena here! nevertheless, old version (delme)
    #print hex(self.ttena)
    if calsel==0:
      self.enadis(6, 0) ; self.enadis(5, 1)
    elif calsel==1:
      self.enadis(6, 0) ; self.enadis(5, 1)
    elif calsel==2:
      self.enadis(6, 1) ; self.enadis(5, 0)
    elif calsel==3:
      self.enadis(6, 0) ; self.enadis(5, 0)
    else:
      auxfuns.printError("cal_setmode(%d) not defined (int. error)"%calsel)
    #print hex(self.ttena)
  def cal_phreject(self, phrej):
    if phrej==1:
      self.board.SetBit(self.ctrln, 18)
      phtx="on"
    else:
      self.board.ResetBit(self.ctrln, 18)
      phtx="off"
    print("PH_reject_ena:",phtx)
  def cal_rate(self, newrate=None, ncals=None):
    """ 1. show registers (newrate==None) or
2. set registers (if newrate !=None): newrate, ncals as a text!
"""
    if newrate==None:   # show
      pp_n= self.board.Read(self.tn+".PP_bc_n")
      cal_n= self.board.Read(self.tn+".CAL_bc_n")
      pp_ntx= self.n2str(pp_n)
      cal_ntx= self.n2str(cal_n)
      pprate= self.board.Read(self.tn+".PP_bc_rate")
      calrate= self.board.Read(self.tn+".CAL_bc_rate")
      print("PP_bc_rate:%d/%s CAL_bc_rate:%d/%s"% (pprate,pp_ntx, calrate, cal_ntx))
    else:
      ppcal= self.cal_getmode()
      if ppcal<=1:
        signam="PP_bc"
      else:
        signam="CAL_bc"
      newratebc= auxfuns.fromms(newrate)
      self.set_rn(signam, newratebc, ncals, False)
      ppcaltx= ["pp", "ppcal", "cal"][ppcal]
      print("%s (mode:%s) set to:%s(%s bcs). n:%s"%\
        (signam, ppcaltx, newrate,newratebc, ncals))
  def excmd(self, cmd):
    """ execute interactive cmd e.g.: ["emu", "start" ]. 
'excmd' set is a superset of all commands
see also excmdbatch()  (subset of all commands).
"""
    if len(cmd) < 2:
      self.prtHelp()
    elif cmd[1]=='start':
      rcm= self.start()
      if rcm[1]!="": print(rcm)
    elif cmd[1]=='stop':
      clearhw= False
      if len(cmd) > 2:
        if cmd[2]=='clear': clearhw= True
      self.stop(clearhw)
    elif cmd[1]=='reset':
      emubit=0x1
      rsttxt= "emu"
      if len(cmd) > 2:
        rsttxt= cmd[2]
        if cmd[2]=='outc':
          emubit=0x10
        elif cmd[2]=='orbit':
          emubit=0x20
        else:
          print("reset of what? (outc or orbit allowed)")
          rsttxt= None
      if rsttxt!=None:
        print("%s reset done."%rsttxt)
        self.emureset(emubit)
    elif cmd[1]=='resetoutc':
      self.emureset(0x10)
    elif 'bstatus'.find(cmd[1])==0:
      print(self.bstatus(fmt='h'))
      print(self.bstatus())
    elif cmd[1]=='status':
      self.status()
    elif cmd[1]=='config':
      print(self.trigs_config())
    elif cmd[1]=='Tcos':
        tdgssm= "tdg"
        trc= self.ltudevs[tdgssm].start_recording()
        self.emureset(0x20)
        self.ltudevs[tdgssm].stop_recording()
        print("recording started at %.6fs, orbit cleared recording stopped."%trc)
    elif cmd[1]=='sss' or cmd[1]=='Tsss' or cmd[1]=='Ssss':
      tdgssm= None
      if cmd[1][0]=='T':
        tdgssm= "tdg"
      if cmd[1][0]=='S':
        tdgssm= "ssm"
      if len(cmd)>=3:
        ms=int(cmd[2])
      else:
        ms= 10
      if tdgssm:
        trc= self.ltudevs[tdgssm].start_recording()
        self.start("recording started at %.6fs on %s..."%(trc,tdgssm))   # start emu
      else:
        self.start()   # start emu
      print("sleep(%d ms)..."%ms)   # start emu
      time.sleep(ms/1000.)
      self.stop()
      if tdgssm:
        time.sleep(0.023)
        print("Stopping recording on %s, after 23ms to get a last TF with EOX"%(tdgssm))
        self.ltudevs[tdgssm].stop_recording()
    elif cmd[1]=='hbr':
      self.prtHelp("hbr")
    else:
      rc= self.excmdBatch(cmd, help=True)
      if rc!=0:
        print("Error:", rc)
  def excmdBatch(self,cmd, help=False):
    """ cmd: ["emu", "ph_bc", "2hz", "30"]
help: True: show help text if bad cmd, False: no stdout (only rc>0) in case of error
rc: 0 ok
    4 error in 'emu signale rate...' command
    8 'emu ttc' error
    16 bad command
Motivation: here only 'batch' commands
  available for ECS over grpc, i.e. in .pd or .par file
"""
    rc=0
    if len(cmd) < 2:
      return 0
    if cmd[1]=='trig':
      self.setsoxbits("SOTEOT")
    #elif cmd[1]=='start':
    #  self.start()
    #elif cmd[1]=='stop':
    #  clearhw= False
    #  if len(cmd) > 2:
    #    if cmd[2]=='clear': clearhw= True
    #  self.stop(clearhw)
    elif cmd[1]=='cont':
      self.setsoxbits("SOCEOC")
    elif cmd[1]=='nosox':
      self.setsoxbits("NOSOX")
    elif cmd[1]=='rnc':
      self.rnc()
    #elif cmd[1]=='pulserin':
    #  self.options["pulserin"].cmd(cmd[1:])
    elif cmd[1] in self.options:
      #plsr_sel: ctpemu.ctrl[7]: 0-plsr_in, 1-fastlm_in
      self.options[cmd[1]].cmd(cmd[1:])
      #print "not done yet"
    elif cmd[1]=='ferst':
      if len(cmd)>2:
        if cmd[2]=="off":
          tfs= None
        else:
          tfs= int(cmd[2])
      else:
        tfs= 1
      self.setferst(tfs)
    elif 'hbaccept'.find(cmd[1])==0:
      # emu hbaccept 44%
      if len(cmd)== 3:
        np= cmd[2]
        if np[-1]=='%':
          fperc= float(np[:-1])
          rateint= self.hbr_perc2n(fperc)
          rate= str(rateint)
        else:
          rate= int(np)
          if rate>self.TForbitsActual: rate= self.TForbitsActual
          if rate<0: rate= "0"
      else:
        rate= "255"
      self.set_rn("HBr_orbit", rate, "inf", True)
    elif cmd[1]=='bcm':
      ed= ""
      if len(cmd)>3:
        # emu bcm TOF enable
        ed= cmd[3]
        if bcmask.symname.count(cmd[2])!=1:
          print("""
emu bcm NAME e/d     -enable/disable mask
bcm NAME             -show mask definition

Note: CAL,PP masks always enabled. See 'bcm' for list of NAMEs.
""")
        mskid= bcmask.symname.index(cmd[2])
        if "enable".find(ed)==0:
          self.board.SetBit(self.tn+".BCmask_ena", mskid)
        elif "disable".find(ed)==0:
          self.board.ResetBit(self.tn+".BCmask_ena", mskid)
        else:
          pass
      print(self.showMasks())
    elif cmd[1]=='cal':
      if len(cmd)>=3:
        if cmd[2]=='status':
          self.cal_status()
        elif cmd[2]=='delay': # CAL_protection_delay: PHreject-PP distance
          if len(cmd)>=4:
            delay= int(cmd[3])&0x1fff   # in BCs 12..0
            self.setCAL_protection_time(delay)
          else:
            delay= self.board.Read(self.tn+".CAL_protection_time")
            print("CAL_protection_time: %d"%delay)
        elif cmd[2]=='pp':    # only pp
          self.cal_setmode(0)
          self.cal_getmode()
        elif cmd[2]=='cal':   # only cal
          self.cal_setmode(2)
          self.cal_getmode()
        elif cmd[2]=='ppcal': # pp+cal
          self.cal_setmode(1)
          self.cal_getmode()
        elif cmd[2]=='off': # noppcal
          self.cal_setmode(3)
          self.cal_getmode()
        elif cmd[2]=='phreject': # [off] [on]:default
          if len(cmd)>=4:
            if cmd[3]=='off':
              self.cal_phreject(0)
            else:
              self.cal_phreject(1)
          else:
            self.cal_phreject(1)
        elif cmd[2]=='rate':   # set pp(pp or ppcal) resp. cal rate (cal mode)
          if len(cmd)>=4:
            rate= cmd[3]   # 
            ncal= "inf"
            if len(cmd)>=5: ncal= cmd[4] # emu cal rate 20 NMAX
            self.cal_rate(rate, ncal)
          else:
            self.cal_rate()
        else:
          self.prtHelp("cal")
      else:
        self.prtHelp("cal")
    elif cmd[1]=='ttc':
      if self.tnttc==None:
        print("ttc part of CTPemu not available.")
        return 8
      mode= 0
      if len(cmd)==2:
        self.prtHelp("oldttc")
        return 8
      else:
        sleepms= 100   # was 20
        if (cmd[2]=='0') or (cmd[2]=='ttcpon'): pass
        elif cmd[2]=='status':
          mode= -3
          if len(cmd)>3:
            self.ttcstatusbits.prtme()
            #status= self.board.Read(self.tnttc+".status")   # <v16_00
            status= self.board.Read(self.tnttc+".statusbits")
            self.ttcstatusbits.prt(status)
            self.ttcerrors.prtme()
            #status= self.board.Read(self.tnttc+".status")   # <v16_00
            status= self.board.Read(self.tnttc+".errorbits")
            self.ttcerrors.prt(status)
          else:
            self.ttcstatus()
        elif cmd[2]=='init':
          mode= -3
          if len(cmd)>3:
            sleepms= int(cmd[3])
          resetctrl= int(sleepms)*0.001
          self.board.Write(self.tnttc+".chbdata", 0x800006ff) # ttcrx reset
          if sleepms>0:
            time.sleep(resetctrl)
          self.board.Write(self.tnttc+".chbdata", 0x80000000) # fine delay1: 3126ps
          self.board.Write(self.tnttc+".chbdata", 0x80000180) # fine delay2: 16459ps
          self.board.Write(self.tnttc+".chbdata", 0x80000200) # coarse delay: 0
          self.board.Write(self.tnttc+".chbdata", 0x800003f9) # activate B
          self.board.Write(self.tnttc+".chbdata", 0x80018000) # reset BC counter
          print("TTCinit done (%d ms sleep after TTCrx reset)"%sleepms)
        elif cmd[2]=='feecmd':
          # emu ttc feecmd code data
          mode= -3
          fedata= 0
          fecmd= 12 ; nrep= 1
          if len(cmd)>3: fecmd= eval(cmd[3])
          if len(cmd)>4: fedata= eval(cmd[4])
          if (fecmd<12) or (fecmd>15):
            auxfuns.printError("Bad fecmd:%s, 12..15 expected."%cmd[3])
          else:
            fecmddata= (fecmd<<12) | (fedata&0xfff)
            dw= 0x80010000 | fecmddata
            if nrep==1:
              self.board.Write(self.tnttc+".chbdata", dw) # FEE cmd
              print("feecmd:%d data:0x%x sent over ttc."%(fecmd, fedata))
            else:
              print("feecmd:%d data:0x%x sending %dx (1ms apart) over ttc..."%\
                (fecmd, fedata, nrep))
              for ix in range(nrep):
                self.board.Write(self.tnttc+".chbdata", dw) # FEE cmd
                time.sleep(0.001)
        elif cmd[2]=='nobch':
          mode= -3
          self.board.Write(self.tnttc+".chbdata", 0x800006ff) # ttcrx reset
          if len(cmd)>3:
            sleepms= int(cmd[3])
          resetctrl= int(sleepms)*0.001
          time.sleep(resetctrl)
          # 99 - no parallel out bus, no B-serial data
          # the must (seems ttcrx reset itself does not destroy control reg):
          self.board.Write(self.tnttc+".chbdata", 0x80000399) # de activate B
          print("B-channel deactivated, %d ms sleep before contr. register programming"%sleepms)
        elif cmd[2]=='oldttc': mode=1
        elif cmd[2]=='trd': mode=2
        elif cmd[2]=='cpv': mode=3
        elif cmd[2]=='busy': 
          if len(cmd) == 4:
            ed= cmd[3]
            if "enable".find(ed)==0:
              bsyenadis= 1
            elif "disable".find(ed)==0:
              bsyenadis= 0
            elif "lvds".find(ed)==0:
              bsyenadis= 2    # -2 = lvdsppon bit
            elif "pon".find(ed)==0:
              bsyenadis= 3
            else:
              print("emu ttc busy e[nable] or d[isable] expected")
              return 8
          else:
            print("emu ttc busy e[nable] or d[isable] expected")
            return 8
          mode=-1
        elif cmd[2]=="reset":
          mode=-2
          if len(cmd)>3:
            if "sticky".find(cmd[3])==0:
              mode=-4
        elif cmd[2]=='l1reject':
          mode= -3; self.setttcreg(self.tnttc+".L1reject", cmd)
        elif cmd[2]=='llprot':
          mode= -3; self.setttcreg(self.tnttc+".LLprot", cmd)
        elif cmd[2]=='calrate':
          mode= -3; self.setttcreg(self.tnttc+".CAL_rate", cmd)
        elif 'delay'.find(cmd[2])==0:
          mode= -3
          if (len(cmd)==5) and (cmd[3] in self.delchans):
            # emu ttc delay A d
            self.ttc_setdel(cmd[3], cmd[4])
          elif len(cmd)==7:
            # emu ttc delay d 0 1 4
            self.ttc_setdelall(cmd[3:])   # a b 1 2
          elif len(cmd)==4:
            # emu ttc delay A      i.e. show
            self.printDelay(cmd[3])
          elif len(cmd)==3:
            # emu ttc delay      i.e. show
            masks= self.board.Read(self.tnttc+".masks")
            for chan in self.delchans:
              self.printDelay(chan, masks)
          else:
            print("""
Show:
emu ttc delay
emu ttc delay A|B|1|2
Set just one value:
emu ttc delay A|B|1|2 value
Set 4 values by one line:
emu ttc delay Aval Bval 1val 2val

where value Aval Bval 1val or 2val is one of:  d    0    1   2   3   4   5
 corresponding 6-bits value in ttc.masks reg:  0 0x10 0x20 0x1 0x2 0x4 0x8

d    -disable channel
0..5 -delay for given channel in 1/6 BC
""")
            masks= self.board.Read(self.tnttc+".masks")
            self.Sttcdelays.prt(masks)
        elif cmd[2]=='ctrl':
          mode=-3 ; rw= self.setttcreg(self.tnttc+".ctrl", cmd)
          if rw!=None: self.ttcctrl= rw
        elif cmd[2]=='defaults':
          mode=-3 ; rw= self.ttc_setdefaultsHW()
        else:
          print("bad mode")
          return 8
      rw= self.board.Read(self.tnttc+".ctrl")
      if mode>=0:
        #rw= (rw & 0xff0f) | (mode<<4)   # <v16_00
        rw= (rw & 0xffff0fff) | (mode<<12)
        self.board.Write(self.tnttc+".ctrl", rw) ; self.ttcctrl= rw
        print("Mode in .ctrl reg: %d %s"%(mode, self.DETMODETX[mode]))
      elif mode==-1:   # "busy"
        #rw= (rw & 0xfffffd) | (bsyenadis<<1)   #<v16_00
        lvdspontx= ("LVDS", "PON")
        if bsyenadis<=1:
          rw= (rw & 0xffffffef) | (bsyenadis<<4)
          lvdspon= (rw&0x20)>>5
        else:
          lvdspon= bsyenadis-2
          rw= (rw & 0xffffffdf) | (lvdspon<<5)
          bsyenadis= (rw&0xef)>>4
        print("Detbusy input now:%d source:%s"%(bsyenadis,lvdspontx[lvdspon]))
        self.board.Write(self.tnttc+".ctrl", rw) ; self.ttcctrl= rw
      elif mode==-2:  # "reset"
        self.ttc_reset()
        #crst= rw | 0x1
        #self.board.Write(self.tnttc+".ctrl", crst)
        #rw= rw & 0xfffffffe
        #self.board.Write(self.tnttc+".ctrl", rw) ; self.ttcctrl= rw
        print("reset: write 1 + write 0")
      elif mode==-3:  # other cmd
        pass
      elif mode==-4:  # "reset sticky"
        self.ttc_resetsticky()
        #crst= rw | 0x2
        #self.board.Write(self.tnttc+".ctrl", crst)
        #rw= rw & 0xfffffffd
        #self.board.Write(self.tnttc+".ctrl", rw) ; self.ttcctrl= rw
        print("reset sticky: write 1 + write 0")
    elif cmd[1] in self.signames: 
      #print "excmdBatch:", cmd 
      # excmdBatch: ['emu', 'PH_bc', '34', 'inf']
      rate,nn,immediate,Pp= check_rn(cmd)
      signam= cmd[1]
      if Pp:
        self.invertpulser(signam)
        self.prt_rn(signam)
        if rate==None: return 0   # emu PH_bc P
      if rate==None: return 4   # bad rate argument
      self.set_rn(signam, rate, nn, immediate)
    else:
      if help: self.prtHelp()
      rc= 16  # bad batch cmd
    return rc
  def prtHelp(self, part=None):
    if part==None:
      print("""
emu start                 toggle rst_N_cnts + reset+set TTena bits in hw using USRena + set EMU_run bit
    rnc                   toggle rstN_cnts only (0xffffffff). ONLY when emulation active!
    stop [clear]          0 -> EMU_run + TTena bits in hw
                          clear: clear also hw _rate/_N registers for each trigger signal
    status                show current status
    config                show current config when emulation is active
                          (can be used for .par definition for stdalone run)
    reset                 write 1+0 to bit 0x1 ctpemu.ctrl (reset emu+scalers+HB_counter)
    cal                   calibration status/control
    reset orbit           orbit -clear Orbit counter (ctpemu.ctrl bit 0x20)
    reset outc            outc  -reset outcnts counters ((ctpemu.ctrl bit 0x10)
    sss [milisecs]        start emulation, sleep(defalut:10 ms), stop emulation
    Tsss [milisecs]         - start TDG recording before sss
    Ssss [milisecs]         - start SSM recording before sss
    Tcos                  start TDG, clear orbit, stop tdg  
    trig | cont |nosox    set SOXEOX mode to SOT/EOT, SOC/EOC or noSOX with 
                          start/stop emulation
    bcm MASK e | d        enable/disable MASK (PP CAL TOF see 'bcm' for all names...)
    pulserin [pulser | fastlm]  -choose pulser input, i.e.
                          Front Panel PLSR/ecl or FASTLM/lvds input
    pulserdet [strobe|edge] -pulser signal detcted by strobe or edge
    ferst [N]             FEErst_tf_delay register to N (0: behaves like 1) + enable ferst
                          when run started, FEreset is generated N TFs before SOX (default: 1)
    ferst off             FEreset trigger disabled, FEErst_tf_delay left untouched
    SIGNAME [ rate [N [i|P]] ]
                          -set USRena bit (can be reset by N:0 for a given signal)
                          -set 'SIGNAME rate N' (default N: inf) in hwregs _n,_rate
                          rate:  e.g.for ~orbit rate: 3564 89us 11245khz
                             For _bc signals:
                             number: distance between 2 signals in BCs (0: 40.079 mhz)
                             number followed by: us(usecs) ms(milisecs) s(secs)   
                                                 hz khz mhz
                          N: number of signals
                            0   -disable,i.e. USRena bit cleared also
                                 (TTena will not be set at next start)
                            inf -infinite number (0xffffffff)
                          i: -> immediate action, i.e. change rate/n during active emulation 
                          P: -> connect/disconnet to PULSER input
    SIGNAME P             -connect/disconnect SIGNAME to PULSER (ecl input). I.e.
                           the resulting rate will be a sum of:
                           - the rate choosen by 'SIGNAME rate N' +
                           - the rate of signals coming over the PULSER input
                           'P' option is allowed only for signals:
                           PH_bc PH_rnd PP_bc CAL_bc TPC_SYNC_orbit TPC_RST_bc
SIGNAME is one of: %s
Note about TOF: TOF can be generated also with a BCmask (OR) see: h bcm

=== Usefull:
outcnts r   -check number of output signals generated from last reading.
             Signals counted BEFORE tdomux in stdalone mode!
tdomux      -check/set Trigger Data Output mux (should be emu or emutrd)

emu ttc     -get help for oldTTC
emu hbr     -get help on 'rate' meaning for HBr_orbit signal
fpga init   -init LTU board

"""%(" ".join(list(self.signames.keys()))))
    elif part=='oldttc':
      if self.tnttc==None:
        detbusybit= "-"
      else:
        modebits= (self.ttcctrl&0xf000)>>12
        #detbusybit= str((self.ttcctrl&0x2)>>1)  #<v16_00
        detbusybit= str((self.ttcctrl&0x10)>>4)
        busysrcbit= str((self.ttcctrl&0x20)>>5)
      print("""
=== oldTTC detectors:    
    ttc    status [bits]  -show status. bits: show all bits in
                           ttc.statusbits/errorbits registers
           MODE           -MODE: one of: ttcpon oldttc trd cpv (now:%d)
           busy e[nable]  -enable busy input
                d[isable] -disable busy input (ENABUSY bit in CTRL reg, now:%s)
                l[vds]    -busy source (ctrl[5] bit): LVDS or PON, now:%s
                p[on]
           reset          -counters reset + FIFO clear
           reset sticky   -reset sticky bits (i.e. also TTC-[AB12]tick bits)
           init [ms]      -TTCinit (i.e.: ttcrx reset,sleep(ms), set fine delays,
                           activate B-channel (0xf9 in ctrl. reg 3), resetBC)
           nobch [ms]     -deactivate B channel (i.e.: ttcrx reset,sleep(ms),
                           write 0x99 to TTCrx control register 3)
           feecmd hdr data -fee cmd word sent over B-channel
                           hdr: 12..15  (default:12)
                           data: 12 bits data (default 0x0)
           llprot VALUE   -L0-L1 protection in BC (min 8)
                           L0-L0 protection in case of TRD
           l1reject VALUE -reject rate, e.g. 3: 1/3*100= 33.33%% (0:no reject)
           calrate VALUE  -e.g.: 2:each 3rd PH trigger is CAL trigger
                           TRD calibration rate
           delay [A|B|1|2] D -set delay for TTC-A, TTC-B, LVDS1 or LVDS2.
                           D is one of d 0 1 2 3 4 5  
                             d -disable channel
                             0..5 -set delay in 1/6 BCs
                             Values in ttc.masks register:
                             0: channel disabled
                             1:2/6BC 2:3/6BC 4:4/6BC 8:5/6BC
                             0x10:no delay 0x20:1/6BC 
           delay A B 1 2   set all 4 dealys in one go. E.g. disable all:
                          emu ttc delay d d d d
           ctrl [0xABC0]  -16 low bits. A:mode (0:ttcpon,1:ttc,2:trd,3:cpv)
                           B: 1:disable XOX over LVDS/TTC -used by TRD+CPV
                           C: BUSY source+disabled/enabled
                              0: lvds+disabled    1: lvds+enabled
                              2:  pon+disabled    3:  pon+enabled
                           0: always 0 (controlled by 'reset [sticky]' command)
           defaults       -set llprot,delays,calrate and ctrl to default values
                           (ltuIP_ADDR.cfg parameters: LLprotection, DETMODE)
"""%(modebits,detbusybit,busysrcbit))
    elif part=='hbr':
      print("""HBr_orbit 'rate' register contains a value in interval 0..256.
Notes:
1. Outside of the run (EMU not active): all 256 HB triggers in each TF
   are HBr triggers (i.e. also a HB with TF trigger)
2. During an active run (i.e. between SOX/EOX, when EMU active)
   TF trigger is always accepted, i.e. never accompanied with HBr bit
   Note: in v13_00 the HB following a TF-HB is accepted
3. HBr are generated at the end of TF, i.e.:
   T00....1T00....1   -for rate=255 (T: TF, 0:HBa   1:HBr)
   T00...11T00...11   -for rate=254
   T01...11T01...11   -for rate=2
4. Number of HBr triggers in a TF: 256-rate, i.e.:

      triggers in TF
rate  HBr    HBa
----  ----------------------------
   0  256    0          (see Rule 2. above)
   1  255    1
   2  254    2
   3  253    3
 ...  ...
 254  2      254
 255  1      255
 256  0      256
 257  0      256
For convenience, use hbaccept command to set HBr_orbit register:
hbaccept N%
where N is a percentage of HB triggers accepted (0..100). E.g.:
hbaccept 25%  sets HBr_orbit rate register to 64.
""")
    elif part=='cal':
      print("""Calibration status/control:
emu cal status   show status. 3 possible modes:
                 pp    -only PP generated
                 ppcal -PP trigger followed by CAL trigger
                 cal   -only CAL triggr generated
    cal delay N  -set start of the protection time when PH triggers 
                  are disabled
                  N: BCs before PP trigger when PH protection starts
                     till CAL trigger generation
    cal pp       -set mode 'pp only'
    cal ppcal    -set mode 'pp followed by cal trigger'
    cal cal      -set mode 'cal trigger only'
    cal off      - no PP neither CAL
    cal phreject on|off
                 off: PH triggers allowed before CAL trigger
                 on: PH triggers disabled 'delay' BCs before PP
    cal rate [N] -set the distance in BCs between 2 calib. requests
                  i.e. set PP_bc_rate(pp or ppcal mode) or
                  CAL_bc_rate(in cal mode) register.
                  N: limit the number of signals (inf defaut)
Notes:
PP/CAL triggers are sent in beforehand defined BC by PP/CAL BCmask
If corresponding BCmask is not set (3564L), triggers 
are not generated.

See also:
bcm PP           -check/set PP/CAL BCmask for PP/CAL generation
    CAL
""")
    else: 
      print("help text not available for ",part)

