#!/usr/libexec/platform-python

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import range
from builtins import object
from past.utils import old_div
import time #,pdb
import hexdump,colors
from .partdef import getN4tdimux
from auxfuns import isinargs,printError
def bit_not(n, numbits=32):
    return (1 << numbits) - 1 - n

class Sysmon(object):
  """
FPGA monitoring. 
ug580-ultrascale-sysmon.pdf, atb: SYSMONE1
ADC: 10 bits, 0.2msps
['temperature', 'vccaux', 'vccint', 'vccbram', 'alarms']
0:   a042   9a07   5134   511b   8100  8000 ? 0100: Vuser0 upper
                                 8100 -always ON
                intermitent:     8182  0002: VCCINT upper+lower
                                 818a  0008:VCCBRAM upper+lower
                                 ..8.  OR of ...7f (7 bitov) ok
                                 .1..  Vuser0 upper+lower
alarms: see p.90 22 
        from 9_02 should read 0
vccint  -feeding fpga core
vccbram -feeding bram memory blocks inside fpga core
   aux  -aux voltage

alarms: not sticky (in fpga ver 9_02) ! -i.e. difficult to catch if intermittent
"""
  def __init__(self, board, devname):
    self.ipsl= "ctrl"
    self.ipslstatus= "ctrl.status"
    self.devname= devname
    self.board= board
    self.regnames= ["temperature", "vccaux", "vccint", "vccbram", "alarms"]
    self.lturst= {"pll":3, "outcnts":4, "ttccnts":5,"icnts":6,"":7,"":8, "":9}
    self.ctprst= {"pll":3, "outcnts":4, "gcnts":6,"emus":7,"ir":8, "irsm":9}
    self.regs= [0, 0, 0, 0, 0]
    #ctrl= self.board.Read("ctrl.ctrl")
    #self.Write("ctrl.ctrl", 0)  let's keep (global) mode
    #ctrl2= self.board.Read("ctrl.ctrl")
    #if (ctrl2 & 0x3) == 0:
    #  print colors.color("ctrl.ctrl: 0x%x, After 0 written:0x%x -ok"%(ctrl, ctrl2), "fg_green")
    #else:
    #  print colors.color("ctrl.ctrl: 0x%x, After 0 written:0x%x -error (0 expected)"%(ctrl, ctrl2), "fg_red")
    # common for ltu/ctp:
    self.ipsloi= "ctrl.ctrl"   # for orbitin cmd
    self.orbitinbit=5       # v19_00 ctrl.ctrl ORBIT_sel: 0 int, 1 ext
    self.resetbit=2
    self.ipsloio= None  # is in ctrl.ctrl[19:8] also in ctp_v8
    if self.board.isfiwa((self.board.FW_CTP_LOGIC,)):
      self.orbitset(1)
      print("fpga: CTP board, external orbit is used.")
    elif self.board.isfiwa(self.board.FW_LTU):
      if self.board.isfiverhe(0x19):
        ctrl2= self.board.Read("ctrl.ctrl")
        gsint= (ctrl2>>4) & 0x1
        print("LTU Mode: %s ctrl.ctrl:0x%x"%(("stdalone","global")[gsint], ctrl2))
      elif self.board.isfiverhe(0x17):
        self.orbitinbit= 6     # bit in ctpemu.ctrl
        self.resetbit= 5
        self.ipsloi= "ctpemu.ctrl"
        self.ipsloio= "ctpemu.ORBIT_offset"   # for orbitin cmd (orbit offset reg)
      else:
        pass
      detn= getN4tdimux(self.board.bname)   # 1
      if detn==None: 
        self.orbitset(1)
        print("""fpga: $CTP3_DBCTP/ctp.net not available, using det1 for global.
Orbitin: internal.
""")
        
        detn=1
      else:
        self.orbitset(1)
        print("""fpga: $CTP3_DBCTP/ctp.net: tdimux value: %d (1,2,4-> det1,2,3)for global.
Orbitin: external."""%detn)
      self.myctp123= detn   # 1 2 4 -> det1 det2 or det3
    else:
      pass   # not CTP neither LTU board
  def finishltuinit(self):   # applied only for LTU
    # this could not be done in time of __init__ (td?mux was not yet created)
    #print "FINISHINIT"
    self.tdimux= self.board.atbdevs["tdimux"]
    self.tdomux= self.board.atbdevs["tdomux"]
    if isinargs("emuinit"):
      self.setStdaloneMode()
  def finishctpinit(self):   # applied only for CTP
    self.tdomux= self.board.atbdevs["tdomux"]
  def fpgainit(self):
    """ 
ltu: set stdalone + emu init + hbmb init
ctp: all emus init
rc: 0,"" -ok
"""
    if self.board.isfiwa(self.board.FW_LTU):
      rcmsg= self.setStdaloneMode()
      if rcmsg[0]!=0: return rcmsg
      self.board.atbdevs["emu"].emuinit()   # includes also bcms init
      rcmsg= self.board.atbdevs["hbmb"].hbmbinit()
    else:
      maxemu= 18
      for ix in range(1,maxemu+1):
        self.board.atbdevs["emu%d"%ix].emuinit()
      rcmsg= (0,"")
    return rcmsg
  def readAll(self):
    for ix in range(len(self.regnames)):
      rn= self.ipsl+".fpga_"+self.regnames[ix]
      self.regs[ix]= self.board.Read(rn)
  def tempDec(self, adc16):
    # p.40: 10 MSB, on-chip reference
    return old_div(((adc16>>6)*501.3743),1024)-273.6777
  def powDec(self, adc16):
    # p.41: 16 bits:
    return 3*adc16/65536.
  def gettemp(self):
    rn= self.ipsl+".fpga_"+self.regnames[0]
    self.regs[0]= self.board.Read(rn)
    return self.tempDec(self.regs[0])
  #def set10(self, regname, bit): see board.Set10
  #  reg= self.board.Read(regname)
  #  reg= reg | (1<<bit)
  #  self.board.Write(regname, reg)
  #  reg= reg & (~(1<<bit))
  #  self.board.Write(regname, reg)
  #  print "%s bit%d set to 1 and 0"%(regname,bit)
  def prtTempVolts(self):
    self.readAll()
    #hexdump.hd_dump(self.regs, 0, len(self.regs)-1)
    temp= self.tempDec(self.regs[0])
    txc= "fg_green"
    if temp>48: txc= "fg_red"
    line= "fpga Temp:%8.4fC Volts:"%temp
    for ix in range(1,4):
      line= line + "%8.4f"%self.powDec(self.regs[ix])
    print(colors.color(line, txc))
  def orbitset(self, enadis, shift=None):
    """
enadis: 1: enable external orbit 0: disable  None: do not enable neither disable
shift: 0..3563 set orbit shift
       3564 reset orbit counter
reset to 0: orbitset(None, 3564)
"""
    # valid for fi>=v17_00
    if shift!=None:
      if shift==3564:
        self.board.Set10(self.ipsloi, self.resetbit)
      else:
        if self.ipsloio:   # older versions
          self.board.Write(self.ipsloio, shift)
        else:
          self.board.SetBits(self.ipsloi, shift, 0xfff, 8)
    if enadis==1:
      self.board.SetBit(self.ipsloi, self.orbitinbit)
    elif enadis==0:
      self.board.ResetBit(self.ipsloi, self.orbitinbit)
    else:
      pass
  def orbitshow(self):
    disena= (self.board.Read(self.ipsloi)>>self.orbitinbit)& 1
    disenatx= ("disabled","enabled")[disena]
    if self.board.isfiverhe(0x19):   # also ctp
      shift= (self.board.Read("ctrl.ctrl")>>8 ) & 0xfff
      print("Orbitin %s, shift: %d"%(disenatx, shift))
    elif self.board.isfiverhe(0x17):
      shift=  self.board.Read(self.ipsloio)
      print("Orbitin %s, shift: %d"%(disenatx, shift))
    else:
      print("Ext. orbit not available for fi < ltu_v17_00")
  def getMode(self, prt=False):
    """ return: 0|1,""   0: stdalone 1:global  "" -no error
                8   "err message"
"""
    errmsg= ""
    sgint= (self.board.Read(self.ipsloi)>>4) & 0x1
    muxi= self.tdimux.getmux()
    muxis= self.tdimux.getmuxsym(muxi)
    muxo= self.tdomux.getmux()   
    muxos= self.tdomux.getmuxsym(muxo)
    if sgint==0:   # std
      if (muxi!=0) or (muxo!=self.tdomux.tdodef):
        errmsg= "Bad tdimux (%d expected:%d) or tdomux (%d expected:%d)"%\
          (muxi, 0, muxo, self.tdomux.tdodef)
    else:          # global
      if (muxi!=self.myctp123) or (muxo!=8):
        errmsg= "Bad tdimux (%d expected:%d) or tdomux (%d expected:%d)"%\
          (muxi, self.myctp123, muxo, 8)
    if prt:
      modetx= ("stdalone","global")[sgint]
      if errmsg=="":
        print("LTU Mode: %s, tdimux:%d(%s) ok, tdomux:%d(%s) ok"%\
          (modetx, muxi,muxis, muxo,muxos))
      else:
        print("LTU mode: %s, "%modetx+errmsg)
    if errmsg: sgint= 8
    return (sgint,errmsg)
  def setGlobalMode(self,runn=0xffffffff):
    rcmsg= (0,"")
    if self.board.isfiwa(self.board.FW_LTU):
      if self.checkemu():
        rcmsg=(8, "CTP emulation started, change to global not done")
      else:
        # set tdi/tdomux global:
        #tdimux= self.board.atbdevs["tdimux"]
        self.tdimux.setmux(self.myctp123)
        #tdomux= self.board.atbdevs["tdomux"]
        #tdomux.setmux(8)
        gval= self.tdomux.setGlobal()
        # ORBIT_OFFSET_OLDTTC -to be set differently for s/g in LTU, does not exist in CTP
        self.board.SetBits(self.ipsloi, 0xdb5, 0xfff, 20)
        self.board.SetBit(self.ipsloi, 4)
        #print "dbg tdimux not set:", self.myctp123
        #emu= self.board.atbdevs["emu"]
        #emu.ttc_setgs(1)
        self.setDport(runn)
        print("stdalone->global done (tdimux:%s, tdomux:%s). Use tdimux cmd if detX not correct."%\
          (self.tdimux.getmuxsym(self.myctp123), self.tdomux.getmuxsym(gval)))
        # read counters BeforeSOX (also in ctpemu.start)
    else:
      print("needs LTU...")
      rcmsg= (12,"needs LTU...")
    return rcmsg
  def setStdaloneMode(self):
    rcmsg= (0,"")
    if self.board.isfiwa(self.board.FW_LTU):
      #tdimux= self.board.atbdevs["tdimux"]
      self.tdimux.setmux(0)   # rawonu data (3 detectors)
      # read counters AfterSOX (also in ctpemu.stop)
      self.tdomux.setStdalone()
      # ORBIT_OFFSET_OLDTTC -to be set differently for s/g in LTU, does not exist in CTP
      self.board.SetBits(self.ipsloi, 3564-61, 0xfff, 20) # =0xdaf from ltu_v1d_0c
      self.board.ResetBit(self.ipsloi, 4)
      if self.checkemu():
        errmsg="global->stdalone done. CTP emulation was/is started (internal error?)"
        rcmsg=(0, errmsg)
      else:
        print("global->stdalone done. tdimux:onu tdomux:CTPemu")
        #emu= self.board.atbdevs["emu"]
        #emu.ttc_setgs(0)
      self.setDport(0xffffffff)   # ltud active/no run started
    else:
      print("needs LTU...")
      rcmsg= (12,"needs LTU...")
    return rcmsg
  def setDport(self, dport):
    # dport (value in ctrl.pid_1):
    # LTU board:
    # -run number (global or stdalone)
    #    0: run not active ltud OFF (emu could be started by qtltu/atb)
    #    0xffffffff run not active (as 0) ltud ON
    # CTP board:
    # 0: ctpd OFF
    # >0 only meaningfull is 0xffffffff: ctpd ON
    self.board.Write("ctrl.pid_1", dport)
  def getDport(self):
    return self.board.Read("ctrl.pid_1")
  def checkemu(self):
    emu= self.board.atbdevs["emu"]
    if emu.emuactive():
      return True
    else:
      return False
  def excmd(self, cmd):
    if len(cmd) < 2:
      self.prtHelp()
      return
    if cmd[1]=='rr':
      reps= 1
      if len(cmd) > 2:
        reps= eval(cmd[2]) 
      print(self.regnames)
      for rep in range(reps):
        self.readAll()
        hexdump.hd_dump(self.regs, 0, len(self.regs)-1)
        line= "%8.4f"%self.tempDec(self.regs[0])
        for ix in range(1,4):
          line= line + "%8.4f"%self.powDec(self.regs[ix])
        print(line)
    elif cmd[1]=='reset':
      rn= self.ipsl+".rst" ; bit= None
      if len(cmd) <= 2:
        self.prtHelp()
      else:
        if cmd[2]=='ipbus':
          bit= 0 ; rn= self.ipsl+".ctrl"
        elif cmd[2]=='soft':
          bit= 1 ; rn= self.ipsl+".ctrl"
        elif cmd[2]=='i2c':
          bit= 10 ; rn= self.ipsl+".rst"
        elif cmd[2]=='hbmb':
          bit= 11 ; rn= self.ipsl+".rst"
        elif (cmd[2] in self.lturst) and self.board.isfiwa(self.board.FW_LTU):
          # todo: ctrl.rst allows simultaneous resets (all resets in 1 register)
          bit= self.lturst[cmd[2]] 
        elif (cmd[2] in self.ctprst) and self.board.isfiwa((self.board.FW_CTP_LOGIC,)):
          if ((self.board.fwinfo>>8)&0xffff)>=0x9:
            bit= self.ctprst[cmd[2]] 
          else:
            bit= None
        else:
          print("reset %d not allowed."%cmd[2])
          return
        if bit==None:
          print("reset %s not available before ctp_v9_xx"%cmd[2])
        else:
          self.board.Set10(rn, bit)
          print("1+0 written in bit %s[%d]."%(rn, bit))
    elif 'global'.find(cmd[1])==0:
      rcmsg= self.setGlobalMode()
      printError("Be aware: ltud puts run number (run being loaded) in ctrl.pid_1")
      print(rcmsg)
    elif 'stdalone'.find(cmd[1])==0:
      rcmsg= self.setStdaloneMode()
      print(rcmsg)
      printError("Be aware: ltud puts 0xffffffff in ctrl.pid_1")
    elif 'init'.find(cmd[1])==0:
      rcmsg= self.fpgainit()
      print(rcmsg)
    elif 'status'.find(cmd[1])==0:
      errmsg= ""
      dport= self.getDport()
      if self.board.isfiwa(self.board.FW_LTU):
        sgint,errmsg= self.getMode(prt=True)
        #print "ltud port: %d"%(dport)
        if dport==0xffffffff:
          print("ltud ON, no active run (ctrl_pid1:0x%x)"%dport)
        elif dport>0:
          print("ltud ON, active run (ctrl_pid1:%d)"%dport)
        else:
          print("ltud OFF (ctrl.pid_1: %d)"%(dport))
      else:
        if dport==0xffffffff:
          print("ctpd ON (dport:0x%x)"%dport)
        else:
          print("ctpd OFF? ctrl.pid_1:0x%x"%dport)
        muxos= self.tdomux.getmuxsym()
        errmsg= "tdomux: %s"%muxos
        print(errmsg)
      # ORBIT_OFFSET_OLDTTC -to be set differently for s/g in LTU, does not exist in CTP
      ooo= (self.board.Read(self.ipsloi)>>20)&0xfff
      print("ORBIT_OFFSET_OLDTTC: 0x%x"%ooo)
      # todo valid for ctp_vc_10: ctrl.ir_state_machine_codes[11..0]
      # it should be 1 (other states last only for 1 or 2 clock cycles.
      errTick= (self.board.Read(self.ipslstatus)>>2) & 0x1
      print("errTick:%d (not sticky), see also 'gcnts checktick' "%(errTick))
      self.orbitshow()
    elif 'orbitin'.find(cmd[1])==0:
      if len(cmd) > 2:
        enadis=shift= None
        if 'enable'.find(cmd[2])==0:
          enadis= 1   # ext. orbit
        elif 'disable'.find(cmd[2])==0:
          enadis= 0
        elif 'reset'.find(cmd[2])==0:
          shift= 3564
        else:
          shift= int(cmd[2])
          if (shift>=0) and (shift<=3563):
            enadis= 1  # ext. orbit with offset
          else:
            shift= None
            print("Error: orbit shift must be 0..3563")
        self.orbitset(enadis, shift)
        self.orbitshow()
      else:
        self.orbitshow()
    else:
      self.prtHelp()
  def prtHelp(self):
    """ Print help, consists of 2 parts:
-common for LTU+CTP
-LTU-specific or CTP-specific
"""
    if self.board.isfiwa(self.board.FW_LTU):
      specifichlp="""
reset ttccnts   -reset oldTTC counters
reset icnts     -reset icnts counters

init            -initialize board i.e. stdalone + init (ltuIP.cfg) emu,bcms,hbmb.
                 'emuinit' arg performs the same action.
global          -set global mode
stdalone        -    stdalone mode
"""
    else:   # CTP
      specifichlp="""
reset gcnts     -(*) reset gcnts counters
reset emus      -(*) reset emu* state machines + all scalers
                 (before global run)
reset ir        -(*) reset IR buffer
reset irsm      -(*) reset IR state machine

(*) -available from ctp_v9_00
"""
    print("""
status
orbitin         -show ext. orbit status
orbitin enable | disable 
                -enable/disable ext. orbit
orbitin 0..3563 -set the shift for ext. orbit + enable it
orbitin reset   -reset orbit counter to 0
reset ipbus     -hard ipbus reset ctrl.ctrl[0]: 0-> all IPBUS registers
reset soft      -soft ipbus reset ctrl.ctrl[1]  
reset i2c       -write 1+0 to ctrl.rst[10]  
reset hbmb      -write 1+0 to ctrl.rst[11]  
reset pll       -PLL load needed after
reset outcnts   -reset all outcnts counters
%s
rr [rep]        -show %s regs
                 (raw data), repeat rep times
"""%(specifichlp, str(self.regnames)))
