from __future__ import print_function
from builtins import object
import sys,traceback
#!/usr/bin/python3
import ipbuswrap,symbreg
from atb import counters
class Gbt(object):
  AVAGO=0x100  # bit9(0x200) on: force Data header bit to 1 for both receivers
  VTRX=0x2100  # 
  def __init__(self, SFPinst, gbtn=1, dest=None):   # dest: vtrx sfp sfpbsr
    # experimental:
    #self.board= atbwrap.OpenATBoard(SFPinst.b.board.bname) ?
    self.sfpinst= SFPinst
    self.board= SFPinst.b.board
    if gbtn==0:
      self.slave="gbt"
    else:
      self.slave="gbt%d"%gbtn
    self.gbtn= gbtn
    self.dest= dest
    #self.slave="gbt"   # v5 has just 1 slave "gbt"
    self.ok= False
    if self.board.nodExists(self.slave, True):
      #self.init()
      self.reg_status= symbreg.Register("status",
        TxAlignedComp=0,
        TxAlignedBank=1,
        MgtReady=2,
        rxBitSlips=(3,8),
        FrameReady=11,
        ErrDetected=12,
        RxDataF=13,
        GbtRxReady=17,
        GbtBankTxReady=18)
      self.reg_txc= symbreg.Register("txctrl",
        TxShift=(0,8),
        TxEnc= 8,      # 0:WIDE_BUS, 1: FRAME (we use FRAME for itsmft and ctp)
        Txls=9,        # 0:idle header pattern, 1: data hp
        ClkMuxSel=12,
        GbtTxPol=13)   # 0: normal, 1: Inverted (vtrx)
      self.reg_rxc= symbreg.Register("rxctrl",
        rxEnc=0, 
        rxBitSlipRst=4,
        Loopback=(5,3),
        GbtRxPol=8)    # 0: normal, 1: Inverted (vtrx)
      #print "%s -> %s"%(self.slave, dest)
    else:
      return
    if dest != None: self.initdest(dest)
    #self.showstatus()
    if self.board.nodExists(self.slave+".Word_count", False):
      #print "Word_count exists, counters available (only in itsmft firmware?)"
      self.cnts= counters.CountersGBT(self.board, self.slave)
    else:
      self.cnts= None
      print("gbt: %s.Word_count does not exist."%self.slave)
    self.ok= True
  def initdest(self, dest):
    #print "Gbt.initdest:",dest
    try:
      rc= self.board.Read(self.slave+".txctrl")
      self.exprxctrl= 0x1
      if dest=="vtrx":
        if rc != self.VTRX:
          print("%s -> vtrx: setting .txctrl register to: 0x%x"%(self.slave, self.VTRX))
          self.board.Write(self.slave+".txctrl",self.VTRX)  
        else:
          print("%s -> vtrx: .txctrl register ok: 0x%x"%(self.slave, self.VTRX))
      elif dest=="sfp" or dest=="sfpbsr":
        if rc != self.AVAGO:
          print("%s -> %s: setting .txctrl register to: 0x%x"%(self.slave, dest, 
            self.AVAGO))
          self.board.Write(self.slave+".txctrl",self.AVAGO)  
        else:
          print("%s -> %s: .txctrl register ok: 0x%x"%(self.slave, dest, self.AVAGO))
        if dest=="sfpbsr":
          self.exprxctrl= 0x11
        else:
          self.exprxctrl= 0x1
        rc= self.board.Read(self.slave+".rxctrl")
        if rc== self.exprxctrl:
          print("%s -> %s: .rxctrl register ok: 0x%x"%(self.slave, dest, 
            self.exprxctrl))
        else:
          print("%s -> %s: setting .rxctrl register to: 0x%x"%(self.slave, dest, 
            self.exprxctrl))
          self.board.Write(self.slave+".rxctrl",self.exprxctrl)  
      else:
        print("%s -> %s ???"%(self.slave, dest))
    except:
      traceback.print_exc(file=sys.stdout)
  def init(self, loopback=0):
    dest= self.dest
    # config tx for avago/vldb: 0x100
    # another option: cofig tx for vtrx/vldb: 0x2100
    if dest=="sfp":
      receiver= self.AVAGO
      deste="sfp, downstream+upstream"
    elif dest=="vtrx":
      receiver=self.VTRX
      deste="vtrx, downstream only"
    else:
      print("Error: %s init: bad destination: "%self.slave, dest)
      return
    print("%s: sending to %s."%(self.slave,deste))
    self.board.Write(self.slave+".txctrl",receiver)  
    rc= self.board.Read(self.slave+".txctrl")
    if rc != receiver:
      print("%s Init error txctrl: 0x%x instead 0x%x"%(self.slave, rc, receiver))
    # loopback: 0: default (no loopback)
    # r3doc/GBT_in_ATB_v5.pdf
    #https://www.xilinx.com/support/documentation/user_guides/ug576-ultrascale-gth-transceivers.pdf (Pag 86)
    # 
    self.board.Write(self.slave+".rxctrl",(self.exprxctrl|(loopback<<5)))  
    rc= self.board.Read(self.slave+".rxctrl")
    if rc != self.exprxctrl:
      print("Warning: %s Init rxctrl: 0x%x instead 0x%x"%(self.slave, rc, self.exprxctrl))
    self.reset()
    rc= self.board.Read(self.slave+".status")
    if dest=="sfp":
      if rc != 0x60807:
        print("%s Init error (status): 0x%x instead 0x60807"%(self.slave,rc))
      else:
        print("%s Init (SFP) ok"%self.slave)
    elif dest=="vtrx":
      if (rc&0x41003) != 0x41003:
        print("%s Init error (status): 0x%x (0x41003 bits expected)"%(self.slave,rc))
      else:
        print("%s Init (VTRX) ok"%self.slave)
    else:
      print("%s Init Error: "%self.slave,dest)
  def reset(self, bit="cpu"):
    if bit=="cpu": 
      msk=0x1
    elif bit=="rx": 
      msk=0x4
    elif bit=="tx": 
      msk=0x2
    else:
      print("Error: %s, only cpu rx or tx resets available", bit)
      return
    # write only, read no sense, rr= self.board.Read(self.slave+".resets")
    rw1=  msk
    self.board.Write(self.slave+".resets",rw1)  # ->1
    rw2= 0
    self.board.Write(self.slave+".resets",rw2)  # ->0
    print("gbt reset 0x%x 0x%x"%(rw1,rw2))
  # todo: check + repair gbt bidirectional connection:
  # if status not stable:
  #   olt reset    -on both sides!
  #   if status not ok (0x60807), but stable:
  #     gbt N reset rx
  #   elif status not stable:
  #     gbt N reset cpu
  def settxrx(self, gv):   # "sfp" or "vtrx"
    if gv=="sfp":
      self.board.Write(self.slave+".txctrl",self.AVAGO)  
    elif gv=="vtrx":
      self.board.Write(self.slave+".txctrl",self.VTRX)  
    elif gv=="lb0":
      rx= self.board.Read(self.slave+".rxctrl")
      rx= (rx & 0xffffff1f ) # no loopback
      self.board.Write(self.slave+".rxctrl", rx)
    elif gv=="lb1":
      rx= self.board.Read(self.slave+".rxctrl")
      rx= (rx & 0xffffff1f ) | 0x20  # internal loopback 1
      self.board.Write(self.slave+".rxctrl", rx)
    elif gv=="bsr0":   # rxBitSlipRst off
      rx= self.board.Read(self.slave+".rxctrl")
      rx= (rx & 0xffffffef )
      self.board.Write(self.slave+".rxctrl", rx)
    elif gv=="bsr1":   # rxBitSlipRst on
      rx= self.board.Read(self.slave+".rxctrl")
      rx= (rx | 0x10 )
      self.board.Write(self.slave+".rxctrl", rx)
    else:
      print("fp or vtrx expected...")
      return
    tx= self.board.Read(self.slave+".txctrl")
    self.reg_txc.prt(tx)
  def status(self):
    st= self.board.Read(self.slave+".status")
    tx= self.board.Read(self.slave+".txctrl")
    rx= self.board.Read(self.slave+".rxctrl")
    return st,tx,rx
  def showbitdefs(self, bits=None):
    if bits==None:
      self.reg_status.prtme()
      self.reg_txc.prtme()
      self.reg_rxc.prtme()
    else:
      self.reg_status.prt(bits)
      self.reg_txc.prt(bits)
      self.reg_rxc.prt(bits)
  def showstatus(self,detailed=True, hdr=False):
    st,tx,rx= self.status()
    #self.reg_status.prtme()
    if detailed:
      self.reg_status.prt(st, "%s(%s)"%(self.slave,self.dest))
      self.reg_txc.prt(tx)
      self.reg_rxc.prt(rx)
    else:
      if hdr:
        print("%5s(%4s): %8s %8s %8s"%("gbtN","dest", "status","txctrl","rxctrl"))
      print("%5s(%4s): %8x %8x %8x"%(self.slave,self.dest, st,tx,rx))
  def prthelp(self):
    print("""
status
init
reset [cpu|rx|tx] -reset GBT link
                   'reset cpu': is propagated in fi to ALL GBT channels
bits [pattern]    -show bits defs
sfp | vtrx        -destination opt.component (sfp/bidirectional  or
                   vtrx/downstream)
lb0 | lb1         -loopback off or set loopback1 in .rxctrl
bsr0 | bsr1       -BitSlipRst bit off/on in .rxctrl

cnts              -examine counters (if available in .xml)

Note: 127ns latency
tx: latency optimazed rx: standard mode (all ITSMFT links receiver not used anyhow)
    the same for gbt1 (readout) link
""")    
  def excmd(self,cmd):
    if len(cmd)<=1:
      self.prthelp()
      return
    if cmd[1]=='status':
      self.showstatus()
    elif cmd[1]=='reset':
      bitx="cpu"
      if len(cmd)>2:  # gbt N reset rx
        bitx= cmd[2]
      self.reset(bitx)
      print("%s reset done."%self.slave)
    elif cmd[1]=='bits':
      if len(cmd)>2: bits= eval(cmd[2])
      else: bits= None
      self.showbitdefs(bits)
    elif cmd[1]=='init':
      self.init()
    elif (cmd[1] == 'sfp') or (cmd[1] == 'vtrx') or (cmd[1]=='lb0') or\
         (cmd[1] == 'lb1') or (cmd[1] == 'bsr0') or (cmd[1]=='bsr1'): 
      self.settxrx(cmd[1])
    elif cmd[1]=='cnts':
      if self.cnts!=None:
        rc= self.cnts.excmd(cmd[1:], stdprint="rc")
        print("%5s: %s"%(self.slave, rc))
      else:
        print("GBT module not inserted or %s slave does not supply counters"%self.slave)
    else:
      self.prthelp()
#!/usr/bin/python3
#!/usr/bin/python3
