#!/usr/libexec/platform-python
from __future__ import print_function
from builtins import input
from builtins import str
from builtins import range
from builtins import object
import sys,string,time,os
import ipbuswrap,auxfuns
import ttcpon.netfile as netfile
import ilogger
from auxfuns import printError

#!/usr/bin/python3
def read_ipbus_mem(node, ixlow, ixhigh):
  mem=ReadBlock(node, ixhigh-ixlow, ixlow)
  return mem

class HBMB(object):
  pos_map= "hbm_present_pos_map.data"
  hppm= "hbm_present_pos_map.data"   # 64*9 x 18bits
  nolts= 9
  nonus= 64
  NOT_PRESENT_F=0x200
  MASKED_F= 0x400
  TRD_F= 0x800
  TRD_SHIFT= 12
  HBM_MAP_ADDR_F= 0x1ff
  c_error= 0xff
  c_not_configured= 0x1
  c_configured= 0x8           # 32 bits
  c_idle_rx_link_err_mem= 0   # 3 bits (but only 2 bit from ctrl[15..14]
  pos_set_clear_rx_link_err_mem= 4
  pos_set_read_rx_link_err_mem= 5
  disable_error_message_not_connected_cru= 10
  """ hbm_present_pos_map item bits (self.pos_mapd:
 8..0 address of hbm map %3d   ONUaddress                         AAA
 9 = '1' CRU not present            0:.  1:T(TRD) or P(not TRD)   X
         i.e. 0x200 for not present  <- this is used for hbmbinit()
              0x400 present but ignored in hbmb?
10 = '1' CRU is masked in hbm_map   0:' ' 1:m                     M
11 = '1' CRU is TRD      see 9
17 .. 12 is TRD address of TRDbusyMapRegister (0..35)  %2d        YY
         YYMXAAA
         00..000
         11.T000
         01mP122
tips for 'hbmb show':
olt1: TP...............
      M

oltrx_lec bits:
15     0x8000 -rx_cnt_received  (was response)
14..8         -rx_monitor_cnt   (as received from the link)
 7..0         -rx_link_err      error count does not roll over

the length of the error code data here:
https://gitlab.cern.ch/aliceCTP3/ctp3-fw-common_logic/-/blob/HBM_builder/RTL/HBM_builder/pack_hbm_builder.vhd#L193
and copied here:
constant error_auto_configure_check_cru_not_connected_lines_to_read : integer:= 3;
constant error_cru_not_connected_but_data_valid_lines_to_read : integer:= 2;
constant error_hbf_ack_out_of_time_lines_to_read : integer:= 2;
constant error_hbf_ack_read_write_count_inconsistency_lines_to_read : integer:= 2;
constant error_hbf_count_between_tf_not_as_expected_lines_to_read : integer:= 2;
"""
  def __init__(self, board, devname="hbmb", trd=False):
    self.board= board
    self.devname= devname
    self.trd= trd
    self.printdbg= False
    self.emu= self.board.atbdevs["emu"]
    if self.emu.DETMODE==3:
      print("emuDETMODE:", self.emu.DETMODE, " self.trd forced to 3 (was:%d)..."%self.trd)
      self.trd= True
    self.Run= None    # run not started
    self.hcs= "hbm_control_status"          # 0..3 control, 4..7 status (ro)
    self.hcsregs= ("run_control", "numb_hbf_timeout", "count_hbm_ack_threshold",
      "hbf_count_between_tf_expected",
      "state_machine_codes",
      "trd_busy_reg_low", "trd_busy_reg_high", "write_error_address_pointer")
    self.smc= ("reset", 
      "not_configured",    # after sw reset
      "auto_configure_clear_cru_not_connected_masked",
      "auto_configure_fill_cru_not_connected_masked",
      "auto_configure_count_cru_not_connected_masked",
      "auto_configure_check_cru_not_connected",
      "auto_configure_clear_hbm_ram",
      "auto_configure_clear_error_ram",
      "configured",  # 8
      "hc_start",
      "hc_active",
      "hc_receive",
      "hc_end",
      "run_start",
      "run_active",   # 14
      "run_receive",
      "run_end")   #=0x10, error = 0xFF"/>
    self.errcodes= ("no_error",
      "auto_configure_check_cru_not_connected", #1
      "cru_not_connected_but_data_valid",
      "hbf_ack_out_of_time",
      "hbf_ack_read_write_count_inconsistency",
      "hbf_count_between_tf_not_as_expected")
    self.errlen= (None, 3, 2, 2, 2, 2)
    self.herm= "hbm_error_mem.data"          # 256 x 32bits ?
    self.hhcm= "hbm_hc_mem.data"             # 64*9 x 10bits
    self.hlecm= "oltrx_lec.data"             # 64*9 x 10bits
    # 1. 1 or 2:
    #self.pos_mapd= self.create_nc_pos_map()
    #print "empty pos_mapd created (not read from the board)."
    # 2.
    self.pos_mapd= self.readhw_block()
    self.validmap= self.calctrdmap()
    print("hbmb: pos_mapd read from the board.")
    #if self.trd:    -no need, done with each emu.start for trd
    #  if auxfuns.isinargs("emuinit"):
    #    self.hbmbinit("trd.net")
  #def clear_pos_map(self):
  #  dw= self.onu_nc()
  #  for ix in range(self.nolts*self.nonus):
  #    self.pos_mapd[ix]= dw
  #  self.board.WriteBlock(self.hppm, self.pos_mapd)
  def report(self, str):
    if self.printdbg: print(str)
  def create_nc_pos_map(self):
    pos_mapd= []
    dw= self.onu_nc()
    for iolt in range(self.nolts):
      for ionu in range(self.nonus):
        pos_mapd.append(dw)
    return pos_mapd
  def readhw_block(self, node=hppm, length=None):
    if length==None:
      length= self.nolts*self.nonus
    print("reading %d words in %s..."%(length,node))
    rc= self.board.ReadBlock(node, length)
    return rc
  #
  def onu_nc(self):   # not connected ONU
    if self.trd:
      #dw= self.NOT_PRESENT_F | self.TRD_F    does not work
      # i.e. if not present: TRD_f not set! even for TRD detector
      dw= self.NOT_PRESENT_F
    else:
      dw= self.NOT_PRESENT_F
    return dw
  def loadnet(self, net=""):
    """ net: 
"" -load empty config -i.e. disable all
file.net -read file into memory, update in hw
"""
    if net=="" or net=="disable":
      posmapd= self.create_nc_pos_map()   # all ONUs not connected by default
    elif net[-4:]!=".net":
      self.tbconfig_hbm_present_pos_map(net)
      return(8)
    else:
      logger= ilogger.ilogger()
      nf= netfile.NetFile(net, logger)
      if nf.waserror: return(8)
      print("%s read in."%net)
      posmapd= self.create_nc_pos_map()   # all ONUs not connected by default
      olts= nf.getolts()   # e.g: [['olt1', '13', 'atb_55:55'], ['olt2', ' ']]
      # todo: replace onuseqn by a number form interval reserved for each detector
      onuseqn= 0   # 0,1,...,Number_of_all_onus-1.
      trdseqn= 0   # 0,1,...35  -bit number in 
      for olt in olts:
        #print olt
        oltn= int(olt[0][3])-1   # 0,1,...,8
        noo= len(olt)-1          # Number Of Onus for olt
        if (noo==1) and (olt[1]==' '): continue # OLT without CRUs
        ixbase= oltn*self.nonus
        for ix in range(noo):
          onuinfo= olt[ix+1].split(':')
          if len(onuinfo)==1:
            onuaddr= int(onuinfo[0])
          else:
            onuaddr= int(onuinfo[1])
          if self.trd:
            dw= (trdseqn<<self.TRD_SHIFT) | self.TRD_F | onuseqn
            trdseqn= trdseqn+1
          else:
            dw= onuseqn
          #posmapd[ixbase+ix]= dw
          posmapd[ixbase+onuaddr]= dw
          onuseqn= onuseqn+1
    self.update_pmd(posmapd)
    return 0
  def update_pmd(self, posmapd):
    self.pos_mapd= posmapd
    self.validmap= self.calctrdmap()
    self.writehw()
  def writehw(self, posmapd=None):
    if posmapd==None: posmapd= self.pos_mapd
    print("updating %s..."%self.hppm)
    self.board.WriteBlock(self.hppm, posmapd)
  def hbmbinit(self,netname=None):
    """
netname: "" -noTRD: disable all CRUs (loadnet("")
             TRD:`  load trd.net (if exists)
         None: do not touch hbmb pos_mapd
rc: (0,""): ok
"""
    rcmsg= [0,""]
    if netname!=None: 
      if netname=="" and self.trd:
        cfgdir= os.getenv("CTP3_DBCTP")
        if os.path.exists(os.path.join(cfgdir,"nets/trd.net")):
          netname= "trd.net"
      if self.loadnet(netname)!=0:
        return 16,"hbmb: cannot load "+str(netname)
      else:
        rcmsg[1]= "hbmb: '{}' loaded".format(netname)
    if self.swreset() != self.c_not_configured:
      return 8,"hbmb: swreset did not succeed"
    if self.configure() != self.c_configured:
      return 12, "hbmb: configure did not succeed"
    self.clear_rx_link_err_mem()
    if self.rxle("start")!=0:
      return 20,"hbmb: rxle start did not succeed"
    return rcmsg
  def trdstop(self):
    self.rxle("stop")
  def calctrdmap(self):
    """rc: 0 if error found in self.pos_mapd
"""
    bits36= 0   # 36 bits
    errs= 0
    for ix in range(len(self.pos_mapd)):
      dw= self.pos_mapd[ix]
      if (dw & self.NOT_PRESENT_F)==self.NOT_PRESENT_F:   # notpresent
        continue
      self.report("calctrdmap: %d 0x%x"%(ix, dw))
      if (dw & self.MASKED_F)==self.MASKED_F: 
        continue
      if (dw & self.TRD_F)==0:           # not TRD
        errs= errs+1
        continue
      trdbit= (dw>>self.TRD_SHIFT) & 0x3f
      if trdbit>35:
        printError("CRU address %d > 35, ignored"%trdbit)
        errs= errs+1
        continue
      bits36= bits36 | (1<<trdbit)
    if errs>0:
      printError("hbmb calctrdmap: errs:%d"%errs)
      bits36= 0   # 0xfffffffff
    return bits36
  def getBusy(self):
    low32= self.rctrl(5)   # has sense after hbmbinit() ?
    high3= self.rctrl(6)&0x3
    bits= (low32 | (high3<<32)) & self.validmap
    return bits   # 0 if not busy
  def show1(self,oltn,onun, prt="all"):
    dw= self.pos_mapd[(oltn-1)*self.nonus + onun-1]
    addr= dw&self.HBM_MAP_ADDR_F
    st= "%d "%(addr)
    if dw & self.NOT_PRESENT_F:
      st= st+"NOT_PRESENT "
      present= '.'
    else:
      if dw & self.TRD_F:
        st= st+"TRD "
        present= 'T'
      else:
        st= st+"nonTRD "
        present= 'P'
    if dw & self.MASKED_F:
      st= st+"MASKED "
      msk= 'm'
    else:
      msk= '.'
    trdaddr= (dw>>12)&0x3f
    st= st+ "TRDaddr:%d"%(trdaddr)
    asci= "%2.2d%c%c%3.3d"%(trdaddr, msk, present, addr)
    if prt=="all":
      print("olt%d onu%d:0x%x = %s %s"%(asci,oltn, dw, onun,st))
    return asci
  def show(self, details=""):
    """ show memory
details:
oltn onun -print info about 1 CRU 
d -2 lines/olt i.e. hexa + ascii
         7825       19 . .
      07.T005  00.P025 . .
a -1 line/olt i.e. ascii only
      07.T005  00.P025 . .
x - 1 line/olt hexadecimal

"""
    if self.trd: 
      print("TRD LTU, validmap: 0x%x -36bits, '1' for each active+not masked CRU"%(self.validmap))
    else:
      printError("not TRD LTU.")
    #for ix in range(self.nolts*self.nonus):
    if len(details.split())==2:
      oo= details.split()
      oltn= int(oo[0]) ; onun= int(oo[1])
      if oltn>0 and oltn<10 and onun>0 and onun<65:
        self.show1(oltn,onun)
      else:
        print("Error: hbmb show OLT ONU expected (OLT:1..9, ONU:1..64)")
      return
    for oltn in range(self.nolts):
      onuseqn= -1   # 0..63 consecutive hmb map addr for each olt
      curonus= 0
      detailed1= "     "
      detailed2= "     "
      detailedx= "     "
      maxperolt= 3
      olterrs= 0
      validlen= len(detailed1)
      for onun in range(self.nonus):
        onuseqn=onuseqn+1 
        dw= self.pos_mapd[oltn*self.nonus + onun]
        if dw&self.NOT_PRESENT_F:
          detailed1= detailed1+". "
          detailed2= detailed2+". "
        else:
          detailed1= detailed1+"%8x "%dw
          detailed2= detailed2+" %s "%self.show1(oltn+1, onun+1, "asci")
          validlen= len(detailed1)
        detailedx= detailedx+"%8x "%dw
        if dw & self.NOT_PRESENT_F:   # check if ONU present
          #break     # when checking only till first 'empty one'
          continue   # check all...
        # check 
        curonus= curonus+1
        err= ""
        if False: # self.trd:
          if not (dw & self.TRD_F):
            err= "TRD_F not set"
        #if (dw & self.HBM_MAP_ADDR_F) != onuseqn:
        #  err= "seqN: 0x%x exp.: 0x%x"%((dw & self.HBM_MAP_ADDR_F), onuseqn)
        if err!="":
          if olterrs<maxperolt:
            print("Error: olt%d onu%d: 0x%x %s"%(oltn+1, onun+1, dw, err))
          olterrs= olterrs+1
      if details=="all" or curonus>0 or details=='x':
        print("olt%d: %d"%(oltn+1, curonus))   # numb. of present onus
      #if validlen==None:
      # detailed
      if curonus>0 or details=='x':
        if details=="d":
          print(detailed1[:validlen])
          print(detailed2[:validlen])
        elif details=="x":
          print(detailedx)
        elif details=="a":
          print(detailed2[:validlen])
        else:
          pass
  def excmd(self, cmd):
    #import auxfuns.cmdarg as cmdarg
    #from auxfuns import cmdarg  #-ok, but does not see global cmd
    def cmdarg(n, default=""):
      if len(cmd)>n: 
        return(cmd[n])
      else:
        #print "default arg:", default
        return(default)
    if len(cmd) < 2:
      self.prtHelp()
    elif cmd[1]=='dbgon':
      self.printdbg= True
    elif cmd[1]=='dbgoff':
      self.printdbg= False
    elif 'load'.find(cmd[1])==0:
      self.loadnet(cmdarg(2, "test.net"))
    elif 'gethw'.find(cmd[1])==0:
      self.pos_mapd= self.readhw_block(self.hppm)
      self.validmap= self.calctrdmap()
    elif 'status'.find(cmd[1])==0:
      self.status(cmdarg(2, ""))
    elif 'show'.find(cmd[1])==0:
      #self.show(cmdarg(2, ""))
      self.show(" ".join(cmd[2:]))
    elif 'tb'.find(cmd[1])==0:
      self.tb1125(cmdarg(2, "0"))
    elif 'reset'== cmd[1]:
      self.swreset()
    elif 'run'== cmd[1]:
      if len(cmd)>2:
       if ('start'.find(cmd[2])==0) and (self.Run==None):
         self.startRun()
       elif 'stop'.find(cmd[2])==0:
         self.stopRun()
       else:
         print("Run started:", self.Run)
    elif 'configure'.find(cmd[1])==0:
      self.configure()
    elif 'clear'.find(cmd[1])==0:
      self.clear_wea()
    elif 'rem'.find(cmd[1])==0:
      n= None
      if len(cmd)>2: n= int(cmd[2])
      self.read_error_mem(n)
    elif 'rxle'.find(cmd[1])==0:
      # hbmb rxle c[lear] | r[ead] | stop
      # hbmb rxle start [stopus]
      if len(cmd)>2:
        if "clear".find(cmd[2])==0:
          self.clear_rx_link_err_mem()
        elif "start".find(cmd[2])==0:
          if len(cmd)>3:
            self.rxle("start", int(cmd[3]))
          else:
            self.rxle("start")
        elif "stop".find(cmd[2])==0:
          self.rxle("stop")
        elif "read".find(cmd[2])==0:
          self.read_rx_link_err_mem()
        else:
          printError("Expected: hbmb rxle c[lear] | r[ead] | start [stopus] | stop")
    elif 'hbmbinit'.find(cmd[1])==0:
      rc= self.hbmbinit(cmdarg(2,None))
      print("rc:%d"%rc)
    elif cmd[1]=='busy':
      print("validmap: 0x{:x} input busys:0x{:x}".format(self.validmap,self.getBusy()))
    elif 'busyout'.find(cmd[1])==0:
      self.busyout(cmdarg(2, "e"))
    elif cmd[1]=='setc':
      #hbmb setc 0 0x20
      # 0    1   2   3
      if len(cmd)<4:     # show
        if len(cmd)<3:   # show all
          rr= list(range(8))   # 8 regs
        else:
          creg= cmd[2]
          if creg.isdigit() and int(creg)>7:
            printError("reg addr (0..7) expected)")
            return
          creg= int(creg)
          rr= list(range(creg, creg+1))
        # show value of control reg
        for ix in rr:
          if ix>3:
            rwro="ro"
          else:
            rwro="rw"
          val= self.rctrl(ix)
          print("%d: 0x%8x %s (%s)"%(ix, val, self.hcsregs[ix], rwro))
        if len(cmd)<3:   # show all
          self.prtHelp("setc")
      else:
        creg= int(cmd[2])
        if creg>3:
          printError("reg addr (0..3) and value (0x... or decimal) expected)")
          return
        self.wctrl(creg, eval(cmd[3]))
    else:
      self.prtHelp()
  def prtHelp(self, part=None):
    if part==None:
      print("""hbmb cmmands (dbgon|dbgoff:{}):
load f.net      -read CFG/nets/f.net config, check consistency with 
                 plugged OLTs and update hw with new config
                 or read WORK/file.txt. E.g.:
     hbm.txt    = hbm_present_pos_map_few_cru_active.txt
     hbmerr.txt = hbm_present_pos_map_error.txt
     disable    -disable all (0x200)
gethw           -read hw (hbm_present_pos_map) to memory
show            -show pos_map memory content (hw not read)
show d          -detailed print (2 lines/olt: hexa+ascii)
     a          -detailed print (1 line/olt: ascii)
     x          -raw hexa
     oltN onuN  -show specific pos_map oltN/onuN item
tb [0|1|2]      -tb_hbm_builder.vhd lines 1125.. (reset_i 1/0/1)
                 0: hbm_present_pos_map_few_cru_active.txt
                 1: hbm_present_pos_map_error.txt + '0'
                 2: do not load hbm_present_pos_map from file (use the current one in hbmb)
reset           -sw reset
config          -set hbm_builder to configured (ctrl0=2)
run start 
    stop
rem N           -read_error_mem() N words (1 item usually 2 words)
clear           -clear write_error_address set 0x40 bit to 1/0 in ctrl0

rxle clear       clear link error counters -usually before start.
rxle start [us]  start link error counters monitor, stop after us (do not stop by default)
                 Indicated in ctrlreg0(0x20) -if must be ON. OFF: hbmb output is not valid.
rxle stop        stop link error counters monitor.
rxle read        read link error counters. Usually after 'stop'. Can be read also
                 'on the fly' between start /stop commands.

hbmbinit [name]  load (trd.net) + reset + configure + rxle clear + rxle start
                 started before SOX (see ctpemu.py emuinit, start)
busyout 1|0|t   -force hbmb busy out to 1, 0 or toggle
        e[nable]-enable hbmb busy out (i.e. normal operation)
setc [reg]      -show control reg
setc reg value  -write value to hbm_control_status register (reg: 0..3).
busy            -show 36 bits -hbmb inputs
""".format(self.printdbg))
    elif part=="setc":
      print(""" run control/status registers:
0: run_control
  0x001 -sw reset (1+0)
  0x002 -set configured
  0x010 -set_clear_rx_link_err_mem (1+0)
  0x020 -set_read_rx_link_err_mem 1: hbmbOUT reflects upstream inputs
                                  0: hbmbOUT does not reflect upstream inputs
  0x040 -set_clear_write_error_address  (1+0)
  0x008 -start run, stays ON while run_active 
        -0: stop run (ctrl4: expected to go to c_configured)
  0x380 -value toggle forced
         xx0 -normal operation
         001 -forced to 0
         x11 -forced to toggle
         101 -forced to 1
  0x400 -disable_error_message_not_connected_cru (normal operation:1)
  0x800 -enable_reading_non_connected_olt (normal operation:0)
         1: all OLTs, registered in present_pos_map will be read.
            Non registered but responding CRUs nevertheless will be
            flagged in the hbm_error_memory
         0: OLT data which have no registered CRUs in the present_pos_map
            will not be read even if their corresponding strobe signal
            is active.This is to avoid that if non-populated OLTs 
            create floating inputs to the TTCpon FW and consequently
            by random active strobe data are read by the HBMB
1: numb_hbf_timeout
2: count_hbm_ack_threshold (bs_:31..16 hback_: 15..0)
3: hbf_count_between_tf_expected
""")
    else:
      print("help text not available for ",part)      
  # used by tb1125:
  def rctrl(self, adr):
    """ read control reg. adr: 0..3: rw 4..7:ro
"""
    slv= self.hcs+"."+self.hcsregs[adr]
    return self.board.Read(slv)
  def wctrl(self, adr,val):
    slv= self.hcs+"."+self.hcsregs[adr]
    if adr>3:
      printError("wctrl(%d,... forbiden, addr>3"%adr)
    else:
      self.board.Write(slv,val)
  def wctrlbitmask(self, adr,bitmask, onezero=1):
    cr= self.rctrl(adr)
    if onezero==1:
      ncr= cr| bitmask
    else:
      ncr= cr & (~bitmask)
    self.wctrl(adr, ncr)
    self.report("wctrlbitmask: %d: 0x%x -> 0x%x"%(adr,cr,ncr))
  def busyout2txt(self, val):
    """ val: 3 bits value
"""
    if val&7==0x1:
      msg= "forced to 0"
    elif val&3==0x3:
      msg= "forced to toggle"
    elif val&7==0x5:
      msg= "forced to 1"
    elif val&1==0:
      msg= "normal operation"
    else:
      msg= "strange operation: %d"%val
    return msg
  def busyout(self,e01):
    """ set hbmb busy out bits in ctrl0
'e' -enable
'1' -force to 1
'0' -force to 0
't' -force to toggle
i.e. ctrl0 reg bits 87:
x0 -normal operation
01 -forced to 0
11 -forced to 1

from 1d_0f:
i.e. ctrl0 reg bits 987:
xx0 -normal operation
001 -forced to 0
x11 -forced to toggle
101 -forced to 1

"""
    ctrl0= self.rctrl(0)
    if e01=="0":
      val= 1   # val= 0x1
    elif e01=="1":
      val= 0x5   #val= 0x3
    elif "enable".find(e01)==0:
      val= 0
    elif "toggle".find(e01)==0:
      val= 3
    else:
      printError("e[nable] t[oggle] 1 or 0 expected")
      return
    msg= self.busyout2txt(val)
    ctrl0new= (ctrl0 & 0xfffffc7f) | (val<<7)
    self.wctrl(0, ctrl0new)
    print("ctr0:0x%x -> 0x%x -busy out %s"%(ctrl0,ctrl0new,msg)) 
  def getsymsc(self, sc=None):
    if sc==None: sc= self.rctrl(4)
    stco= sc&0x1fff
    if stco < len(self.smc):
      symsc= self.smc[stco]
    elif stco== 0xff:
      symsc= "error"
    else:
      symsc= "unknown state code"
    return "%d=%s"%(stco,symsc)
  def status(self, arg=""):
    run_control= self.rctrl(0)
    val= (run_control>>7) & 0x7          # 9..7  forced_value force_toggle force_toggleORvalue
    busyo= self.busyout2txt(val)
    #errdisabled= (run_control>>9) & 0x1
    errdisabled= (run_control>>10) & 0x1 # 10:disable_error_message_not_connected_cru
    # from 1d_0e:
    #ernc= (run_control>>10) & 0x1       # 10 enable_reading_non_connected_olt
    ernc= (run_control>>11) & 0x1        # from ltu_v1d_0f
    #
    hbf_timeout= self.rctrl(1)
    thrs= self.rctrl(2)
    hbfintf= self.rctrl(3)
    wem= self.rctrl(7)
    #sc= self.rctrl(4)&0x1fff
    symsc= self.getsymsc()
    print("""
run_control:0x%x   busyout:%s   disable_error_message_not_connected_cru:%d
  enable_reading_non_connected_olt:%d   (from ltu_v1d_0e) 
numb_hbf_timeout:%d
count_hbm_ack_hback_threshold & count_hbm_ack_bs_threshold:0x%x
hbf_count_between_tf_expected:%d
write_error_mem:%d
state_code:%s
"""%(run_control, busyo, errdisabled, ernc, hbf_timeout, thrs, hbfintf, wem, symsc))
  def swreset(self):
    #self.report(" send SW reset_cmd (0:1, 0:0)...")
    # seems following not ok (1/0 value i.e. all bot have to be overwritten)
    #self.wctrlbitmask(0, 1)
    #self.wctrlbitmask(0, 1, 0)   # release  reset_cmd
    self.wctrl(0, 1)
    self.wctrl(0, 0)   # release  reset_cmd
    #rc= self.rctrl(0)
    state_code= self.rctrl(4) & 0x1fff
    # this sw reset not working with ltu_v1d_04 (status_code: 0 here, 1 expected)
    # ok with ltu_1d_05
    if (state_code!= self.c_not_configured):
      self.report("state_code after sw reset: %d != 1, try 'fpga reset hbmb' before"%state_code) 
    else:
      self.report("state_code after sw reset: %d, ok"%state_code)
    self.wctrl(0, 0x400)
    self.report("ctrl0: 0x400 -do not report error for not connected CRUs")
    return state_code
  def configure(self):
    """
Notes: during this, error_mem (see read_error_mem) is cleared sometimes ???
"""
    self.report(" set hbm_builder to configured (0:2)...")
    self.wctrlbitmask(0, 0x402)
    #wait for 5_000 ns;
    #-- poll status register showing present state and wait for error state or config state
    #-- firmware counts number of active CRUs in cru_not_connected_masked in hbm_present_pos_map file
    #-- and verifies that no duplicate address pointers are active
    #-- in case of discrepancy --> state machine goes to error and stays there
    #self.report(" start polling status register and wait for c_configured or c_error...")
    state_code=0xffffffff ; iloops= 0
    while (state_code!= self.c_error) and (state_code!=self.c_configured):
      status_value= self.rctrl(4)
      state_code= status_value & 0x1fff
      #wait for 5_000 ns;   #1230
      time.sleep(0.000005)
      iloops= iloops+1
      if iloops>100: 
        self.report(" timeout(loops:%d) waiting for c_configured or c_error"%iloops)
        break
    if state_code == self.c_configured:
      self.report(" configuration successful")
      #self.report("Setting: disable_error_message_not_connected_cru")
      #wctrlbitmask(self, 0,0x400, onezero=1):
    elif state_code==self.c_error:
      self.report("configuration reports error, read write_err_address pointer... ")   #1281
      write_error_address= self.rctrl(7) & 0xff   # hbm_error_mem_address_width= 8
      self.report(" write_error_address (cs7): 0x%x"%write_error_address)
      #todo: read_error_mem (ipb_in_error_mem, ipb_out_error_mem, write_error_address);
      self.read_error_mem();
    else:
      self.report(" Internal error: state_code:%d %s"%(state_code,self.getsymsc(state_code)))
    return state_code
  def tbconfig_hbm_present_pos_map(self,fname):
    """ read file to self.pos_mapd and update in hw
rc: self.validmap  (i.e. 0 in case of error)
"""
    #wkdir= os.getenv('CTP3_WORK_DIRECTORY')
    lines= auxfuns.getfile(fname)
    if lines==None: return 0
    print("tbconfig_hbm_present_pos_map %d lines read from file:%s"%(len(lines),fname))
    posmapd= self.create_nc_pos_map()
    rows= 0
    for iline in range(3, len(lines)):
      #print lines[iline]
      ws= lines[iline].split() ; row= ""
      for ir in range(self.nonus):
        hv= eval("0x"+ws[ir])
        row= row+"%x"%hv+' '
        posmapd[rows*self.nonus + ir]= hv
      print(row)
      rows= rows+1
      if rows>=9: break
    self.update_pmd(posmapd)
    return self.validmap
  def clear_wea(self):
    #-- set set_clear_write_error_address to '1'
    #-- sets error write pointer to 0, practically clears memory
    self.report(" set set_clear_write_error_address to '1' ")
    #write_control_reg (addr_control0,x"00000040",ipb_in_status_control); 
    self.wctrlbitmask(0, 0x40)
    #wait for 100 ns;
    #-- release set_clear_write_error_address to '1'
    self.report(" set set_clear_write_error_address to '0' ")
    #write_control_reg (addr_control0,x"00000000",ipb_in_status_control); 
    self.wctrlbitmask(0,0x40, 0)

    #wait for 100 ns;
  def read_error_memDELME(self, write_error_address):   # 8 bits 
    """ write_error_address: address of highest occupied memory cell in error_mem
Note: the memory is cleared by self.configure()

called 2x:
1. after hbm_pos_map configuration error
2. after run stopped

Seems 1st word is never checked, and the last one is checked 3 times
reading 6 words in hbm_error_mem.data...
loop_read_error_mem: 1
loop_read_error_mem: 2
loop_read_error_mem: 3
ipb_in_error_mem_int_ipb_addr herm[2]:0x41
loop_read_error_mem: 4
ipb_in_error_mem_int_ipb_addr herm[3]:0x0
loop_read_error_mem: 5
ipb_in_error_mem_int_ipb_addr herm[4]:0x0
loop_read_error_mem: 6
ipb_in_error_mem_int_ipb_addr herm[5]:0x0
loop_read_error_mem: 6
ipb_in_error_mem_int_ipb_addr herm[5]:0x0
loop_read_error_mem: 6
ipb_in_error_mem_int_ipb_addr herm[5]:0x0

atb_35>hbmb reset
state_code after sw reset == 1, ok
atb_35>hbmb configure
 set hbm_builder to configured (0:2)...
 start polling status register and wait for c_configured or c_error...
 configuration successful
atb_35>hbmb rem
reading 4 words in hbm_error_mem.data...
loop_read_error_mem: 1
loop_read_error_mem: 2
loop_read_error_mem: 3
ipb_in_error_mem_int_ipb_addr herm[2]:0x0
loop_read_error_mem: 4
ipb_in_error_mem_int_ipb_addr herm[3]:0x0
loop_read_error_mem: 4
ipb_in_error_mem_int_ipb_addr herm[3]:0x0
loop_read_error_mem: 4
ipb_in_error_mem_int_ipb_addr herm[3]:0x0

"""
    #275
    active_error= 0
    read_error_mem_active= 0
    read_error_mem_active_s1= 0
    read_error_mem_active_s2= 0
    lines_to_read= 0
    loop_read_error_mem= 0
    wea= self.rctrl(7) & 0xff   # hbm_error_mem_address_width= 8
    print(" write_error_address (cs7): 0x%x required:%d..."%(wea,write_error_address))
    print("read_error_mem: reading %d words..."%(write_error_address+1))
    herm= self.readhw_block(self.herm, write_error_address+1)   # ??? 256
    while (loop_read_error_mem <= write_error_address) or (read_error_mem_active_s1 == 1):
      #308 ??? wait until (clk_ipbus'event and clk_ipbus='1');
      read_error_mem_active_s2= read_error_mem_active_s1
      read_error_mem_active_s1= read_error_mem_active
      if loop_read_error_mem <= write_error_address:
        read_error_mem_active= 1
        ipb_in_error_mem_int_ipb_addr= loop_read_error_mem
        loop_read_error_mem= loop_read_error_mem+1
      else:
        read_error_mem_active= 0
      #print "loop_read_error_mem:",loop_read_error_mem
      if read_error_mem_active_s2 == 1:
        #read_error_mem_data := ipb_out_error_mem_int.ipb_rdata;
        # ???
        read_error_mem_data= herm[ipb_in_error_mem_int_ipb_addr]
        print("ipb_in_error_mem_int_ipb_addr herm[%d]:0x%x"%(ipb_in_error_mem_int_ipb_addr, read_error_mem_data))
        #--- following repeated for ERN 1,2,3,4,5
        for ERN in (1,2,3,4,5):
          if read_error_mem_data == ERN and lines_to_read==0:
            self.report(self.errcodes[ERN])
            lines_to_read= self.errlen[ERN]
            active_error= ERN
          elif lines_to_read != 0 and active_error== ERN:
            lines_to_read= lines_to_read-1
            self.report("from CRU:"+str(read_error_mem_data&0x7fffffff))
        #---
  def read_error_mem(self, write_error_address=None):   # 8 bits 
    wea= self.rctrl(7) & 0xff   # hbm_error_mem_address_width= 8
    if write_error_address==None:
      write_error_address=wea
    print(" write_error_address (cs7): 0x%x required:%d..."%(wea,write_error_address))
    if write_error_address<=1:
      print("hbm_error_mem empty.")
      return
    print("read_error_mem: reading %d words..."%(write_error_address+1))
    herm= self.readhw_block(self.herm, write_error_address+1)   # ??? 256
    addr= 0
    print("ADDR E      CRU [CRU2] Error")
    while (addr <= write_error_address) and (addr<wea):
      errcode= herm[addr+0]
      if (errcode >= len(self.errcodes)) or (errcode==0):
        printError("Unknown error code:%d at error_mem[%d]"%(errcode, addr))
        return
      lines_to_read= self.errlen[errcode]
      if (addr+lines_to_read)>write_error_address: break
      cru1= herm[addr+1]
      if lines_to_read == 2:
        print("%3d: %d %8d %s"%(addr, errcode, cru1, self.errcodes[errcode]))
      elif lines_to_read == 3:
        cru2= herm[addr+2]
        print("%3d: %d %8d %d %s"%(addr, errcode, cru1, cru2, self.errcodes[errcode]))
      else:
        printError("Internal error: read_error_mem to big lines_to_read for at error_mem[%d]"%addr)
        return
      addr= addr+lines_to_read
  def read_rx_link_err_mem(self):
    self.report(" reading rx_link_err_mem ")
    #
    # loop over cru_pos_active[9]64] each word 24 bits
    # address_read_ipbus_mem_low <=  0;
    # address_read_ipbus_mem_high <=  2**rx_link_err_mem_addr_width-1;  2**10-1
    # read_ipbus_mem (ipb_in_rx_link_err_mem,                     seems oltrx_lec
    #  ipb_out_rx_link_err_mem,                                   0
    #  address_read_ipbus_mem_low, address_read_ipbus_mem_high,   1023
    #  read_ipbus_mem_data);                                      read here
    # read_ipbus_mem_data[1024] of 32 bits
    read_ipbus_mem_data= self.readhw_block(self.hlecm)
    for i in range(self.nolts):
      for j in range(self.nonus):
        #if cru_pos_active(i)(j)(cru_not_connected_pos) = '0' then -- CRU connected bit9
        cru_pos_activeij= self.pos_mapd[i*64+j]
        dataij= read_ipbus_mem_data[i*64+j]
        if (cru_pos_activeij & 0x200)== 0:
          #if (read_ipbus_mem_data(i*64+j)(pos_rx_cnt_received) = '0') then --  but no response
          if (dataij & 0x8000 == 0):
            self.report(" CRU%d_%d connected but no response"%(i,j))
          else:
            self.report(" CRU%d_%d connected errors:%d rx_count:%d"%(i,j,dataij&0xff, (dataij>>8)&0x7f))
        else:
          if (dataij & 0x8000 == 1):
            self.report(" CRU%d_%d not connected but responded with errors:%d rx_count:%d"%\
              (i,j,dataij&0xff, (dataij>>8)&0x7f))
  def clear_rx_link_err_mem(self):
    """
@@ in order to start hc or run set_clear_rx_link_err_mem needs to be activated before
"""
    #1370  Clearing rx_link_err_mem
    self.report(" clear_rx_link_err_mem, 1+0 -> ctrl[0][bit4]...")
    self.wctrlbitmask(0,0x10)   # 0x10: 1<<pos_set_clear_rx_link_err_mem
    time.sleep(0.000001)   # 100ns in vhd
    self.wctrlbitmask(0,0x10, 0)
    # oltrxle is cleared now
    notneeded_here="""   see Email From alex from 25.3 before noon: .py should check only status_code bits 12..0
    self.report(" start polling status register and wait for s_idle_rx_link_err_mem ")
    state_code=0xffffffff ; iloops= 0
    while (state_code!= self.c_error) and (state_code!=self.c_idle_rx_link_err_mem):
      status_value= self.rctrl(4)
      # r14= pos_state_rx_cnt_min
      # m3= (pos_state_rx_cnt_max-pos_state_rx_cnt_min+1)**2-1 
      state_code= (status_value>>14) & 0x3
      #wait for 1_000 ns;   #1402
      time.sleep(0.000001)
      iloops= iloops+1
      if iloops>10: 
        self.report("timeout waiting for c_idle_rx_link_err_mem ")
        break
    self.report(" status_value:0x%x state_code = addr_status0(15 downto 14):0x%x"%(status_value,state_code))
"""
  def rxle(self, action="", stopus=None):
    """ Alex: mail from 10.5.2021:AT PRESENT rx_link_err_mem CAN ONLY BE ACTIVATED DURING S_CONFIGURE BUT NOT IN RUNS
action:
start -start acquiring. if stopus: stop after stopus us
stop  -stop acquiring
rc: ok
"""
    sc= self.rctrl(4)&0x1fff
    if sc !=8:
      self.report(" state_code:%s, but 8 expected. Link error counters reading abondoned.")
      return 8
    usecs= 5
    sc= self.getsymsc(sc)
    self.report(" state_code:%s."%sc)
    if action=="start":
      self.report(" set_read_rx_link_err_mem = 1")
      # @@ from this moment on trd_busy is acquired by the hbmb, irrespective of whether you start the run.
      self.wctrlbitmask(0,0x20)   # 1<<pos_set_read_rx_link_err_mem
      if stopus!=None:
        time.sleep(0.000001*stopus)
        self.report(" set_read_rx_link_err_mem = 0 after %dus"%stopus)
        self.wctrlbitmask(0,0x20,0)
    elif action=="stop":
      #-- stops to monitor rx_link_errors
      #--> stops link evaluating => @@@ AT PRESENT rx_link_err_mem CAN ONLY BE ACTIVAGTED DURING S_CONFIGURE BUT NOT IN RUNS
      self.report(" set_read_rx_link_err_mem = 0 ")
      self.wctrlbitmask(0,0x20,0)
      #--trd_busy is transmitted during rx_link error acquisition but arrival is not checked by tb
    else:
      self.report("start or stop expected.")
      return 16
    return 0
  def startRun(self):
    """
state_code after ctrl0 bit 0x8 set to start_run):  14=run_active
setsoxbits(SOCEOC): ttena:0xe14 -> 0xe14 0s:fffffe7f 1s:600
ctpemu start. Time: 1620585389.404239s, TTena set to 0xe14, EMU_run:1.
state_code after SOC sent:  14=run_active
waiting 368ms (max. time: (15+1)*23 15TF FEErst + 1TG SOX... 
state_code:  15=run_receive
"""
    self.wctrlbitmask(0,0x8)
    print()
    print("state_code after ctrl0 bit 0x8 set to start_run): ",self.getsymsc())
    #self.report("waiting 10ms... ")   #& time'image(now) ;
    #time.sleep(0.01)
    #wait for 100 ns;
    #-- in simulation simulated CTP sends SOx, ...
    start_run= 1
    self.emu.setsoxbits("SOCEOC")
    rc= self.emu.start()
    if rc!="":
      print(rc)
      return rc
    print("state_code after SOC sent: ",self.getsymsc())
    #--@@@**** implement automatic check for TRD busy and TRD busy reg
    #-- verifies that trd_busy is responding for simulation
    self.report("waiting 368ms (max. time: (15+1)*23 15TF FEErst + 1TG SOX... ")   #16*23= 368
    time.sleep(0.368)
    # perhaps todo: take into account 
    #  ctpemu.TTena[12]: 1- enabled + ctpemu.FEErst_tf_delay (max 15)
    print("state_code: ",self.getsymsc())
    print()
    self.Run= True
    return rc
  def stopRun(self):
    """
state_code after ctrl0 bit 0x8 cleared to stop_run:  15=run_receive
emu stop. TTena set to 0 (but TF_orbit XOX HBr_orbit bits kept), EMU_run:0
 start polling status register and wait for c_configured after EOx 
c_configured, ok after 5ms
 write_err_address pointer:4
 write_error_address (cs7): 0x4 required:4...
"""
    rc= 0
    print()
    #write_control_reg (addr_control0,x"00000000",ipb_in_status_control); 
    self.wctrlbitmask(0,0x8,0)
    print("state_code after ctrl0 bit 0x8 cleared to stop_run: ",self.getsymsc())
    #wait for 100 ns;
    #-- simulation sends EOx
    start_run= 0
    #self.report(" run stopped EOx sent via TTCpon down signals ")
    #wait for 1_000 ns;
    self.emu.stop()
    self.report(" start polling status register and wait for c_configured after EOx ")
    state_code= 0xffffffff ; iloops= 0 ; waitinterval= 5
    while (state_code != self.c_error) and (state_code != self.c_configured):
      #read_status_reg (addr_status0, status_value, ipb_in_status_control, ipb_out_status_control);
      status_value= self.rctrl(4)
      state_code= status_value & 0x1fff
      #state_code(pos_state_code_max-pos_state_code_min downto 0) <= status_value(pos_state_code_max downto pos_state_code_min);
      time.sleep(0.001*waitinterval) #wait for 5 ms;
      iloops= iloops+1
      if iloops>10: 
        self.report("timeout waiting %d loops (each %dms) for c_configured after EOx "%(iloops,waitinterval))
        return state_code
    waited= "%dms"%(iloops*waitinterval)
    if state_code == self.c_configured:
      self.report("c_configured, ok after %s"%waited)
    if state_code == self.c_error:
      self.report("c_error severity failure after %s"%waited)
    self.Run= None
    return state_code
  def tb1125(self, arg="0"):
    HBM_hc_mem_address_width= 10
    max_ipbus_size= 1024
    read_ipbus_mem_data= [0]*max_ipbus_size
    numb_ltu_olt= 9
    numb_max_olt= 64
    HBM_present_pos_map_data_width_sim=24  # bits
    state_codes= {self.c_error:"error", self.c_configured:"successful"}
    # hw reset:
    #1147 reset_i	<= '1', '0' after 10 ns, '1' after 35 ns;
    # fpga reset hbmb # ctrl.rst[11]
    run_config_hbm_present_pos_map_with_error= arg
    #-- 1st run writes configuration file with error into HBMb
    #-- 2nd run is reset by reset_cmd and writes correct file
    for ii in (1,2):    #1150
      if ((ii==1) and (run_config_hbm_present_pos_map_with_error=="1")) or (ii==2):
        #if ii==1:  no sw reset when pos_map with errors ???
        if ii==10:  # always sw reset (with/without errors) -seems this is ok
          pass
        else:
          if (self.swreset()!= self.c_not_configured): return
        if ii==1:
          self.report("config_hbm_present_pos_map with errors...")
          #config_hbm_present_pos_map(ipb_in_HBM_present_pos_map,trig_config_hbm_present_pos_map_finished,"./input_files/hbm_present_pos_map_error.txt", cru_pos_active);
          self.tbconfig_hbm_present_pos_map("hbm_present_pos_map_error.txt")
          # trig_config_hbm_present_pos_map_finished -out. signal
          # cru_pos_active[9][64] -each word 24 bits
        else:
          if run_config_hbm_present_pos_map_with_error=="0":
            self.report("config_hbm_present_pos_map without errors...")
            #config_hbm_present_pos_map(ipb_in_HBM_present_pos_map,trig_config_hbm_present_pos_map_finished,"./input_files/hbm_present_pos_map_few_cru_active.txt", cru_pos_active);
          
            self.tbconfig_hbm_present_pos_map("hbm_present_pos_map_few_cru_active.txt")
          else:
            self.report("Using current hbm_pos_map (no load from file)")
        self.report("set numb_hbf_timeout: (1:8)...")
        #write_control_reg (addr_control1,x"00000008",ipb_in_status_control); 
        self.wctrl(1, 8)
        #-- set count_hbm_ack_hback_threshold & count_hbm_ack_bs_threshold -- defines number of pos HBFack required to qualify for pos HBF decision messages
        self.report("set count_hbm_ack_hback_threshold & count_hbm_ack_bs_threshold (2:0x40002)...")
        #write_control_reg (addr_control2,x"00040002",ipb_in_status_control); 	
        self.wctrl(2, 0x40002)
        #-- set hbf_count_between_tf_expected -- sets number of HBF in one TF
        hbftf= 256
        self.report("set hbf_count_between_tf_expected (3:%d)..."%hbftf)
        #write_control_reg (addr_control3,x"0000000A",ipb_in_status_control);
        self.wctrl(3, hbftf)
        if (self.configure()!= self.c_configured): return
    #state_code=0xffffffff    #1249
    #1302 
    calibrate_ttcpon= 1
    #1335 (1267) run_hc: skipped
    self.clear_rx_link_err_mem()
    self.rxle("start", stopus=5)
    #wait for 5_000 ns;   #1412
    #self.rxle("stop")
    self.read_rx_link_err_mem()
    #time.sleep(0.000002)
    #1536 Starting run
    if self.startRun()!="": return
    #self.report("for trd_busy'event 100ms... ")   #& time'image(now) ;
    #wait until trd_busy'event;
    #self.report("reading trd_busy_reg0 at time : ")   # & time'image(now) ;
    #read_status_reg (addr_status1, status_value, ipb_in_status_control, ipb_out_status_control);
    reg1= self.rctrl(5)
    #wait for 1000 ns;
    #self.report("reading trd_busy_reg1 at time : ")   # & time'image(now) ;
    #read_status_reg (addr_status2, status_value, ipb_in_status_control, ipb_out_status_control);
    reg2= self.rctrl(6)
    self.report("trd_busy_reg_high low:0x%x 0x%x"%(reg2,reg1))
    time.sleep(0.00006) #wait for 60_000 ns;
    #-- Stopping run
    sc= self.stopRun()
    if sc != self.c_configured: return
    #-- verify whether errors have been reported 
    status_value= self.rctrl(7)
    write_error_address= status_value&0xff -1   # seems # of words to be read -1
    self.report(" write_err_address pointer:%d"%write_error_address)
    #-- read error memory 
    #read_error_mem (ipb_in_error_mem, ipb_out_error_mem, write_error_address);
    if write_error_address>10:
      self.report("hbm_error_mem.data not cleared, use rem %d for whole memory reading."%(write_error_address-1))
    else:
      self.read_error_mem()
    #wait for 33_000 ns;
    #self.clear_wea()
    time.sleep(0.000033)

if __name__ == '__main__':
  ipbuswrap.Init()
  board= ipbuswrap.Board("atb_2")
  #board= atbwrap.OpenATBoard("atb_2")
  hbmb= HBMB(board)
  if len(sys.argv) < 2:
    print("h -display help text")
    while True:
      line= input(">"); 
      if line=="": break
      cmd= string.split(line)   # cmd cmdname a1
      if cmd[0]=="q": break
      hbmb.excmd(['hbmb']+cmd)
  else:
    #hbmb.loadnet("test.net")
    hbmb.excmd(sys.argv)

