#!/usr/libexec/platform-python
from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
import os, string, sys, types
from ctpemu import check_emubatchcmd
from auxfuns import parseline,printError,getfile,isinargs
from atb.bcmasktxt import BCmask
from atb.bcmask import symname

#!/usr/bin/python3
#!/usr/bin/python3
LTUS=None

def getN4tdimux(ltuboardname):
  """rc: 1,2, or 4  (det1 det2 det3) for tdimux/global in LTU
None: no config available
""" 
  cfgdir= os.getenv("CTP3_DBCTP")
  if cfgdir==None: return None
  netcfg= getfile(os.path.join(cfgdir,"ctp.net"), errprint=False)
  if netcfg==None: return None
  linen= -1
  for ixl in range(len(netcfg)):
    #atb_64
    #1 atb_37 atb_38 atb_39
    #2 atb_40 atb_41 atb_43
    #3 atb_51 atb_50 atb_44
    #4 atb_52 # atb_53
    #5 atb_45 atb_46
    #6 atb_47 atb_48 atb_49
    line= parseline(netcfg[ixl])
    if line==None: continue
    if linen==-1:    # 1st line of netcfg
      linen= 0
      ctpbname= line[0]
      continue
    if not (line[0] in ('1','2','3','4','5','6')): continue   # gbtX...
    oltn= int(line[0])
    locemun= 1   # 1,2 or 3
    for ixbn in range(len(line[1:])):
      ltubname= line[ixbn+1]
      if ltubname==ctpbname:
        print("getN4tdimux MONONU for OLT%d, ignored"%oltn)
        continue
      #if ltu==None:
      #  print("Unknown board:", ltubname, " in ctp.net")
      #  continue
      emun= (oltn-1)*3 + locemun
      if ltubname==ltuboardname:
        # emun: 1..18  %3 -> 1,2,3  but bit for mux -> 1 2 4
        det124= ((emun-1) % 3)+1
        if det124==3: det124= 4
        return det124
      locemun= locemun+1
  return None
def findEmuN(ltuname):
  """ check if ltuname has a ttcpon connection to the ctp
rc: None
    1..18 in case connected
"""
  if ltuname in LTUS.ltus:
    if LTUS.ltus[ltuname]== None: return None
    return LTUS.ltus[ltuname].emun
  printError("Internal: findEmuN(%s)"%ltuname)
  return 0
class Ltu(object):
  def __init__(self, name, bname):
    self.name= name
    self.bname= bname
    self.emun= None   # 1..18, None: not present in ctp.net, i.e. ltu not wired/wired
    self.ltustatus= None # available, active (std/global?), locked (by?)
class Ltus(object):
  def __init__(self):
    """
I:
1. ctpboards.cfg
2. ctp.net
outcome:
self.ltus: {} of Ltu refs, key: ltu name
self.ltun: [] of Ltu refs ordered by emu number
self.ctpbname the name of ctp board:
     -the ctp-line in ctpboards.net is read, but ctpbname overwritten by
     -value in 1st line in ctp.net (if ctp.net exists)
"""
    #print("partdef.LTUS constructor")
    self.ctpbname= None   # ctpboards.cfg overwritten by ctp.net
    self.ltus={ "trd":None, "hmp":None, "phs":None, "emc":None, "cpv":None,
      "fv0":None, "ft0":None, "fdd":None, "its":None, "tof":None,
      "mft":None, "mid":None, "mch":None, "tpc":None, "zdc":None,
      "daq":None}
    cfgdir= os.getenv("CTP3_DBCTP")
    if cfgdir==None: return
    netcfg= getfile(os.path.join(cfgdir,"ctpboards.cfg"), errprint=False)
    if netcfg==None: return
    #print("$CTP3_DBCTP dir processing...")
    for ixl in range(len(netcfg)):
      #trd 45 c26t 101 102 lvpecl
      #spare -
      line= parseline(netcfg[ixl])
      if line==None: continue
      if len(line)<2: continue
      if line[1]=='0': continue
      name= line[0]
      if name in self.ltus:
        self.ltus[name]= Ltu(name, "atb_"+line[1])
      elif name == "ctp":
        self.ctpbname= "atb_"+line[1]
      else:
        print("strange LTU name in ctpboards.cfg:{}".format(name))
    #
    netcfg= getfile(os.path.join(cfgdir,"ctp.net"))
    if netcfg==None: return
    linen= -1
    for ixl in range(len(netcfg)):
      #atb_64
      #1 atb_37 atb_38 atb_39
      #2 atb_40 atb_41 atb_43
      #3 atb_51 atb_50 atb_44
      #4 atb_52 # atb_53
      #5 atb_45 atb_46
      #6 atb_47 atb_48 atb_49
      line= parseline(netcfg[ixl])
      if line==None: continue
      if linen==-1:    # 1st line of netcfg
        linen= 0
        self.ctpbname= line[0]   # higher priority if ctp.net exists
        continue
      if not (line[0] in ('1','2','3','4','5','6')): continue   # ignore gbtX...
      oltn= int(line[0])
      locemun= 1   # 1,2 or 3
      for ixbn in range(len(line[1:])):
        ltubname= line[ixbn+1]
        if ltubname==self.ctpbname:
          #print("MONONU for OLT%d, ignored"%oltn)
          continue
        ltu= self.findBN(ltubname)
        if ltu==None:
          printError("Unknown board: {} in ctp.net".format(ltubname))
          continue
        ltu.emun= (oltn-1)*3 + locemun
        locemun= locemun+1
    #self.prt()
  def findDN(self, detname, wired=False):
    """ check if detname is:
wired  check
------------
False  valid:   i.e. fv0, tpc,hmp,... is a valid name
True   wired:   valid + wired with ctp  (stronger test)
rc: bname       -ok, ltu wired to the ctp
    atb_detname -board not assigned (missing in ctpboards.cfg)
    None        -detname not defined in Alice
"""
    if detname in self.ltus:
      # detname is valid, i.e. exists in ctpboards.cfg
      if self.ltus[detname]==None:
        # but board not wired/connected to the ctp
        if wired:
          return None
        return "atb_"+detname   # valid, board not assigned, i.e. cannot be wired
      else:
        if wired:
          if self.ltus[detname].emun==None:
            return None   # not wired i.e. not present in ctp.net
        return self.ltus[detname].bname
    return None
  def findBN(self, bname):   # rc: Ltu ref or None
    for ltuname in self.ltus:
      if self.ltus[ltuname]==None: continue
      if self.ltus[ltuname].bname==bname: return self.ltus[ltuname]
    return None
  def prt(self):
    line= ""
    for ltun in self.ltus:
      ltu= self.ltus[ltun]
      if ltu:
        line= line+"%s %s:%s, "%(ltu.name, ltu.bname, str(ltu.emun))
    print("Ltus:",line)
class BCM(object):
  def __init__(self, name):
    self.maskdef= None
    self.name= name
  def setMaskDef(self, maskdef):
    # check and set mask definition
    msk= BCmask()   # check syntax
    msk.update(maskdef)
    if msk.errmsg=="":
      #self.maskdef= maskdef
      self.maskdef= msk.bcmstr
    else:
      printError(msk.errmsg)
class LTG(object):
  def __init__(self,ltgline, partinst):
    print("LTG:", ltgline)
    self.partinst= partinst
    self.detname= ltgline[0]
    self.loadcmds= []    # list of cmds to load LTG
    # emuname= findEmuN()
  def loadcmdsAppend(self, cmd):
    emsg= check_emubatchcmd(cmd)  # emuX par1 par2
    if emsg!="":
      self.partinst.regError(emsg+" "+" ".join(cmd))
    else:
      self.loadcmds.append(cmd)   # emuX par1 par2 ...
    return emsg
class Cluster(object):
  def __init__(self, cluline):
    global LTUS
    print("Cluster:",cluline)
    self.cluname= cluline[0]
    self.detnames= []
    self.classes= []
    grundets= 0   # 18 bits
    for ixd in range(1,len(cluline)):
      detname= cluline[ixd]
      #print detname,ixd
      if (detname in LTUS.ltus) and (LTUS.ltus[detname]!=None):
        emun= LTUS.ltus[detname].emun
        if emun==None:
          printError("Cluster %s: detector:%s not connected"%(self.cluname,detname))
        else:
          grundets= grundets | (1<<(emun-1))
          self.detnames.append(detname)
        #print "Cluster det:", detname, hex(grundets)
      else:
        printError("Cluster %s: unknown detector:%s"%(self.cluname,detname))
    if grundets==0: grundets= None
    self.grundets= grundets
class Partition(object):
  STOPPED=5    # stopped, being deleted
  LOADED=4     # partition def. read in from db, not started
  LOADEDHW=2   # LOADED + loaded in hw
  PAUSED= 1
  RUNNING=0    # active (see ctpd.proto RunReply)
  def __init__(self, pname, runn, detectors=""):   
    """
pname: filename[.pd] (in $CTP3_DBCTP/pd/) or
       part. definition (list of strings/lines or long string containing \n)
Out: self.status LOADED
"""
    self.clusters= []
    self.ltgs= {}   # key: detname
    self.masks= {}
    self.status= None   #  will become LOADED after successful load 
    self.errs= 0    # number of Errors (all in self.allerrs)
    self.allerrs= ""
    self.runn= runn
    if type(pname)==types.ListType:
      netcfg= pname
      pname= "run{}".format(runn)   # direct list of strings
    elif len(pname)>24:
      netcfg= pname.split('\n')
      pname= "run{}".format(runn)   # direct string with NLs
    else:
      #if len(pname)>24:
      #  self.regError("Too long part. name %24s...".format(pname))
      #  return
      if pname[-3:]==".pd": pname= pname[:-3]
      cfgdir= os.getenv("CTP3_DBCTP")
      fname= os.path.join(cfgdir,"pd",pname+".pd")
      netcfg= getfile(fname)
      if netcfg==None:
        self.regError("{} does not exist.".format(fname))
        return
    self.pname= pname
    self.detectors= detectors
    self.activesec= None   # ref to an active Cluster or LTG instance
    for ixl in range(len(netcfg)):
      #cluster name det1 det2 ...
      self.line= parseline(netcfg[ixl])
      if self.line==None: continue
      #
      # sections in fixed order:
      # None (common)
      # LTG
      # Cluster
      #
      # LTG: bcm,...   cluster ->finish LTG
      # todo: class
      if self.activesec== None:   # ====== common section
        # bcm ltg cluster
        if self.line[0]=='bcm':
          # todo: check self.masks if already in
          if len(self.line)==1:
            self.regError("bcm name missing in: {}".format(self.line))
          maskname= self.line[1]
          if not (maskname in symname):
            self.regError("Incorrect mask name: {}".format(maskname))   
          bcm= BCM(self.line[1])
          bcm.setMaskDef(" ".join(self.line[2:]))
          if bcm.maskdef!=None:
            self.masks[maskname]= bcm
          else:
            self.regError("Incorrect definition of {} mask in:{}".format(maskname,self.line))
        elif self.keyw('ltg', self.line[0]):
          if self.procltg()==None: break
        elif self.line[0]=='cluster':
          if self.proccluster()==None: break
        else:
          self.regError("bcm, ltg or cluster key expected in line:%d"%ixl)
          #print("Line:",self.line)
      elif isinstance(self.activesec, LTG):       # ====== ltg section
        # bcm TOF enable
        # ltg detname
        # cluster cluname det1 det2 ...
        #print "LTG %s active. Line:"%self.activesec.detname,self.line
        dn= self.activesec.detname
        n= findEmuN(dn)
        if n==None:
          self.regError("Unknown detector {} in:{}".format(dn,self.line))
          continue
        emuname= "emu%d"%n
        if self.line[0]=='bcm':
          # special (should probably be removed) 'bcm TOF enable'
          # correct is 'emu bcm TOF enable'
          if len(self.line)==1:
            self.regError("bcm name missing for ltg "+dn)
          else:
            # "emuX bcm TOF enable"
            cmd= (emuname, "bcm", self.line[1],"enable")
            self.activesec.loadcmdsAppend(cmd)   # emuX...
        elif self.line[0]=='emu':
          cmd= (emuname,)+ tuple(self.line[1:])
          self.activesec.loadcmdsAppend(cmd)   # emuX...
        elif self.keyw('ltg', self.line[0]):
          if self.procltg()==None: break
        elif self.line[0]=='cluster':
          if self.proccluster()==None: break
        else:
          #self.regError("ltg %s definition wrong (only 'bcm emu' lines allowed for now)"%\
          #  self.activesec.detname)
          cmd= (emuname,)+ tuple(self.line[0:])
          self.activesec.loadcmdsAppend(cmd)   # emuX...
      elif isinstance(self.activesec, Cluster):   # ====== cluster
        if self.line[0]=='cluster':
          if self.proccluster()==None: break
        else:
          self.regError("cluster %s definition: body(i.e.classes) not allowed yet"%\
            self.activesec.cluname)
      else:
        # no classes yet...
        self.regError("Internal error (unknown activesec ?) in:"+self.line)
    if self.errs== 0:
      self.setStatus(self.LOADED)
  def keyw(self, resword, word):
    if resword==word.lower(): return True
    return False
  def regError(self, msg):
    self.allerrs= self.allerrs + msg + "\n"
    #printError(msg+". Partition:"+self.pname+", in line:\n"+string.join(self.line))
    self.errs+= 1
  def getDets18(self):
    allclus= 0
    for ix in range(len(self.clusters)):
      allclus= allclus | self.clusters[ix].grundets
    return allclus
  def getCluDets(self):   # rc: "clu1:d1,d2 clu2:d1.d2:
    allclus= ""
    for ix in range(len(self.clusters)):
      allclus= allclus + self.clusters[ix].cluname + ":" +\
        ",".join(self.clusters[ix].detnames) + " "
    return allclus
  def getDets(self):   # rc: "clu1:d1,d2 clu2:d1.d2:
    alldets= ""
    for lname in self.ltgs:
      alldets= alldets + lname + " "
    return alldets.strip()
  def procltg(self):
    if len(self.line)==1:
      self.regError("ltg name missing in:"+self.line)
      return None
    lname= self.line[1]
    if lname in self.ltgs:
      print("ltg %s already defined once, extending..."%lname)
      ltg= self.ltgs[lname]
    else:
      if LTUS.findDN(lname)==None:
        self.regError("Unknown LTG name {} in:{}".format(lname,' '.join(self.line)))
        return None
      # check if wired
      if findEmuN(lname)==None:
        self.regError("LTG {} not wired in line:{}".format(lname,' '.join(self.line)))
        return None
      ltg= LTG(self.line[1:], self)
      self.ltgs[lname]= ltg
    self.activesec= ltg
    return ltg
  def proccluster(self):
    if len(self.line)<2:
      self.regError("cluster cluster_name det1 [det2 ...] expected in line:"+\
        self.line)
      return None
    lname= self.line[1]
    clu= self.findclu(lname)
    if (clu != None):
      if len(self.line) == 2:
        print("cluster %s already defined once, extending..."%lname)
        # i.e. allowed: cluster anyname    (det list must be empty)
      else:
        self.regError("Detectors ignored in line {}".format(" ".join(self.line)))
    else:
      clu= Cluster(self.line[1:])
      if clu.grundets!=None:
        self.clusters.append(clu)
        self.activesec= clu   # current one
      else:
        clu= None
    return clu
  def findclu(self,cluname):
    for ix in range(len(self.clusters)):
      if self.clusters[ix].cluname==cluname:
        return self.clusters[ix]
    return None
  def printPart(self):
    print("Partition:{} run:{} status:{}".format(self.pname,self.runn, self.getStatus()))
    if self.errs>0:
      print(self.allerrs)
    else:
      print("Masks:", list(self.masks.keys()))
      print("CLusters:", self.getCluDets()) 
      print("LTGs:", list(self.ltgs.keys()))
  def setStatus(self, newstatus):
      self.status= newstatus
  def getStatus(self):
    if self.status == self.RUNNING:
      st= "R"
    elif self.status == self.LOADED:
      st= "l"
    elif self.status == self.PAUSED:
      st= "P"
    elif self.status == self.LOADEDHW:
      st= "L"
    else:
      st= "?"
    return st
class Partitions(object):
  def __init__(self):
    self.parts= {}   # self.parts[runnumber] -Partition instance
  def addPart(self, pname, runn, dets):
    """
rc: Partition object or None
"""
    #if pname in self.parts:
    #  printError("Partition %s already in, replacing..."%pname)
    np= Partition(pname, runn, dets)
    if np.status==np.LOADED:
      self.parts[runn]= np
    else:
      #print("addPart(%s) errors:"%pname, np.errs, '\n', np.allerrs)
      #np= None
      pass
    return np
  def delPart(self, runn):
    if runn in self.parts:
      del self.parts[runn]
    else:
      printError("%d not loaded/started"%runn)
  def getN(self):
    return len(self.parts)
  def getPart(self, runn):
    if runn in self.parts:
      return self.parts[runn]
    else:
      return None
  def listParts(self):
    print("RunN  ", "Name   ", "Status")
    for npn in self.parts:
      np= self.parts[npn]
      st= np.getStatus()
      print(npn, np.pname, st)

if LTUS==None: LTUS= Ltus()

def main():
  # parted.py partname
  if len(sys.argv) < 2 or (len(sys.argv)==2 and sys.argv[1]=="help"):
    print("""
Usage: partdef.py pname [runn] [detectors]
pname     -partition definition in CTP3_DBCTP/pd/pname.pd
runn      -runn (0 or missing: -> validate partition )

todo:
1.
detectors -filter out given detectors

2. interactive cmdline mode: check more partitions -i.e. should be able to keep a replica
of what is loaded in ctp

""")
    sys.exit()    
  pname= sys.argv[1]
  runn= 0
  dets= ""
  if len(sys.argv) > 2: runn= int(sys.argv[2])
  if len(sys.argv) > 3: dets= " ".join(sys.argv[3:])
  parts= Partitions()
  part= parts.addPart(pname, runn, dets)
  print()
  #part.printPart()
  if part.status!=part.LOADED:
    sys.exit(8)
  else:
    print("ok")
  #  part.printPart()
  
if __name__ == "__main__":
  main()

