#!/usr/libexec/platform-python
""" 
devname 
         class TDIOmux:
tdomux   class TDOmux(TDIOmux):
tdomuxup class TDOmuxup(TDIOmux):
tdimux   class TDImux(TDIOmux):
tdomux   class CTPTDOmux:

"""
from __future__ import print_function
from builtins import range
from builtins import object
class TDIOmux(object):
  def __init__(self, board):
    self.board= board
    #if self.mux_name=="TDOmux":
    #  self.setmux(1)   # emu default
  def setmux(self, inpi):
    if not( inpi in (0,1,2,4,8)):
      print("Int.Error setmux:", inpi)
      return
    val= self.board.Read("ctrl."+self.muxs)
    #&0xfffffff0
    val= val & ~(0xf<<self.ctrlshift)
    val= val | ((inpi&0xf)<< self.ctrlshift)
    self.board.Write("ctrl."+self.muxs, val)
  def getmuxsym(self, val=None):
    if val==None: val= self.getmux()
    return self.mux_docsbrief[val]
  def getmux(self):
    """ rc: LTU: low 4 bits in ctrl.tdo_mux
"""
    #fwinfo= self.board.Read("ctrl.fwinfo")
    #if ((((fwinfo>>8) & 0xffff) == 9) and ((fwinfo & 0xff) <= 6)) or\
    #  (((fwinfo>>8) & 0xffff) < 9):
    #  val= self.board.Read("ttcponolt_u1.uregs", 128)&7
    #else:
    #  val= self.board.Read("ctrl.tdo_mux")&7   # from fi 9_07
    #val= self.board.Read("ctrl."+self.muxs)&0xf   # from fi 9_07
    val= (self.board.Read("ctrl."+self.muxs)>>self.ctrlshift)&0xf   # from fi 9_07
    return val
  def setcmd(self, inpsi):
    """ int 0 1 2 4 8
str 0 1 2 4 8
str ttcpon emu emutrd tdg onu
"""
    if type(inpsi) is int:
      if inpsi in self.mux_docsbrief:
        self.setmux(inpsi)
      else:
        print("%s: bad setting:%d, not set"%(self.mux_name,inpsi))
    elif type(inpsi) is str:
      if inpsi in ('0','1','2','4','8'):
        self.setmux(int(inpsi))
      #elif inpsi in self.mux_docsbrief.values():
      else:
        found= False
        for ki in list(self.mux_docsbrief.keys()):
          #print "dbg tdomux:",ki,self.mux_docsbrief[ki]
          if self.mux_docsbrief[ki]==inpsi:
            self.setmux(ki)
            found= True ; break
        if not found:
          print("%s: bad setting:"%self.mux_name,inpsi,"not set")
    else:
      print("%s: bad setting:"%self.mux_name,inpsi,"not set")
  def prtStatus(self):
    val= self.getmux()
    if val in self.mux_docs:
      rv= self.mux_docs[val]
    else:
      rv= "?0x%x"%val
    print("%s: %d = %s"%(self.mux_name,val,rv))
  def prtHelp(self):
    muxcmd= self.mux_name.lower()
    print("""
%s      -show current setting
%s %s   -set mux output to one of the sources shown 
"""%(muxcmd, muxcmd," | ".join(list(self.mux_docsbrief.values()))))
    self.prtSpecHelp()
  def prtSpecHelp(self):
    pass
  def excmd(self, cmd):
    if len(cmd)==2:
      if cmd[1] in list(self.mux_docsbrief.values()):
        self.setcmd(cmd[1])
        val= self.getmux()
        print("%s: %d = %s"%(self.mux_name,val,self.mux_docs[val]))
      else:
        self.prtHelp()
    elif len(cmd)==1:
      self.prtStatus()
    else:
      self.prtHelp()
class TDOmux(TDIOmux):
  mux_name= "TDOmux"
  muxs= "tdo_mux"   # reference to .xml
  mux_docsbrief= {0:"ttcpon", 1:"emu", 2:"emutrd", 4:"tdg", 8:"onu"}
  mux_docs= {0:"ttcpon firmware", 1:"CTP emulator", 2:"CTP emulator for TRD/CPV",
    4:"TDG (Trigger Data generator)", 8:"ONU data"}
  def __init__(self, board, mode=1):
    TDIOmux.__init__(self, board)
    self.ctrlshift= 0
    self.tdodef= mode    # default mode when in stdalone. 2:TRD/CPV 1: CRU
    #
    ctrl2= self.board.Read("ctrl.ctrl")
    gsint= (ctrl2>>4) & 0x1
    #print "LTU Mode: %s ctrl.ctrl:0x%x"%(("stdalone","global")[gsint], ctrl2)
    if gsint==1:
      self.setmux(8)   # onu data (global)
    else:
      self.setmux(self.tdodef)   # stdalone
  def setStdalone(self):
    self.setmux(self.tdodef)   # stdalone
  def setGlobal(self):
    self.setmux(8)   #  global
    return 8
  def prtSpecHelp(self):
    print("""
ttcpon: always OLT1 connected, even in case of empty OLT1 slot.
""")
class TDOmuxup(TDIOmux):
  mux_name= "TDOmuxup"
  muxs= "tdo_mux"   # reference to .xml
  mux_docsbrief= {0:"ttcpon", 1:"hbmb", 2:"ssm"}
  mux_docs= {0:"ttcpon firmware", 1:"HBM builder", 2:"SSM generator, upstream"}
  def __init__(self, board):
    TDIOmux.__init__(self, board)
    self.ctrlshift= 4
    self.setmux(1)   # HBM builder default
  def prtSpecHelp(self):
    print("""
ttcpon: ???
""")
class TDImux(TDIOmux):
  mux_name= "TDImux"
  muxs= "tdi_mux"   # reference to .xml
  mux_docsbrief= {0:"onu", 1:"det1", 2:"det2", 4:"det3", 8:"SSM"}
  mux_docs= {0:"raw ONU data", 1:"DET1 data", 2:"DET2 data", 4:"DET3 data",
    8:"SSMgen data"}
  def __init__(self, board):
    TDIOmux.__init__(self, board)
    self.ctrlshift= 0
    ctrl2= self.board.Read("ctrl.ctrl")
    gsint= (ctrl2>>4) & 0x1
    #print "LTU Mode: %s ctrl.ctrl:0x%x"%(("stdalone","global")[gsint], ctrl2)
    if gsint==0:
      self.setmux(0)   # raw onu default for stdalone, do not change if global
    # in global the connections known in P2, but in lab not fixed.
class CTPTDOmux(object):
  mux_name= "CTPTDOmux"
  muxs= "tdo_mux"   # reference to .xml. ctp_logic_v3: tdo_mux_olt1..6
  def __init__(self, board):
    self.board= board
    self.muxcmd= "tdomux"   #self.mux_name.lower()
    self.setcmd("cccccc")   # default is ctp
  def tcp2n(self, tcp):
    if tcp=='t':  # tdg
      val= 2
    elif tcp=='c':  # ctp
      val= 1
    elif tcp=='p':  # ttcpon
      val= 0
    else:
      val=None
    return val
  def setmux(self, valall):
    """ valall: 6x4 bits for 6 olts: 654321
"""
    #self.board.Write("ctrl."+self.muxs, valall)   # _v2
    # _v3
    for ix in range(6):
      val= (valall>>(ix*4))& 0xf
      self.board.Write("ctrl.%s_olt%d"%(self.muxs,ix+1), val)
  def setcmd(self, inpstr, oltn= None):
    """ inpstr: string of length 1 or 6
Exmple:
tttccp -set TDG for olt1..3 and CTP for olt4..5 and TTCPON for olt6
t      -set just for oltn (oltn has to be !=None)
"""
    #val= self.board.Read("ctrl."+self.muxs)&0xffffff
    if len(inpstr)==1:   # just 1 olt
      val= self.tcp2n(inpstr)
      if val!=None and ((oltn!=None) and ((oltn>='1')and(oltn<='6'))):
        self.board.Write("ctrl.%s_olt%d"%(self.muxs,oltn), val)
      else:
        print("Bad option for CTPTDOmux:%s, olt:"%inpstr, oltn)
    elif len(inpstr)!=6:
      self.prtHelp()
      valall= None
    else:
      valall= 0
      for ix in range(6):
        oltn= ix+1
        val= self.tcp2n(inpstr[ix])
        if val==None:
          print("Bad option ofor CTPTDOmux:%s."%inpstr)
          return None
        valall= valall | (val<<(ix*4))
      self.setmux(valall)
    #return valall
  def getmuxsym(self, val=None):
    if val==None: val= self.getmux()
    return self.val2str(val)
  def getmux(self):
    """
_v2: valall= self.board.Read("ctrl."+self.muxs)&0xffffff  obsolete
_v3: return all to 6x4 bits in 1 word/24 for olts 654321
"""
    valall= 0
    for ix in range(6):
      val= self.board.Read("ctrl.%s_olt%d"%(self.muxs,ix+1))&0xf
      #ixbits= (5-ix)*4
      ixbits= ix*4
      valall= valall | (val<<ixbits)
    return valall
  def val2str(self, valall):
    outstr=""
    for ix in range(6):
      val= (valall >> (ix*4)) & 0xf
      if val==2:
        outstr= outstr + 't'
      elif val==1:
        outstr= outstr + 'c'
      elif val==0:
        outstr= outstr +'p'
      else:
        outstr= outstr + "%X"%val
    return outstr
  def prtStatus(self):
    val= self.getmux()
    valstr= self.val2str(val)
    print("%s: 0x%x = %s"%(self.mux_name, val, valstr))
  def prtHelp(self):
    print("""
%s          -show current setting
%s xxxxxx   -set mux for all OLTs (1..6). where x is t|c|p
%s N x      -set mux for OLTN
             t:TDGgen   c:CTP   p:TTCPON
"""%(self.muxcmd, self.muxcmd, self.muxcmd))
  def excmd(self, cmd):
    if len(cmd)==3:
      if (cmd[1]>="1") and (cmd[1]<="6"):
        self.setcmd(cmd[2], int(cmd[1]))
        print("tdo_mux_olt%s: %s"%(cmd[1], cmd[2]))
      else:
        print("Incorrect cmd")
    elif len(cmd)==2:
      if cmd[1]=="test":
        val=0x012345
        valstr= self.val2str(val)
        print("%s TEST: 0x%x = %s"%(self.muxcmd,val, valstr))
      elif cmd[1]=="h" or cmd[1]=="?":
        self.prtHelp()
      else:
        self.setcmd(cmd[1])
        val= self.getmux()
        valstr= self.val2str(val)
        print("%s: 0x%x = %s"%(self.muxcmd,val, valstr))
    elif len(cmd)==1:
      self.prtStatus()
    else:
      self.prtHelp()
