#!/usr/libexec/platform-python
from __future__ import print_function
from builtins import range
from builtins import object
class TICG1(object):
  BRS= ("b1","b2","r1","r2")
  GNAMES= ("bc_1","bc_2","rnd_1","rnd_2")
  def __init__(self, board, i, valid=True):
    self.board= board
    self.tig= i
    self.nn= "tig_%d"%i
    self.valid= valid
  def b03(self,br):
    try:
      rc= self.BRS.index(br)
    except:
      rc= None
    return rc
  def getgen(self, br):   # rc: content of corresponding generator
    n03= self.b03(br)
    if n03==None: return None
    if self.valid==False:
      return None
    rc= self.board.Read(self.nn+"."+self.GNAMES[n03])
    return rc
  def valtx(self, br):   # rc: ? - . 1
    if self.valid==False:
      return "-"
    ctr= self.board.Read(self.nn+".control")
    bitena= self.b03(br)
    if bitena==None: return "?"
    if (ctr & (1<<bitena)) != 0:
      return "1"
    else:
      return "."
  def enadis(self,ed,gen):   # gen: b1 b2 r1 r2. None: all
    if self.valid==False:
      return
    if gen==None:
      gens= list(range(4))
    else:
      g= self.b03(gen)
      if g==None: return
      gens= (g,)
    ctr= self.board.Read(self.nn+".control")
    for ixg in gens:
      if ed=='e':
        ctr= ctr | (1<<ixg)
      elif ed=='d':
        ctr= ctr & (~(1<<ixg))
      else:
        return
    self.board.Write(self.nn+".control", ctr)
  def setgen(self,br,N):
    if self.valid==False:
      return
    g= self.b03(br)
    nname= self.nn+"."+self.GNAMES[g]
    self.board.Write(nname,N)
class TICG(object):
  NINPS= 48
  def __init__(self, board):
    self.tigs= []
    for ix in range(self.NINPS):
      valid= True
      ix1= ix+1
      if ix1==11 or ix1==12: valid= False
      self.tigs.append(TICG1(board, ix1, valid))
  def show(self, br):   # br: b1 b2 r1 r2
    line= ""
    for ix in range(self.NINPS):
      line= line+self.tigs[ix].valtx(br)
    #rint "    1...,...10....,...20....,...30....,...40....,.48"
    #      b1: -.1
    print("%2s: %s"%(br, line))
  def showall(self):   # br: b1 b2 r1 r2
    print("    1...,...10....,...20....,...30....,...40....,.48")
    for br in TICG1.BRS:
      self.show(br)
  def showgens(self):
    print("    %9s %9s %9s %9s"%TICG1.GNAMES)
    for ix in range(self.NINPS):
      if self.tigs[ix].valid==False: continue
      #      '22: '
      line= "%2d: "%(ix+1)
      for br in TICG1.BRS:
        gv= self.tigs[ix].getgen(br)
        line= line+"%9d "%(gv)
      print(line)
  def nns2tns(self,nns):
    if nns==None:
      tns= list(range(self.NINPS))
    else:
      tns= [int(x)-1 for x in nns]
    return tns
  def enadis(self,ed, gen=None, nns=None):
    """
ed: 'e' or 'd'
gen: b1 b2 r1 r2
nns: 1 2 ... 48
"""
    tns= self.nns2tns(nns)
    for ix in tns:
      self.tigs[ix].enadis(ed,gen)
  def setgen(self, br, N, nns):
    tns= self.nns2tns(nns)
    for ix in tns:
      self.tigs[ix].setgen(br,N)
  def excmd(self, cmd):
    if len(cmd) < 2:
      self.prtHelp()
    elif cmd[1]=='s':
      self.showall()
    elif cmd[1]=='g':
      self.showgens()
    elif cmd[1]=='e' or cmd[1]=='d':
      gen= None
      if len(cmd)>2: gen= cmd[2]
      nns= None
      if len(cmd)>3: nns= cmd[3:]
      self.enadis(cmd[1], gen, nns)
    elif cmd[1] in TICG1.BRS:
      if len(cmd)<3: return
      N= int(cmd[2])
      nns= None
      if len(cmd)>3: nns= cmd[3:]
      self.setgen(cmd[1], N, nns)
    else:
      pass
  def prtHelp(self):
    print("""
ticg s           -show enabled/disabled inputs
     g           -show generators
     e|d|r       -enable|disable|reset all generators for all inputs
                                 reset: NOT DONE YET
     e b1        -enable b1 for all inputs
     e b1 n...   -enable b1 for n...  inputs
       b2 n...
       r1 n...
       r2 n...
     b1|b2|r1|r2 N n...  -set generator to N for n... inputs (all without n)
""") 
