#!/usr/libexec/platform-python
#!/usr/bin/python3
#!/usr/bin/python3
from __future__ import print_function
from builtins import hex
from builtins import range
from builtins import object
import sys,time,traceback
import ipbuswrap,saxipbus,i2c,colors,specdevs
#!/usr/bin/python3
from auxfuns import isinargs,printError,ltucfgpars
import spi
import spiatlas as spiatlas

import atb.fpga
import atb.hbmb
import atb.ticg
import atb.ddrmem
import atb.counters as counters
import atb.ctp as ctp
import atb.ctpemu as ctpemu
import atb.tdomux as tdomux

AllBoards= {}
def printError(incolor, msg=""):
  print(colors.color(incolor, "fg_red"), msg)
class BaseOption(object):
  def __init__(self, board, ipslave, poslen, name, **kw):
    """ name: option name
ipslave: None: do not update in hw
poslen: (4,5) corresponds to mask 0x1f0
**kw: optA=val1, optB=val2
Operation:
"""
    self.name= name
    self.board= board
    self.ipslave= ipslave
    self.poslen= poslen
    self.opts= {}
    #for key in kw.keys():
    #  self.opts[key]= kw[key]
    self.opts= dict(**kw)
  def cmd(self, arglist):
    """
arglist: (name, arg1, arg2,...)
Base cmds:
set:
arg1: a name defined in self.opts.keys(). arg2... ignored in MUXcase
show: arglist:(name) i.e. only name provided
"""
    if len(arglist)==1:   # show name: symbvalue(value)
      val= (self.board.Read(self.ipslave) >> self.poslen[0]) & (2**(self.poslen[1]-1))
      #if val in self.opts.values():
      self.showValue(val)
    else:   # set arglist[1]
      self.setValue(arglist)
class MuxOption(BaseOption):
  def __init__(self, board, ipslave, poslen, name, **kw):
    BaseOption.__init__(self, board, ipslave, poslen, name, **kw)
  def setValue(self, args):
    if args[1] in self.opts:
      bitmask= 2**(self.poslen[1]-1)
      val= self.opts[args[1]]
      self.board.SetBits(self.ipslave, val, bitmask, self.poslen[0]) 
   #else if todo: int or hex
    else:
      print(args[1],"- unknown option, use one of:", list(self.opts.keys()))
  def showValue(self, val):
    k= "None"
    for kv in list(self.opts.items()):
      if kv[1]== val:
        k= kv[0] ; break
    print("%s: %s (%d)"%(self.name, k, val))    
      
class ABoard(ipbuswrap.Board):
  def __init__(self, boardname, pllmode=None):
    """ pllmode:
Special:
None - choose by fwinfo
pllnocheck
pllcheckonly
pll name: pllin1,...

"""
    print("===   ABoard(%s) instance..."%boardname)
    ipbuswrap.Board.__init__(self, boardname)
    self.deverr= ""   # or err message
    self.pllmode= pllmode
    self.i2cinuse= None ;     self.i2cs= []   # 2x i2c bus:
    # i2fs
    # i2cg
    self.i2cg= None   # i2c bus (list of i2c devices)
    self.tc74= None   # TC74 device
    self.spic= [] ; self.spicinuse= None
    # see setGVar, evalGVar
    self.gvars= {"tf":3649536, "2tf":7299072, "3tf":10948608, "4tf":14598144,
      "lflast":None, "orbit":14256}
    self.atbdevs= {}
    self.checkBase()
  def setdbgmode(self, mode=0):
    self.dbgmode= mode
  def ltucfgpar(self, par):
    return ltucfgpars(self.bname, par)
  def checkBase(self):   # common (any board) devices
    print(colors.color("checkBase: checking IPBUS slaves...", "bold"))
    fwinfo="not present"; boardid="not present"
    if self.fwinfo:
      #self.fwinfo= self.board.fwinfo
      fwinfo= hex(self.fwinfo)
    #if self.board.nodExists("ctrl.boardid"):
    if self.nodExists("ctrl.boardid"):
      #boardid= self.board.Read("ctrl.boardid") & 0xff
      boardid= self.Read("ctrl.boardid") & 0xff
      boardid= "08-00-30-00-2A-%x"%boardid
    #print "%s: boardid:MAC:%s fwinfo:%s"%(self.board.bname, boardid, fwinfo)
    print("%s: boardid:MAC:%s fwinfo:%s"%(self.bname, boardid, fwinfo))
    tnn= "fpga"   # monitoring
    if self.nodExists("ctrl.fpga_temperature"):
      self.atbdevs[tnn]= atb.fpga.Sysmon(self, tnn)   # see also Sysmon.finishltuinit
      self.atbdevs[tnn].prtTempVolts()
      #self.atbdevs[tnn].excmd(("any","rr","1"))
    #self.odevs= []   # 'other' devices (not connected to i2c|spi bus)
    tnn="leds"
    if self.nodExists("ctrl.ledenable"):
      #self.odevs.append(specdevs.Leds(self, "LEDS"))
      self.atbdevs[tnn]= specdevs.Leds(self, tnn)
    tnn="sfpleds"
    ledreg= "ctrl.sfpledenable"
    if self.nodExists(ledreg):
      self.atbdevs[tnn]= specdevs.Leds(self, tnn, ledreg=ledreg, nleds=24)
    # i2c:
    self.pllwasloaded= False   # will be set to True in checkandload
    self.plldev= None 
    if self.nodExists("i2cpll", True) and (fwinfo!="not present"):    # pll should be first
      # Note: in Si5345 class, oltreset() can be used only when also i2cmaintttc (i2cg) done!
      i2cpll= i2c.I2Ccore(self,"i2cpll")   # only 1 device on this bus
      self.i2cs.append(i2cpll) ; self.i2cinuse= len(self.i2cs)-1
      # add devices
      pll= i2c.Si534x(i2cpll)
      checkok= False
      if self.pllmode=="pllnocheck":
        print(colors.color("Si5345 config NOT checked (pllnocheck arg).", "fg_red"))
        checkok= True
      elif self.pllmode=="pllcheckonly":
        if pll.check(): checkok= True
      else:
        if pll.checkandload(self.pllmode):
          checkok= True
      if checkok:
        i2cpll.addDevice(pll)
        self.plldev= pll
    else:
      i2cpll= None
      print(colors.color("i2cpll not found.", "fg_red"))
    i2cmuxs= (0x74, 0x76)
    if self.nodExists("i2cmainttc", True):  # ttcpon (csaba/EP) i2c
      self.i2cg= i2c.I2Ctpcore(self,"i2cmainttc",i2cmuxs)  # i2c slave on both (ctpltu kcu105)
      #self.Write("i2cmainttc.mux", 11)   # 1..9: olt1..9   10:onu   11:ipbus
      self.i2cg.seti2cttcmux("ipbus")
      print("i2cmainttc bus, i2cttc mux set to ipbus access.")
      self.i2cs.append(self.i2cg) ; self.i2cinuse= len(self.i2cs)-1
    elif self.nodExists("i2cmain", True):   # Wishbone i2c
      self.i2cg= i2c.I2Ccore(self,"i2cmain",i2cmuxs)  # i2c slave on both (ctpltu kcu105)
      self.i2cs.append(self.i2cg) ; self.i2cinuse= len(self.i2cs)-1
    else:
      self.i2cg= None
      print("NO i2cmain[ttc] present!")
    if self.i2cinuse:
      if len(self.i2cs)>1:
        # initialize 40mhz tick if needed (i.e. when pll was just loaded above):
        #pll.oltreset todo
        self.i2cinuse= 1   # let leave it on i2cg...
      print("Using %s..."%self.i2cs[self.i2cinuse].i2cslave)
    #if self.nodExists("ctrl.ctrl"):
    # ctrl.ctrl  bits (27.5.2018):   moved to fpga.py 20.2.2021
    #  ctrl= self.Read("ctrl.ctrl")
    #  self.Write("ctrl.ctrl", 0)
    #  ctrl2= self.Read("ctrl.ctrl")
    #  if (ctrl2 & 0x3) == 0:
    #    print colors.color("ctrl.ctrl: 0x%x, After 0 written:0x%x -ok"%(ctrl, ctrl2), "fg_green")
    #  else:
    #    print colors.color("ctrl.ctrl: 0x%x, After 0 written:0x%x -error (0 expected)"%(ctrl, ctrl2), "fg_red")
    if self.i2cg:
      # add devices on mux_u45 (0x74)
      self.tc74= i2c.TC74(self.i2cg, "tc74", mux=(0x76,0, 0x74,1<<0))   # used by getSlow
      rc=self.i2cg.addDevice(self.tc74)
      if rc==None:
        printError("Seems I2C bus broken...")
      #print "SFP (ipbus) not checked..."
      rc= self.i2cg.addDevice(i2c.SFP(self.i2cg, "SFP", "ipbus", mux=(0x76,0, 0x74,1<<1)))   # ipbus connection
      if rc==None:
        printError("I2C access to SFP device (IPBUS ethernet) bus broken...")
        printError("Consider a reset or FPGA reload from FM (i.e. 'fpga reset soft' or 'bootfgpa').")
        #time.sleep(1)
      # add devices on mux_u59 (0x76)
      # FMC       addr   eeprom type    datasheet
      # ttcpon    0x50   24AA02548T     r3doc/20002124E-254617.pdf
      # FM-S18    0x50   M24C02         r3doc/21202h.pdf
      # dodo             CAT24C02       https://twiki.cern.ch/twiki/pub/ALICE/FMCCards/CAT24C01-D.PDF
      # dodoINV          CAT24C02
      # KAYA      0x50   AT24C01B       r3doc/doc5156.pdf
      # TTC              24LC02B        r3doc/21709c.pdf
      #print "fmceeprom not checked..."
      # 0x74: mux_u45   0x76: mux_u59
      try:
        eeprom= i2c.M24C0x(self.i2cg, "fmceeprom", 0x50, mux=(0x74, 0, 0x76, 1<<0)) #SM-18
        if eeprom.getid()!=None:
          self.i2cg.addDevice(eeprom)
          print("eeprom.getid:",eeprom.getid())
      except:
        print("eeprom cannot be checked")
      #
      # adding onboard SFPs SFP1..14: 
      chkall= isinargs("all")
      if chkall or ((self.fwinfo>>24) == self.FW_LTU_ITSMFT_LOGIC):
        print("Checking all SFPs (all in parameters or ltu_logic_itsmft)...")
      else:
        #if ((self.fwinfo>>24) == self.FW_LTU_LOGIC) or ((self.fwinfo>>24) == self.FW_CTP_LOGIC):
        print(colors.color("ctp/ltu_logic,", "bold")+\
          " checking only 5 SFPs: ONU, OLT1, OLT2, GBT1, GBT2... (use 'all' if needed)")
      for role in list(self.fwsfps.keys()):   # olt1..19 onu gbt1..14
        rolesfp= self.fwsfps[role]      # SFP1..12 FSFP0..3
        if rolesfp[0]=='F': continue    # do not check FMC (done in LTUBoard)
        rolemux= self.i2cgmuxes[rolesfp]
        if (self.fwinfo>>24) == self.FW_LTU_ITSMFT_LOGIC:
          sfp= i2c.SFP(self.i2cg, rolesfp, role, mux=rolemux)
          #print "rolesfp:",rolesfp,"role:",role,"rolemux:",rolemux
        else:
          if chkall:
            sfp= i2c.SFP(self.i2cg, rolesfp, role, mux=rolemux)
          else:
            if (role=="onu") or (role=="olt1") or (role=="olt2") or\
              (role=="gbt1") or (role=="gbt2"):
              sfp= i2c.SFP(self.i2cg, rolesfp, role, mux=rolemux)
            else:
              continue
        #print "rolesfp:",colors.color(rolesfp+" "+str(sfp.purpose),"bold"),"role:",role,"rolemux:",rolemux
        if sfp.purpose!=None:
          print(colors.color(rolesfp+" ","bold"),"role:",role,"rolemux:",rolemux)
        if sfp.purpose:
          self.i2cg.addDevice(sfp)
          #print "ok"
    # spi:
    # check if spi0 available before instantiating spic[xx]!
    # i.e. see: ublaze/ttcpon.py DriverIpbus getNodeNames...
    # ctrl.ctrl bit2: valid for SPI0 (StartUpE3) NOT used from ltu_logic_v2
    # 0:used over SPI  1: not used over SPI
    # spi0/1 on both ltukcu + proto_*:
    if self.nodExists("spi0", True):
      print("spi0 found, creating...")
      self.spic.append(spi.SPIcore(self,"spi0"))  # StartUpE3 spi
      self.spic[0].addDevice(spi.N25Q(self.spic[0], "fm_cfg", "n25q128"))
      self.spicinuse= 0
    if self.nodExists("FLASH_SPI_Ram_0",True):
      print("Flash_SPI_Ram_0 found, creating...")
      self.spic.append(spiatlas.SPIatlas())    # here seems error: also node name should be passed to SPIatlas...
      # assuming that either spi0 or FLASH_SPI exlusively is present
      #print "hw:",self.getBoardNumber()
      fmname="n25q128"
      memsize=0x1000000
      if self.getBoardNumber() >= 60:
          fmname="n25q256"
          memsize=0x2000000
      print("FM 0 size:",hex(memsize))
      self.spic[len(self.spic)-1].addDevice(spiatlas.N25Q(self.spic[len(self.spic)-1],
        "fm_cfg", fmname,self.getHW(),"FLASH_SPI_Ram_0",memsize))
      # check if ok:
      devid= self.spic[len(self.spic)-1].connected[0].getid()
      print("spi dev id:", devid)
      if devid!= None:
        self.spicinuse= len(self.spic)-1
      else:
        del self.spic[len(self.spic)-1]
        print(colors.color("Flash_SPI_Ram_0 removed","fg_red"))
    if self.nodExists("spi1", True):
      self.spic.append(spi.SPIcore(self,"spi1"))  # spi1 (no StartUpE3)
      self.spic[len(self.spic)-1].addDevice(spi.N25Q(self.spic[len(self.spic)-1],
        "fm_semip", "n25q128"))
      self.spicinuse= len(self.spic)-1
    if self.nodExists("FLASH_SPI_Ram_1",True):
      print("Flash_SPI_Ram_1 found, creating...")
      self.spic.append(spiatlas.SPIatlas())
      # assuming that either spi1 or FLASH_SPI exlusively is present
      fmname="n25q1268"
      memsize=0x1000000
      if self.getBoardNumber() >= 60:
          fmname="n25q256"
          memsize=0x2000000
      print("FM 1 size:",hex(memsize))    
      self.spic[len(self.spic)-1].addDevice(spiatlas.N25Q(self.spic[len(self.spic)-1],
        "fm_semip", fmname,self.getHW(),"FLASH_SPI_Ram_1",memsize))
      # check if ok:
      devid= self.spic[len(self.spic)-1].connected[0].getid()
      print("spi dev id:", devid)
      if devid!= None:
        self.spicinuse= len(self.spic)-1
      else:
        del self.spic[len(self.spic)-1]
        print(colors.color("Flash_SPI_Ram_1 removed","fg_red"))
    if self.nodExists("spiadc", True):
      if self.nodExists("ctrl.adc_test_data"):
        # adc: ONLY if spiadc + ctrl.adc_test_data available in .xml: 
        self.spic.append(spi.SPIcore(self,"spiadc", sslogic=True))
        self.spicinuse= len(self.spic)-1
        self.spic[len(self.spic)-1].addDevice(spi.MAX11661(self.spic[len(self.spic)-1],
          "adc", "MAX11661"))
        print("spiadc added. len(spic):",len(self.spic))
    if self.spicinuse==0:
      print("Using spi%d (StartUpE3)"%self.spicinuse)
    elif self.spicinuse>0:
      print("Using spi%d"%self.spicinuse)
    else:
      print("No spi ?")
    if self.nodExists("bcmasks", topnode=True):
      self.atbdevs["bcm"]= atb.bcmask.BCmasks(self)
    print("DDR4s:")
    for tnn in ("ssm","tdg"):
      if self.nodExists(tnn, topnode=True):
        # ltu/ctp/ttcit firmwares: equipped also with a 
        # "ssm|tdgctrl" IPbus slave node (BC clock domain
        # allowing recording/generation).
        ctrlslave= tnn+"ctrl"
        if not self.nodExists(ctrlslave, topnode=True):
          ctrlslave= None
        self.atbdevs[tnn]= atb.ddrmem.DDR3(self, tnn, ctrlslave)
    if self.nodExists("tig_1", topnode=True):
      self.atbdevs["ticg"]= atb.ticg.TICG(self)
  def setGVar(self, varname, value):
    self.gvars[varname]= value
  def evalGVar(self, param):
    """ input:
$varname -return stored value
varname  -just evaluate given parameter
"""
    if param[0]=='$':
      if param[1:] in self.gvars:
        rc= self.gvars[param[1:]]
      else:
        rc= None
    else:
      rc= eval(param)
    return rc
  def muxOption(self, ipslave, poslen, name, **kw):
    return MuxOption(self, ipslave, poslen, name, **kw)
  def geti2cinuseslave(self):
    if self.i2cinuse!=None:
      return self.i2cs[self.i2cinuse].i2cslave
    else:
      return "None"
  def add_ctpemu(self, emuname):
    ctpemulator= None
    if self.nodExists(emuname, topnode=True):
      if emuname=="ctpemu":
        emun=""
      else:
        emun=emuname[6:]
      try:
        ctpemulator= ctpemu.CTPemu(self,self.atbdevs, emuname)
        self.atbdevs["emu"+emun]= ctpemulator
        #print "CTP emulator:", emuname, " added"
      except:
        #print sys.exc_info()
        traceback.print_exc(file=sys.stdout)
        ctpemulator= None
        #self.deverr= "CTP emulator not available"
        printError("CTP emulator %s not available"%emuname)
    else:
      self.deverr= "CTP emulator %s not available"%emuname
      printError("CTP emulator %s not available"%emuname)
      #raise Exception("CTP emulator not available")
    return ctpemulator
  def add_outcnts(self, tnn, ttccnts=None, cnttype=None, devname=None):
    """ tnn: top node name (string): outcnts, outcnts[1..18], ttcitcnts or gcnts
ttccnts: none or
         ttccnts instance("ttc" top node name in .xml). I.e. if they are present,
         they are always together with LTU's "outcnts" counters
cnttype:
"""
    if self.nodExists(tnn, topnode=True):
      if cnttype==None:
        self.atbdevs[tnn]= counters.CountersCTPEMU(self, tnn, ttccs=ttccnts) 
      else:
        if devname!=None:
          self.atbdevs[devname]= cnttype(self, tnn, devname=devname)
        else:
          self.atbdevs[tnn]= cnttype(self, tnn)
      #print "CTPEMU counters:", tnn
    else:
      printError("%s not available"%tnn)
  def printDevices(self):
    """ print: all devices spi* i2c*
"""
    print(colors.color("Detected devices:","bold"))
    for ix in range(len(self.spic)):
      #if self.spic[ix].spislave != "spiadc":
      #  print "aboard.py printDevices spi skipping(causing error?):", self.spic[ix].connected[0].devname,\
      #    "slave:", self.spic[ix].spislave
      #  continue
      self.spic[ix].printDevs()
    for ix in range(len(self.i2cs)):
      self.i2cs[ix].printDevs()
    print("Other devices:")
    #for ix in range(len(self.odevs)):
    #  print "  "+self.odevs[ix].getid()
    for devname in list(self.atbdevs.keys()):
      print("  "+devname)   #self.atbdevs[devname].getid()
  def findDevice(self, devname):
    """ find all devices connected to spi* i2c* or odevs
when last character of devname is '*': rturn the list of all devices
"""
    for ix in range(len(self.spic)):
      if len(self.spic[ix].connected)==0: continue
      if self.spic[ix].connected[0].devname == devname:   # only 1 device on spi!
        return self.spic[ix].connected[0]
    for ix in range(len(self.i2cs)):
      i2cb= self.i2cs[ix]
      for ixc in range(len(i2cb.connected)):
        if i2cb.connected[ixc].devname == devname:
          return i2cb.connected[ixc]
    #for ix in range(len(self.odevs)):
    #  if self.odevs[ix].devname == devname:
    #    return self.odevs[ix]
    for devn in list(self.atbdevs.keys()):
      #print "findDevice:", devn
      if devn == devname:
        return self.atbdevs[devn]
    return None
  def findDeviceSFPS(self, sfpname=None):
    """ sfpname: find only 1 sfp (e.g. SFP1)
rc: list of found SFPs
"""
    sfps= []
    for ix in range(len(self.i2cs)):
      i2cb= self.i2cs[ix]
      for ixc in range(len(i2cb.connected)):
        dn= i2cb.connected[ixc].devname
        if (len(dn) >=4) and ( (dn[:4]=="FSFP") or (dn[:3]=="SFP") ) :
          if sfpname==None:
            sfps.append(i2cb.connected[ixc])
          else:
            if sfpname==dn:
              sfps.append(i2cb.connected[ixc])
              break
    return sfps
  def getSlow(self):
    """ rc: (errmsg,(tstamp,val1,val2,...))
        errmsg: "" -ok
                error message during measurement
"""
    tstamp= time.time()
    errmsg= ""
    try:
      tc74= self.tc74.gettemp()
    except:
      # seems gets stucked here until tc74 not read in 'atb tc74 id'
      tc74= None
      print("except in atbwrap.py %s getSlow: empty tc74"%(self.bname))
      errmsg= "empty %s.tc74"%(self.bname)
    try:
      ft= self.atbdevs["fpga"].gettemp()
    except:
      ft= None
      print("except in atbwrap.py getSlow: empty fpga temp")
      if errmsg: errmsg=errmsg+"\n"
      errmsg= errmsg+"empty %s.fpga"%self.bname
    return (errmsg,(tstamp, tc74, ft))

class CTPBoard(ABoard):
  def __init__(self, boardname, pllmode=None):
    print("=== CTPBoard(%s) instance..."%boardname)
    ABoard.__init__(self, boardname, pllmode)
    if self.deverr !="": return
    print("=== CTPBoard(%s) instance checkDevices..."%boardname)
    self.checkDevices()   # LTU devices
  def checkDevices(self):
    maxemu= 18
    for ix in range(1,maxemu+1):
      self.add_ctpemu("ctpemu%d"%ix)
    for ix in range(1,maxemu+1):
      self.add_outcnts("outcnts%d"%ix)
    self.add_outcnts("gcnts", None, counters.CountersCTP)
    tnn= "outcnts"
    self.atbdevs[tnn]= counters.CtpEmuCounters(self)
    print("All emus output counters:", tnn)
    tnn= "ctp"
    self.atbdevs[tnn]= ctp.Ctp(self)
    print("ctp:", tnn)
    tnn= "tdomux"
    self.atbdevs[tnn]= tdomux.CTPTDOmux(self)
    print("CTPTDO mux:", tnn)
    self.atbdevs["fpga"].finishctpinit()
    if (self.pllwasloaded or isinargs("emuinit")):
      rcmsg= self.atbdevs["fpga"].fpgainit()
      print("fpgainit:", rcmsg)
class TTCITBoard(ABoard):
  def __init__(self, boardname):
    print("=== TTCITBoard(%s) instance..."%boardname)
    ABoard.__init__(self, boardname)
    if self.deverr !="": return
    print("=== TTCITBoard(%s) instance checkDevices..."%boardname)
    self.checkDevices()   # TTCIT devices
  def checkDevices(self):
    self.add_outcnts("ttcitcnts", None, counters.CountersTTCIT)   # ttcit
class LTUBoard(ABoard):
  def __init__(self, boardname, pllmode=None):
    """ todo:
stdalone: tdimux: det1..det3, tdomux: CTPemu
global:   tdimux:n/a          tdomux: onu data
"""
    print("=== LTUBoard(%s) instance..."%boardname)
    #ipbuswrap.Board.__init__(self, boardname)
    ABoard.__init__(self, boardname, pllmode)
    if self.deverr !="": return
    print("=== LTUBoard(%s) instance checkDevices..."%boardname)
    self.checkDevices()   # LTU devices
    #if self.isfiwa(self.FW_LTU):
    #  self.checkLtuDevices()
    #elif self.isfiwa((self.FW_CTP_LOGIC,)):
    #  self.checkCtpDevices()
    #else:
    #  printError("Only ctp_ or ltu_logic_ allowed")
  def checkDevices(self):
    if self.nodExists("i2cfmcsfp", True):
      i2fs= i2c.I2Ccore(self,"i2cfmcsfp")  # only ltu
      if i2fs:
        print("FMC i2c bus ok (node: i2cfmcsfp)")
        self.i2cs.append(i2fs)  # ; self.i2cinuse= len(self.i2cs)-1
      else:
        print("FMC i2c bus failed (node: i2cfmcsfp)")
    else:
      i2fs= None
    # GBT SFPs on FMC connected over i2cfmcsfp:
    if i2fs:          # Front view channel numbers (to be tested)
      if (self.fwinfo>>24) == self.FW_LTU_ITSMFT_LOGIC:
        ##i2fs.addDevice(i2c.SFP(i2fs, "FSFP0", mux=(0,0)))   # 3/228  7/227
        ##i2fs.addDevice(i2c.SFP(i2fs, "FSFP1", mux=(0,1)))   # 2/228  6/227
        ##i2fs.addDevice(i2c.SFP(i2fs, "FSFP2", mux=(0,2)))   # 1/228  5/227
        ###i2fs.addDevice(i2c.SFP(i2fs, "FSFP3", mux=(0,3)))  # 0/228  4/227/not connected
        ##i2fs.addDevice(i2c.SFP(i2fs, "FSFP4", mux=(0,4)))   # not connected ?
        # 13 gbts:
        #i2fs.addDevice(i2c.SFP(i2fs, "FSFP5", "gbt9",mux=(0,5)))
        #i2fs.addDevice(i2c.SFP(i2fs, "FSFP6", "gbt10",mux=(0,6)))
        #i2fs.addDevice(i2c.SFP(i2fs, "FSFP7", "gbt11",mux=(0,7)))
        # 14 gbts:
        sfp= i2c.SFP(i2fs, "FSFP0", "gbt9",  mux=(0,0))
        if sfp.purpose: i2fs.addDevice(sfp)
        sfp= i2c.SFP(i2fs, "FSFP1", "gbt10",  mux=(0,1))
        if sfp.purpose: i2fs.addDevice(sfp)
        sfp= i2c.SFP(i2fs, "FSFP2", "gbt11",  mux=(0,2))
        if sfp.purpose: i2fs.addDevice(sfp)
        sfp= i2c.SFP(i2fs, "FSFP3", "gbt12",  mux=(0,3))
        if sfp.purpose: i2fs.addDevice(sfp)
      else:
        print("FMC devices not checked (not itsmft logic)")
    #
    # ctpemu: must preceed tdomux device (TRD)
    ctpemu= self.add_ctpemu("ctpemu")
    emumode= ctpemu.getDETMODE()
    tnn= "ttccnts" ; ttccnts= None
    if self.nodExists("ttccnts", topnode=True):
      # fi >=v14: ttccnts node for new TTC counters
      try:
        self.atbdevs[tnn]= ttccnts= counters.CountersTTC(self, "ttccnts", tnn)
        print("TTC counters:", tnn)
      except:
        if tnn in self.atbdevs: del self.atbdevs[tnn]
        traceback.print_exc(file=sys.stdout)
    elif self.nodExists("ttc", topnode=True):
      # fi< v14: ttccnts node does not exist
      try:
        self.atbdevs[tnn]= ttccnts= counters.CountersTTC_13(self, "ttc", tnn)
        print("TTC counters(obsolete fi < v14 !):", tnn)
      except:
        if tnn in self.atbdevs: del self.atbdevs[tnn]
        traceback.print_exc(file=sys.stdout)
    else:
      printError("%s not available"%tnn, "(seems old version of .xml or firmware)")
    if emumode==1: ttccnts= None    # ttccnts not needed for CRU, although available with 'ttccnts' cmd
    self.add_outcnts("outcnts", ttccnts)
    tnn= "cnts"
    self.add_outcnts("cnts", None, counters.CountersLTU, devname="icnts")
    print("LTU ph. inputs counters:", "icnts")
    #
    tnn= "leco1"
    if self.nodExists("leco1", topnode=True):
      tnn= "leco1"
    elif self.nodExists("oltrx_lec", topnode=True):
      tnn= "oltrx_lec"
    else:
      printError("Link error (upstream) counters not available (leco1, resp. oltrx_lec for ltu_logic fi>14)")
    try:
      self.atbdevs[tnn]= counters.CountersLECO(self, tnn) 
      print("Link error (upstream) counters:", tnn)
    except:
      if tnn in self.atbdevs: del self.atbdevs[tnn]
      traceback.print_exc(file=sys.stdout)
    #
    # emumode: 1/CRU 2/TTC 3/TRD 4/CPV and corresponding stdalone tdomux:
    # tdomux:  1     1     2     2
    if emumode>2:
      stddef= 2
    else:
      stddef= 1
    tnn= "tdomux"
    self.atbdevs[tnn]= tdomux.TDOmux(self, stddef)
    print("TDO mux:", tnn)
    tnn= "tdomuxup"
    self.atbdevs[tnn]= tdomux.TDOmuxup(self)
    print("TDO mux upstream:", tnn)
    tnn= "tdimux"
    self.atbdevs[tnn]= tdomux.TDImux(self)
    print("TDI mux:", tnn)
    self.atbdevs["fpga"].finishltuinit()
    tnn= "hbmb"    # must be after emu
    if self.nodExists("hbm_present_pos_map", topnode=True):
      self.atbdevs[tnn]= atb.hbmb.HBMB(self)
      print("Heart Beat Manager Builder:", tnn)
    if (self.pllwasloaded or isinargs("emuinit")):
      rcmsg= self.atbdevs["fpga"].fpgainit()
      print("fpgainit:", rcmsg)

def Init():
  return ipbuswrap.Init()
def OpenATBoard(bname, role=None, pllmode=None):
  """ 
role: class name: (ABoard,CTPBoard, LTUBoard,... )
      None: use alice_cf file and .xml name to to choose the instance
pllmode: pllnocheck, pllcheckonly or one of names defined in i2c.plls
  "s", "g": should not be used here -allowed only in checkgs() call
rc: None if board not found
"""
  if bname in AllBoards:
    #print "=== OpenATBoard(%s) using existing instance."%bname
    pass
  else:
    if role!= None:
      binstance= role(bname, pllmode)
    else:
      atable= saxipbus.findAddrTable(bname)
      if atable==None:
        return None
      if atable.find("ltu_logic") == 0:
        binstance= LTUBoard(bname, pllmode)
      elif atable.find("ctp_logic") == 0:
        binstance= CTPBoard(bname, pllmode)   # in case we want pllin1o0lvds
      elif atable.find("ttcit_logic") == 0:
        binstance= TTCITBoard(bname)
      else:
        binstance= ABoard(bname)
    AllBoards[bname]= binstance
  return AllBoards[bname]

def main():
  if len(sys.argv) < 2 or (len(sys.argv)==2 and sys.argv[1]=="help"):
    print("""
Usage:
atbwrap.py atb_name

""")
    sys.exit()
  phbname= sys.argv[1]
  ipbuswrap.Init()
  atb= OpenATBoard(phbname)
  print("tc74 fpga temperatures",atb.bname,atb.getSlow())

if __name__ == "__main__":
  main()

