#!/usr/libexec/platform-python
#!/usr/bin/python3
from __future__ import print_function
from builtins import hex
from builtins import range
from builtins import object
DumpFormats= {
"dpon":(("TType","Trigger Type",0,32),
  ("BCID", "BCID", 32, 12),
  ("L", "Trigger Level/Spare", 44, 4),
  ("Orbit", "Orbit", 48, 32),
  ("V", "Trigger Type Valid", 119, 1),
  ("H", "Header Flag", 120, 1),
  ("WC", "Word Count", 121, 7)),
  #("HBDRp", "HBDR payload", 128, 40)),
  #("V", "HBDR Valid", 199, 1)),
"upon":(
  ("ONU1", "ONUad -TTCPON ONU address filled by LTU", 0, 8),
  ("HBid", "HBid: Orbit of ACKed/NACKed or current", 8, 32),
  ("CRUid", "CRU id ALICE-wide", 40, 10),
  ("A", "CRU acknowledge", 50, 1),
  ("B", "Buffer status", 51, 2),
  ("LC", "Link Monitor Counter", 55, 7),
  ("T", "TRD busy", 62, 1),
  ("V", "Valid bit", 63, 1)),
"upon2":(
  ("ONU1", "ONUad -TTCPON ONU address filled by LTU", 0, 8),
  ("HBid", "HBid: Orbit of ACKed/NACKed or current", 8, 32),
  ("CRUid", "CRU id, ALICE-wide", 40, 10),
  ("A", "CRU acknowledge", 50, 1),
  ("B", "Buffer status", 51, 2),
  ("LC", "Link Monitor Counter", 55, 7),
  ("T", "TRD busy", 62, 1),
  ("V", "Valid bit", 63, 1),
  ("ONU2", "ONUad -filled by LTU. Upper part only with more OLTs!", 64, 8),
  ("HBi2", "HBid: Orbit of ACKed/NACKed or current", 16+56, 32),
  ("CRUi2", "CRU id, ALICE-wide", 16+32+56, 10),
  ("A2", "CRU acknowledge", 16+42+56, 1),
  ("B2", "Buffer status", 16+43+56, 2),
  ("LC2", "Link Monitor Counter", 16+47+56, 7),
  ("T2", "TRD busy", 16+54+56, 1),
  ("V2", "Valid bit", 16+55+56, 1)),
"ctpltu":(("BCID", "BCID", 0, 12),
  ("Orbit", "Orbit", 12, 32),
  ("TT1", "Trigger Type det1", 44, 32),
  ("L1", "Level det1", 76, 4),
  ("V1", "Valid det1", 80, 1),
  ("TT2", "Trigger Type det2", 81, 32),
  ("L2", "Level det2", 113, 4),
  ("V2", "Valid det2", 117, 1),
  ("TT3_10", "Trigger Type det3(low10)", 118, 10)),   # 32
  #("L3", "Level det3", 150, 4),
  #("V3", "Valid det3", 154, 1)),
"ttcit":(("D", "ttc_data_i.clk40_gated", 0, 1),
  ("DL", "Low 16bits", 1, 16),
  ("E", "Address enable bit", 17, 1),
  ("DH", "High 14 bits", 18, 14),
  ("S", "ttc_data_i.adr_strobe", 32, 1),
  ("BR", "BRC bits: be d4 t2= xxPTRAxx", 33, 8),
  ("BRstr", "BRC strobe", 41, 1),
  ("1", "ttc_data_i.l1accept", 42, 1),
  ("ST", "Status bits: ready err_sng err_dbl err_comm div_nrst", 43, 5),
  ("A", "A-channel _proc bits: CAL_signal L1_signal L0_signal", 48, 3),
  ("TType", "Trigger Type (32bits)", 51, 32),
  ("Orbit", "Orbit", 83, 32),
  ("BCid", "BC id 12 bits", 115, 12),
  ("F", "SSM_flag", 127, 1)),
"hbmb":(("H", "HBr_request", 0,1),
  ("B", "Busy_OR /125ns", 1, 1),
  ("BsyVect", "Busy Vector 35:0 /125ns",2,36),
  ("S", "Busy strobe /125ns", 38, 1),
  ("Garbage", "garbage", 39, 97)),
"gbt1tx":(("User", "User data 79:0", 0, 80),
  ("SC", "Slow Control", 80, 4),
  ("S", "TxIsDataSel", 84,1),
  ("0", "Zeros", 85, 43)),
"gbt1rx":(("User", "User data 79:0", 0, 80),
  ("SC", "Slow Control", 80, 4),
  ("S", "RxIsDataSel", 84,1),
  ("T", "GBT Rx Tick", 85, 1),
  ("0", "Zeros", 86, 42)),
"gbt1rxu3":(("U0", "User data 31:0", 0, 32),
  ("U1", "User data 63:32", 32, 32),
  ("U2", "User data 79:64", 64, 16),
  ("SC", "Slow Control", 80, 4),
  ("S", "RxIsDataSel", 84,1),
  ("T", "GBT Rx Tick", 85, 1),
  ("0", "Zeros", 86, 42)),
"cpv":(("TType", "Trigger Type", 0, 32),  # first 80 bits of dpon
  ("BCID", "BCID", 32, 12),
  ("L", "Level", 44, 4),
  ("Orbit", "Orbit", 48, 32)),
"ctpcore":(("I32", "CTP inputs 32..1", 0, 32),
  ("I48", "CTP inputs 48..33", 32, 16),
  ("LMGrb", "LM Generators rnd2 rnd1 bc2 bc1", 48, 4),
  ("LMb", "LMclass before vetos", 52, 6),
  ("LMa", "LMclass after", 58, 6),
  ("LMclu", "LM cluster", 64, 6),
  ("L0Grb", "L0 Generators rnd2 rnd1 bc2 bc1", 70, 4),
  ("L0b", "L0class before vetos", 74, 6),
  ("L0a", "L0class after", 80, 6),
  ("L0clu", "L0 cluster", 86, 6),
  ("L1Grb", "L1 Generators rnd2 rnd1 bc2 bc1", 92, 4),
  ("L1b", "L1class before vetos", 96, 6),
  ("L1a", "L1class after",102, 6),
  ("L1clu", "L1 cluster",108, 6))
}
DumpFormats["gbt2tx"]= DumpFormats["gbt1tx"]
DumpFormats["gbt2rx"]= DumpFormats["gbt1rx"]
ttbits= ("TOF","TPC_RST","TPC_SYNC","GAP2","GAP1","","","", # 31..24
    "", "", "", "", "", "", "", "",                         # 23..16
    "", "RS", "RT", "FErst", "TF", "EOC", "SOC", "EOT",     # 15..8
    "SOT", "CAL", "PP", "PH", "HC", "HBr", "HB", "ORB")     # 7..0

#!/usr/bin/python3
Valids= { "dpon":(119), "upon":(55), "upon2":(55,55+64),
  "ctpltu":(80,117,154) }

def prtbits(bitsdef, ebits):
  ib= 31 ; rc= ""
  while ib>=0:
    if ebits & (1<<ib):
      tx= bitsdef[31-ib]
      if tx!="":
        rc= rc+tx+" "
    ib= ib-1
  return rc

def GetField(fmt, fieldn):
  ''' fmt: format name  (i.e. "dpon","ttcit",...
field: fieldname: name of the field in format
rc: word(0..3), rel_position_in_word, bmask
    bmask: inside the word. 
DOES NOT WORK FOR FIELD IN MORE WORDS!
'''
  w3= None; bitpw= None; msk= None
  if fmt in DumpFormats:
    for ix in range(len(DumpFormats[fmt])):
      field= DumpFormats[fmt][ix]
      if field[0]==fieldn:
        w3= 3- (field[2]//32)
        #w3= field[2]//32
        bitpw= field[2]%32    # position in word
        maxbits= 32- bitpw    # max. bits in this word
        if maxbits<field[3]:
          print("Warning: only lowest %d bits considered in word %d."%(maxbits,w3))
        else:
          maxbits= field[3]
        msk= (2**maxbits-1) << bitpw
        print("dumpssm.GetField(%s,%s) rc:"%(fmt,fieldn), w3,bitpw, hex(msk))
        return w3, bitpw, msk
    print("dumpssm.py: field %s not found for format %s"%(fieldn, fmt))
  else:
    print("dumpssm.py: %s not found"%(fmt))
  return w3, bitpw, msk
class Dump(object):
  def __init__(self, fdef, filterc=None, full=False):
    """fdef: format item -one of items in DumpFormats dictionary
filterc: string "& namval namval ..."     log. and of signals
                "| namval namval ..."     log. or of signals
                where namval is:
                name=val    -compare if equal
                name        -compare if !=0
                name!val    -check if at least 1 bit set
         only lines for which filterc is True are displayed with getLine()
  full: False -supress equal lines
"""
    #print "dumpssm.Dump:",fname
    #self.flds= DumpFormats[fname]
    # self.flds: list of items. item: (name, description, position, length)
    # the list must be ordered by increasing position
    self.flds= fdef
    self.full= full
    # filterc -> filterf,conds 
    self.conds= []       # list of names defined conditions, empty if no conditions defined
    self.condsTF= []     # list of defined conditions True/False
    self.compare= []     # True: compare value, False: compare if 1 bit set
    self.values= []      # values used when compare needed (None: no value given)
    self.filterf= None   # & or | if filter active
    self.finalcond= None # None: not decided yet, True: ok show line, False: ok discard line
    self.lineN= 0   # number of lines returned
    self.lastdw= None    # the last (previous getLine) TType value
    self.ttls= {}   # ttls["legend"]: number of line (lineN) when printed
    if filterc!=None:
      filterc_l= filterc.split()
      self.filterf= "&"  # default (i.e. if forgotten)
      if filterc_l[0]=="|":
        self.filterf= "|"
      # prepare conditions:
      for nameval in filterc_l:
         if nameval in ('&','|'): continue
         compset= True
         value= None ; eqex= None
         if nameval.find("=")>0:
           eqex= '='
         elif nameval.find("!")>0:
           eqex= '!'
           compset= False
         else:
           pass
         if eqex != None:
           nava= nameval.split(eqex)
           name= nava[0]
           if (len(nava)==2):
             if len(nava[1])>0:
               value= eval(nava[1])
           else:
             print("Bad field name=val (too many '=' or '!' ?): %s"%nameval)
         else:
           name= nameval
         if name in [sn[0] for sn in self.flds]:
           #print "preparing conds:",name
           #self.conds[name]= False  
           self.conds.append(name)
           self.compare.append(compset)
           self.values.append(value)   # compare against this value
           self.condsTF.append(False)
         else:
           print("Unknown field name: %s"%name)
  def clearConditions(self):
    self.finalcond= None
    self.false_conds_met= 0
    #for name in self.conds:
    for ix in range(len(self.conds)):
       self.condsTF[ix]= False  
  def getFL(self, ixd):
    """ get a field length of item self.flds[ixd]
"""
    flen= (self.flds[ixd][3]+3) // 4    # field data length
    if len(self.flds[ixd][0]) > flen:   # field name length
      flen= len(self.flds[ixd][0])
    return flen
  def getHeader(self):
    hdr= ""
    for ixd in range(len(self.flds)-1,-1,-1):
      #flen= (self.flds[ixd][3]+3) // 4    # data length
      flen= self.getFL(ixd)
      fmt= "%%s %%%d.%ds"%(flen,flen)
      #print "hdr fmt:", fmt
      hdr= fmt%(hdr, self.flds[ixd][0])
    return hdr
  def getHelp(self):
    #htx= ""
    htx= "%8.8s%4s %3s %s\n"%("SHORT  ", "FROM", "LEN", "MEANING_OF_THE_BITS")
    for ixi in range(len(self.flds)):
      field= self.flds[ixi]
      #print "dumpssm.py:",field
      #      shortName from len long
      line= "%8.8s %3d %3d %s"%(field[0], field[2], field[3],field[1])
      htx= htx+line+'\n'
    return htx
  def checkname(self, ix, dw):
    #if self.values[name]!=None:
    if self.values[ix]!=None:
      #compv= self.values[name]
      compv= self.values[ix]
      if self.compare[ix]:
        # = -> compare if equal
        if dw==compv:
          return True
        else:
          return False
      else:
        # ! -> check if at least 1 bit set
        if (dw&compv)!=0:
          return True
        else:
          return False
    else:
      compv= 0
      if dw==compv:
        return False
      else:
        return True
  def getLine(self, data):
    """ data: [wordhighest, wordhigh, wordlow, wordlowest]
                           [wordhigh, wordlow, wordlowest]
                                     [wordlow, wordlowest]
                                              [wordlowest]
max 4 words, each 32 bits
rc: output line or
    "" in case condition not fullfilled
    " ..." in case when full=False and:
      -filter not applied AND
      -dpon + TType bits the same as in previous getLine call
"""
    ttypelegend= ""
    li= len(data)-1
    w128= data[li]   # 3
    for ixd in range(li-1, -1, -1): # 2,1,0
      nwshifted= data[ixd]<<(32*(li-ixd))
      w128= nwshifted | w128
      #print "%d: %x-> w128 result: %x"%(ixd, nwshifted, w128)
    line= ""
    self.clearConditions()
    for ixd in range(len(self.flds)-1,-1,-1):
      #flen= (self.flds[ixd][3]+3) // 4
      flen= self.getFL(ixd)
      fblen= self.flds[ixd][3]
      fname= self.flds[ixd][0]
      dw= (w128 >> self.flds[ixd][2]) & eval("0b"+"1"*fblen)
      #if (fname in self.conds) and (self.finalcond==None):
      if self.finalcond==None:
       for ix in range(len(self.conds)):
        if fname != self.conds[ix]: continue
        #   checked            and  result not known yet
        #print "cond:",fname,"dw:",hex(dw)
        if self.filterf=="&":   # &: at least 1 false ->discard line
          #if dw==0:
          #if not self.checkname(fname, dw):
          if not self.checkname(ix, dw):
            #print "final False (at least one in & cond), discarding..."
            return ""
        else:                   # |: all false ->discard line
          #if dw!=0:
          #if self.checkname(fname, dw):
          if self.checkname(ix, dw):
            #self.conds[fname]= True
            self.condsTF[ix]= True
            self.finalcond= True   # no need to check remaining (|)
            #print "final True (at least one in | cond), ..."
          else:
            self.false_conds_met= self.false_conds_met+1
      #print "dumpssm fname:",fname
      if fname=="TType":
        if (self.filterf==None) and (self.lastdw==dw) and (not self.full):
          # Apply ONLY without filter and full=True
          return " ..."
        else:
          self.lastdw= dw
          ttypelegend= prtbits(ttbits, dw)
      fmt= "%%s %%%dx"%flen
      line= fmt%(line, dw)
    if self.filterf=="|": 
      if (len(self.conds)>0) and (self.false_conds_met== len(self.conds)):
        #print "final False (all False in | cond), ..."
        return ""
    if ttypelegend!="":
      if ttypelegend in self.ttls:
        if (self.lineN-10) > self.ttls[ttypelegend]:
          self.ttls[ttypelegend]= self.lineN
          line= line+" "+ttypelegend
      else:
        self.ttls[ttypelegend]= self.lineN
        line= line+" "+ttypelegend
    self.lineN= self.lineN+1
    #print "dumpssm line:",line
    return line

def mymain():
  t= (("tt","Trigger type", 0, 32),
      ("databitlen9", "data bits of len 9", 32, 9), 
      ("V", "Valid bits", 64, 4)) 
  d= Dump(t)
  d= Dump(t, "| V=1 V=0x7")   # 0x1 or 0x7 in 4 bits from 64
  print(d.getHelp())
  print(d.getHeader())
  print(d.getLine((0x05050505,0x12345671, 0x12345678,0x345678)))
  print(d.getLine((0x05050505,0x12345673, 0x12345678,0x345678)))
  print(d.getLine((0x05050505,0x12345673, 0x12345678,0x345678)))
  print(d.getLine((0x05050505,0x12345673, 0x12345678,0x345678)))
  print(d.getLine((0x05050505,0x12345677, 0x12345178,0x37)))
  print(" dpon.TType:",GetField("dpon", "TType")) 
  print(" TTCit.BCid:",GetField("ttcit", "BCid")) 
  print(" TTCit.BR:",GetField("ttcit", "BR")) 
  print(" TTCit.Orbit:",GetField("ttcit", "Orbit")) 
  print(" TTCit.F:",GetField("ttcit", "F")) 
if __name__ == "__main__":
  mymain()

