#!/usr/libexec/platform-python
from __future__ import print_function
from builtins import input
from builtins import range
import os, sys, time
import fcntl
""" Mutual lock (advisory) among processes on one computer
See also .c version at cwrap/examples/myflock.c
Usage:
import fcntlflock
rc= fcntlflock.acquire(bname, timeout)   # non blocking (0) by default
    True: acquired
    False: not acquired, reason: timeout or error (the same process)
rc= fcntlflock.release(bname)
    True: released

Notes:
1. when lock requested for already locked file from other process:
(<type 'exceptions.IOError'>, IOError(11, 'Resource temporarily unavailable'))
2. when locked from afs account: cannot release from another (non afs) account
3. double lock (from the same process) allowed 
   -prevented by semp dictionary (also for different thread)
4. double lock from other process not allowed (the same account, different shell)
5. as 3. for release
6. as 4 for release
7. lock released automatically when process owning the lock finishes

py:
./fcntlflock.py

shell:
who executed when /tmp/atbff unlocked, waiting 3s for non-locked state:
flock -x -w 3 /tmp/atbff who  (wait 3 secs)

c:
./myflock   (see flock.c)
"""
waitslot= 0.05   # secs
semp= {}
def acquire(bname="atb_0", waitsecs=0):
  """ waitsecs: <0 blocking lock
                >=0: non blocking lock (waiting waitsecs)
  rc: True:  ok, acquired
      False: not acquired (for non-blocking), or an error (e.g. already acquired)
  """
  global semp
  if (bname in semp)!=False:
    print("Error: already acquired, the same process. ")
    #print "Warning: already acquired, the same process (may be different thread). "
    return False
  fn= "/tmp/"+ bname +".lock"
  if not os.access(fn, os.W_OK):
    if not os.access(fn, os.F_OK):
      f= open(fn,"w")
      f.close()
      os.chmod(fn, 0o666)
    else:
      print("Error: %s not writtable"%fn)
  sempfd= open(fn, "w+")
  if waitsecs<0:   # blocking lock
    rc= True
    fcntl.flock(sempfd, fcntl.LOCK_EX)  # returns immediately in the same process 
    semp[bname]= sempfd                 # even if in different thread
    return rc
  sleeptime= 0
  #print "acquire", fn
  while sleeptime <= waitsecs:
    try:
      rc= True
      fcntl.flock(sempfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
      semp[bname]= sempfd
      break
    except IOError:
      #print "IOError exception:",sys.exc_info()[1].errno, "\n", sys.exc_info()
      if sys.exc_info()[1].errno==11:
        rc= False
        time.sleep(waitslot)
        sleeptime= sleeptime + waitslot
    except:
      print("acquire exception:", "\n", sys.exc_info())
      rc= False
      break
  return rc
def release(bname="atb_0"):
  """ release lock. rc: True: ok, released. False: cannot release
  """
  global semp
  if (bname in semp)==False:
    print("Error: cannot release (not acquired)")
    return False
  #fn= "/tmp/"+ bname +".lock"
  #semp[bname]= open(fn, "w+")   open in acquire!
  fcntl.flock(semp[bname], fcntl.LOCK_UN)
  semp[bname].close()
  del semp[bname]
  return True
def main():
  print("""
a secs bname -acquire lock, wait secs if not available (-1: forever),
              file: /tmp/bname.lock
r bname      -release
b bname N    -bench mark:  for ix in range(N): 'acquire -1 bname ; r bname'

Examples:
a            -blocking lock atb_0
a -1 atb_ff  -blocking lock atb_ff
a 0 atb_11   -non-blocking lock atb_11, no timeout
a 2          -non-blocking lock of atb_0, timeout 2 secs
""")
  while True:
    line=input("a[cquire] [secs [bname]] r[elease] (q to quit):")
    if len(line)==0 :continue
    ls= line.split()
    if ls[0]=='q':break
    try:
      if ls[0]=='r':
        if len(ls)>1:
          bname= ls[1]
        else:
          bname= "atb_0"
        rc= release(bname)
        print("release rc:",rc)
      elif ls[0]=='a':
        tout= -1
        bname= "atb_0"
        if len(ls)>1:
          tout= float(ls[1])
        if len(ls)>2:
          bname= ls[2]
        #print "acquire ",tout, bname, "..."
        rc= acquire(bname, tout)
        print("acquire", tout," %s rc:"%bname,rc)
      elif ls[0]=='b':   # b bname N
        if len(ls)>2:
          bname= ls[1]
          N= eval(ls[2])
          t1= time.time()
          for ix in range(N):
            rc= acquire(bname, -1)
            if rc: 
              rc= release(bname)
            else:
              print("Error: %s not acquired."%bname)
              break
          t2= time.time()
          print("acquire/release %s loops: %d time: %.2f secs"%(bname,ix,(t2-t1)))
        else:
          print("b bname N expected")
      else:
        print("bad cmd.")
    except:
      #print "%s exception:"%ls[0],"\n", sys.exc_info()[:2]
      print("%s exception:"%ls[0],"\n", sys.exc_info())
if __name__ == "__main__":
   main()

