#!/usr/libexec/platform-python
"""
14.4.2018 ipbuswrap v2. goal: shorter (fornon-interactive use ?), used by checkboards.py.
"""
from __future__ import print_function
from builtins import str
from builtins import object
import uhal,sys,os,time,types
import findat
manager= None
CTP3_ROOT= os.getenv("CTP3_ROOT")
AliceBoards= os.path.join(os.getenv("CTP3_CONFIG_DIRECTORY"),"alice_boards.xml")
class Board(object):
  """ Usage:
  import ibw
  ibw.Init("path to connection_file") 
  board= ibw.Board("boardname")   # one of names in connection file
  # single register operations:
  readval= board.Read("regname")   # one of "SINGLE" names in
  board.Write("regname",value)     # board's .xml file
  # block operations, i.e. read/write array of values:
  values= board.ReadBlock("blockname",N,offset)   # mode="block"
  board.WriteBlock(blockname,values, offset)

  """
  def __init__(self, boardname):
    self.bname= boardname
    self.idslist=[]   # sorted node-names [node1, node1.id1, node1.id2, node2, node2.id1,...]
    self.nodlist={}   # key-access to each node nodes+leaves
    self.tn_inuse=""
    self.level= 0
    self.dispisin= True
    self.hw = manager.getDevice(boardname)
    self.idslist= self.hw.getNodes(".*")   # all nodes (id id.id ...)
    self.idslist.sort()
    # ['ctrl_reg', 'ctrl_reg.id', 'ctrl_reg.rst', 'pram', 'pram.addr', 'pram.data', 'ram', 'reg']
    #print "idslist:", self.idslist
    ixnod=0
    while ixnod < len(self.idslist):
      # ixnod:  points now to item to be processed
      ixnod= self.scannode("", ixnod)
    #print "nodlist:", self.nodlist
  def scannode(self, parentname, ixnod):
    # ixnod: start processing from self.idslist[ixnod]
    # rc: index to idslist pointing after 'last processed item'
    DBGLIST= False
    while ixnod < len(self.idslist):
      name= self.idslist[ixnod]
      nod= self.hw.getNode(name)
      mode= nod.getMode()
      descr= nod.getDescription()
      if str(mode)=='HIERARCHICAL':
        if DBGLIST:
          print("HIERARCHICAL: %s %s:"%(name, nod), descr)
        self.nodlist[name]= nod
        self.level=+ 1
        ixnod= self.scannode(name, ixnod+1)   # pointing after last processed item
      else:
        if name.startswith(parentname):
          if str(mode)=="INCREMENTAL":
            dims= "[%d]"%nod.getSize()
            if DBGLIST:
              print(" "*self.level + "%s %s %s:"%(mode, name, str(nod)+dims), descr)
          else:
            if DBGLIST:
              print(" "*self.level + "%s %s %s:"%(mode, name, nod), descr)
          self.nodlist[name]= nod
        else:
          self.level=- 1
          break
        ixnod += 1
    return ixnod   # end of list reached
  def nodExists(self,nname,topnode=False):
    if nname in self.nodlist:
      mode= self.nodlist[nname].getMode()
      #print "nodExists %s mode:%s: type:"%(nname, mode), type(mode)
      if topnode:
        if str(mode) == "HIERARCHICAL":
          return True
        else:
          return False
      else:
        return True
    else:
      return False
  def applyTN(self, name):
    pref= ""
    if self.tn_inuse != "":
      names= name.split(".")
      #if (len(names) > 1) and ( self.tn_inuse != names[0]):
      if len(names) == 1:
        # apply correction only for single names
        pref= self.tn_inuse+"."
    return pref+name
  def mydispatch(self):
      if self.dispisin: self.hw.dispatch()
  def dispatch(self):
      self.hw.dispatch()
  def Read(self,name, offset=None):
    regname= self.applyTN(name)
    if offset==None:
      register= self.nodlist[regname].read()
      self.mydispatch()
      readval= register.value()
    else:
      block= self.nodlist[regname].readBlockOffset(1,offset)
      self.mydispatch()
      readval= block[0]
    return readval
  def Write(self,name, value, offset=None):
    self.Write_nd(name,value, offset)
    self.mydispatch()
  def Write_nd(self,name, value, offset=None):
    regname= self.applyTN(name)
    if offset==None:
      self.nodlist[regname].write(value)
    else:
      block= []; block.append(value)
      self.nodlist[regname].writeBlockOffset(block, offset)
  def ReadBlock(self,name,N,offset):
    #print "ReadBlock:", name, N, offset
    block= self.nodlist[self.applyTN(name)].readBlockOffset(N,offset)
    self.mydispatch()
    return block
  def WriteBlock_nd(self,name,block,offset):
    self.nodlist[self.applyTN(name)].writeBlockOffset(block, offset)
  def WriteBlock(self,name,block,offset):
    self.WriteBlock_nd(name,block,offset)
    self.mydispatch()
  def WriteFifo(self,name,block):
    self.nodlist[self.applyTN(name)].writeBlock(block)
    self.hw.dispatch()
  def ReadFifo(self,name,N):
    block= self.nodlist[self.applyTN(name)].readBlock(N)
    self.hw.dispatch()
    return block
def Init(phbname=None):
  """ Set global manager  pointing to a connection file with used board(s)
rc: global manager
    None: ipbus init problem
"""
  global manager
  if manager: 
    print("ibw.Init: already called...")
    return manager
  #uhal.disableLogging()
  uhal.setLogLevelTo( uhal.LogLevel.WARNING ) # DEBUG INFO NOTICE WARNING ERROR FATAL
  if phbname=="alice_cf":
    atbublaze= os.path.join(os.getenv("CTP3_CONFIG_DIRECTORY"),"alice_cf.xml")
    manager= uhal.ConnectionManager("file://"+atbublaze)
    print("ibw.Init: using STATIC %s connection file..."%atbublaze)
    return manager
  if phbname=="cwd":
    atbcwd= os.path.join(os.getcwd(), "alice_boards.xml")
    manager= uhal.ConnectionManager("file://"+atbcwd)
    print("ibw.Init: using STATIC %s connection file..."%atbcwd)
    return manager
  #uhal.setLogLevelFromEnvironment( env_variable_name )
  #
  # check boardid/fwinfo
  manager= uhal.ConnectionManager("file://"+AliceBoards)
  if phbname==None: return manager
  board= Board(phbname)
  boardid=fwinfo=None
  cstat="ctrl."
  if board.nodExists(cstat+"boardid"):
    boardid= board.Read(cstat+"boardid") & 0xff
  if board.nodExists(cstat+"fwinfo"):
    fwinfo= board.Read(cstat+"fwinfo"); 
  #print "%s: boardid:%d fwinfo:0x%x"%(board.bname, boardid, fwinfo)
  if (boardid == None) or (fwinfo == None):
    print("Board %s: cannot read %sboardid:"%(phbname,cstat),boardid," or %sfwinfo:"%cstat, fwinfo)
    manager= None
    return None
  #
  # create phbname.xml, use fi_numbers.cfg
  fwname= fwinfo>>24 ; fwver= (fwinfo>>8) & 0xffff
  # 5 1 -> test_logic_v1.xml
  adxml= findat.FindAT(fwname, fwver)
  if adxml==None:
    print("Board %s: fwinfo: 0x%x not known (cannot assign an address table)"%(phbname, fwinfo))
    print("Modify pywrap/findat.py or add an entry in %s"%AliceBoards)
    manager= None
    return
  remainsfromipbuswrap="""
  addrtable= os.path.join(os.getenv("CTP3_CONFIG_DIRECTORY"), adxml)
  phxml="<connections>
  <connection id="%s" uri="chtcp-2.0://localhost:10203?target=%s:50001" address_table="file://%s"/>
</connections>
" %(phbname, ip, addrtable)
  cnout= open(phbname+".xml", "w")   # opening in ~/WORK directory
  cnout.write(phxml); cnout.close()
  print "Connection file:", os.path.join(os.getcwd(), (phbname+".xml"))
  print "Address table:", addrtable
  manager= uhal.ConnectionManager("file://"+phbname+".xml")
"""
  print("%s: boardid: 0x%x fwinfo: 0x%x adrtable: %s"%(phbname, boardid, fwinfo, adxml))
  return manager
def main():
  if len(sys.argv) < 2 or (len(sys.argv)==2 and sys.argv[1]=="help"):
    print("""
Program usage:
import ibw
ibw.Init("path_to_connection_file")
board= ibw.Board("boardname");
value= board.Read("reg_name")
board.Write("reg_name", value)
...
""")
    sys.exit()
  bname= sys.argv[1]
  Init(bname)
if __name__ == "__main__":
  main()

