#!/usr/libexec/platform-python
from __future__ import division
from __future__ import print_function
from builtins import str
from builtins import hex
from builtins import map
from builtins import range
from builtins import object
from past.utils import old_div
import string,types,random,symbreg
import struct
import time

helptext="""
spi.py: let's keep here
- SPIcore: general spi access
- classes for SPI devices, i.e.
  N25Q (flash memory)

SPI commands:
si N  -choose which spi is used 0:StartUpE3 or 1: direct
ss    -show regs (CTRL, SS, Divider)
ss01 [0|1] -set Slave Select (when device not used: 0)

Not valid(i.e. not tested):
sr    -read 1 byte (NBITS bits)

Usage:
spim= SPIcore(board, "spi0")

"""
#!/usr/bin/python3
def prtErr(msg):
  print("ERROR:%s"%msg)
class SPIcore(object):
  Adata= 0   #"spiX.d0"
  Actrl=Astatus= 4
  Adivider= 5
  Ass= 6
  SSauto= 1<<13
  LSBfirst= 1<<11
  TXneg= 1<<10   # with CLK falling edge
  RXneg= 1<<9
  Bgobusy= 1<<8
  low= 0xffff
  def __init__(self, board, spislave="spi1", sslogic=True):
    """ board: parent board
    spislave: "spi0" or "spi1"
"""
    self.board= board   # has to be connected
    self.spislave= spislave   # has to be connected
    # Note: bit ctrl.ctrl 0x2: has to be 0 for spi0 
    self.spir= "%s.rs"%(self.spislave)
    self.connected= []   # connected devices, seems max. 1 device in ctpltu case
    self.defdev= 0   # default device
    self.sspositive= sslogic # True: positive logic for SS (seen inverted in ILA!)
    # spiclk= (wclk/(divider+1)/2, i.e.:
    # wclk div spiclk
    # 32M   0   32/2. = 16M
    # 32M   1   32/4. = 8M     ?
    # 32M   2   32/6. = 5...M  ?
    # max. clk for N25Q128A: 108Mhz
    # For system clock (wbclk) 32mhz (50M during powerup ?),
    # serial clock (sclk) is 16M (25M ?) when divider:0
    self.divider_val= 1 #0x7f   
    self.nbits= 8  
    #self.defbits= self.LSBfirst | self.SSauto | self.nbits
    #
    # RXneg: p17: 
    # def(0): Input data is latched in on the rising edge of the clock,
    # TXneg:
    # def(0): Output data is changed on the raising edge of the clock
    self.defbits=  self.nbits | self.RXneg | self.TXneg
    self.configured= False
    self.ctrl= symbreg.Register("CTRL", ASS=13, IE=12, LSBfirst=11, TXneg=10, RXneg=9,
      GObusy=8, NBITS=(0,7))
    #self.ctrl.prtme()
    #self.ctrl.prt(0xfff, "0xfff:")
    self.config()
  def addDevice(self, device):
    if len(self.connected)>0:
      prtERR("can't add device:%s (only 1 spi device supported)"%device.devname)
    else: 
      self.connected.append(device)
  def printDevs(self, ix=""):
    if len(self.connected)>0:
      dev= self.connected[0]   # 1 dev only on 1 bus!
      print("%s \t%s: \t%s/%s \t%s (%d)"%(ix,self.spislave, 
        dev.devname,   
        dev.model, 
        dev.getid(),
        len(self.connected)))
    else:
      print("%s: no device connected"%(self.spislave))
  def config(self):
    self.board.Write(self.spir,self.divider_val,self.Adivider)
    #
    #self.board.Write(self.spir,0,self.Ass)   # manual ASS
    self.setSS(0, False)
    # CTRL
    self.board.Write(self.spir,self.defbits,self.Actrl)
    #self.status()
    self.configured = True
  def setSS(self, zo, log=False):
    # zo:   0: unselect(deactivate)    1:select (activate)
    if self.sspositive:
      if zo==1:
        ss= 1<<self.defdev
      else:
        ss= 0
    else:
      if zo==1:
        ss= (0xffffffff & (~(1<<self.defdev)))
      else:
        ss= 0xffffffff
    self.board.Write(self.spir,ss,self.Ass) 
    if self.sspositive:
      logic="pos"
    else:
      logic="neg"
    if log:
      print("            %s SS: %d logic:%s -> 0x%x"%(self.spislave,zo, logic, ss))
    #if (self.sspositive and (zo==1)) or \
    #   ((not self.sspositive) and (zo==0)):
    #  ss= 1<<self.defdev
    #else:
    #  ss= (0xffffffff & (~(1<<self.defdev)))
    #self.board.Write(self.spir,ss,self.Ass) 
  def status(self, text=" "):
    ctrld = self.board.Read(self.spir, self.Actrl)
    ss = self.board.Read(self.spir, self.Ass)
    divider = self.board.Read(self.spir, self.Adivider)
    if self.sspositive:
      logic="pos"
    else:
      logic="neg"
    print("%s SS (logic:%s):0x%x Divider:0x%x"%(self.spislave,
      logic, ss, divider))
    self.ctrl.prt(ctrld)
  def stran(self, doutp=None, datalen=8):
    """ start transmition of 
    doutp-bytes or 
    dummydata-byte (when doutp==None)
    datalen: 8, 32, 64, 96, 128   -number of bits in 1 ipbus packet
    doutp: 
      None: in case of READ operation
      List of bytes in case of write op (max. length: 16x 1byte).
    Operation:
    - set data register (doutp(W) or dummydata(R))
    - set R/W bit (doutp:None -> R i.e. 0x4) -not needed!
    - config SPI data length +
    - start transition (change bit GO_BSY 0->1)
    - wait for end of transition (test GO_BSY)
    - read data register
    - rc: None in case of write op or timeout (i.e persistent BUSY), or 
          the read data:
          word  -in case of datalen ==8
          list of 1, 2,3 or 4 32bits words, byte order:
          [0123,4567,8 9 10 11, 12 13 14 15]
          i.e.:
          datalen: 8, 32, 64, 96, 128   -in bits
          words:   1   1   2   3    4
"""
    dummydata=0xc7c7c7c7 ; dout= []
    dummydata=0x12345678
    words= old_div((datalen-1),32) + 1   # 1 2 3 or 4
    if doutp!=None:   # WRITE op
      ixo=0
      for ix in range(words):
        d= 0
        for ixo in range(4):               #  bytes in a word
          if ixo>= len(doutp): break
          #d= d | (doutp[ixo]<<(8*ixo))    # [1,2,3,4] -> 04030201
          d= (d<<8) | doutp[4*ix + ixo]       # [1,2,3,4] -> 01020304
        dout.append(d)
      txdat= str(list(map(hex,doutp)))+" W"
      #rwbit= 0x0
    else:             # READ op
      for ix in range(words):
        dout.append(dummydata)
      txdat= str(list(map(hex,dout)))+" R"
      #rwbit= 0x4      
    # dout: max 4 words (16 bytes)
    #print "stran:dout:", map(hex,dout)
    for ix in range(words):
      self.board.Write_nd(self.spir,dout[words-1-ix],self.Adata+ix)   # reversed!
    #
    #self.board.Write("ctrl.ctrl", rwbit)   # not needed even for spi0 (StartUpE3)    
    # change GO_BSY: 0->1
    defbits= (self.defbits & 0xffffff80) | datalen
    self.board.Write(self.spir,defbits,self.Actrl)   # 0
    self.board.Write_nd(self.spir,(defbits| self.Bgobusy),self.Actrl)   # 1
    #time.sleep(0.3)
    busy= True ; loops= 0 ; datain= None
    while busy:
      status= self.board.Read(self.spir, self.Astatus)
      busy = (status & SPIcore.Bgobusy) > 0
      loops += 1
      if loops > 10000:
        prtErr("stran timeout, loops:%d"%loops) ; break
    if not busy:
      #if words==1:
      if datalen==8:
        datain = self.board.Read(self.spir, self.Adata) & 0xff
        #print "stran out:%s loops:%d datain:"%(txdat,loops), hex(datain)
      else:
        datain= []; temp = self.board.ReadBlock(self.spir, words, self.Adata)
        # 4 words (16 bytes) - words are coming in reversed order, also for 2,3 words
        for ix in range(words):
          datain.append(temp[words-ix-1])
        #print "stran out:%s loops:%d datain(reversed):"%(txdat,loops), map(hex,datain)
    else:
      print("stran timeout out:%s loops:%d"%(txdat,loops))
    return datain
  def cmdaddr(self, cmdad, datalen=32):
    # write instruction (4x 8 bit). maddr: 3 bytes addr
    #self.stran(cmdad[0])
    #self.stran(cmdad[1])
    #self.stran(cmdad[2])
    #self.stran(cmdad[3])
    # write instruction (1x 32 bits). maddr: 3 bytes addr
    self.stran(cmdad, datalen)
#!/usr/bin/python3
class GenDev(object):
  def __init__(self, spi, devname, model):
    self.spi= spi    # a SPIcore instance
    self.devname= devname
    self.model= model # may be overwritten by real model info from chip
  def getid(self):   # overwritten by device's method
    """ should return None or string composed from something read from device
"""
    return "not implemented"

class MAX11661(GenDev):
  """ MAX11661: 8 bits, 1 channel,
500ksps Conversion Rate, No Pipeline Delay
"""
  DEFDLEN= 16
  def __init__(self, spi, devname="adc", model=""):    # only ctplu
    GenDev.__init__(self, spi, devname, model)
  def prthelp(self):
    print("""
get -get                     - get current value (11 bits)
rreg dlen loops dummy        -read reg. 
     dlen: length in bits, default:16, 
     loops: in a loop (1)
     dummy: dummy read before reading reg

test [ms]            -write/read 32 values: 0,1,3,7,f,...,ffffffff
                      wait ms (10: default) milisecs between set/read
wttc val loops       -write (lopps times) val to ctrl.adc_test_data reg
""")
  def getid(self):
    adc= self.rreg(self.DEFDLEN, dummyr1=True)
    return "%s: %s"%(self.model, str(adc))
  def excmd(self,cmd):
    if len(cmd)<=1:
      self.prthelp()
      return
    if cmd[1]=='get':
      #if len(cmd)>2:
      #  data= self.rreg(self.DEFDLEN, True)
      #else:
      #  data= self.rreg(self.DEFDLEN, False)
      #self.prtbyte(data)
      data= self.rreg(self.DEFDLEN)[0]
      data= 0x7ff & (data>>4)
      print("11 bits: 0x%x"%data)
    elif cmd[1]=='rreg':
      dlen=self.DEFDLEN
      dummyr=False ; loops= 1
      if len(cmd)>2: dlen= eval(cmd[2])
      if len(cmd)>3: loops= eval(cmd[3])
      if len(cmd)>4: dummyr= True
      for ix in range(loops):
        if (ix%10000)==9999: print("loop:",ix)
        byte= self.rreg(dlen, dummyr)
      self.prtbyte(byte)
    elif cmd[1]=='wttc':
      loops= 1
      val= eval(cmd[2])
      if len(cmd)>3: loops= eval(cmd[3])
      for ix in range(loops):
        if (ix%10000)==9999: print("loop:",ix)
        self.spi.board.Write("ttc.adc_test_data",val)
    elif cmd[1]=='test':
      ms= 10
      if len(cmd)>2: ms= int(cmd[2])
      self.test(ms)
    else:
      self.prthelp()
  def rreg(self, dlen, dummyr1=False):
    self.spi.setSS(1)
    if dummyr1:   # 1 dummy read before instruction sent out
      self.spi.stran(None,dlen)
      #self.spi.board.rawin()
    #                        sslogic:   True          False
    #byte= self.spi.stran(None,8)   #-> 0xc0 (p_1)    0x80(p_1)
    #                                   0xc0(p_2)        0(p_2)
    #byte= self.spi.stran(None,16)  #-> [0xc7c70000]  [0xc7c70000] (p_2)
    #                                   0xc7c7c (p_1)
    byte= self.spi.stran(None,dlen)
    self.spi.setSS(0)
    # read data
    datatx="""
    for ixd in range((inbytes-1)/self.nb1t +1):  # 20b->3w  16b->2w
      datain= self.spi.stran(None,self.nb1t*8)
      #if self.nb1t<=4:
      if self.nb1t==1:
        datatx= datatx + "%x "%datain
      else:
        for ixw in range(self.nb1t/4):
          datatx= datatx + "%x "%datain[ixw]
      #if ixd < (inbytes-1): selfspi..board.rawin()
    self.spi.setSS(0)
    print "adc out:0x%x datain:0x%s"%(cmd,datatx)
"""
    return byte
  def prtbyte(self, byte):
    if type(byte) == list:
      print("MAX11661.rreg:",list(map(hex,byte)))
    else:
      print("MAX11661.rreg:",hex(byte))
  def test(self,ms=10):
    """
analog           -> ADC out
0x       0       -> 0x0
0x       f       -> 0x7c
0x      ff       -> 0xfc
0x     fff       -> 0x180
0x    ffff       -> 0x200
0x   fffff       -> 0x280
0x  ffffff       -> 0x300
0x fffffff       -> 0x380
0xffffffff       -> 0x3fc

>dev adc test 100
analog           -> ADC out
0x       0       -> 0x0
0x       1       -> 0x20
0x       3       -> 0x40
0x       7       -> 0x5c
0x       f       -> 0x80
0x      1f       -> 0xa0
0x      3f       -> 0xc0
0x      7f       -> 0xe0
0x      ff       -> 0x100
0x     1ff       -> 0x120
0x     3ff       -> 0x140
0x     7ff       -> 0x160
0x     fff       -> 0x180
0x    1fff       -> 0x1a0
0x    3fff       -> 0x1c0
0x    7fff       -> 0x1e0
0x    ffff       -> 0x200
0x   1ffff       -> 0x220
0x   3ffff       -> 0x240
0x   7ffff       -> 0x260
0x   fffff       -> 0x280
0x  1fffff       -> 0x2a0
0x  3fffff       -> 0x2c0
0x  7fffff       -> 0x2e0
0x  ffffff       -> 0x300
0x 1ffffff       -> 0x320
0x 3ffffff       -> 0x340
0x 7ffffff       -> 0x360
0x fffffff       -> 0x380
0x1fffffff       -> 0x3a0
0x3fffffff       -> 0x3c0
0x7fffffff       -> 0x3e0
0xffffffff       -> 0x3fc

"""
    print("analog           -> ADC out")
    #for x in (0,0xf,0xff,0xfff,0xffff,0xfffff,0xffffff,0xfffffff,0xffffffff):
    # or with smaller step:
    x= 0
    for ix in range(33):
      #self.spi.board.Write("ttc.write_ADC",x)  (Jakub)
      self.spi.board.Write("ctrl.adc_test_data",x)
      time.sleep(ms/1000.)
      data= self.rreg(self.DEFDLEN)[0] ; data= 0x7ff & (data>>4)
      print("0x%8x \t -> 0x%x"%(x, data))
      x= x | (1<<ix)
class N25Q(GenDev):   # _see spiatlas.py:
  SHSL2= 0.000001   # 50ns
  WRITE_ENABLE=0x06
  WRITE_DISABLE=0x04
  READ_FLAG_STATUS=0x70
  READ_STATUS=0x05
  SECTOR_ERASE=0xd8
  BULK_ERASE=0xc7
  SECTOR_LEN=0x10000
   
  def __init__(self, spi, devname="n25q128", model=""):    # kcu105:256mb   ctpltu:128mb
    GenDev.__init__(self, spi, devname, model)
    self.nb1t= 16    # number of bytes in 1 ipbus packet (1, 4, 8, 12, max 16)
    self.seed= None  # != None with last 'wmem seed'
    if model=="n25q128":
      self.MAXADDR= 0x1000000   # for 128Mbit memory on ctpltu board
      self.LASTSECTOR= 255
    else:
      self.MAXADDR= 0x2000000   # for 256Mb memory on KCU105 (ltukcu) board
      self.LASTSECTOR= 511
  def getid(self):
    return "id:%s"%self.rreg(0x9e,20)
  def rr1(self, cmd, inbytes=1):
    # read register
    # cmd: command
    # inbytes: 1,2,3 or 4 (i.e. foreseen for register readings)
    # rc: register content or None if error
    self.spi.setSS(1)
    self.spi.stran([cmd])   # rc: None always (output)
    datain= 0
    for ixd in range(inbytes):
      d1= self.spi.stran()
      #if d1==None: return None
      datain= (datain<<8) | d1
    self.spi.setSS(0)
    return datain
  def rreg(self, cmd=0x9e, inbytes=20, txt=None, dummyr1=False):   # p.39, 0x9e: READID o ROTP
    # foreseen for ID and OTP memory readings
    self.spi.setSS(1)
    if dummyr1:   # 1 dummy read before instruction sent out
      self.spi.stran()
      #self.spi.board.rawin()
    # write instruction
    self.spi.stran((cmd,))
    # read data
    datatx=""
    for ixd in range(old_div((inbytes-1),self.nb1t) +1):  # 20b->3w  16b->2w
      datain= self.spi.stran(None,self.nb1t*8)
      #if self.nb1t<=4:
      if self.nb1t==1:
        datatx= datatx + "%x "%datain
      else:
        for ixw in range(old_div(self.nb1t,4)):
          datatx= datatx + "%x "%datain[ixw]
      #if ixd < (inbytes-1): selfspi..board.rawin()
    self.spi.setSS(0)
    #print "rreg:", txt, datatx
    if txt != None:
      print("%s out:0x%x datain:0x%s"%(txt, cmd,datatx))
    return "%s"%(datatx)
    #self.spi.status()
  def rmem(self, maddr=0, inbytes=20, check=False):   # p.43
    self.spi.setSS(1)
    cmd= 0x03 # READ instruction (0d: FAST READ not available with ipbus spi core)
    self.spi.cmdaddr([cmd, (maddr>>16)&0xff, (maddr>> 8)&0xff, maddr&0xff])
    # read data
    nfirst=16   # has to be multiple od self.nb1t
    nlast=16
    errs=0 ; maxerrs=4
    if check:
      if (self.seed!= None): random.seed(self.seed)
    else:
      dataout=[]
    datatx="" ; t1= time.time()
    loops= old_div((inbytes-1),self.nb1t) +1   # loops: number of ipbus-readings (blocks)
    wrds= old_div(self.nb1t,4)                 # 1 2 3 or 4
    for ixd in range(loops):                     # 1 4   8     12         16
      datain= self.spi.stran(None,self.nb1t*8)   # w [w] [w,w] [w,w,w] or [w,w,w,w]
      if (ixd*self.nb1t < nfirst) or (ixd*self.nb1t >= (inbytes-nlast)):
        if self.nb1t==1:
          datatx= datatx + "%x "%(datain)
        #elif self.nb1t==4:
        #  datatx= datatx + "%x "%(datain)
        else:
          for ixw in range(wrds):
            datatx= datatx + "%x "%datain[ixw]
      elif ixd*self.nb1t == nfirst:
        datatx= datatx + " ... "
      else:
        pass
      if check:
        if (ixd % 1000)==0:
          if self.nb1t==1:    #if self.nb1t<=4:
            print("%d: "%(ixd), hex(datain))
          else:
            print("%d: "%(ixd), list(map(hex,datain)))
      #if ixd < (inbytes-1): selfspi..board.rawin()
      if check:
        for ixw in range(wrds):
          rn= 0
          rnr= datain[ixw]
          for ix1 in range(4):
            reladr= 4*wrds*ixd + 4*ixw + ix1
            if self.seed:
              rn1= random.randint(0,255)
            else:
              rn1= (maddr + reladr)&0xff
            if reladr>=inbytes:   # correction for data not to be checked
              rn1= (rnr>>((3-ix1)*8)) & 0xff
            rn= (rn<<8) | rn1
          if rn != rnr:
            errs= errs+1
            if errs<maxerrs:
              print("Error: loop:%d ixw:%d read:0x%x exp:0x%x"%(ixd, ixw, rnr, rn))
            elif errs==maxerrs:
              print("Error: too many errors (printing to stdout suppressed)...")
            else:
              pass
          #else:
          #    print "OK   : loop:%d ixw:%d read:0x%x exp:0x%x"%(ixd, ixw, rnr, rn)
      else:   # return
        #for ixw in range(wrds):
        dataout= dataout + datain
    self.spi.setSS(0)
    t2= time.time() ; t= t2-t1; average= old_div(t*1024,inbytes)
    sector= maddr>>16  ; ainsec= maddr &0xffff
    if check:
      print("Errors:", errs)
      #if (page % 256)==0:
      #      print "writing sector %d..."%((maddr>>16) + page/256)
      print("rmem %d bytes from:0x%x(sec:%d adinsec:0x%x) datain:0x%s %.1f secs, %.2f secs/kB"%\
        (inbytes,maddr, sector, ainsec, datatx, t, average))
    else:
      if errs>0: print("Errors:", errs)
      print("rmem %d bytes from:0x%x(sec:%d adinsec:0x%x) datain:0x%s %.1f secs, %.2f secs/kB"%\
        (inbytes,maddr, sector, ainsec, datatx, t, average))
      return dataout
    #self.spi.status()
  def wmem(self, fraddr=0, Nbytes=2, seed=None):
    """
    fraddr: write from this address
    Nbytes:     nuber of bytes to be written
    seed:  None: use 0xXY for addrXY -i.e. low 8 bits of address for value written
           N     use random number 0..255, using seed N
    Notes:
    sector:    64kB  0x10000
    subsector:  4kB  0x1000
    page:     256B   0x100
"""
    txt="wmem"
    cmd= 0x02   # p.53 PROGRAM operations.  max. 256 bytes in 1 PROGRAM operation!
    # note: setSS(1/0) -MUST be done with each COMMAND, i.e. 3 times:
    # 1. wr-enable   2. wr-bytes   3. wr-disable
    ##self.spi.setSS(1)
    ##self.spi.stran([self.WRITE_ENABLE])
    ##self.spi.setSS(0)
    #
    if seed != None:
      self.seed= seed
      random.seed(seed)
    else:
      self.seed= None
    pages= old_div((Nbytes-1),256) + 1
    maddr= fraddr
    N= 256 ; loops= old_div((N-1),self.nb1t) +1
    datatx="" ; t1= time.time()
    for page in range(pages):   # 1 page/1 program op.
      if (page % 256)==0:
          print("writing sector %d..."%((maddr>>16) + old_div(page,256)))
      if page== pages-1: 
        N= Nbytes - page*256
        loops= old_div((N-1),self.nb1t) +1
      self.spi.setSS(1)
      self.spi.stran([self.WRITE_ENABLE])
      self.spi.setSS(0)
      self.spi.setSS(1)
      self.spi.cmdaddr([cmd, (maddr>>16)&0xff, (maddr>> 8)&0xff, maddr&0xff])
      for ixd in range(loops):
        dbs= []
        if seed != None:
          for ixw in range(self.nb1t):
            dbs.append(random.randint(0,255))
        else:
          for ixw in range(self.nb1t):
            #reladr= 4*wrds*ixd + 4*ixw + ix1  -from rmem
            #reladr= ixd*self.nb1t + ixw
            dbs.append((maddr + ixd*self.nb1t + ixw)&0xff)
        self.spi.stran(dbs,self.nb1t*8)
        datatx= datatx + " " + str(list(map(hex,dbs)))
        #if ixd < (N-1): selfspi..board.rawin()
      self.spi.setSS(0)
      maddr= maddr+256
    #self.allstatus()
    self.spi.setSS(1)
    self.spi.stran([self.WRITE_DISABLE])
    self.spi.setSS(0)
    t2= time.time() ; t= t2-t1; average= old_div(t*1024,Nbytes)
    #print "%s out:0x%x addr:0x%x %d bytes out:0x%s %.1f secs %.2f secs/kB"%\
    #  (txt, cmd,maddr, N, datatx, t, average )
    print("%s addr:0x%x %d, %.1f secs %.2f secs/kB"%\
      (txt, fraddr, Nbytes, t, average ))
    #self.spi.status()
  def allstatus(self, txt=None):
    if txt==None: note="%s, st stf vol evol nvol ear"%self.spi.spislave
    else: note= txt
    print("Flash on %s, using %d bytes for ipbus packet."%(self.spi.spislave, self.nb1t))
    stf= self.rr1(self.READ_FLAG_STATUS)         # p.27
    st= self.rr1(self.READ_STATUS)          # p.21
    volcfgr= self.rr1(0x85)     # p.23
    evolcfgr= self.rr1(0x65)    # p.26
    nvolcfgr= self.rr1(0xb5,2)  # p.22
    ear= self.rr1(0xc8)
    print("%s:0x%x 0x%x 0x%x 0x%x 0x%x 0x%x"%\
      (note, st, stf, volcfgr, evolcfgr, nvolcfgr, ear))
  def erasesec(self, maddrf,nsecs=1):   # p.59
    """ 
maddrf:0 Erase 1st sector (set 0xff... in first 64kB)
maddrf:0x10000 -2nd sector
maddrf:0x20000 -3nd sector
       0xff0000 -last one (for 128Mbit memory on ctpltu board)
nsecs: number of sectors
seems status+flag status regs availabe during erase operation:
>f erase
erase started...:0x3 0x0 0x0 0x0 0x0 0x0
erase finished...:0x0 0x80 0xfb 0xdf 0xffff 0x0
erasesec loops: 182
"""
    maddr= maddrf
    for ixsec in range(nsecs):
      maddr= maddr+ ixsec*self.SECTOR_LEN
      if maddr > self.MAXADDR: break
      # 0xd8: sector erase, 0x20 subsector erase, 0xc7: bulk erase
      self.spi.setSS(1)   # write enable cmd: must be here before each 'sector erase' cmd
      self.spi.stran([self.WRITE_ENABLE])
      self.spi.setSS(0)
      self.spi.setSS(1)
      self.spi.cmdaddr([self.SECTOR_ERASE, (maddr>>16)&0xff, (maddr>> 8)&0xff, maddr&0xff])
      self.spi.setSS(0)   # now erase op. starting...
      self.allstatus("erasesec %d started..."%ixsec)
      loops=0
      time.sleep(0.7)
      while True:
        st= self.rr1(self.READ_FLAG_STATUS)
        if st & 0x80: break   # 1: ready (erase finished)
        loops= loops+1
        if loops>1000: break  # sector: cca 180 loops (nosleep) 0.7 - 3 secs
        # bulk erase: 240s - 480s !    Look at p.82 for typical times...
        time.sleep(0.1)
      print("erasesec %d loops: %d"%(ixsec,loops))
    self.allstatus("erase finished, WRITE_DISABLE cmd...")
    self.spi.setSS(1)
    self.spi.stran([self.WRITE_DISABLE])
    self.spi.setSS(0)
  def erasebulk(self):   # p.59
    self.spi.setSS(1)
    self.spi.stran([self.WRITE_ENABLE])
    self.spi.setSS(0)
    self.spi.setSS(1)
    # 0xd8: sector erase, 0x20 subsector erase, 0xc7: bulk erase
    self.spi.cmdaddr([self.BULK_ERASE], 8)
    self.spi.setSS(0)   # now erase op. starting...
    self.allstatus("erasebulk started...")
    loops=0
    while True:
      st= self.rr1(self.READ_FLAG_STATUS)
      if st & 0x80: break   # 1: ready (erase finished)
      loops= loops+1
      if loops>1000: break
      # bulk erase: 240s - 480s !    Look at p.82 for typical times...
      time.sleep(1)
    self.allstatus("erasebulk finished after %d loops, WRITE_DISABLE..."%loops)
    self.spi.setSS(1)
    self.spi.stran([self.WRITE_DISABLE])
    self.spi.setSS(0)
  def reset(self):   # p.64
    self.spi.setSS(1)
    self.spi.stran([0x66])   # RESET ENABLE
    self.spi.setSS(0)
    time.sleep(self.SHSL2)
    self.spi.setSS(1)
    self.spi.stran([0x99])   # RESET MEMORY
    time.sleep(0.001)   #self.SHSL3?
    self.spi.setSS(0)
  def prthelp(self):
    print("""
id      -read ID, i.e. write 0x9e and read back 20 bytes ID (N25Q READ ID cmd)
status  -read status/flag/volatile/non volatile/enhanced volatile/extended addr regs
reset   -N25Qxxx reset cmd
rotp    -read 64 OTP memory bytes
rmem [fromad [N] [c]]  -read N bytes from fromad  
                        (c -check data, written before using wmem)
wmem [fromad [N] [r]]  -write N (def. 2) bytes from fromad (def. 0)
                        value written: last 8 bits of address (default) or random (r)

dump filename [bytes]  -dump to binary file (only first 'bytes' if given)
load filename          -load from binary file NOT PREPARED YET

erase fromadd [N]   -erase N sectors from sector fromadd  (default N:1 -> cca 0.7s)
      add     sector
      ---     ------
      0        0
      0x10000  1
      0xff0000 255   i.e. the last one on ctpltu 
                     On ltukcu 2x more:0x1ff0000 = sector 511
erasebulk           -erase all sectors (BULK ERASE cmd 0x7h), 4 minutes with 128Mb
set N   -1 4 8 12 or 16 bytes length of data in IPBUS packet for r/w operations (now:%d)
help
"""%(self.nb1t))
  def excmd(self,cmd):
    if len(cmd)<=1:
      self.prthelp()
      return
    if cmd[1]=='id':
      print(self.getid())
    # N25Q256A11ESF40F DUAL QUAD SPI 256Mb flash (2xon kcu105, 128Mb on LTU)
    # spi1:
    # rreg out:0x9e datain:0x20 bb 19 10-00 00 23 33-00 06 03 00-19 00 29 31-08 13 c9 ef
    # manufacturer bb 19: 256Mb 10: 16 bytes follow...
    # 00 00: extended device id ->volatile cfg reg required!
    # rreg out:0x9f datain:0x20 bb 19 10 00 00 23 33 00 06 03 00 19 00 29 31 08 13 c9 ef
    # spi0  (using startupe3):
    # rreg out:0x9e datain:0x20 bb 19 10 00 00 23 33 10 15 02 00 08 00 11 04 09 13 90 da
    # rreg out:0x9f datain:0x20 bb 19 10 00 00 23 33 10 15 02 00 08 00 11 04 09 13 90 da
    #
    elif cmd[1]=='set':
      if len(cmd)==3:
        l= eval(cmd[2])
        if (l==1) or (l==4) or (l==8) or (l==12) or (l==16):
          self.nb1t= l
          print("nb1t:", l)
    elif cmd[1]=='status':
      self.allstatus()
    elif cmd[1]=='rotp':   # p.65
      self.rreg(0x4b, 64, "OTP")
    elif cmd[1]=='rmem':
      fromad= 0; byts= 20; checkdata= False
      if len(cmd) > 2: fromad= eval(cmd[2])
      if len(cmd) > 3: byts= eval(cmd[3])
      if len(cmd) > 4: 
        if cmd[4] == 'c':
          checkdata= True
      start = time.time()
      self.rmem(fromad, byts, checkdata)
      end = time.time()
      print("Duration: ",end-start," secs")
    elif cmd[1]=='wmem':   # p.53...
      fromad= 0; byts= 2; seed= None
      if len(cmd) > 2: fromad= eval(cmd[2])
      if len(cmd) > 3: byts= eval(cmd[3])
      if len(cmd) > 4: seed= 97
      start = time.time()
      self.wmem(fromad,byts, seed)
      end = time.time()
      print("Duration: ",end-start," secs")
    elif cmd[1]=='dump':
      if len(cmd) > 2:
        fn= cmd[2]
        dfile= open(fn,"wb")
        if self.nb1t>1:
          sectorwords= old_div(self.SECTOR_LEN,4)    # 16384
        else:
          sectorwords= self.SECTOR_LEN      # not tested!
        reqbytes= self.SECTOR_LEN*(self.LASTSECTOR+1)
        if len(cmd) > 3:
          reqbytes= int(cmd[3])   # only multiple of 4 tested
        lastsn= old_div(reqbytes,self.SECTOR_LEN)
        lastslw= old_div((reqbytes%self.SECTOR_LEN),4)   # must be multiple of 4
        print("Dumping %d sectors and %d words to %s..."%(lastsn, lastslw, fn))
        wwords= 0
        for secn in range(lastsn+1):
          #if secn>2: break
          print("reading sector %d..."%secn)
          fromad= secn*self.SECTOR_LEN
          sector= self.rmem(fromad, self.SECTOR_LEN)
          if secn == lastsn:
            words= lastslw
          else:
            words= sectorwords
          #print "dumping sector (>I) %d (%d words, expected:%d)..."%\
          #  (secn,len(sector),sectorwords)
          #print "dumping sector (>I) %d ..."%(secn)
          for ixw in range(words):
            dfile.write(struct.pack('>I',sector[ixw]))   # >I: big endian, unsigned 4 bytes int
            # >I: big endian i.e. DCBA because .bin file is a big endian and we want to compare
            #(<I would write little endian: ABCD)
          wwords= wwords+ words
        dfile.close()
        print("%d bytes written to %s"%(wwords*4,fn))
      else: 
        print("filename missing")
    elif cmd[1]=='load':
      if len(cmd) > 2:
        fn= cmd[2]
        dfile= open(fn,"rb")
        print("not implemented yet")
        return 
        ixw= 0
        while (not dfile.eof()):
          mw= self.rmem(fromad, 1)
          w= struct.unpack('>I',dfile.read(4))   # >I: big endian, unsigned 4 bytes int
          print("mem:%x file:%x"%(mw, w))
          ixw= ixw+1
          if ixw>4: break
      else: 
        print("filename missing")
      dfile.close()
    elif cmd[1]=='erase':
      fromad= 0; nsec= 1
      if len(cmd) > 2: fromad= eval(cmd[2])
      if len(cmd) > 3: nsec= eval(cmd[3])
      self.erasesec(fromad,nsec)
    elif cmd[1]=='erasebulk':
      self.erasebulk()
    elif cmd[1]=='help':
      self.prthelp()
    elif cmd[1]=='reset':
      self.reset()
    else:
      print("???")

