#!/usr/libexec/platform-python
#!/usr/bin/python3
from __future__ import print_function
import os,json,time
class WSclient:
  """ 2 types of connection:
ltud ---wsup---> awmtors <---ws---> web_browser

wsup: one way
ws:   bidirectional
"""
  def __init__(self, netname, contype="wsup"):
    host= os.uname()[1]
    self.wspath= "ws://{0}:8888/{1}/{2}".format(host,contype,netname)
    #wspath= "ws://{0}:8888/{1}/{2}".format("zenaj","wsup","test")
    self.contype= contype
    self.mid="{0}/{1}".format(contype,netname)
    self.ws= None
    self.ws_open()
  def ws_open(self):
    self.nsir= 0   # not sent in row
    rcc= self.ws_close()   # todo perhaps: replace by pingpong
    rc= 0; msg= "ok"
    if self.ws==None:
      try:
        from websocket import create_connection
        self.ws = create_connection(self.wspath)
        # here we should update ONU statuses in awm
        msg= "ok"
      except:
        msg= "Error: cannot open wspath: " + self.wspath
        rc= 8
        self.ws= None
    else:
      rc= 4
      msg= "Warning: ws already open wspath:"+ self.wspath
    #print(msg)
    return(rc,msg)
  def ws_close(self):
    rc= 0; msg= "ok"
    if self.ws:
      self.ws.close()
      self.ws= None
    else:
      msg="{0} was closed already".format(self.wspath)
    return(rc,msg)
  def ws_send(self, msgo):
    #if self.ws.connected:   # always True (even with broken connection)
    # todo check if open
    if self.ws:
      rc= self.ws.send(json.dumps(msgo))
      print("{0}.send rc:{1} {2}".format(self.mid,rc, msgo))
      if self.contype=="wsup":
        ackmsg = self.ws.recv()
        print("< %s"%ackmsg)
      self.nsir= 0
    else:
      self.nsir= self.nsir+1
      if (self.nsir < 4) or ((self.nsir % 1000)== 0):
        print("ws_send to {0}: not sent (not connected)".format(self.mid))
class AWMclient(WSclient):
  def __init__(self, netname, contype="wsup"):
    WSclient.__init__(self, netname, contype)
    self.allonus= {}   # "olt?_*_dsmon":[onuN, last_dsmon_value, state]
    # ?: olt number 1...     *: onuAddr 1...63
    # onuN: order number 1,2,... 
    # last_dsmon_value: [fecS,fecD,sc_crc_detected]
    # state: 0 (on), 1(err), 2(not known)
    self.changedonus=[]  # [(oltN,onuN,2/1/0),...]
    self.dbgchonu= False   # True: dbg mode
    self.timestamp= 0. # time.time()   # latest message sent (sendawm)
  def update_cfg(self, oltname, onuaddr, ixonu1):
    self.allonus[oltname+"_"+onuaddr+"_dsmon"]= [ixonu1,[0,0,0], 2]
  def update(self, keyname, dsmon_v):
    """ Op: compare old tripple [fecS, fecD, sc_crc] with a new one (dsmon_v)
Outcome: set in self.cahngedonus+self.allonus
rc: (0,msg) 0:ok update done 
"""
    newstate= self.allonus[keyname][2]
    if self.dbgchonu == True:
      # in dbg mode: state= state+1
      msg= "dbg {0}: {1}-> ".format(keyname, str(newstate))
      if newstate==2:
        newstate= 0
      else:
        newstate= newstate+1
    else:
      msg= "{0}: {1}-> ".format(keyname, str(newstate))
      #print("AWMclient update...")
      if dsmon_v==None:
        newstate= 2   # unknown
      else:
        if self.allonus[keyname][1]!=dsmon_v:
          newstate= 1    # err
        else:
          newstate= 0    # ok
    msg= msg+str(newstate)
    self.allonus[keyname][1]= dsmon_v
    if newstate != self.allonus[keyname][2]:
      oltn= int(keyname[3])
      self.changedonus.append((oltn,self.allonus[keyname][0],newstate))
      self.allonus[keyname][2]= newstate
    #print("tri_ws update:",self.changedonus)
    return(0, msg)
  def get_statuses(self):
    """ to get the state of all onus from tud
Invoked from: ltud_server (CmdRequest.WS_ONUSTATS) initiated from ltud_client
  rc: prepared for CmdReply message
"""
    states= []
    for dsk in self.allonus:
      ks= dsk.split("_")
      oltn= int(ks[0][3])
      onua= ks[1]
      onun= self.allonus[dsk][0]
      onus= self.allonus[dsk][2]
      states.append((oltn, onun, onus))
    print("get_statuses:", states)
    msgo= {"excmd":"", "onus":states, "epochtime":self.timestamp}
    # seems bad idea:
    return(0,"dbgchonu:{} ".format(self.dbgchonu)+json.dumps(msgo))
    #return(0,json.dumps(msgo))
  def sendawm(self,force=False):
    """ send ws message downstream (timestamp at least)
Invoked from: 
-ttcponnet.py measureSlow(), after all ONU statuses updated
-self.forceupdate()
"""
    if force:
      changedonus= []
      for keyname in self.allonus:
        oltn= int(keyname[3])
        state= self.allonus[keyname][2]
        #print("sendawm force:",keyname,oltn,self.allonus[keyname][0],state)
        self.changedonus.append((oltn,self.allonus[keyname][0],state))
    else:
      self.timestamp= time.time()
    if len(self.changedonus)>0:
      #send 
      msgo= {"excmd":"", "onus":self.changedonus, "epochtime":self.timestamp}
      self.ws_send(msgo)
      self.changedonus=[]
    else:
      self.ws_send({"epochtime":self.timestamp})
  def forceupdate(self):
    """ used for:
- debug (ltud_client ows cmd)
- in time of awmtors.py start 'ows' is invoked opening ws+forceupdate
"""
    # todo: test if following works:
    # start awmtors, ltud update awmtors.py file
    # 
    rc= self.ws_open()
    print("forceupdate ws_open rc:",rc)
    if rc[0]<=4:
      self.sendawm(True)
    else:
      return (8,"cannot open ws")
    return(0,"ok")
  def forceunlock(self):
    self.ws_send({"excmd":'lockaction("unlockme")'})
    return(0,"ok")
  def set_dbgchonu(self):
    if self.dbgchonu == True:
      self.dbgchonu= False
    else:
      self.dbgchonu= True
    return(0,"ok dbgchonu:{0}".format(str(self.dbgchonu)))
