#!/usr/libexec/platform-python
from __future__ import absolute_import
from __future__ import print_function

import os.path   #,string
from auxfuns import parseline
class NetFile:
  def __init__(self, netnamefull, logger):
    self.name= netnamefull.split('.')[0]
    self.logger= logger
    self.waserror= False
    self.gbts= []
    self.allolts= []  # list of OLTs
    self.allonus= []  # and corresponding list of ONU lists
    self.ttcits= []   # list of all ttcits (items: "atbp_7 or "0x8a1000")
    self.maxonus= 0
    cfgdir= os.getenv("CTP3_CONFIG_DIRECTORY")
    fpath= os.path.join(cfgdir,"nets", netnamefull)
    if not os.path.exists(fpath):
      self.logger.critical("%s does not exist",fpath)
      self.waserror= True
      return
    linen= -1
    fd= open(fpath,"r") ; netcfg= fd.readlines() ; fd.close()
    for ixl in range(len(netcfg)):
      #ixhs= netcfg[ixl].find('#')
      #if ixhs>-1:
      #  linenc= netcfg[ixl][:ixhs]
      #else:
      #  linenc= netcfg[ixl]
      #line= linenc.strip().split()
      #if len(line)==0: continue
      line= parseline(netcfg[ixl])
      if line==None: continue
      if linen==-1:    # 1st line of netcfg
        #----------------------------------------------------- # 1st line of netcfg
        self.masterbn= line[0] 
        maxgbtn=14    # todo: consider version (ltu_logic vs ltu_logic_itsmft)...
        linen= 0
        self.atbmode= 's'
        atbmodetx= 'standalone'
        if (len(line) > 1) and ((line[1]=='g') or (line[1]=="global")): 
          self.atbmode= 'g'
          atbmodetx= 'global'
          # special onu instance. -1: do not touch onu address when instantiating
          # needs to be instantiated even without ONU SFP (exd_freeze_cdr_test_mode) ?
          # Let's instantiate only in global mode!
          #
          # self.onufromctp abandoned, from 26.6.2019: there is a spec. instance of OLT1
          # for each onu in superior(ctp) net.
          # for each ONU on ATB board regardless of:
          # - ATB is in role of CRU (debugging ie OLT physically does not have to be present) or
          # - ATB in role of LTU
          # self.onufromctp= OnuMlinkExdsgAlice(masterbn, '', -1, False)
        continue
      if (len(line[0])>3) and (line[0].find("gbt")==0):   # gbt line
        #----------------------------------------------------- GBT line
        #print("gbt line", line)
        try:
          gslot= int(line[0][3:])
        except:
          self.logger.critical("Error in .net: %s",netcfg[ixl])
          self.waserror= True
          continue
        if gslot > maxgbtn:
          self.logger.critical("Error too high gbt nmuber (max:%d): %s",(maxgbtn,netcfg[ixl]))
          continue
        gbtname= "gbt%d"%gslot
        #if gbtboard.nodExists(gbtname):
        # todo: optional check if exists in .xml file 
        alreadyin= False
        for ixgbts in range(len(self.gbts)):
          if self.gbts[ixgbts][0] == gslot:
            print("Error: {0} already defined in .net file, later one ignored".format(gbtname))
            alreadyin= True
            break
        if alreadyin: continue
        if (len(line)<2) or ((line[1]!="vtrx") and (line[1]!="sfp") and (line[1]!="sfpbsr")):
          self.logger.critical("Error in .net (gbtN expected with vtrx sfp or sfpbsr):"+netcfg[ixl])
          continue
        #self.gbts.append(Gbt(SFPinst, gslot, line[1]))
        self.gbts.append([gslot, line[1]])
        continue
      if (line[0]== "ttcit"):   # ttcit line
        #----------------------------------------------------- ttcit line
        # 2 examples with old and new ttcit board:
        # ttcit 0x8a2000
        # ttcit atbp_7
        #print("ttcit line", line)
        if len(line)<2:
          self.logger.warning("Error empty ttcit line in .net: %s",netcfg[ixl])
          continue
        for ttcit in line[1:]:
          self.ttcits.append(ttcit)
        continue
      #----------------------------------------------------- OLT line
      lane= int(line[0])-1   # lane: an olt slot: 0..8 corresponding to olt1..olt9
      #olt= OltMlinkExdsgAlice(masterbn, '', lane=lane, mode= atbmode)
      self.allolts.append(lane)
      self.allonus.append([])   # each OLT line accompanied by a list of ONUs
      #print("OLT{0:d}:".format((lane+1)), self.masterbn, atbmodetx)
      onus= self.allonus[-1]
      onusnames={}   # check if ONU already defined in an OLT line
      onuadds=[]    # check the uniqueness of an ONU address in an OLT line
      # todo: check if atb appeared in multiple OLT lines...
      maxonus= 0
      for idaix in range(len(line[1:])):   # start after OLTN (1st in line with fed onus)
        ida= line[1+idaix]    # atb_10 atbp_7 atb_8:20 20 kc:1 kcu:4
        if len(onus)>=64:
          self.logger.critical("Error: %s ignored (>64 onus)"%(ida))
          continue 
        if (ida in onusnames) and (onusnames[ida]==True):
          self.logger.warning("Warning: %s ignored (given already in the same line)"%(ida))
          continue 
        mononu= False
        #if type(ida) is int:   # CRU
        if (ida[0]>'0') and (ida[0]<='9'):   # CRU -integer>1
          onuadd= int(ida)
          #onus.append(OnuMlinkExdsgAlice('', '', onuadd))
          onus.append((None, onuadd,None))
        elif ida[:4]=="atb_" or ida[:5]=="atbp_":   # atb_addr or atb_N:addr
          ida01= ida.split(":")
          # default: atb in role of LTU. 
          # If 3rd param added (eg: atb_11:7:cru), ol2ndlevel is not allocated,
          # i.e. no onu2olt good phase done)
          ida0= ida01[0]
          if ida0[:4]=="atb_":       # let's take atb number for ONU addr
            onuadd= int(ida0[4:])
          else:
            onuadd= int(ida0[5:])
          if len(ida01)>=2:          # e.g. atb_20:7
            if ida01[1]!="":
              onuadd= int(ida01[1])   # explicitly defined, i.e. atb_20:7
          if onuadd in onuadds:
            self.logger.critical("Error: ONU address %d already used in the same .net line"%onuadd)
            self.waserror= True
            continue
          if ida0 == self.masterbn:
            # monitoring ONU special: GOOD_PHASE first, and do not load PLL (loaded for OLT already)!
            if idaix==0: 
              if self.atbmode=='g':
                self.logger.critical("Error: monitoring ONU %s allowed only in stdalone mode."%ida)
                #return None,None,None,None,None
                self.waserror= True ; continue
              else:
                mononu= True
            else: 
              self.logger.critical("Error: monitoring ONU %s has to be 1st one in the list of ONUs."%ida)
              #return None,None,None,None,None
              self.waserror= True ; continue
          # add 1 special olt instance for onu2olt phase. Not needed if ATB in CRU role,
          # but allocated anyhow: it can be differentiated in .net file by e.g. by using:
          # atb_24:cru  -for atb_24 in role of CRU
          # atb_24      -for atb_24 in role of LTU  (default -normally, atb boards are in LTU role!)
          # if roleCRU: olt2ndlevel= None
          if mononu:
            olt2ndlevel= None
          elif (len(ida01)==3):
            if (ida01[2]!='CRU') and (ida01[2]!='cru'):
              self.logger.critical("Error: cru expected, (not %s). %s role: LTU"%(ida01[0],ida01[2]))
              olt2ndlevel= True
            else:
              olt2ndlevel= None
          else:
            olt2ndlevel= True
          onus.append((ida0, onuadd,olt2ndlevel))
        else:
          self.logger.critical("Error: bad ONU: %s in .net file"%ida)
          self.waserror= True
        #print("ONU{0:d}:{1:s}".format(len(onus), ida), onus[len(onus)-1], " mononu:",mononu)
        onusnames[ida]= True
        onuadds.append(onuadd)
        maxonus= maxonus+1
      if maxonus> self.maxonus: self.maxonus= maxonus
      if len(onus)==0: onus.append((None, " ",None))   # placeholder for 'no onus olt'
    if not self.waserror: self.logger.info(netnamefull+" ok")
  def getoverview(self):
    """ return a list of
netname
masteratb
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
"""
    nltus=ncrus=ncrues=0
    for ixolt in range(len(self.allolts)):
      if self.allolts[ixolt]==None: continue
      for ixonu in range(len(self.allonus[ixolt])):
        onu= self.allonus[ixolt][ixonu]
        #print("onu%d:"%ixonu, onu)
        if onu==None: continue
        if self.allonus[ixolt][ixonu][0]==None:
          if self.allonus[ixolt][ixonu][1]==' ': continue   # olt without onus
          ncrus= ncrus+1   # real CRU
        else:   # LTU or CRU in role of LTU
          if self.allonus[ixolt][ixonu][2]:  # LTU
            nltus= nltus+1
          else:
            ncrues= ncrues+1     # CRU emulated by LTU
    return(self.name, self.masterbn,len(self.allolts),
      nltus, ncrus, ncrues, len(self.gbts), len(self.ttcits))
  def getolts(self):
    """ rc example:
[['olt1', '1', 'atb_20:3:LTU', 'atb_21:4'], ['olt3', 'atb_27:27', '3']]

"""
    rc= []
    for ixo in range(len(self.allolts)):
      o1= ["olt%d"%(self.allolts[ixo]+1)]
      for ixn in range(len(self.allonus[ixo])):
          # [atb_xx:]addr[:LTU]
          atbadrltu= self.allonus[ixo][ixn][0]
          if atbadrltu == None:
            atbadrltu= ""
          else:
            atbadrltu= atbadrltu+':'
          atbadrltu= atbadrltu + str(self.allonus[ixo][ixn][1])
          if self.allonus[ixo][ixn][2]:
            atbadrltu= atbadrltu + ":LTU"
          o1.append(atbadrltu)
      rc.append(o1)
    return rc
if __name__ == '__main__':
  import sys, ilogger
  # nf.waserror True -syntax error
  # nf.masterbn  .atbmode .maxonus:
  # masterbn: "atb_29"
  # maxonus: max. umber of ONUs for any OLT
  # atbmode: "s" or "g"
  #
  # nf.gbts .allolts .allonus .ttcits:
  # gbts: [gbtslot,"sfp|vtrx"],[],...          -list of gbts
  # allolts: [ 0, 3,...]                       -list of olt numbers -1
  # allonus: list of 3-item lists [atbname, onu_addr, is_ltu]
  #   len(allolts) corresponds to len(allonus), i.e.
  #   allolts[X] controls allonux[X] onus:
  # [ None, " ", None]      -no onu (e.g. for olt without onus)
  # [ None, addr, None]     -CRU  (addr: integer 1..63
  # [ "atb_27", addr, True] -LTU
  # [ "atb_27", addr, None] -LTU in role of CRU
  # ttcits: e.g.: [ "atbp_7", "0x8a1000"] list of all connected TTCit boards
  if len(sys.argv)==1:
    print("""
name.net -show name.net
ov       -overview of all nets (i.e. showing the number of OLTs, LTUs, CRUs, ...)
""")
  elif sys.argv[1]=='ov':
    cfgdir= os.getenv("CTP3_CONFIG_DIRECTORY")
    logger= ilogger.ilogger()
    netspath= os.path.join(cfgdir,"nets")
    rc= ("netname","atb","olts","ltus","crus","ec","gbts","ttcits")
    #fmt= "{:>10} {:>7} {:4} {:4} {:4} {:2} {:4} {:4}"
    fmt= "{:10} {:7} {:4} {:4} {:4} {:2} {:4} {:6}"
    print(fmt.format(*rc))
    for netname in os.listdir(netspath):
      if netname.find('.net')== (len(netname)-4):   # only .net files
        nf= NetFile(netname, logger)
        rc= nf.getoverview()
        #return(self.name, self.masterbn,len(self.allolts),
        #  nltus, ncrus, ncrues, len(self.gbts), len(self.ttcits))
        #print(rc)
        print(fmt.format(*rc))
    print("""
atb: the name of master ATB driving the whole net
ec:  the number of Emulated CRUs, i.e. LTUs in role of CRU + MONOUs.""")
  else:
    netname= sys.argv[1].split('.')[0]
    logname= os.path.join(os.getenv("CTP3_WORK_DIRECTORY"),"netfile_%s.log"%netname)
    logger= ilogger.ilogger(logname, netname)
    nf= NetFile(sys.argv[1], logger)
    if nf.waserror: sys.exit(8)
    print(netname, "masterbn:", nf.masterbn,"maxonus:", nf.maxonus, 
      "\nolts:", nf.getolts(),"\nallonus:", nf.allonus)

