#!/bin/bash
function helpexit() {
cat - <<-EOF
Configure ctp3-ipbus files in config directory:

Prerequisities:
--------------
Environment variables:
CTP3_ROOT             ctp3-ipbus software (read only directory)
CTP3_CONFIG_DIRECTORY config directory (defaults to ~/CFG)
CTP3_WORK_DIRECTORY   config directory (defaults to ~/WORK)

A file \$CTP3_CONFIG_DIRECTORY/myboards.txt
containing one line per LTU board:

bn ip

where:
bn:  LTU board number
ip:  IP address of the board

Usage:
-----
ctp3-init.sh
Use this when started 1st time, or with a new LTU board

ctp3-init.sh update
Use 'update' to update the CFG directory files
E.g. before a firmware update or a change of IP address

Operation:
---------
- create file: ethers (from myboards.txt), only when 'update'
- call makemycfg.py

EOF
exit 1
}

echo "Doing ctp3-init as user '$USER' with args '$*'"
cfgdir=${CTP3_CONFIG_DIRECTORY-"~/CFG"}
wrkdir=${CTP3_WORK_DIRECTORY-"~/WORK"}
#
if [ ! -d $CTP3_ROOT ] ;then
  cat - <<-EOF

\$CTP3_ROOT does not exist.

EOF
  exit 1
fi
if [ ! -e $cfgdir/myboards.txt ] ;then
  cat - <<-EOF

myboards.txt missing in config directory

EOF
  exit 1
fi
if [ ! -d $wrkdir ] ;then
  mkdir $wrkdir
  echo "Working directory $wrkdir created."
fi
#
if [ $# -eq 1 ] ;then
  if [ "$1" != "update" ] ;then
    helpexit
  fi
else
  # new board(s), create ethers file:
  while read -a line ;do
    bni=${line[0]}
    bipi=${line[1]}
    printf -v hexbn "%x" $bni
    echo "08:00:30:00:2a:"$hexbn $bipi >>$cfgdir/ethers
  done < $cfgdir/myboards.txt
  cat - <<-EOF

Recommended lines in ~/.bashrc:
  export CTP3_ROOT="..."
  export CTP3_WORK_DIRECTORY=$wrkdir
  export CTP3_CONFIG_DIRECTORY=$cfgdir
  source \$CTP3_ROOT/etc/profile.d/ctp3-ipbus.sh

Use:
ttcpon to calibrate/check status of TTCPON network
qtltu  to control the CTP emulator

EOF
fi
# update part:
source $CTP3_ROOT/etc/this.sh
cd $CTP3_CONFIG_DIRECTORY
$CTP3_ROOT/scripts/bin/makemycfg.py check

