#!/bin/bash
function helpexit() {
cat - <<-EOF
Configure ctp3-ipbus environment (should be started from trigger account):

Usage:
Assuming ctp3-ipbus software is installed, i.e. CTP3_ROOT env. variable is set.

ctp3-init.sh eth bn ip
             eth myboards_file
             update
1.
eth: ethernet interface where LTU board is connected
bn:  LTU board number
ip:  IP address of the board

2.
pair 'bn ip' can be given in a myboards_file file -necessary when more boards
present

3.
Use 'update' to update the CFG directory (e.g. before firmware update),
created before 

Operation:
- create: ~/CFG/ ~/WORK/
  copy/create files: ->CFG/myboards.txt ethers
- ping ip
  if ok:
    - set CTP3_CONFIG_DIRECTORY, CTP3_WORK_DIRECTORY
    - msg: ping ok...
    - start: makemycfg.py    (preparing some files in CFG directory)

EOF
exit 1
}

echo "Doing ctp3-init as user '$USER' with args '$*'"
cfgdir=~/CFG   #my
wrkdir=~/WORK  #my
mkdir -p $cfgdir $wrkdir
if [ $# -eq 3 ] ;then    # eth bn ip
  ethif=$1
  bn=$2
  bip=$3
  printf -v hexbn "%x" $bn
  echo "08:00:30:00:2a:"$hexbn $bip >$cfgdir/ethers
  echo $bn $bip >$cfgdir/myboards.txt
elif [ $# -eq 2 ] ;then  # eth myboards.txt
  if [ ! -e "$2" ] ;then
    echo "$2 not found"
    exit 1
  fi
  ethif=$1
  #myfile=$2
  cp $2 /tmp/temp_myboards.txt
  myfile="/tmp/temp_myboards.txt"
  lnum=0 ; bn=0
  while read -a line ;do
    lnum=$(($lnum+1));
    bni=${line[0]}
    bipi=${line[1]}
    #echo $lnum;
    printf -v hexbn "%x" $bni
    if [ $lnum -eq 1 ] ;then
      bn=$bni   # when more boards, check oly 1st one with a ping later
      bip=$bipi
      echo "08:00:30:00:2a:"$hexbn $bipi >$cfgdir/ethers
      echo $bn $bip >$cfgdir/myboards.txt
    else
      echo "08:00:30:00:2a:"$hexbn $bipi >>$cfgdir/ethers
      echo $bni $bipi >>$cfgdir/myboards.txt
    fi
  done < $myfile
elif [ $# -eq 1 ] ;then  # update
  if [ "$1" != "update" ] ;then
    helpexit
  fi
else
  helpexit
fi
#
# check if LTU respondig:
#export CTP3_ETHER_FILE=$cfgdir/ethers
#ctp3-docker service start
#sleep 1
ping -c 1 $bip
if [ $? -ne 0 ] ;then
  cat - <<-EOF
atb_$bn, ip:$bip does not respond, check network config for $ethif

EOF
exit 1
fi

echo
echo "atb_$bn, ip:$bip ping ok, updating CFG directory..."
#
export CTP3_CONFIG_DIRECTORY=$cfgdir
export CTP3_WORK_DIRECTORY=$wrkdir
#
source $CTP3_ROOT/etc/this.sh
cd $CTP3_CONFIG_DIRECTORY
$CTP3_ROOT/scripts/bin/makemycfg.py check
#
cat - <<-EOF

Recommended lines in ~/.bashrc:
  export CTP3_ROOT="..."
  export CTP3_WORK_DIRECTORY=$wrkdir
  export CTP3_CONFIG_DIRECTORY=$cfgdir
  source \$CTP3_ROOT/etc/profile.d/ctp3-ipbus.sh

Log off/on (to get above lines executed).
Use:
ttcpon to calibrate/check status of TTCPON network
qtltu  to control the CTP emulator

EOF

