#!/bin/bash
function help() {
cat - <<-EOF
Usage:
1. install docker and ctp3-docker (root account):
ctp3install install

2. configure environments for service/user dockers (trigger account):
ctp3install eth bn ip
            eth myboards.txt
eth: ethernet interface where LTU board is connected
bn:  LTU board number
ip:  IP address of the board

Operation:
- create: ~/CFG/ ~/WORK/
  copy/create ->CFG/myboards.txt ethers
- ctp3-docker service start
  ping ip
  if ok:
    - set CTP3_CONFIG_DIRECTORY, CTP3_WORK_DIRECTORY
    - msg: ping ok...
    - start: ctp3-docker prod run makemycfg.py
    - msg: start 'ctp3-docker prod run gui user'
EOF
exit 1
}
# echo pars:$#
[ $# -gt 0 ] || help
if [ `whoami` = 'root' -a $1 = 'install' ] ;then
  yum install -y docker
  systemctl enable --now docker
  yum install yum-plugin-copr -y
  yum copr enable mvala/ctp -y
  yum install ctp3-docker -y
  #echo "useradd trigger..."
  #useradd -m -s/bin/bash -d/home/trigger -c'Trigger account' trigger
  if [ ! -f /etc/sudoers.d/trigger ] ;then
    cat - <<-EOF >>/etc/sudoers.d/trigger
trigger	ALL=(ALL)       NOPASSWD: /usr/bin/docker
EOF
  fi
  exit 0
fi
#
cfgdir=~/CFG
wrkdir=~/WORK
mkdir -p $cfgdir $wrkdir
if [ $# -eq 3 ] ;then    # eth bn ip
  ethif=$1
  bn=$2
  bip=$3
  printf -v hexbn "%x" $bn
  echo "08:00:30:00:2a:"$hexbn $bip >$cfgdir/ethers
  echo $bn $bip >$cfgdir/myboards.txt
elif [ $# -eq 2 ] ;then  # eth myboards.txt
  if [ ! -e "$2" ] ;then
    echo "$2 not found"
    exit 1
  fi
  ethif=$1
  #myfile=$2
  cp $2 /tmp/temp_myboards.txt
  myfile="/tmp/temp_myboards.txt"
  lnum=0 ; bn=0
  while read -a line ;do
    lnum=$(($lnum+1));
    bni=${line[0]}
    bipi=${line[1]}
    #echo $lnum;
    printf -v hexbn "%x" $bni
    if [ $lnum -eq 1 ] ;then
      bn=$bni   # when more boards, check oly 1st one with a ping later
      bip=$bipi
      echo "08:00:30:00:2a:"$hexbn $bipi >$cfgdir/ethers
      echo $bn $bip >$cfgdir/myboards.txt
    else
      echo "08:00:30:00:2a:"$hexbn $bipi >>$cfgdir/ethers
      echo $bni $bipi >>$cfgdir/myboards.txt
    fi
  done < $myfile
else
  help
fi
#
# check if LTU respondig:
export CTP3_ETHER_FILE=$cfgdir/ethers
ctp3-docker service start
sleep 1
ping -c 1 $bip
if [ $? -ne 0 ] ;then
  cat - <<-EOF
atb_$bn, ip:$bip does not respond, check network config for $ethif

EOF
exit 1
fi

echo
echo "atb_$bn, ip:$bip ping ok, preparing CFG directory..."
#
export CTP3_CONFIG_DIRECTORY=$cfgdir
export CTP3_WORK_DIRECTORY=$wrkdir
#
# neviem ako vykonat script v dockeri (tak aby mal nastavene CTP3_ premenne) napr:
#~/bin/ctp3-docker prod user runcmd './makemycfg.py check'
#
# Toto bavi:
sudo docker run -it --rm -v /home/trigger:/home/user -ti --stop-signal SIGRTMIN+3 --tmpfs /run --tmpfs /tmp -v /sys/fs/cgroup:/sys/fs/cgroup:ro -e 'container=docker' --cap-add=ALL  --net=host -v /home/trigger/CFG:/home/user/ctp3-cfg -e CTP3_CONFIG_DIRECTORY=/home/user/ctp3-cfg -v /home/trigger/WORK:/home/user/ctp3-wk -e CTP3_WORK_DIRECTORY=/home/user/ctp3-wk --user user -w /home/user gitlab-registry.cern.ch/alicectp3/ctp3-ipbus:latest /bin/bash -c 'source /opt/ctp3-ipbus/etc/this.sh ; cd ~/CFG ; ~/bin/makemycfg.py check'
#
cat - <<-EOF

Enter the docker environment by issuing following command:
ctp3-docker prod run user gui

EOF

