#!/bin/bash

PROJECT_DIR="$(dirname $(dirname $(readlink -m $0)))"

PRINT_DEBUG=false
MY_CMAKE_OPTS=""
MY_MAKE_OPTS=""
MY_BUILDSYS="make"
MY_CMAKE="cmake"
MY_PREFIX=${MY_PREFIX-$PROJECT_DIR}

[ -f /usr/bin/cmake3 ] && MY_CMAKE="cmake3"

CXX=""
CC=""

for ARG in $@
do
	case $ARG in
		"clean")
			echo "Cleaning up build directory"
			rm -rf ${PROJECT_DIR}/build
			;;
		"release")
			echo "Building release version"
			MY_CMAKE_OPTS="${MY_CMAKE_OPTS} -DCMAKE_BUILD_TYPE=Release"
			BUILDING_DOC=true;
			;;
		"install")
			echo "Will install after compilation"
			MY_MAKE_OPTS="${MY_MAKE_OPTS} install"
			;;
		"rpm")
			echo "Building rpm"
			MY_MAKE_OPTS="${MY_MAKE_OPTS} rpm"
			;;
        "grpc")
			echo "Building gdpr"
                        MY_CMAKE_OPTS="${MY_CMAKE_OPTS} -DGRPC=ON"
			;;
		*)
			echo "Unknown argument! Exiting..."
			exit 1
	esac
done

if [[ ! -d $PROJECT_DIR/build ]]
then
	echo "Creating build directory"
	mkdir $PROJECT_DIR/build
fi

MY_PROG_NAME="ctp3-ipbus"
if [ -n "$MY_PROJECT_VER" ];then
        MY_VER_MAJOR=$(echo "$MY_PROJECT_VER" | cut -d '.' -f1)
        MY_VER_MINOR=$(echo "$MY_PROJECT_VER" | cut -d '.' -f2)
        MY_VER_PATCH=$(echo "$MY_PROJECT_VER" | cut -d '.' -f3 | cut -d '-' -f1)
        MY_VER_RELEASE=$(echo "$MY_PROJECT_VER" | cut -sd '-' -f2)
        [ -n "$MY_VER_RELEASE" ] || MY_VER_RELEASE="1"
        [[ $MY_VER_RELEASE =~ ^[[:digit:]] ]] || MY_VER_RELEASE="0.1.$MY_VER_RELEASE"
        sed -i 's/^project(.*/project('$MY_PROG_NAME' VERSION '$MY_VER_MAJOR'.'$MY_VER_MINOR'.'$MY_VER_PATCH' DESCRIPTION "CTP3 ipbus - ctp software for ALICE RUN 3")/' CMakeLists.txt
        sed -i 's/^set(PROJECT_VERSION_RELEASE.*/set(PROJECT_VERSION_RELEASE '$MY_VER_RELEASE')/' CMakeLists.txt
        echo "Custom version : $MY_VER_MAJOR.$MY_VER_MINOR.$MY_VER_PATCH-$MY_VER_RELEASE"
fi

cd $PROJECT_DIR/build || { echo "Missing build directory!"; exit 1; }

$MY_CMAKE -DCMAKE_INSTALL_PREFIX=${MY_PREFIX} ${MY_CMAKE_OPTS} ../
N=$(nproc)
${MY_BUILDSYS} -j$N ${MY_MAKE_OPTS}
