#!/usr/bin/python3

from __future__ import print_function
import csv
import datetime
import math
import time
import logging
from random import randint

from . pon_fmc_i2c_program import program_clkmux
from . pon_fmc_i2c_program import program_si570
from . pon_fmc_i2c_program import program_si5344
from . pon_fmc_i2c_program import monitor_sfp
from . pon_fmc_i2c_program import program_leds
from . pon_fmc_i2c_program import set_phase_si5344
from . pon_fmc_i2c_program import get_status_si5344 as get_status_pll

logger_pon = logging.getLogger('ttc_pon')

#!/usr/bin/python3
c_MAX_TIMEOUT_MGT_READY = 20

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

def check_onu_locked(onu):
    """
        ***************************************************************************
        Check if ONU design is locked and phase_good for ONU phase good status
        ***************************************************************************

        ***************************************************************************
        arg:
            onu              : Object of OnuCore class	

        return:
           [locked, phase_good]
        ***************************************************************************
    """
    # Check ONU status before doing phase alignment
    # ONU must be locked in the downstream direction
    status_onu = onu.exd_read_onu_status()
    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
    if(mgt_rxpll_lock==0):
        logger_pon.warn("ONU MGT RX_PLL not locked!")
        return [0,0]
    elif(mgt_rx_ready==0):
        logger_pon.warn("ONU MGT RX not ready!")
        return [0,0]
    elif(rx_locked==0):
        logger_pon.warn("ONU downstream not locked!")
        return [0,0]

    # Check PLL is locked
    pll_locked = 0
    while(pll_locked==0):
        if onu.conver=="ublaze":
          pll_status = get_status_pll(onu)
        else:
          pll_status= onu.ctrlcon.plldev.get_status()
        #print "check_onu_locked:", pll_status
        if(pll_status[0]==0):
            logger_pon.warn("IIC communication problem during get status of PLL")	
        pll_locked = not (pll_status[2])
        pll_loss_of_signal = pll_status[1]
        if(pll_loss_of_signal):
            logger_pon.warn("ONU PLL lost signal, there might be an issue in the downstream! Check problem!")			
            return [0, 0]

    # If the function arrive here the PLL must be already locked
    # Now we should check that the MGT TX_PLL is locked and the MGT_TX is ready
    timeout_cntr   = 0
    while(not (mgt_txpll_lock and mgt_tx_ready)):
        status_onu = onu.exd_read_onu_status()
        [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.1)
        if(timeout_cntr > c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("TX is not getting ready after timeout! Check problem!")						
            return [0, 0]

    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu

    return [1, phase_good]

def onu_init(onu, onu_addr=None, cascaded_nstandalone=1, pllload=False ):   # +alice
    """
        ***************************************************************************
        ONU local initialization algorithm
        ***************************************************************************

        ***************************************************************************
        arg:
            onu              : Object of OnuCore class
            onu_addr         : Address of ONU for upstream TDM (1-254)
            cascaded_nstandalone : cascaded is the common ONU case
			           standalone is when a monitoring ONU is used in the same board and 
                                     OLT clock is the same as ONU clock, all clocks come from oscillator
        return:
           success
        ***************************************************************************
    """
    if onu_addr==None: onu_addr= onu.onun   # +alice
    #print "onu_init:",onu_addr,pllload,lhc, onu.conver
    # Program FMC devices
    if onu.conver=="ublaze":
      if(cascaded_nstandalone) : program_clkmux(onu, 0x37)
      else : program_clkmux(onu, 0xFF)
      program_si570(onu, 1)
      if pllload: program_si5344(onu, 1, cascaded_nstandalone)
    else:
      #print "aboard atbp_7    dev si5345 loadpll"
      if pllload: onu.ctrlcon.plldev.checkandload()   # check (load if different) default onu config

    # Wait some time to PLL output to be stable
    if pllload: time.sleep(0.6)
    onu.exd_core_reset()
    time.sleep(0.5)

    # Program ONU upstream TDM address
    onu.exd_set_addr(onu_addr)

    # Check ONU status before doing phase alignment - ONU must be locked
    status_onu = onu.exd_read_onu_status()
    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
    timeout_cntr = 0
    print("[RX40_LOCKED, phase_good, RX_LOCKED, operational, MGT_TX_READY, MGT_RX_READY, MGT_TXPLL_LOCK, MGT_RXPLL_LOCK]")
    while((not (mgt_rx_ready and mgt_rxpll_lock and rx40_locked and rx_locked and mgt_txpll_lock and mgt_tx_ready))\
           and timeout_cntr< c_MAX_TIMEOUT_MGT_READY):
        status_onu = onu.exd_read_onu_status()
        print("onu_init %d:"%timeout_cntr, status_onu)
        [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
        timeout_cntr = timeout_cntr + 1
        time.sleep(1) #0.2)
        if(timeout_cntr >= c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("ONU is not getting ready after timeout, check ONU status (phase_good:0 in case of CRU role without BCin)!")		
            return 0
    if phase_good==0:
      logger_pon_warn("if CDC (fi>11) & CRU role & no BCin signal: phase_good is 0")
    if onu.conver=="ublaze":
      program_leds(onu, 0x00)
    monitor_sfp(onu,1)

    return 1


