#!/usr/bin/python3

from __future__ import print_function
from builtins import str
import csv
import datetime
import math
import time
import logging
from random import randint

from . pon_fmc_i2c_program import program_clkmux
from . pon_fmc_i2c_program import program_si570
from . pon_fmc_i2c_program import program_si5344
from . pon_fmc_i2c_program import monitor_sfp
from . pon_fmc_i2c_program import program_leds
from . pon_fmc_i2c_program import set_phase_si5344
from . pon_fmc_i2c_program import get_status_si5344 as get_status_pll

logger_pon = logging.getLogger('ttc_pon')

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
c_STEP_PHASE = 5

#!/usr/bin/python3
c_MAX_TIMEOUT_MGT_READY = 80

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
def check_onu_mlinkolt_locked(onu, olt):
    """
        ***************************************************************************
        Check if ONU_MLINK_OLT design is locked and phase_good for ONU2OLT CDC status
        ***************************************************************************

        ***************************************************************************
        arg:
            onu              : Object of OnuCore class
            olt              : Object of OltCore class			

        return:
           [locked, phase_good_onu2olt, waittime]
        ***************************************************************************
    """
    # Check ONU status before doing phase alignment
    # ONU must be locked in the downstream direction
    wt= 0.0
    status_onu = onu.exd_read_onu_status()
    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
    if(mgt_rxpll_lock==0):
        logger_pon.warn("ONU MGT RX_PLL not locked!")
        return [0,0,wt]
    elif(mgt_rx_ready==0):
        logger_pon.warn("ONU MGT RX not ready!")
        return [0,0,wt]
    elif(rx_locked==0):
        logger_pon.warn("ONU downstream not locked!")
        return [0,0,wt]

    # Check PLL is locked
    pll_locked = 0
    while(pll_locked==0):
        if onu.conver=="ublaze":
          pll_status = get_status_pll(onu)
        else:
          pll_status = onu.ctrlcon.plldev.get_status()
        if(pll_status[0]==0):
            logger_pon.warn("IIC communication problem during get status of PLL")	
        pll_locked = not (pll_status[2])
        pll_loss_of_signal = pll_status[1]
        if(pll_loss_of_signal):
            logger_pon.warn("ONU PLL lost signal, there might be an issue in the downstream! Check problem!")			
            return [0, 0, wt]
    # If the function arrive here the PLL must be already locked
    # Now we should check that the MGT TX_PLL is locked and the MGT_TX is ready
    timeout_cntr   = 0
    #print "[rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock]"
    while(not (mgt_txpll_lock and mgt_tx_ready)):
        status_onu = onu.exd_read_onu_status()
        [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
        #print "check_onu_mlink_locked %d:"%timeout_cntr, status_onu
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.5) ; wt= wt+0.5
        if (timeout_cntr % 10)==0:
          print("check_onu_mlinkolt_locked: %.2fs onustatus:"%(wt), status_onu)
        if(timeout_cntr > c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("ONU TX is not getting ready after timeout(%.2fs)! Check problem!"%wt)
            return [0, 0, wt]

    # If the function arrive here the PLL must be already locked
    # Now we should check that the OLT Tx/Rx is ready and OLT Tx PLL is locked
    olt.exd_clear_sticky()
    time.sleep(0.1)	
    status_olt = olt.exd_read_olt_status_sticky()
    [mgt_pll_lock, mgt_tx_ready, mgt_rx_ready, phase_good_onu2olt] = status_olt
    timeout_cntr   = 0
    while(not (mgt_pll_lock and mgt_tx_ready and mgt_rx_ready)):
        olt.exd_clear_sticky()
        time.sleep(0.1)  ; wt= wt+0.1
        status_olt = olt.exd_read_olt_status_sticky()
        [mgt_pll_lock, mgt_tx_ready, mgt_rx_ready, phase_good_onu2olt] = status_olt
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.1) ; wt= wt+0.1
        if (timeout_cntr % 10)==0:
          print("check_onu_mlinkolt_locked: %.2fs oltstatus:"%(wt), status_olt)
        if(timeout_cntr > c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("OLT Tx is not getting ready after timeout(%.2fs)! Check problem!"%wt)						
            return [0, 0, wt]

    return [1, phase_good_onu2olt, wt]

def onu2olt_mode(onu, olt, mode):
    """
        ***************************************************************************
        Check if ONU_MLINK_OLT design is locked and phase_good for ONU2OLT CDC status
        ***************************************************************************

        ***************************************************************************
        arg:
            onu              : Object of OnuCore class
            olt              : Object of OltCore class			
            mode             : 0=STANDALONE (OLT clock comes from oscillator), 1=CASCADED (OLT clock comes from ONU external PLL)
        return:
           [locked, phase_good_onu2olt]
        ***************************************************************************
    """
    if(mode==1):   # Cascaded
        program_clkmux(onu, 0x36)
        onu.exd_core_reset()
        time.sleep(1)
        olt.exd_core_reset()
        time.sleep(1)		
    elif(mode==0): # Standalone
        program_clkmux(onu, 0x37)
        olt.exd_core_reset()
        time.sleep(1)
        onu.exd_core_reset()
        time.sleep(1)			
    else:
        logger_pon.warn("Invalid Configuration")
        return 0

def automatic_phase_good_onu2olt(onu, olt):
    """
        ***************************************************************************
        Algorithm to find phase good for the ONU 2 OLT CDC (for more information about the algorithm, check the comments on the header of the file)
        ***************************************************************************

        ***************************************************************************
        arg:
            onu              : Object of OnuCore class
            olt              : Object of OltCore class			
            mode             : 0=STANDALONE (OLT clock comes from oscillator), 1=CASCADED (OLT clock comes from ONU external PLL)
        return:
           success
        ***************************************************************************
    """
    found_phase         = 0
    found_ris_edge      = 0
    phase               = 0
    first_iter          = 1
    previous_phase_good_onu2olt = 0
    #logger_pon.info('PHASE'+ ' --- ' + 'GOOD?')	
    wt1=wt2=0.0
    while(not found_phase):
        # Set PLL phase
        onu.exd_phase_offset40(phase)

        # Check ONU and OLT are properly locked
        [onu_mlinkolt_ok, phase_good_onu2olt,wt1] = check_onu_mlinkolt_locked(onu, olt)
        if (onu_mlinkolt_ok==0): return 0

        # Algorithm to stop iteration
        # Finds rising edge of phase_good signal
        # Then finds falling edge
        # Then it positions in the middle		
        if(not first_iter):
            if(previous_phase_good_onu2olt!=phase_good_onu2olt):
                if(phase_good_onu2olt==1): # Found a rising edge
                    found_ris_edge = 1
                    phase_ris_edge = phase
                    previous_phase_good_onu2olt = phase_good_onu2olt					
                elif(found_ris_edge):      # Found a falling edge and already found rising edge
                    # Position phase in the middle of ris->fall window				
                    phase = (phase + phase_ris_edge)//2
                    # Set PLL phase
                    onu.exd_phase_offset40(phase)
                    # Check status
                    [onu_mlinkolt_ok, phase_good_onu2olt,wt2] = check_onu_mlinkolt_locked(onu, olt)
                    if (onu_mlinkolt_ok==0): return 0					
                    logger_pon.info("onu2olt: Phase adjusted at:" + str(phase) + " " + str(wt2) + "s")
                    found_phase = 1
                else:
                    previous_phase_good_onu2olt = phase_good_onu2olt
            else:
                previous_phase_good_onu2olt = phase_good_onu2olt			
        else:
            previous_phase_good_onu2olt = phase_good_onu2olt

        first_iter = 0
        # note: wt1 values range from 3 to 14 s (measured 31.7.2019 on atb_27)
        str_print = "onu2olt: %5d --- %5d (%2.2fs)" % (phase, phase_good_onu2olt, wt1)
        #print "phase_good_onu2olt:", str_print
        logger_pon.info(str_print)

        phase = phase + c_STEP_PHASE

        if(phase>280*2):
            logger_pon.warn("Not found phase good for ONU2OLT CDC")		
            return 0

    return 1
