#!/usr/bin/python3

from __future__ import print_function
from builtins import object
import socket
#!/usr/bin/python3
import time
import csv
from random import randint
#!/usr/bin/python3
import atbwrap,decimal,i2c,auxfuns
#!/usr/bin/python3
#!/usr/bin/python3
decimal.getcontext().prec=16

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

DBGPRT= False
class DriverComm(object):
  """
        This class is the driver-layer of the TTC-PON firmware. It is responsible for communicating from Python to the FPGA using the methods write_ctrl_reg and read_ctrl_reg
        The higher level methods implemented in TTC-PON in the OltCore and OnuCore classes make use of those driver-layer methods

        ***************************************************************************************************************
        ***   In case a user is implementing TTC-PON in its CUSTOM BOARD:                                           ***
        ***       option1: this class shall be replaced by the equivalent control interface used for his/her board  *** 
        ***       option2: the user should overwrite the methods read_ctrl_reg and write_ctrl_reg                   *** 
        ***************************************************************************************************************

  """
  def __init__(self, ipubl, port, olt=False, olt2ndlevel=None):
    """ 3 possible ways how ONU is connected:
ublaze: ip, port    ttcpon fi:v010
cru:    ip:'' port:'' (i.e. CRUs ONU)  fi: does not matter
ipbus:  ip: 'boardname' port:''        ttcpon fi: v010
        I.e. separate ipbus connection for each onu/olt 
             POSSIBLE common ipbus connection for onu/olt (multiolt)
olt2ndlevel: None in case of ONU when ATB in CRU role
             inst. of OltMlinkExdsgAlice (see ttcponnet.py non CRU role)
"""
    self.multiolt= -1
    self.conver= None
    if port=='':   # cru or ipbus type
      if ipubl!='':   # ipbus type
        if (not olt) and (olt2ndlevel==None):
          self.ctrlcon= DriverIpbus(ipubl, olt,mode="g")
        else:
          self.ctrlcon= DriverIpbus(ipubl, olt)   # "s" default
        if self.ctrlcon.board==None:
          self.ctrlcon= None
          return
        self.conver= "ipbus"
        self.name= self.ctrlcon.board.bname
      else:
        self.ctrlcon= None
        self.conver= "cru"
        self.name= "cru"
    else:
      self.ctrlcon= DriverUblaze(ipubl, port)
      self.conver= "ublaze"
      self.name= "ublaze_"+ipubl
      #self.read_ctrl_reg= self.ctrlcon.read_ctrl_reg incorrect
    if olt: self.multiolt= 1   # default v4 kcu105 fi, using 1st OLT
    #self.logger = logging.getLogger('ttc_pon') moved to OltMlinkExdsgAlice
    #self.logger.debug('Opening socket at port:' + str(port) + ", ip:" + ipubl+",ver:"+self.conver)
  def getoltn(self):
    return self.multiolt
  def getoltcsvn(self):
    csvname= "_%d_%s"%(self.getoltn(), self.name)
    return csvname
  def getuname(self):
    if self.ctrlcon.olt:
      return "%s.olt%d"%(self.name, self.multiolt)
    else:
      return "%s.onu"%(self.name)
  def setolt(self, n):
    """ Must be called after OltCore instantiated.
Set OLT for all operations (i.e. only 1 OLT in a time can be used).
"""
    if (n>0) and (n<10):
      if (self.multiolt!=0):   # -1, 1..9
        self.multiolt= n
        if self.conver=="ipbus":
          self.ctrlcon.setolt(n)
          sfpname= self.ctrlcon.board.fwsfps["olt%d"%n]
          insts= self.ctrlcon.board.findDeviceSFPS(sfpname)
          if len(insts)==1:
            self.sfpinst= insts[0]
          else:
            self.sfpinst= None
        else:
          print("Error: driver_comm setolt: %d for ublaze ?"%n)
      else:
        print("Error: setolt not available (fi v2)")
    else:
      print("Error: setolt %d not allowed (1..9 available)"%n)
  def read_ctrl_reg(self, reg_addr):
    if DBGPRT: print("r 0x%x"%(reg_addr))
    val= self.ctrlcon.read_ctrl_reg(reg_addr)
    if DBGPRT: print("r 0x%x: 0x%x"%(reg_addr, val))
    return(val)
  def write_ctrl_reg(self, reg_addr, reg_value):
    if DBGPRT: print("w 0x%x = 0x%x"%(reg_addr,reg_value))
    return(self.ctrlcon.write_ctrl_reg(reg_addr, reg_value))

URGS= 0x4000  # User regs addres bit set in addr! Was 0x40 before fi v4, v4: 0x400
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
class DriverIpbus(object):
  def __init__(self,bn,olt, mode=None):
    # mode: None (pllin1) or g(pllin0)
    if (mode=="s") or (mode==None):
      mode= auxfuns.ltucfgpars(bn,"pllcfg")
    self.board= atbwrap.OpenATBoard(bn, None, mode)   
    self.olt= olt
    if self.board== None: return
    # ltu_kcu105: si5345 is on i2c
    # ltu_logic: si5345 is on i2cpll
    #
    #nnlist= self.board.getNodeNames()
    #try:
    #  i= nnlist.index("ttcpon")
    #except ValueError:
    #  i= None
    #print "i:",i, "nnlist:",nnlist
    # -1: only at the very start 
    #  0: fi v2 kcu105, 
    # default  1..9: fi v4 or 010 (multilinkolt)
    #if i==None:   # v4 fi
    #von if self.board.nodExists("i2cpll", True):
    #  i2cpll= i2c.I2Ccore(self.board,"i2cpll")
    #  self.plldev= i2c.Si534x(i2cpll, "si5345")
    #  pllmodel= self.plldev.getid()
    #  #print "DriverIpbus pll:%s"%pllmodel
    #else:
    #  self.plldev= None
    self.plldev= self.board.plldev
    if not olt or olt:
      # temperature of the ATB board:
      #i2cmuxs= (0x74, 0x76)
      #i2cg= i2c.I2Ctpcore(self.board,"i2cmainttc", i2cmuxs)
      ##self.board.Write("i2cmainttc.mux", 11) has to be done before tc74.gettemp()
      #self.tc74= i2c.TC74(i2cg, "tc74", mux=(0x76,0, 0x74,1<<0))
      ##print "DriverIpbus:",olt,bn,self.tc74.getid()
      # new (atbwrap way):
      self.tc74= self.board.tc74
    if olt:
      # check PLL for global or stdalone according to requested mode):
      #  self.setolt(0)   # has to be done 1st time (-1 -> 0)
      self.setolt(1)   # v010, using 1st OLT
      #print "setolt: ipbus, using OLT1..."
    else:   # onu
      # check PLL for global:
      self.CORE_REGS= "ttcpononu_c.cregs"
      self.EDL_REGS= "ttcpononu_u.uregs"
      self.cl_offset= 0x500
      self.ul_offset= 0x4500
  def setolt(self, n):
    if (n>0) and (n<10):
      #print "DBG: driver_comm setolt:", n
      self.CORE_REGS= "ttcponolt_c%d.cregs"%n
      self.EDL_REGS= "ttcponolt_u%d.uregs"%n
      #
      self.cl_offset= (n-1)*0x80
      self.ul_offset= (n-1)*0x80 + 0x4000
      #self.cl_offset= 0    # cl/ul offset considered in nodename (CORE/EDL_REGS)
      #self.ul_offset= 0x4000
    else:
      print("ERROR: driver_comm setolt:", n)
  def write_ctrl_reg(self, reg_addr, reg_value):
    #print "write: 0x%x <- 0x%x"%(reg_addr,reg_value)
    blockdata= [reg_value]
    #if reg_addr & URGS:
    #  self.board.WriteBlock(self.EDL_REGS, blockdata, reg_addr&0xffffbfff)
    #else:
    #  self.board.WriteBlock(self.CORE_REGS, blockdata, reg_addr)
    if reg_addr >= URGS:
      rega= reg_addr - self.ul_offset
      self.board.WriteBlock(self.EDL_REGS, blockdata, rega)
    else:
      rega= reg_addr - self.cl_offset
      self.board.WriteBlock(self.CORE_REGS, blockdata, rega)
    data= 1   #reg_value
    # read back (as ublaze version is):
    #data= self.board.Read("ttcpon.oltonuregs", reg_addr)
    return int(data)
  def read_ctrl_reg(self, reg_addr):
    #print "read: 0x%x"%reg_addr
    #if reg_addr & URGS:
    #  blockdata= self.board.ReadBlock(self.EDL_REGS, 1, reg_addr&0xffffbfff)
    #else:
    #  blockdata= self.board.ReadBlock(self.CORE_REGS, 1, reg_addr)
    if reg_addr >= URGS:
      rega= reg_addr - self.ul_offset
      blockdata= self.board.ReadBlock(self.EDL_REGS, 1, rega)
    else:
      rega= reg_addr - self.cl_offset
      blockdata= self.board.ReadBlock(self.CORE_REGS, 1, rega)
    data= blockdata[0]
    return int(data)

class DriverUblaze(object):
    def __init__(self, ip='192.168.1.2', port=9001):
        """
            ***************************************************************************
            Class Constructor
            ***************************************************************************

            ***************************************************************************
            arg:
                ip: string with the IP address of the board being controlled.
                    The board IP in the TTC-PON example design can be changed via the software implemented in the microblaze, i.e. by using the Vivado SDK
                port: integer with the port number
            ***************************************************************************
        """	
        self.ip= ip
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.connect((ip, port))
        #self.logger = logging.getLogger('ttc_pon')
        #self.logger.debug('Opening socket at port:' + str(port) + ", ip:" + ip)
        #slow_down is only used for TTC-PON demo in order to slow down the communication and perform demonstrations of the system, user do not have to care about that
        self.slow_down = 0

    def write_ctrl_reg(self, reg_addr, reg_value):
        """
            ***************************************************************************		
            Write register via microblaze interface for the example design
			
            !!! This method depends on the way the board is controlled !!!
            This method shall be replaced by the equivalent read method in Python in case a user implements PON in a custom board           
            ***************************************************************************

            ***************************************************************************
            arg: reg_addr and reg_value are, respectively, the address register and value to be written to the interface
            return: 1 if success
            ***************************************************************************
			
            ***************************************************************************
            In our example design, two slaves are present: CORE SLAVE + EXDSG SLAVE
            -> To send a message to the CORE slave:
            we send strings in the format: "w 0x0008 0x00000390" to indicate write "0x00000390" to control reg 2 (byte addressing) of the core
            -> To send a message to the EXDSG slave, an offset 0x1000 is given:
            we send strings in the format: "w 0x1008 0x00000390" to indicate write "0x00000390" to control reg 2 (byte addressing) of the exdsg
            
            The method write_ctrl_reg, multiplies the reg_addr by four in order to send to the microblaze
            If the user wants to write to the slave, it should write to => slave_reg_local_addr + 0x1000/4
            It is: if you want to write to register 2 of the exdsg, call this method as: self.write_ctrl_reg(2+1024, value)
            ***************************************************************************
        """		
        message = b"w 0x%0.4X 0x%0.8X\n" % (4 * reg_addr, reg_value)
        if(self.slow_down) : time.sleep(0.1)
        self.s.send(message)
        data = self.s.recv(2048)
        data = int(data, 16)	
        return data


    def read_ctrl_reg(self, reg_addr):
        """
            ***************************************************************************
            Reads register via microblaze interface for the example design

            !!! This method depends on the way the board is controlled !!!
            This method shall be replaced by the equivalent read method in Python in case a user implements PON in a custom board
            ***************************************************************************

            ***************************************************************************
            arg: reg_addr is the address register to be read
            return: register_value as an integer
            ***************************************************************************

            ***************************************************************************
            In our example design, two slaves are present: CORE SLAVE + EXDSG SLAVE
            -> To send a message to the CORE slave:
            we send strings in the format: "r 0x008" to indicate read from control reg 2 (byte addressing) of the core
            -> To send a message to the EXDSG slave, an offset 0x1000 is given:
            we send strings in the format: "r 0x1008" to indicate read from control reg 2 (byte addressing) of the exdsg
            
            The method read_ctrl_reg, multiplies the reg_addr by four in order to send to the microblaze
            If the user wants to read from the slave, it should read from => slave_reg_local_addr + 0x1000/4
            It is: if you want to read to register 2 of the exdsg logic, call this method as: self.write_ctrl_reg(2+1024, value)
            ***************************************************************************
        """		
        message = b"r 0x%0.4X\n" % (4 * reg_addr)	
        #print("r 0x%x        "%reg_addr+message)		
        self.s.send(message)
        data = self.s.recv(2048)
        data = int(data, 16)
        return(data)
