#!/usr/bin/python3

from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import range
import time,sys

from . olt_core import OltCore
from . onu_core import OnuCore
#!/usr/bin/python3
#!/usr/bin/python3
from . import regnames
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

class OltExdsg(OltCore):

    """
        Class to control the OLT_USER_CORE example design user logic
        This class does not act on OLT_CORE ip

        Replace this class with the equivalent for an user custom design
    """

    def __init__(self, ip='192.168.1.2', port=9001):
        """
            ***************************************************************************
            Class Constructor
            ***************************************************************************

            ***************************************************************************
            arg:
                ip: string with the IP address of the board being controlled.
                    The board IP in the TTC-PON example design can be changed via the software implemented in the microblaze, i.e. by using the Vivado SDK
                port: integer with the port number
            ***************************************************************************
        """		
        OltCore.__init__(self, ip, port)
        self.addr_offset = 1024

    def exd_core_reset(self):
        """
            ***************************************************************************
            Reset OLT core
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """		
        stat = self.read_ctrl_reg(self.addr_offset + 0)
        stat = stat | 1
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        return 1

    def exd_bert_set_heartbeat(self, number_onu):
        """
            ***************************************************************************
            Set number of active ONU's after calibration (required for BERT)
            ***************************************************************************

            ***************************************************************************
            arg:
                number_onu: number of ONU's connected to the system after calibration
            return:
                1
            ***************************************************************************
        """		
        self.write_ctrl_reg(self.addr_offset + 1, (number_onu * self.c_BURST_LENGTH_OLTRXCLK) << 2)
        return 1

    def exd_bert_clear(self):
        """
            ***************************************************************************
            Clear BERT counters
            It is important to set heartbeat using the method exd_bert_set_heartbeat before clearing counters
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """		
        stat = self.read_ctrl_reg(self.addr_offset + 1)
        stat = stat | 1
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        return 1

    def exd_bert_latch(self):
        """
            ***************************************************************************
            Latch BER counters
            It is important to clear at least once the counters with the method exd_bert_clear after calibration
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """			
        stat = self.read_ctrl_reg(self.addr_offset + 1)
        stat = stat | (1 << 1)
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        stat = stat ^ (1 << 1)
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        return 1

    def exd_bert_select_channel(self, channel):
        """
            ***************************************************************************
            Select channel for BERT reading
            The OLT example design firmware contains 16 BBERT units in parallel checking addresses 1-16 (channels 0-15)
            ***************************************************************************

            ***************************************************************************
            arg:
                channel: 0-15 corresponding to addresses 1-16 			
            return:
                1
            ***************************************************************************
        """			
        #print "exd_bert_select_channel:", self.addr_offset,channel
        #if channel>15:   is ok (i.e. up to 64) in rev101
        #  print "Error: exd_bert_select_channel: %d > 15, only low 4 bits valid!"%channel
        self.write_ctrl_reg(self.addr_offset + 10, channel)
        return 1

    def exd_read_error_sum(self):
        """
            ***************************************************************************
            Read BBERT error counter values (BBERT assumes that ONU is transmitting a PRBS-7 pattern)
            ***************************************************************************

            ***************************************************************************
            return:
                number of errors
            ***************************************************************************
        """				
        msb = self.read_ctrl_reg(self.addr_offset + 7)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 6)
        return (lsb + msb)

    def exd_read_bit_sum(self):
        """
            ***************************************************************************
            Read BBERT bit counter values (BBERT assumes that ONU is transmitting a PRBS-7 pattern)
            ***************************************************************************

            ***************************************************************************
            return:
                number of bits
            ***************************************************************************
        """			
        msb = self.read_ctrl_reg(self.addr_offset + 5)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 4)
        return (lsb + msb)

    def exd_read_ploss_sum(self):
        """
            ***************************************************************************
            Read BBERT packet loss counter values (BBERT assumes that ONU is transmitting a PRBS-7 pattern)
            ***************************************************************************

            ***************************************************************************
            return:
                number of packet lost
            ***************************************************************************
        """			
        msb = self.read_ctrl_reg(self.addr_offset + 9)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 8)
        return (lsb + msb)

    def exd_tx_cntr(self):
        """
            ***************************************************************************
            Sets OLT transmitter to transmit a byte-counter pattern (useful for debugging)
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """		
        self.write_ctrl_reg(self.addr_offset + 2, 3)

    def exd_tx_cte(self):
        """
            ***************************************************************************
            Sets OLT transmitter to transmit a constant pattern (useful for debugging)
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """		
        self.write_ctrl_reg(self.addr_offset + 2, 2)
        return 1

    def exd_tx_prbs23(self):
        """
            ***************************************************************************
            Sets OLT transmitter to transmit a PRBS-23 pattern
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """		
        self.write_ctrl_reg(self.addr_offset + 2, 1)		
        return 1

    def exd_tx_prbs7(self):
        """
            ***************************************************************************
            Sets OLT transmitter to transmit a PRBS-7 pattern (for downstream BER tests)
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """		
        self.write_ctrl_reg(self.addr_offset + 2, 0)		
        return 1

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

class OnuExdsg(OnuCore):

    """
        Class to control the ONU_USER_CORE example design user logic
        This class does not act on ONU_CORE ip

        Replace this class with the equivalent for an user custom design
    """

    def __init__(self, ip='192.168.1.11', port=9001, onun= None, olt2ndlevel=None):
        """
            ***************************************************************************
            Class Constructor
            ***************************************************************************

            ***************************************************************************
            arg:
                ip: string with the IP address of the board being controlled.
                    The board IP in the TTC-PON example design can be changed via the software implemented in the microblaze, i.e. by using the Vivado SDK
                port: integer with the port number
            ***************************************************************************
        """		
        OnuCore.__init__(self, ip, port, onun, olt2ndlevel)
        self.addr_offset = 1024 #Example design address offset (related to microblaze implementation)

    def exd_core_reset(self):
        """
            ***************************************************************************
            Reset ONU core
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """		
        stat = self.read_ctrl_reg(self.addr_offset + 0)
        stat = stat | 1
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        return 1

    def exd_read_onu_status(self, print_enable=0):
        """
            ***************************************************************************
            Read port status from ONU
            ***************************************************************************

            ***************************************************************************
            arg:
                print_enable: (0/1) enables status print in the screen			
            return:
                [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock]
            ***************************************************************************
        """
        stat = self.read_ctrl_reg(self.addr_offset + 3)
        rx40_locked    = (stat & 0x00000001) >> 0
        phase_good     = (stat & 0x00000002) >> 1	
        rx_locked      = (stat & 0x00000004) >> 2
        operational    = (stat & 0x00000008) >> 3
        mgt_tx_ready   = (stat & 0x00000010) >> 4
        mgt_rx_ready   = (stat & 0x00000020) >> 5	
        mgt_txpll_lock = (stat & 0x00000040) >> 6
        mgt_rxpll_lock = (stat & 0x00000080) >> 7

        if(print_enable):
            print("ONU_STATUS:")
            print("rx40_locked:"    + str(rx40_locked))
            print("phase_good:"     + str(phase_good)) 
            print("rx_locked:"      + str(rx_locked))
            print("operational:"    + str(operational)) 
            print("mgt_tx_ready:"   + str(mgt_tx_ready))
            print("mgt_rx_ready:"   + str(mgt_rx_ready))
            print("mgt_txpll_lock:" + str(mgt_txpll_lock))
            print("mgt_rxpll_lock:" + str(mgt_rxpll_lock))

        return [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock]

    def exd_set_addr(self, addr_value):
        """
            ***************************************************************************
            Set ONU address (port of ONU CORE) corresponding to the ONU address in the TDM
            ***************************************************************************

            ***************************************************************************
            arg:
                addr_value: onu address (1-254)
            return:
                1
            ***************************************************************************
        """	
        self.write_ctrl_reg(self.addr_offset + 0, addr_value << 1)
        return 1

    def exd_get_addr(self):
        """
            ***************************************************************************
            Get ONU address (port of ONU CORE) corresponding to the ONU address in the TDM
            ***************************************************************************

            ***************************************************************************
            return:
                onu address (1-254)
            ***************************************************************************
        """		
        stat = self.read_ctrl_reg(self.addr_offset + 0)
        addr = (stat>>1)&0xFF
        return addr

    def exd_bert_clear(self):
        """
            ***************************************************************************
            Clear BERT counters
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """			
        stat = self.read_ctrl_reg(self.addr_offset + 1)
        stat = stat | 1
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        return 1

    def exd_bert_latch(self):
        """
            ***************************************************************************
            Latch BER counters
            It is important to clear at least once the counters with the method exd_bert_clear after ONU is locked
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """			
        stat = self.read_ctrl_reg(self.addr_offset + 1)
        stat = stat | (1 << 1)
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        stat = stat ^ (1 << 1)
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        return 1

    def exd_read_derror_correc_sum(self):
        """
            ***************************************************************************
            Read FEC double error corrected counter
            ***************************************************************************

            ***************************************************************************
            return:
                number of double error corrected
            ***************************************************************************
        """			
        msb = self.read_ctrl_reg(self.addr_offset + 11)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 10)
        return (lsb + msb)

    def exd_read_serror_correc_sum(self):
        """
            ***************************************************************************
            Read FEC single error corrected counter
            ***************************************************************************

            ***************************************************************************
            return:
                number of single error corrected
            ***************************************************************************
        """		
        msb = self.read_ctrl_reg(self.addr_offset + 9)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 8)
        return (lsb + msb)

    def exd_read_error_sum(self):
        """
            ***************************************************************************
            Read BER error counter values (BERT assumes that OLT is transmitting a PRBS-7 pattern)
            ***************************************************************************

            ***************************************************************************
            return:
                number of errors
            ***************************************************************************
        """			
        msb = self.read_ctrl_reg(self.addr_offset + 7)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 6)
        return (lsb + msb)

    def exd_read_bit_sum(self):
        """
            ***************************************************************************
            Read BER bit counter values (BERT assumes that OLT is transmitting a PRBS-7 pattern)
            ***************************************************************************

            ***************************************************************************
            return:
                number of bits
            ***************************************************************************
        """		
        msb = self.read_ctrl_reg(self.addr_offset + 5)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 4)
        return (lsb + msb)

    def exd_tx_inject_error(self):
        """
            ***************************************************************************
            Inject transmitter error to test upstream BBERT functionality in OLT side
            ***************************************************************************

            ***************************************************************************
            return:
                1
            ***************************************************************************
        """		
        stat = self.read_ctrl_reg(self.addr_offset + 2)
        stat = stat | (1 << 3)
        self.write_ctrl_reg(self.addr_offset + 2, stat)
        time.sleep(1)
        stat = stat ^ (1 << 3)
        self.write_ctrl_reg(self.addr_offset + 2, stat)
        return 1

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

class OnuFMCExdsg(OnuExdsg):

    """
        Class to control the internal TTC-PON team design where two ONU's are implemented in a KC705 (one in SFP connector and the other in TTC-PON FMC)
    """

    def __init__(self, ip='192.168.1.11', port=9001):
        """
            ***************************************************************************
            Class Constructor
            ***************************************************************************

            ***************************************************************************
            arg:
                ip: string with the IP address of the board being controlled.
                    The board IP in the TTC-PON example design can be changed via the software implemented in the microblaze, i.e. by using the Vivado SDK
                port: integer with the port number
            ***************************************************************************
        """		
        OnuExdsg.__init__(self, ip, port)
        self.fmc_offset = 2048

    def write_ctrl_reg(self, reg_addr, reg_value):
        """
            ***************************************************************************
            Write to FPGA with an offset
            ***************************************************************************

            ***************************************************************************
            arg: reg_addr and reg_value are, respectively, the address register and value to be written to the interface
            return: 1 if success
            ***************************************************************************
        """			
        return OnuExdsg.write_ctrl_reg(self, reg_addr + self.fmc_offset, reg_value)

    def read_ctrl_reg(self, reg_addr):
        """
            ***************************************************************************
            Read from FPGA with an offset
            ***************************************************************************

            ***************************************************************************
            arg: reg_addr is the address register to be written to the interface
            return: register value
            ***************************************************************************
        """			
        return OnuExdsg.read_ctrl_reg(self, reg_addr + self.fmc_offset)

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

class OltMlinkExdsg(OltExdsg):

    """
        Class to control the OLTs in the internal ALICE + TTC-PON team design where 1 ONU + 10 OLTs are implemented in the same FPGA
    """

    def __init__(self, ip='192.168.1.2', port=9001, lane=0):
        """
            ***************************************************************************
            Class Constructor
            ***************************************************************************

            ***************************************************************************
            arg:
                ip: string with the IP address of the board being controlled.
                    The board IP in the TTC-PON example design can be changed via the software implemented in the microblaze, i.e. by using the Vivado SDK
                port: integer with the port number
                lane: 0-9 OLT to be selected
            ***************************************************************************
        """			
        OltExdsg.__init__(self, ip, port)
        self.lane_offset = 128*lane
        self.addr_offset = 16384

    def write_ctrl_reg(self, reg_addr, reg_value):
        """
            ***************************************************************************
            Write to FPGA with an offset
            ***************************************************************************

            ***************************************************************************
            arg: reg_addr and reg_value are, respectively, the address register and value to be written to the interface
            return: 1 if success
            ***************************************************************************
        """		
        return OltExdsg.write_ctrl_reg(self, reg_addr + self.lane_offset, reg_value)

    def read_ctrl_reg(self, reg_addr):
        """
            ***************************************************************************
            Read from FPGA with an offset
            ***************************************************************************

            ***************************************************************************
            arg: reg_addr is the address register to be written to the interface
            return: register value
            ***************************************************************************
        """			
        return OltExdsg.read_ctrl_reg(self, reg_addr + self.lane_offset)

    def exd_clear_sticky(self):
        """
            ***************************************************************************
            Clear sticky status implemented in the OLT user logic of the ONU_MLINKOLT design
            ***************************************************************************

            ***************************************************************************
            return: 1
            ***************************************************************************
        """			
        stat = self.read_ctrl_reg(self.addr_offset + 0)
        stat = stat | 2
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        stat = stat ^ 2
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        return 1

    def exd_read_olt_status_sticky(self):
        """
            ***************************************************************************
            Read sticky status implemented in the OLT user logic of the ONU_MLINKOLT design
            ***************************************************************************

            ***************************************************************************
            return: [mgt_pll_lock, mgt_tx_ready, mgt_rx_ready, phase_good_onu2olt]
            ***************************************************************************
        """			
        stat = self.read_ctrl_reg(self.addr_offset + 3)
        mgt_pll_lock       = (stat>>16)&1
        mgt_tx_ready       = (stat>>17)&1
        mgt_rx_ready       = (stat>>18)&1
        phase_good_onu2olt = (stat>>19)&1
        return [mgt_pll_lock, mgt_tx_ready, mgt_rx_ready, phase_good_onu2olt]

    def exd_tx_cascaded_forward_data(self, forward=1):
        """
            ***************************************************************************
            Selects OLT in cascaded mode to forward data received from ONU
            ***************************************************************************

            ***************************************************************************
            arg: forward = 1, data is forwarded. 0 data comes from local generator
            return: 1
            ***************************************************************************
        """			
        stat = self.read_ctrl_reg(self.addr_offset + 2)
        if(forward) : stat = stat | 4
        else:         stat = stat & 0xFFFB
        self.write_ctrl_reg(self.addr_offset + 2, stat)
        return 1

class OnuExdsgAlice(OnuExdsg):
  def __init__(self, ip='192.168.1.11', port=9001, onun= None, olt2ndlevel=None):
    OnuExdsg.__init__(self, ip, port, onun, olt2ndlevel)
    self.notvalid= False
    self.mux= None    # CRU
    if self.ctrlcon==None: return
    #print "OnuExdsgAlice ip:%s %s self.addr_offset: 0x%x"%\
    #  (ip, self.conver, self.addr_offset)
    if self.conver=="ublaze":
      self.mux=(I2C_FMC_SWITCH_SLAVE_ADDRESS,I2C_FMC_SEL_SFP)   # FMC switch
    elif self.conver=="ipbus":
      self.mux= (0x76,0, 0x74,1<<3) # fixed, only 1 ONU (SFP2) on each alice board
    else:
      pass
    #self.Asel_tx_pattern= 2 + self.addr_offset
  def getura(self, addr):
    return(addr+self.addr_offset)
    #def whoami(self):
    #if (self.onun==3) or (self.onun==6) or (self.onun==7):
    #  return "cru"
    #elif self.onun>=10:
    #  return "atb"
    #else:
    #  return "kc"
    return self.conver
  def onureset(self, wmsk=0x30):   # 0x20: rxReady   0x10: txReady
    if self.conver=="cru":
      print("ONU_%d reset not done (cru)"%self.onun)
    else:
      print("ONU_%d reset..."%self.onun)
      self.exd_core_reset()
      # Eduardo's note:
      # exd_core_reset() is the MGT reset which disables Tx optical transceiver for a while
      nloops= 0 ; slt= 0.5
      while nloops<20:
        Aonu_rx40_locked= self.getura(3)
        val= self.read_ctrl_reg(Aonu_rx40_locked)
        if (val & wmsk) == wmsk:
          print("  onureset ok, (%.2f s): waitmask: 0x%x status:0x%x"%(slt*nloops, wmsk, val))  
          return
        time.sleep(slt)
        nloops= nloops+1
    print("  onureset timeout (%.2f s): waitmask: 0x%x status:0x%x"%(slt*nloops, wmsk, val))  
  def pllreset(self):
    if self.conver=="cru":
      print("ONU_%d pllreset not done (cru)"%self.onun)
    elif self.monionu:
      print("ONU_%d pllreset not done (monitoring onu)"%self.onun)
    else:
      print("ONU_%d pllreset..."%self.onun)
      if self.ctrlcon.plldev:
        self.ctrlcon.plldev.hwreset()
        time.sleep(1)
        self.ctrlcon.plldev.load()
  def onuseta(self):
    # set ONU number
    self.write_ctrl_reg(self.addr_offset, self.onun*2)
  def getonustatus(self, rc=None):
    if rc==None:
      Aonu_rx40_locked= self.getura(3)
      val= self.read_ctrl_reg(Aonu_rx40_locked)
      return val
    else:
      return """RTrtOLP4  -8 bits of ONUuser[3]
       ONU_RX40_LOCKED
      ONU_PHASE_GOOD
     ONU_RX_LOCKED
    ONU_OPERATIONAL
   ONU_MGT_TX_READY
  ONU_MGT_RX_READY
 ONU_MGT_TXPLL_LOCK
ONU_MGT_RXPLL_LOCK
"""

  def onuinit_delme(self):   # not used -> see onu_init in ttcpon_board_config.init_onu
    print("onuinit: %d"%self.onun)
    onu_addr= self.onun
    # Program FMC devices
    #program_clkmux(onu, 0x37)
    #program_si570(onu, 1)
    #program_si5344(onu, 1)
    #
    # Wait some time to PLL output to be stable
    time.sleep(0.6)
    self.exd_core_reset()
    time.sleep(1.0)
    #
    # Program ONU upstream TDM address
    self.exd_set_addr(onu_addr)
    #
    # Check ONU status before doing phase alignment - ONU must be locked
    status_onu = self.exd_read_onu_status()
    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
    timeout_cntr = 0
    while((not (mgt_rx_ready and mgt_rxpll_lock and rx40_locked and rx_locked and mgt_txpll_lock and mgt_tx_ready)) and timeout_cntr< c_MAX_TIMEOUT_MGT_READY):
        status_onu = self.exd_read_onu_status()
        [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.1)
        if(timeout_cntr >= c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("ONU is not getting ready after timeout, check ONU status!")		
            return 0
    #
    for i in range(0,1):	
        logger_pon.info("Phase adjustment nb. "+str(i))
        #stat = dbg_automatic_phase_good(onu,i)
        stat = self.automatic_phase_good(onu)	
        #onu.exd_core_reset()
        #time.sleep(1.0)
    #
    #rogram_leds(onu, 0x00)
    #monitor_sfp(onu,1)
    #
    return stat
  def seti2cmainttcmux(self, val):
      # val: 1..11 corresponding to olt1..9,onu,ipbus
    if self.conver == "ipbus":
      was= self.ctrlcon.board.Read("i2cmainttc.mux")
      print("seti2cmainttcmux: was %d -> %s on onuaddr:%d"%(was,val,self.onun))
      #self.ctrlcon.board.Write("i2cmainttc.mux", val),  better:
      self.ctrlcon.board.i2cg.seti2cttcmux(val)
    else:
      print("Warning seti2cmainttcmux: not ipbus connection:", self.conver)
  def setmux(self, stdout=False):  
    if self.conver=="ipbus":
      self.seti2cmainttcmux("onu")   # 10
    if self.mux: 
      for ix in range(0,len(self.mux),2):
        ma= self.mux[ix] ; val= self.mux[ix+1]
        # like in ttcpon_board_config/pon_fmc_i2c_program.py 
        if stdout: print("onu setmux 0x%x: 0x%x"%(ma, val))
        rc= self.i2c_write(ma, val, val)   # set mux
        if rc!= 1:
          print("setmux 0x%x:0x%x failed..."%(ma,val))
    #self.ctrlcon.board.Write("i2cmainttc.mux", 10)
  def getmux(self):
    """ test (not working in ttcpon v010: i2c implementation does not allow read)"""
    self.seti2cmainttcmux("onu")
    if self.mux:
      for ix in range(0,len(self.mux),2):
        ma= self.mux[ix] ; vale= self.mux[ix+1]
        val= self.i2c_read(ma, ma, 1)
        print("getmux 0x%x:  expected:0x%x got:"%(ma, vale), val)
  def setseltx(self, seltx):
    #Aseltx= 2 | self.uroffset
    #print "setseltx:    0x%x correct:0x%x"%(self.Asel_tx_pattern, 2 + self.addr_offset)
    Asel_tx_pattern= self.getura(2)
    val= self.read_ctrl_reg(Asel_tx_pattern)
    val= (val & 0xfffffffc) | seltx
    self.write_ctrl_reg(Asel_tx_pattern, val)
  def clrDownCounts(self):
    if self.conver == 'ipbus':
      print("ONU_%d bert/errmon clear..."%self.onun)
      self.exd_bert_clear()   # user regs
      self.errmon_clear()    # control regs
    else:
      print("ONU_%d bert/errmon clear NOT DONE (cru)"%self.onun)
  def bertLatch(self):
    #print "ONU%d BERT latch..."%self.onun
    self.exd_bert_latch()
  def getDownCounts(self):
    if self.conver=="cru":   # not availbale for CRU-ONU
      print("getDownCounts: not availbale for CRU-ONU")
      rc=[0]*8
    else:
      self.bertLatch()
      Abs= 4 | self.addr_offset ; rc= []
      for ix in range(4):
        bs2= self.read_ctrl_reg(Abs+2*ix+1) & 0xffff
        bs1= self.read_ctrl_reg(Abs+2*ix) 
        rc.append(bs1); rc.append(bs2)
      #print "OnuCore.getDownCounts:", bs1, bs2
      onuC= self.errmon_read()   #fec_sing8 fec_doub8 sc_crc8
      rc= rc + onuC 
    return rc   # sum32,16 err32,16 fec_sing32,16 fec_doub32,16
  def excluded(self):
    return self.notvalid
  def exclude(self):
    self.notvalid= True
    if self.conver != "ipbus":
      print("Did you set Core reg0 (ONU_MODE) to 0?")
    else:
      onumode= self.read_ctrl_reg(0)
      self.write_ctrl_reg(0, 0)
      print("Core reg0 ONU_MODE set to 0 (excluded), was 0x%x"%(onumode))
  def include(self):
    self.notvalid= False
    if self.conver != "ipbus":
      print("Did you set Core reg0 (ONU_MODE) to 0xe9?")
    else:
      onumode= self.read_ctrl_reg(0)
      self.write_ctrl_reg(0, 0xe9)
      print("Core reg0 ONU_MODE set to 0xe9 (included), was 0x%x"%(onumode))
  def printRegs(self, wregs="cu", only1flags=True):
    if self.conver == "cru":
      print("ONU not accessible (cru)")
      return
    onuc_regs= regnames.onuc_regs_fiv101
    onuu_regs= regnames.onuu_regs_fiv010
    if wregs.find('c')>=0:
      print("ONU_%d CORE:"%self.onun)
      regnames.printTable(self, onuc_regs, 0, 27, only1flags)
    if wregs.find('u')>=0:
      print("ONU_%d USER:"%self.onun)
      #self.printTable(oltu_regs, self.uroffset, 30)
      regnames.printTable(self, onuu_regs, self.addr_offset, 26, only1flags)
      if not only1flags:
        cs= self.getDownCounts()
        print("BITSUM: %d %d   ERRSUM:%d %d   FEC_S:%d %d   FEC_D:%d %d"%(cs[0], cs[1], cs[2], cs[3], cs[4], cs[5], cs[6], cs[7]))
  def cfg_upstream(self):
    if self.conver=="cru":
      print("ONU_%d cfg_upstream not done (cru)"%self.onun)
    else:
      self.onureset()
      self.setseltx(0)   # 0:PRBS7
      # ONU status
      #print("ONU_%d MODE 0x%x:"%(self.onun,self.read_ctrl_reg(0)))
      #print("HEARTBEAT_PERIOD-1: 0x%x"%self.read_ctrl_reg(1)) 
      #print "SEL_TX_PATTERN:0x%x"%self.read_ctrl_reg(self.getura(2))
  def txpa(self, action="m", newval=None):
    self.setmux()
    if action=='g':
      prev_phase= self.get_tx_phase_aligner_status()
      print("ONU_%d get_tx_phase_aligner:"%self.onun, prev_phase)
      return prev_phase
    elif action=='m':
      prev_phase= self.get_tx_phase_aligner_status()
      self.set_tx_phase_aligner(0, 0)
      print("ONU_%d set_tx_phase_aligner(0, 0) done, was:"%self.onun, prev_phase)
      self.onureset(0xb0)   # rxLocked rxReady txReady
      phase= self.get_tx_phase_aligner_status()
      print("ONU_%d get_tx_phase_aligner_status:"%self.onun, phase)
      return phase
    elif action=='s':
      print("ONU_%d set_tx_phase_aligner("%self.onun,newval,", 1)...")
      self.set_tx_phase_aligner(newval, 1)

  def printRegs_v4(self):
    if self.conver == "cru":
      print("ONU not accessible (cru)")
      return
    mode= self.read_ctrl_reg(0)
    hb= self.read_ctrl_reg(1)
    print("ONU%d CORE:MODE[7..0]:%d  HB_PERIOD[31-16]:%d  HB_PERIOD[15-0]:%d"%\
      (self.onun, mode, hb>>16, hb&0xffff))
    sfp= self.read_ctrl_reg(2)
    if self.conver == "ublaze":
      print("     SFP_ENA_DEL:%d  FINE__OFFSET_TDM:%d"%\
        (sfp&0xf, (sfp>>8)&0x1f))
    elif self.conver == "ipbus":   # fy v4 (1 reg. more)
      drpaccess= self.read_ctrl_reg(0xe)
      print("     SFP_ENA_DEL:%d  FINE__OFFSET_TDM:%d DRPACCESS:0x%x"%\
        (sfp&0xf, (sfp>>8)&0x1f, drpaccess&0x7ffffff))
    else:
      print("INTERROR: self.conver:", self.conver)
    print("USER:")
    add= (self.read_ctrl_reg(0+self.addr_offset)>>1)&0xff
    cl= self.read_ctrl_reg(1+self.addr_offset)
    print("ADD:%d  BERT_CLEAR 1[0]:%d  BERT_LATCH 1[1]:%d"%\
      (add, cl&1, (cl>>1)&1))
    tx= self.read_ctrl_reg(2+self.addr_offset)
    txp= tx&0x3; txps=("PRBS7","PRBS23","CONST","BYTECNT")
    print("SEL_TX_PAT 2[1-0]:%d=%s TX_INT_MODE 2[2]:%d  TX_INJECT_ERR 2[3]:%d"%\
      (txp,txps[txp], (tx>>2)&1, (tx>>3)&1))
    mg= self.read_ctrl_reg(3+self.addr_offset)&0xff
    ur3bits=("RX40LOCKED","PHASEGOOD","RXLOCKED","OPER","TXRDY","RXRDY",
      "TXPLLLOCK", "RXPLLLOCK")
    mgs=""
    for ix in range(7,-1,-1):
      if mg & (1<<ix): mgs= mgs+" "+ur3bits[ix]
    print("3[7-0]:0x%x=%s"%(mg, mgs))
    bs1,bs2,es1,es2= self.getDownCounts()
    print("BITSUM: %d %d   ERRSUM:%d %d"%(bs1,bs2,es1,es2))
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

class OnuMlinkExdsg(OnuExdsgAlice):

    """
        Class to control the ONU in the internal ALICE + TTC-PON team design where 1 ONU + 10 OLTs are implemented in the same FPGA
    """

    def __init__(self, ip='192.168.1.2', port=9001, onun=None, olt2ndlevel=None):
        """
            ***************************************************************************
            Class Constructor
            ***************************************************************************

            ***************************************************************************
            arg:
                ip: string with the IP address of the board being controlled.
                    The board IP in the TTC-PON example design can be changed via the software implemented in the microblaze, i.e. by using the Vivado SDK
                port: integer with the port number
            ***************************************************************************
        """			
        OnuExdsgAlice.__init__(self, ip, port, onun, olt2ndlevel)
        self.lane_offset = 128*10
        self.addr_offset = 16384

    def write_ctrl_reg(self, reg_addr, reg_value):
        """
            ***************************************************************************
            Write to FPGA with an offset
            ***************************************************************************

            ***************************************************************************
            arg: reg_addr and reg_value are, respectively, the address register and value to be written to the interface
            return: 1 if success
            ***************************************************************************
        """			
        return OnuExdsg.write_ctrl_reg(self, reg_addr + self.lane_offset, reg_value)

    def read_ctrl_reg(self, reg_addr):
        """
            ***************************************************************************
            Read from FPGA with an offset
            ***************************************************************************

            ***************************************************************************
            arg: reg_addr is the address register to be written to the interface
            return: register value
            ***************************************************************************
        """		
        return OnuExdsg.read_ctrl_reg(self, reg_addr + self.lane_offset)

    def exd_freeze_cdr_test_mode(self, enable=0):
        """
            ***************************************************************************
            Freeze CDR for LTU standalone test-mode
            It is necessary to program external PLL with free-running mode or locked to external clock (BC in SMA) before using this method
            ***************************************************************************

            ***************************************************************************
            arg: enable = 1/0			
            ***************************************************************************
        """
        rxpllclksel = 0b10 # Selects QPLL1 as clock source for ONU Rx CDR clock    (this PLL is used by an OLT in the same channel)
        rxsysclksel = 0b11 # Selects QPLL1 as clock source for ONU Rx system clock (this PLL is used by an OLT in the same channel)
        rxoutclksel = 0b10 # Selects rxoutclkpma reference clock as clock for the rxoutclksel
        rxcdrhold   = 0b1  # Hold ONU CDR

        stat = self.read_ctrl_reg(18) & 0x0001FFFF
        self.write_ctrl_reg(18, (0<<31) + (enable<<17) + (rxpllclksel<<18) + (rxsysclksel<<20) + (rxoutclksel<<22) + (rxcdrhold<<25) + stat)
        self.write_ctrl_reg(18, (1<<31) + (enable<<17) + (rxpllclksel<<18) + (rxsysclksel<<20) + (rxoutclksel<<22) + (rxcdrhold<<25) + stat)

    def exd_phase_offset40(self, offset_value):
        """
            ***************************************************************************
            Provides Phase Offset for 40MHz ONU clock
            ***************************************************************************

            ***************************************************************************
            arg: shift 40MHz ONU clock with phase steps equal to the MMCM phase shift value	
            ***************************************************************************
        """
        stat = self.read_ctrl_reg(18) & 0x0001FFFF    # RX_EQUALIZER_CONTROL core reg
        self.write_ctrl_reg(18, (0<<31) + (offset_value<<18) + stat)
        self.write_ctrl_reg(18, (1<<31) + (offset_value<<18) + stat)
        self.exd_core_reset()
        time.sleep(1)

class OnuMlinkExdsgAlice(OnuMlinkExdsg):
  def __init__(self, ip='192.168.1.2', port=9001, onun=None, mononu= False, olt2ndlevel=None, logger=None):
    """ olt2ndlevel:
None: ONU in an ATB board used in CRU role, or
An instance of (can be without physical OLT plugged in?) OltMlinkExdsgAlice
"""
    OnuMlinkExdsg.__init__(self, ip, port,onun, olt2ndlevel)
    self.logger= logger
    self.monionu= mononu
    self.olt2ndlevel= olt2ndlevel
  def equalizer_adapt(self):
    oad= self.onun
    if oad==-1:
      oadtx="fromctp"
      print("Error calling equalizer_adapt with 'fromctp' onu, should never happen")
    else:
      oadtx="_%d"%oad
    self.setmux() 
    try:
      lem= self.get_lpm_equalizer_monitor()
    except:
      lem= None
      print("ONU_%s error (not plugged in or burned?)"%oadtx)
    print("ONU%s equalizer before adaptation:"%oadtx,lem)
    self.rx_equalizer_force(1, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    self.onureset() # reset hangs without ONU inserted
    [RXLPMHF, RXLPMLF, RXLPMOS, RXLPMAGC]= self.get_lpm_equalizer_monitor()
    print("  new setting: [RXLPMHF, RXLPMLF, RXLPMOS, RXLPMAGC]:",RXLPMHF, RXLPMLF, RXLPMOS, RXLPMAGC)
    self.set_lpm_equalizer_parameters(RXLPMHF, RXLPMLF, RXLPMOS, RXLPMAGC)
    self.rx_equalizer_force(1, 1, 1, 1, 1, 0, 0, 0, 0, 0)
    #onu.setmux() 
    self.onureset()


