#!/usr/libexec/platform-python

import csv
import datetime
import math
import time
import logging
from random import randint

from . pon_fmc_i2c_program import program_clkmux
from . pon_fmc_i2c_program import program_si570
from . pon_fmc_i2c_program import program_si5344
from . pon_fmc_i2c_program import monitor_sfp
from . pon_fmc_i2c_program import program_leds
from . pon_fmc_i2c_program import set_phase_si5344
from . pon_fmc_i2c_program import get_status_si5344 as get_status_pll

logger_pon = logging.getLogger('ttc_pon')

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

c_Mod_ONU_Tx_Period = 114
c_Mod_ONU_Rx_Period = 57

#!/usr/bin/python3
c_MAX_ERR_PLL = 1

#!/usr/bin/python3
c_MAX_TIMEOUT_MGT_READY = 5

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
def set_phase_pll(onu, pll_output, skew):

    set_phase_si5344(onu,pll_output, skew)

    # Max lock time according to datasheet
    time.sleep(0.5)

    # Here we should have only a TX reset (but it is not yet implemented)
    onu.exd_core_reset()
    time.sleep(0.5)

    return 1

def check_onu_locked(onu):
    # Check ONU status before doing phase alignment
    # ONU must be locked in the downstream direction
    status_onu = onu.exd_read_onu_status()
    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
    if(mgt_rxpll_lock==0):
        logger_pon.warn("ONU MGT RX_PLL not locked!")
        return [0,0]
    elif(mgt_rx_ready==0):
        logger_pon.warn("ONU MGT RX not ready!")
        return [0,0]
    elif(rx_locked==0):
        logger_pon.warn("ONU downstream not locked!")
        return [0,0]

    # Check PLL is locked
    pll_locked = 0
    while(pll_locked==0):
        pll_status = get_status_pll(onu)
        if(pll_status[0]==0):
            logger_pon.warn("IIC communication problem during get status of PLL")	
        pll_locked = not (pll_status[2])
        pll_loss_of_signal = pll_status[1]
        if(pll_loss_of_signal):
            logger_pon.warn("ONU PLL lost signal, there might be an issue in the downstream! Check problem!")			
            return [0, 0]

    # If the function arrive here the PLL must be already locked
    # Now we should check that the MGT TX_PLL is locked and the MGT_TX is ready
    timeout_cntr   = 0
    while(not (mgt_txpll_lock and mgt_tx_ready)):
        status_onu = onu.exd_read_onu_status()
        [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.1)
        if(timeout_cntr > c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("TX is not getting ready after timeout! Check problem!")						
            return [0, 0]

    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu

    return [1, phase_good]

def onu_init(onu, onu_addr=None):
    if onu_addr==None: onu_addr= onu.onun
    # Program FMC devices
    program_clkmux(onu, 0x37)
    program_si570(onu, 1)
    program_si5344(onu, 1)

    # Wait some time to PLL output to be stable
    time.sleep(0.6)
    onu.exd_core_reset()
    time.sleep(1.0)

    # Program ONU upstream TDM address
    onu.exd_set_addr(onu_addr)

    # Check ONU status before doing phase alignment - ONU must be locked
    status_onu = onu.exd_read_onu_status()
    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
    timeout_cntr = 0
    while((not (mgt_rx_ready and mgt_rxpll_lock and rx40_locked and rx_locked and mgt_txpll_lock and mgt_tx_ready)) and timeout_cntr< c_MAX_TIMEOUT_MGT_READY):
        status_onu = onu.exd_read_onu_status()
        [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.1)
        if(timeout_cntr >= c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("ONU is not getting ready after timeout, check ONU status!")		
            return 0

    for i in range(0,1):	
        logger_pon.info("Phase adjustment nb. "+str(i))
        #stat = dbg_automatic_phase_good(onu,i)
        stat = automatic_phase_good(onu)	
        #onu.exd_core_reset()
        #time.sleep(1.0)

    program_leds(onu, 0x00)
    monitor_sfp(onu,1)

    return stat

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
def dbg_automatic_phase_good(onu,index):

    phase_save = []
    phase_good_save = []

    for phase in range(0,127):
        # Set PLL phase
        stat = set_phase_pll(onu,1,phase)
        if(stat==0):
            logger_pon.warn("IIC communication problem during set phase of PLL")
            return 0

        [onu_ok, phase_good] = check_onu_locked(onu)
        if (onu_ok==0): return 0

        phase_save.append(phase)
        phase_good_save.append(phase_good)

        # dbg
        # print(str(phase) +" / "+str(phase_good))

    # Save file
    name_file = 'test_phase_pll/phase_swepp_onu' + str(onu.exd_get_addr()) + '_' + str(index)    
    with open(name_file + '.csv', 'w') as csvfile:
        fieldnames = ['phase', 'phase_good']
        for i in range(0, len(fieldnames)):
            # format 20 spaces for each value
            fieldnames[i] = '%20s' % fieldnames[i]
        writer = csv.DictWriter(
            csvfile,
            fieldnames=fieldnames,
            delimiter=',',
            lineterminator='\n')
        writer.writeheader()

        for j in range(0, len(phase_save)):
            dict_write = {	
                fieldnames[0] : ("%20d" % phase_save[j]),
                fieldnames[1] : ("%20d" % phase_good_save[j])}			
            writer.writerow(dict_write)

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
def automatic_phase_good(onu):

    # Binary search based optimal phase finding
    phase_min = int(round(c_Mod_ONU_Rx_Period-c_Mod_ONU_Rx_Period/8))
    phase_max = int(round(c_Mod_ONU_Rx_Period+c_Mod_ONU_Rx_Period/8))

    # Set PLL phase
    stat = set_phase_pll(onu,1,phase_min)
    if(stat==0):
        logger_pon.warn("IIC communication problem during set phase of PLL")
        return 0

    [onu_ok, phase_good_min] = check_onu_locked(onu)
    if (onu_ok==0): return 0

    # Set PLL phase
    stat = set_phase_pll(onu,1,phase_max)
    if(stat==0):
        logger_pon.warn("IIC communication problem during set phase of PLL")
        return 0

    [onu_ok, phase_good_max] = check_onu_locked(onu)
    if (onu_ok==0): return 0

    # dbg
    if(phase_good_max==phase_good_min):
        phase_min = phase_max
        phase_max = c_Mod_ONU_Rx_Period*2
    else:
        phase_min = int(round(c_Mod_ONU_Rx_Period-2*c_Mod_ONU_Rx_Period/8))
        phase_max = int(round(c_Mod_ONU_Rx_Period+2*c_Mod_ONU_Rx_Period/8))
        # Set PLL phase
        stat = set_phase_pll(onu,1,phase_min)
        if(stat==0):
            logger_pon.warn("IIC communication problem during set phase of PLL")
            return 0

        [onu_ok, phase_good_min] = check_onu_locked(onu)
        if (onu_ok==0): return 0
        phase_good_max = 1-phase_good_min

    phase_actual = int(round((phase_min+phase_max)/2))
    err = phase_max - phase_min

	# Find phase good edge using binary search
    while(err > c_MAX_ERR_PLL):
        stat = set_phase_pll(onu,1,phase_actual)
        if(stat==0):
            logger_pon.warn("IIC communication problem during set phase of PLL")
            return 0
	
        [onu_ok, phase_good_actual] = check_onu_locked(onu)
        if (onu_ok==0): return 0
        if(phase_good_actual == phase_good_min):
            phase_min = phase_actual		
        else:
            phase_max = phase_actual

        # dbg
        #print(str(phase_actual) +" / "+str(phase_good_actual))

        phase_actual = int(round((phase_min+phase_max)/2))
        err = phase_max - phase_min

	# Adjust phase good
    if(phase_good_min==0):
        phase_actual = (int(round(phase_actual+c_Mod_ONU_Rx_Period/2))) % c_Mod_ONU_Tx_Period
    else:
        phase_actual = (int(round(phase_actual-c_Mod_ONU_Rx_Period/2))) % c_Mod_ONU_Tx_Period

    stat = set_phase_pll(onu,1,phase_actual)
    if(stat==0):
        logger_pon.warn("IIC communication problem during set phase of PLL")
        return 0

    [onu_ok, phase_good_actual] = check_onu_locked(onu)

    logger_pon.info("Phase adjusted at:" + str(phase_actual))
    return 1


