#!/usr/libexec/platform-python

import csv
import datetime
import math
import time
import logging
from random import randint

from . pon_fmc_i2c_program import program_clkmux
from . pon_fmc_i2c_program import program_si570
from . pon_fmc_i2c_program import program_si5344
from . pon_fmc_i2c_program import monitor_sfp
from . pon_fmc_i2c_program import program_leds
from . pon_fmc_i2c_program import set_phase_si5344
from . pon_fmc_i2c_program import get_status_si5344 as get_status_pll

logger_pon = logging.getLogger('ttc_pon')

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

#!/usr/bin/python3
#!/usr/bin/python3
c_STEP_PHASE = 2

#!/usr/bin/python3
c_MAX_TIMEOUT_MGT_READY = 5

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
def set_phase_pll(onu, olt, pll_output, skew):
    if onu.conver=="ublaze":
      set_phase_si5344(onu,pll_output, skew)
    else:
      onu.ctrlcon.plldev.Sphase(skew, 2)   # OUT6 240MHz N2 divider
    # Max lock time according to datasheet
    time.sleep(0.5)

    # Here we should have only a TX reset for ONU (but it is not yet implemented)
    onu.exd_core_reset()
    olt.exd_core_reset()
    time.sleep(0.5)

    return 1

def check_onu_mlinkolt_locked(onu, olt):
    # Check ONU status before doing phase alignment
    # ONU must be locked in the downstream direction
    status_onu = onu.exd_read_onu_status()
    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
    if(mgt_rxpll_lock==0):
        logger_pon.warn("ONU MGT RX_PLL not locked!")
        return [0,0]
    elif(mgt_rx_ready==0):
        logger_pon.warn("ONU MGT RX not ready!")
        return [0,0]
    elif(rx_locked==0):
        logger_pon.warn("ONU downstream not locked!")
        return [0,0]

    # Check PLL is locked
    pll_locked = 0
    while(pll_locked==0):
        if onu.conver=="ublaze":
          pll_status = get_status_pll(onu)
        else:
          pll_status = onu.ctrlcon.plldev.get_status()
        if(pll_status[0]==0):
            logger_pon.warn("IIC communication problem during get status of PLL")	
        pll_locked = not (pll_status[2])
        pll_loss_of_signal = pll_status[1]
        if(pll_loss_of_signal):
            logger_pon.warn("ONU PLL lost signal, there might be an issue in the downstream! Check problem!")			
            return [0, 0]

    # If the function arrive here the PLL must be already locked
    # Now we should check that the MGT TX_PLL is locked and the MGT_TX is ready
    timeout_cntr   = 0
    while(not (mgt_txpll_lock and mgt_tx_ready)):
        status_onu = onu.exd_read_onu_status()
        [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.1)
        if(timeout_cntr > c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("ONU TX is not getting ready after timeout! Check problem!")						
            return [0, 0]

    # If the function arrive here the PLL must be already locked
    # Now we should check that the OLT Tx/Rx is ready and OLT Tx PLL is locked
    olt.exd_clear_sticky()
    time.sleep(0.1)	
    status_olt = olt.exd_read_olt_status_sticky()
    [mgt_pll_lock, mgt_tx_ready, mgt_rx_ready, phase_good_onu2olt] = status_olt
    timeout_cntr   = 0
    while(not (mgt_pll_lock and mgt_tx_ready and mgt_rx_ready)):
        olt.exd_clear_sticky()
        time.sleep(0.1)		
        status_olt = olt.exd_read_olt_status_sticky()
        [mgt_pll_lock, mgt_tx_ready, mgt_rx_ready, phase_good_onu2olt] = status_olt
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.1)
        if(timeout_cntr > c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("OLT Tx is not getting ready after timeout! Check problem!")						
            return [0, 0]

    return [1, phase_good_onu2olt]

#!/usr/bin/python3
def onu2olt_mode(onu, olt, mode):
    if(mode==1):   # Cascaded
        program_clkmux(onu, 0x36)
        onu.exd_core_reset()
        time.sleep(1)
        olt.exd_core_reset()
        time.sleep(1)		
    elif(mode==0): # Standalone
        program_clkmux(onu, 0x37)
        olt.exd_core_reset()
        time.sleep(1)
        onu.exd_core_reset()
        time.sleep(1)			
    else:
        logger_pon.warn("Invalid Configuration")
        return 0

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
def automatic_phase_good_onu2olt(onu, olt):
    found_phase         = 0
    found_ris_edge      = 0
    phase               = 0
    first_iter          = 1
    previous_phase_good_onu2olt = 0
    logger_pon.info('PHASE'+ ' --- ' + 'GOOD?')	
    while(not found_phase):
        # Set PLL phase
        stat = set_phase_pll(onu, olt,2,phase)
        if(stat==0):
            logger_pon.warn("IIC communication problem during set phase of PLL")
            return 0

        # Check ONU and OLT are properly locked
        [onu_mlinkolt_ok, phase_good_onu2olt] = check_onu_mlinkolt_locked(onu, olt)
        if (onu_mlinkolt_ok==0): return 0

        # Algorithm to stop iteration
        # Finds rising edge of phase_good signal
        # Then finds falling edge
        # Then it positions in the middle		
        if(not first_iter):
            if(previous_phase_good_onu2olt!=phase_good_onu2olt):
                if(phase_good_onu2olt==1): # Found a rising edge
                    found_ris_edge = 1
                    phase_ris_edge = phase
                    previous_phase_good_onu2olt = phase_good_onu2olt					
                elif(found_ris_edge):      # Found a falling edge and already found rising edge
                    # Position phase in the middle of ris->fall window				
                    phase = (phase + phase_ris_edge)//2
                    # Set PLL phase
                    stat = set_phase_pll(onu, olt,2,phase)
                    if(stat==0):
                        logger_pon.warn("IIC communication problem during set phase of PLL")
                        return 0
                    # Check status
                    [onu_mlinkolt_ok, phase_good_onu2olt] = check_onu_mlinkolt_locked(onu, olt)
                    if (onu_mlinkolt_ok==0): return 0					
                    logger_pon.info("Phase adjusted at:" + str(phase))
                    found_phase = 1
                else:
                    previous_phase_good_onu2olt = phase_good_onu2olt
            else:
                previous_phase_good_onu2olt = phase_good_onu2olt			
        else:
            previous_phase_good_onu2olt = phase_good_onu2olt

        first_iter = 0

        str_print = "%5d --- %5d" % (phase, phase_good_onu2olt)
        logger_pon.info(str_print)

        phase = phase + c_STEP_PHASE

        if(phase>127):
            logger_pon.warn("Not found phase good for ONU2OLT CDC")		
            return 0

    return 1
