#!/usr/libexec/platform-python

import os,string
import socket
import logging
import datetime
import time
import csv
import math
from random import randint

from . driver_comm import DriverComm

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3


class OltCore(DriverComm):

    def __init__(self, ip='192.168.1.2', port=9001):
        DriverComm.__init__(self, ip, port, olt=True)

        ### Do not change the parameters below (they might be firmware dependent) ###
        # SFP PARAMETERS		
        self.c_OLT_SFP_PART_NUMBER = 'NOT_INITIALIZED'

        # If this power is detected during calibration, an emergency stop is foreseen and the system should be debugged manually
        # Not yet implemented
        self.c_OLT_RX_PWR_EMERGENCY_STOP = -10 # unit: dBm

        # General
        self.c_BURST_LENGTH_ONURXCLK = 30 # 125 / 4.16
        self.c_BURST_LENGTH_ONUTXCLK = 15 # 125 / 8.33
        self.c_BURST_LENGTH_OLTRXCLK = 15 # 125 / 8.33
        self.c_BURST_LENGTH_OLTTXCLK = 30 # 125 / 4.16
        self.c_BURST_LENGTH_BUNCHCLK = 5  # 125 / 25
        self.c_NUMBER_OLT_REGS = 16

        # Slow Control Parameters
        # Broadcast message to ONU
        self.c_BROADCAST_ADDR = 255
        self.c_OPE_WR = 15
        self.c_OPE_WRACK = 14
        self.c_OPE_RD = 1
        self.c_OPE_RCVD_ERROR = 8

        # Operations c_OPE_WRACK and c_OPE_RD
        self.c_MAX_TRIALS_ONU_ACCESS = 5           # Max Number of trials for read / wrack operations
        self.c_MAX_WAITTIME_PER_ONU_ACCESS = 0.001 # unit: seconds 

        # ONU modes
        self.c_ONU_MODE_DISABLE = 0
        self.c_ONU_MODE_OPERATIONAL = 233
        self.c_ONU_MODE_CALIBRATION = 153
        self.c_ONU_MODE_PRBSBURST = 26
        self.c_ONU_MODE_PRBSCONT = 15

        # ONU detect parameters
        self.c_MAX_TRIALS_ONU_DETECT = 3
        self.c_CONSECUTIVE_SUCCESS_DETECT = 3
        self.c_IDLE_BYTE_ADDR = 31

        # Roundtrip measurement parameters
        self.c_MAX_TRIALS_ONU_MEAS_POS = 5
        self.c_CONSECUTIVE_SUCCESS_MEASPOS = 3
        self.c_RANGE_ACCEPT_MEASPOS = 8
        self.c_FINE_ROUNDTRIP_UNIT = 1    # 1 = 0.104 ns
        self.c_FINE_OFFSET_UNIT = 4       # 0.416 ns
        self.c_COARSE_ROUNDTRIP_UNIT = 80 # 8.33 ns 
        self.c_COARSE_OFFSET_UNIT = 80    # 8.33 ns

        # SFP RX reset calibration parameter
        self.c_MARGIN_SFP_RX_RESET_GFOT = 6 # DEVICE SPECIFIC
        self.c_MARGIN_SFP_RX_RESET = 2      # DEVICE SPECIFIC
        self.c_MASK_NVALID_SFP_RX_RESET = 5
        self.c_MASK_VALID_SFP_RX_RESET = 5

        # SFP EN delay calibration parameter
        self.c_MARGIN_SFP_EN_DELAY = 1
        self.c_MAX_SFP_EN_DELAY = 15

        # SFP SD EN MASK PARAMETERS
        self.c_MAX_SFP_SD_EN_MASK_DELAY = 15
        self.c_MAX_TIME_SFP_SD_COMMA_CORR_STAT = 0.010
        self.c_MIN_COMMA_CORR_STAT_VALID = 200
        self.c_MARGIN_SFP_SD_EN_MASK = 3

        # Missing burst parameters
        self.c_THRESHOLD_MISSING_BURST = 20 # Unit OLT_TX_CLK_PERIOD
		
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

    ### DRP CTRL (begin) ###
    def drp_write(self, addr, value):
        rdy=0	
        while(rdy==0):
            drpen=1
            drpwe=1
            value = (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + ((value&0x0000ffff)<<0)
            self.write_ctrl_reg(17,value)
            #value = (1<<31) + (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + ((value&0x0000ffff)<<0)
            #self.write_ctrl_reg(14,value)

            stat = self.read_ctrl_reg(17)

            value =  (stat&0x0000ffff) >> 0;
            addr  =  (stat&0x01ff0000) >> 16;
            rdy   =  (stat&0x04000000) >> 26;

        return rdy

    def drp_read(self, addr):
        rdy=0	
        while(rdy==0):
            drpen=1
            drpwe=0
            value = (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + 0
            self.write_ctrl_reg(17,value)			
            #value = (1<<31) + (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + 0
            #self.write_ctrl_reg(14,value)

            stat = self.read_ctrl_reg(17)

            value =  (stat&0x0000ffff) >> 0;
            addr  =  (stat&0x01ff0000) >> 16;	
            rdy   =  (stat&0x04000000) >> 26;
        return [addr, value, rdy]

    ### I2C CTRL (begin) ###
    def i2c_write(self, slv_addr, reg_addr, reg_value):
        if self.conver!="ublaze":
          #print "DBG olt.i2c_write 0x%x 0x%x val:0x%x"%(slv_addr, reg_addr, reg_value)
          pass
        slv_addr_format  = slv_addr & 0x7F
        reg_addr_format  = reg_addr & 0xFF
        reg_value_format = reg_value & 0xFF
        if((slv_addr != slv_addr_format) | (reg_addr != reg_addr_format) | (reg_value != reg_value_format)):
            self.logger.warn('Error parameters formatting for IIC write command')		
            return -1
        else:
            message = (reg_value<<18) + (reg_addr<<10) + (slv_addr<<3) + (0<<1) + (0<<1) + (1)	
            self.write_ctrl_reg(13, message)
            time.sleep(0.005)
        i2c_done = 0
        i2c_error = 0
        i2c_drop_req = 0
        while(i2c_done==0 and i2c_error==0 and i2c_drop_req==0):
            i2c_stat     = self.read_ctrl_reg(14)
            i2c_done     = (i2c_stat & 0x00000001) >> 0
            i2c_error    = (i2c_stat & 0x00000002) >> 1
            i2c_drop_req = (i2c_stat & 0x00000004) >> 2

        if(i2c_error):
            self.logger.warn('IIC write error')				
            return 0
        elif(i2c_drop_req):
            self.logger.warn('IIC drop request error')				
            return 0			
        else:
            return 1

    # Read IIC (nb_read_access can be 1 or 2 for double bursty read)
    def i2c_read(self, slv_addr, reg_addr, nb_read_access=1):
        slv_addr_format  = slv_addr & 0x7F;
        reg_addr_format  = reg_addr & 0xFF;
        nb_read_access   = nb_read_access - 1
        nb_read_access_format = nb_read_access & 0x1
        #print "DBG olt.i2c_read 0x%x 0x%x bytes:0x%x"%(slv_addr, reg_addr, nb_read_access)
        if((slv_addr != slv_addr_format) | (reg_addr != reg_addr_format) | (nb_read_access_format != nb_read_access)):
            self.logger.warn('Error parameters formatting for IIC read command')		
            return [-1]
        else:
            message = (0<<18) + (reg_addr<<10) + (slv_addr<<3) + (1<<2) + (nb_read_access<<1) + (1)	
            self.write_ctrl_reg(13, message)
            time.sleep(0.005)   # 5ms -> 10ms/20ms does not help
            # seems improves a bit result from  olt.get_sfp_part_number (see 'olt vendor' cmd)
        i2c_done = 0
        i2c_error = 0
        i2c_drop_req = 0
        reps= 0
        while(i2c_done==0 and i2c_error==0 and i2c_drop_req==0):
            i2c_stat  = self.read_ctrl_reg(14)
            i2c_done     = (i2c_stat & 0x00000001) >> 0
            i2c_error    = (i2c_stat & 0x00000002) >> 1
            i2c_drop_req = (i2c_stat & 0x00000004) >> 2
            i2c_data     = (i2c_stat & 0x0007FFF8) >> 3			
            if reps>10000:
              i2c_error=0xdeadbeaf
            reps= reps+1
        #print "i2c_read: %d stat:0x%x"%(reps, i2c_stat)
        if(i2c_error):
            self.logger.warn('IIC read error')				
            return [-1]
        elif(i2c_drop_req):
            self.logger.warn('IIC drop request error')
            return [-1]			
        else:
            if(nb_read_access > 0):		
                data_return_list = []        		
                for i in range(0,(nb_read_access+1)):
                    data  = (i2c_data >> i*8) & 0xFF
                    data_return_list.append(data)				
                return data_return_list
            else:
                return [i2c_data&0xFF]

    ### TDM arbitration (begin) ###
    # Configure system for number_onu onus
    def tdm_init(self, number_onu):
        self.write_ctrl_reg(0, int(math.ceil(number_onu * self.c_BURST_LENGTH_BUNCHCLK)))
        self.write_ctrl_reg(1, int(math.ceil(number_onu * self.c_BURST_LENGTH_OLTRXCLK)))
        return 1

    ### Internal Communication with ONU ###
    # Send message to ONU
    # ope       : self.c_OPE_RD, self.c_OPE_WR, self.c_OPE_WRACK
    # byte_addr : 0-511
    # byte_value: 0-255, byte
    # onu_addr  : 0-255 (0 is reserved - should not be used by any ONU, and 255 is broadcast)
    def msg2onu(self, ope, byte_addr, byte_value, onu_addr):
        msg_value = onu_addr + (byte_value << 8) + (byte_addr << 16) + (ope << 25)
        self.write_ctrl_reg(2, msg_value)
        #print "msg2onu a o ba bv:%d : %d %d 0x%x"%(onu_addr, ope, byte_addr, byte_value)
        return 1

    # It returns a list of four elements
    # [ope_rcvd,byte_addr_rcvd,byte_value_rcvd,onu_addr_rcvd)]
    def read_onu_msg(self):
        # Read message
        msg_rcvd = self.read_ctrl_reg(6)
        # Data formatting
        onu_addr_rcvd   = (msg_rcvd & 0x000000FF)
        byte_value_rcvd = (msg_rcvd & 0x0000FF00) >> 8
        byte_addr_rcvd  = (msg_rcvd & 0x01FF0000) >> 16
        ope_rcvd        = (msg_rcvd & 0x1E000000) >> 25
        # Return value
        msg_return = [ope_rcvd, byte_addr_rcvd, byte_value_rcvd, onu_addr_rcvd]
        return msg_return

    # robust wrack and rd methods
    def onu_wrack(self, byte_addr, byte_value, onu_addr):
        trial_cntr = 0
        success = 0		
        while(trial_cntr<self.c_MAX_TRIALS_ONU_ACCESS and success ==0):
            self.msg2onu(self.c_OPE_WRACK, byte_addr, byte_value, onu_addr)
            t_start = time.time()
            tdelta=0
            while(tdelta < self.c_MAX_WAITTIME_PER_ONU_ACCESS and success == 0):
                [ope_rcvd, byte_addr_rcvd, byte_value_rcvd, onu_addr_rcvd] = self.read_onu_msg()
                if(ope_rcvd==self.c_OPE_WRACK and byte_addr_rcvd==byte_addr and byte_value_rcvd==byte_value and onu_addr_rcvd==onu_addr):
                    success = 1
                else:
                    t_actual = time.time()
                    tdelta = t_actual - t_start
            trial_cntr = trial_cntr + 1
        return success

    def onu_rd(self, byte_addr, onu_addr):
        trial_cntr = 0
        success = 0		
        while(trial_cntr<self.c_MAX_TRIALS_ONU_ACCESS and success ==0):
            self.msg2onu(self.c_OPE_RD, byte_addr, 0, onu_addr)
            t_start = time.time()
            tdelta=0
            while(tdelta < self.c_MAX_WAITTIME_PER_ONU_ACCESS and success == 0):
                [ope_rcvd, byte_addr_rcvd, byte_value_rcvd, onu_addr_rcvd] = self.read_onu_msg()
                if(ope_rcvd==self.c_OPE_RD and byte_addr_rcvd==byte_addr and onu_addr_rcvd==onu_addr):
                    success = 1
                else:
                    t_actual = time.time()
                    tdelta = t_actual - t_start
            trial_cntr = trial_cntr + 1
        return [success, ope_rcvd, byte_addr_rcvd, byte_value_rcvd, onu_addr_rcvd]

    ### Calibration ###
    # Calibration configuration
    # calib_init : keep to '1' at the beggining of calibration (before taking measurements)
    # calib_on   : keep to '1' during calibration process
    # onu_addr   : address of ONU being calibrated
    def calib_config(self, calib_init, calib_on, onu_addr):
        msg_value = onu_addr + (calib_on << 8) + (calib_init << 9)
        self.write_ctrl_reg(3, msg_value)
        return 1

    # Take position measurement during calibration:
    # It returns a list of two elements [fine_pos , coarse_pos] - fine_pos
    # (0.104ns), coarse_pos(8.33ns)
    def calib_read_pos(self):
        # Latch position
        stat = self.read_ctrl_reg(3)
        stat = stat | (1 << 10)
        self.write_ctrl_reg(3, stat)
        stat = stat ^ (1 << 10)
        self.write_ctrl_reg(3, stat)
        # Read Position
        position = self.read_ctrl_reg(7)
        # Data formatting
        fine_pos = (position & 0x0000007F)
        coarse_pos = (position & 0x007FFF80) >> 7
        # Return value
        pos_return = [fine_pos, coarse_pos]
        return pos_return

    ### BERT ###
    # Fast BERT configuration
    def bert_config(self, onu1_addr, onu2_addr):
        msg_value = onu1_addr + (onu2_addr << 8)
        self.write_ctrl_reg(4, msg_value)
        return 1

    def bert_clear(self):
        stat = self.read_ctrl_reg(4)
        stat = stat | (1 << 16)
        self.write_ctrl_reg(4, stat)
        stat = stat ^ (1 << 16)
        self.write_ctrl_reg(4, stat)
        return 1

    # It returns a list of six elements
    # [onu1_finished,onu1_locked,onu1_errors,onu2_finished,onu2_locked,onu2_errors]
    def bert_read(self):
        # Latch BERT
        stat = self.read_ctrl_reg(4)
        stat = stat | (1 << 17)
        self.write_ctrl_reg(4, stat)
        stat = stat ^ (1 << 17)
        self.write_ctrl_reg(4, stat)
        # Read BERT status
        bert_error = self.read_ctrl_reg(8)
        bert_status = self.read_ctrl_reg(9)
        # Data formatting
        onu1_locked   = (bert_status & 0x00000001)
        onu1_finished = (bert_status & 0x00000004) >> 2
        onu1_errors   = (bert_error & 0x0000FFFF)
        onu2_locked   = (bert_status & 0x00000002) >> 1
        onu2_finished = (bert_status & 0x00000008) >> 3
        onu2_errors   = (bert_error & 0xFFFF0000) >> 16
        # Return value
        bert_return = [onu1_finished,onu1_locked,onu1_errors,onu2_finished,onu2_locked,onu2_errors]
        return bert_return

    ## PHY monitoring ###
    def phymon_clear(self):
        stat = self.read_ctrl_reg(10)
        stat = stat | (1<<2)
        self.write_ctrl_reg(10, stat)
        stat = stat ^ (1<<2)
        self.write_ctrl_reg(10, stat)	
        self.sfp_sd_clear()
        return 1

    # It returns a list of four elements
    # [seen_missing_burst, seen_mgt_tx_nready, seen_mgt_rx_nready, seen_mgt_pll_nlock]
    def phymon_read_seen(self):
        stat = self.read_ctrl_reg(10)
        seen_mgt_tx_nready      = (stat>>16)&0x1
        seen_mgt_rx_nready      = (stat>>17)&0x1
        seen_mgt_pll_nlock      = (stat>>18)&0x1

        stat = self.get_sfp_sd_stat()[1]
        if (stat > self.c_THRESHOLD_MISSING_BURST): seen_missing_burst = 1
        else : seen_missing_burst = 0
        # Return value
        mon_return = [seen_missing_burst, seen_mgt_tx_nready, seen_mgt_rx_nready, seen_mgt_pll_nlock]
        return mon_return

    ###### Error Monitor on-the-fly based on 8b10b error detection capability ######
    def errmon_clear(self):
        stat = self.read_ctrl_reg(15)
        stat = stat | 1
        self.write_ctrl_reg(15, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(15, stat)
        return 1

    # It returns a list of 1 + N elements where the 0th element is 0 or 1(if there was any error seen) and the N optional return values are the addresses of ONUs where an error was seen
    # Examples:	
    # [0]           - if no error was seen
    # [1 3]         - if an error was seen in ONU 3
    # [1 1 2 10 19] - if an error was seen in ONU 1,2,10,19
    def errmon_read_seen_error(self):
        stat = self.read_ctrl_reg(15)	
        error_seen_any = (stat >> 15) & 0x1
        return_list = [error_seen_any]		
        if (error_seen_any!=0) :
            for i in range(0,32):
                # Write page and latch
                stat = self.read_ctrl_reg(15)				
                stat = (stat & 0xFFFFFF83) | (i<<2 | 1<<1)		
                self.write_ctrl_reg(15,stat)    # set 8B10B_ERRDET_PAGESEL
                stat = stat ^ (1<<1)
                self.write_ctrl_reg(15,stat)    # latch
                stat = self.read_ctrl_reg(15)
                error_seen_onu = (stat >> 16) & 0xff
                for j in range(0,8):
                    error_seen = (error_seen_onu>>j)&0x1				
                    if(error_seen==1) : return_list.append(i*8+j)

        return return_list

    def errmon_cntr_config(self, onu_addr):
        msg_value = (onu_addr << 7)
        self.write_ctrl_reg(15, msg_value)
        return 1

    def errmon_get_error_cntr(self):
        # Latch Error Monitoring Status
        stat = self.read_ctrl_reg(15)
        stat = stat | (1 << 1)
        self.write_ctrl_reg(15, stat)
        stat = stat ^ (1 << 1)
        self.write_ctrl_reg(15, stat)
        stat = self.read_ctrl_reg(15)	
        error_cntr = (stat >> 24) & 0xff
        return error_cntr

	# OLT Interrupt
    # Interrupt mask is a list of the first N elements of the interrupt mask vector (N<=31 always)
    # interrupt_async <= (interrupt_mask(0) and mgt_seen_tx_nready_mngrsync_meta) or
    #                    (interrupt_mask(1) and mgt_seen_rx_nready_mngrsync_meta) or
    #                    (interrupt_mask(2) and mgt_seen_pll_nlock_mngrsync_meta) or
    #                    (interrupt_mask(3) and errdet_seen_any);	
    # Ex. set an interrupt with errdet_seen_any:
    # olt.set_interrupt_mask([0, 0, 0, 1])
    # If any value is not given to the interrupt, it assumes 0
    def set_interrupt_mask(self, interrupt_mask):
        message = 0
        for i in range(0,len(interrupt_mask)):
            message = message | (interrupt_mask[i]<<i)
        self.write_ctrl_reg(16, message)
        return 1

    # it returns a list with 31 elements
    def get_interrupt_mask(self):
        interrupt_mask = [0] * 31
        stat = self.read_ctrl_reg(16)
        for i in range(0,31):
            interrupt_mask[i] = ((stat>>i)&0x1)
        return interrupt_mask

    # it returns 0=interrupt not activated/1=interrupt activated
    def get_interrupt(self):
        interrupt = 0
        stat = self.read_ctrl_reg(16)
        interrupt = (stat>>31)&0x1
        return interrupt

    ### SFP_RX_RESET ###
    # Position SFP_RX_Reset:
    # enable: 0/1
    # position: n=0-(mod_position-1)
    def sfp_reset(self, enable, position, mod_position=None):
        if(mod_position == None): mod_position = self.c_BURST_LENGTH_OLTTXCLK
        mod_position = int(math.ceil(mod_position))
        #msg_value = (self.c_BURST_LENGTH_OLTTXCLK-1) + (position << 6) + (enable << 12)
        msg_value = (mod_position-1) + (position << 6) + (enable << 12)        
        self.write_ctrl_reg(5, msg_value)
        return 1

    def get_sfp_reset(self):
        stat = self.read_ctrl_reg(5)
        position = (stat & 0x00000FC0) >> 6
        enable = (stat & 0x00001000) >> 12
        return [enable, position]

    ### TX / RX polarity inversion ###
    def toggle_tx_pol(self):
        stat = self.read_ctrl_reg(10)
        stat = stat ^ 1
        self.write_ctrl_reg(10, stat)
        return 1
    def toggle_tx_match(self):
        addr= 2 + self.addr_offset   # user regs area
        #print "self.addr_offset:", self.addr_offset
        stat = self.read_ctrl_reg(addr)
        if stat&0x8:
          stat= stat & 0xfffffff7
        else:
          stat= stat | 0x8
        self.write_ctrl_reg(addr, stat)
        return 1

    def toggle_rx_pol(self):
        stat = self.read_ctrl_reg(10)
        stat = stat ^ (1 << 2)
        self.write_ctrl_reg(10, stat)
        return 1

    ### SFP_RX_RESET ###
    # Position SFP_RX_Reset:
    # enable: 0/1
    # position: n=0-29)
    def sfp_sd_clear(self):
        stat = self.read_ctrl_reg(11)
        stat = stat | 1		
        self.write_ctrl_reg(11, stat)
        stat = stat ^ 1		
        self.write_ctrl_reg(11, stat)		
        return 1

    def get_sfp_sd_stat(self):
        stat = self.read_ctrl_reg(11)
        stat = stat | 2		
        self.write_ctrl_reg(11, stat)
        stat = stat ^ 2		
        self.write_ctrl_reg(11, stat)	
        stat = self.read_ctrl_reg(12)
        min = (stat & 0x000003FF) >> 0
        max = (stat & 0x000FFC00) >> 10
        cur = (stat & 0x3FF00000) >> 20
        return [min, max, cur]

    ### SFP_RX_RESET ###
    # Position SFP_RX_Reset:
    # enable: 0/1
    # position: n=0-29)
    def sfp_sd_en_mask(self):
        stat = self.read_ctrl_reg(11)
        stat = stat | (1<<4)		
        self.write_ctrl_reg(11, stat)
        return 1

    def sfp_sd_dis_mask(self):
        stat = self.read_ctrl_reg(11)
        stat = stat & (0xFFFFFFEF)		
        self.write_ctrl_reg(11, stat)
        return 1

    def sfp_sd_en_mask_delay(self,delay):
        delay = delay&0xF	
        stat = self.read_ctrl_reg(11)
        stat = stat & 0xFFFFF81F
        stat = stat | (delay<<5)		
        self.write_ctrl_reg(11, stat)
        return 1

    def sfp_sd_comma_corr_clear(self):
        stat = self.read_ctrl_reg(11)
        stat = stat | (1<<2)		
        self.write_ctrl_reg(11, stat)
        stat = stat ^ (1<<2)
        self.write_ctrl_reg(11, stat)
        return 1

    def get_sfp_sd_comma_corr_stat(self):
        # latch data	
        stat = self.read_ctrl_reg(11)
        stat = stat | (1<<3)		
        self.write_ctrl_reg(11, stat)
        stat = stat ^ (1<<3)
        self.write_ctrl_reg(11, stat)
        # read data
        stat = self.read_ctrl_reg(11)
        finished = (stat&0x04000000)>>(16+10)
        corr_factor = (stat&0x03FF0000)>>(16)
        return [finished, corr_factor]

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

    # SFP related methods
    def get_sfp_vendor(self):
        slv_addr = 0x50	
        reg_bank = []
        for reg_addr in range(20,36,2):
            reg_value = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]			
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)
        vendor_name = ''.join(chr(i) for i in reg_bank)
        return [1, string.strip(vendor_name)]

    def get_sfp_part_number(self):
        slv_addr = 0x50		
        reg_bank = []
        for reg_addr in range(40,56,2):		
            reg_value = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]			
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)			
            #time.sleep(0.005)   #does not help
        vendor_part_number = ''.join(chr(i) for i in reg_bank)
        return [1, string.strip(vendor_part_number)]

    def get_sfp_serial_number(self):
        slv_addr = 0x50		
        reg_bank = []
        for reg_addr in range(68,84,2):
            reg_value = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)
        vendor_serial_number = ''.join(chr(i) for i in reg_bank)
        return [1, string.strip(vendor_serial_number)]

    def get_sfp_temp(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 96
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        temp    = (reg_bank[0]<<8) + reg_bank[1]

        # TEMPERATURE:
        # 1) Internally measured transceiver temperature. Represented as a 16 bit signed twos complement value in
        # increments of 1/256 degrees Celsius
        if(temp&0x8000):
            sign = -1
            temp_conv = (temp ^ 0xFFFF) + 1
        else:
            sign = 1
            temp_conv = temp
        temp_conv = sign*temp_conv*(1.0/256.0) # UNIT: degrees Celsius

        return [1, temp_conv]

    def get_sfp_vcc(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 98
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        vcc     = (reg_bank[0]<<8) + reg_bank[1]

        # SUPPLY VOLTAGE:
        # 2) Internally measured transceiver supply voltage. Represented as a 16 bit unsigned integer with the
        # voltage defined as the full 16 bit value (0-65535) with LSB equal to 100 uVolt
        vcc_conv = (vcc*100e-6) # UNIT: V

        return [1, vcc_conv]

    def get_sfp_tx_bias(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 100
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        tx_bias = (reg_bank[0]<<8) + reg_bank[1]	

		# TX BIAS CURRENT:
        # 3) Measured TX bias current in uA. Represented as a 16 bit unsigned integer with the current defined as the
        # full 16 bit value (0-65535) with LSB equal to 2 uA
        # OBS: FOR OLT THIS FACTOR IS DIFFERENT - see datasheet (4uA)		
        tx_bias_conv = (tx_bias*4e-6)*1e3 # UNIT: mA

        return [1, tx_bias_conv]
		
    def get_sfp_tx_pwr(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 102
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        tx_pwr  = (reg_bank[0]<<8) + reg_bank[1]

		# TX OUTPUT POWER:
        # 4) Measured TX output power in mW. Represented as a 16 bit unsigned integer with the power defined as
        # the full 16 bit value (0-65535) with LSB equal to 0.1 uW
        # OBS: FOR OLT THIS FACTOR IS DIFFERENT - see datasheet (0.2uW)		
        tx_pwr_conv = 10*math.log10((tx_pwr * 0.2 * 1e-6)/1e-3) # UNIT: dBm

        return [1, tx_pwr_conv]

    def get_sfp_rx_pwr(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 104
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        rx_pwr  = (reg_bank[0]<<8) + reg_bank[1]
        #print('rx_pwr:' + str(rx_pwr)) - DEBUGGING
		# RX RECEIVED OPTICAL POWER:
        # 5) Measured RX received optical power in mW. Value can represent either average received power or OMA
        # depending upon how bit 3 of byte 92 (A0h) is set. Represented as a 16 bit unsigned integer with the
        # power defined as the full 16 bit value (0-65535) with LSB equal to 0.1 uW
        if(rx_pwr>0):
            rx_pwr_conv = 10*math.log10((rx_pwr * 0.1 * 1e-6)/1e-3) # UNIT: dBm
        else:
            rx_pwr_conv = -1*float('inf')
			
        return [1, rx_pwr_conv]

    ### Set ONU modes ###
    def onu_mode_operational(self, onu_addr):
        self.msg2onu(self.c_OPE_WR, 0, self.c_ONU_MODE_OPERATIONAL, onu_addr)
        return 1

    def onu_mode_disable(self, onu_addr):
        self.msg2onu(self.c_OPE_WR, 0, self.c_ONU_MODE_DISABLE, onu_addr)
        return 1

    def onu_mode_calib(self, onu_addr):
        self.msg2onu(self.c_OPE_WR, 0, self.c_ONU_MODE_CALIBRATION, onu_addr)
        return 1

    def onu_mode_prbsburst(self, onu_addr):
        self.msg2onu(self.c_OPE_WR, 0, self.c_ONU_MODE_PRBSBURST, onu_addr)
        return 1

    def onu_mode_prbscont(self, onu_addr):
        self.msg2onu(self.c_OPE_WR, 0, self.c_ONU_MODE_PRBSCONT, onu_addr)
        return 1

    def onu_heartbeat(self, onu_addr, number_onu):
        heartbeat = int(math.ceil(number_onu * self.c_BURST_LENGTH_ONURXCLK)) - 1
        heartbeat_lsb = (heartbeat & 0x000000FF)
        heartbeat_msb = (heartbeat & 0x0000FF00) >> 8
        self.msg2onu(self.c_OPE_WR, 4, heartbeat_lsb, onu_addr)
        self.msg2onu(self.c_OPE_WR, 5, heartbeat_msb, onu_addr)
        return 1

    def onu_coarse_offset(self, onu_addr, coarse_offset):
        coarse_offset_lsb = (coarse_offset & 0x000000FF)
        coarse_offset_msb = (coarse_offset & 0x0000FF00) >> 8
        self.msg2onu(self.c_OPE_WR, 6, coarse_offset_lsb, onu_addr)
        self.msg2onu(self.c_OPE_WR, 7, coarse_offset_msb, onu_addr)
        return 1

    def onu_sfp_en_delay(self, onu_addr, sfp_en_delay):
        msg_value=sfp_en_delay
        self.msg2onu(self.c_OPE_WR, 8, msg_value, onu_addr)
        return 1

    def onu_fine_offset(self, onu_addr, fine_offset):
        msg_value = fine_offset
        self.msg2onu(self.c_OPE_WR, 9, msg_value, onu_addr)
        return 1

    ### ONU IIC CTRL (begin) ###
    def onu_i2c_write(self, slv_addr, reg_addr, reg_value, onu_addr):
        slv_addr_format  = slv_addr & 0x7F
        reg_addr_format  = reg_addr & 0xFF
        reg_value_format = reg_value & 0xFF
        onu_addr_format = onu_addr &0xFF;		
        if((slv_addr != slv_addr_format) | (reg_addr != reg_addr_format) | (reg_value != reg_value_format) | (onu_addr != onu_addr_format)):
            self.logger.warn('Error parameters formatting for IIC write command')		
            return 0
        else:
            success = self.onu_wrack(13*4+3,0,onu_addr)
            if (not success):
                self.logger.warn('Error OLT2ONU_WRACK')					
                return 0
            message = (1<<31)+(reg_value<<18) + (reg_addr<<10) + (slv_addr<<3) + (0<<1) + (0<<1) + (1)
            message_bytes = [0,0,0,0];
            for i in range(0,4):
                message_bytes[i] = (message>>(i*8))&0xff
                success = self.onu_wrack(13*4+i,message_bytes[i],onu_addr)
                if (not success):
                    self.logger.warn('Error OLT2ONU_WRACK')					
                    return 0
                    					
        i2c_done = 0
        i2c_error = 0
        i2c_drop_req = 0
        while(i2c_done==0 and i2c_error==0 and i2c_drop_req==0):
            i2c_stat=0
            for i in range(0,4):
                [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(14*4+i,onu_addr)
                if (not success):
                    self.logger.warn('Error OLT2ONU_RD in onu_i2c_write')
                    return 0
                i2c_stat = i2c_stat + (reg_value<<(i*8))

            i2c_done     = (i2c_stat & 0x00000001) >> 0
            i2c_error    = (i2c_stat & 0x00000002) >> 1
            i2c_drop_req = (i2c_stat & 0x00000004) >> 2

        if(i2c_error):
            self.logger.warn('IIC write error')				
            return 0
        elif(i2c_drop_req):
            self.logger.warn('IIC drop request error')				
            return 0			
        else:
            return 1

    # Read IIC (nb_read_access can be 1 or 2 for double burst read access)
    def onu_i2c_read(self, slv_addr, reg_addr, nb_read_access=1, onu_addr=1):
        slv_addr_format  = slv_addr & 0x7F
        reg_addr_format  = reg_addr & 0xFF
        nb_read_access   = nb_read_access - 1
        nb_read_access_format = nb_read_access & 0x1
        onu_addr_format  = onu_addr & 0xFF

        if((slv_addr != slv_addr_format) | (reg_addr != reg_addr_format) | (nb_read_access_format != nb_read_access) | (onu_addr_format != onu_addr)):
            self.logger.warn('Error parameters formatting for IIC read command')		
            return [-1]
        else:
            success = self.onu_wrack(13*4+3,0,onu_addr)
            if (not success):
                self.logger.warn('Error OLT2ONU_WRACK')					
                return [-1]
            message = (1<<31)+(0<<18) + (reg_addr<<10) + (slv_addr<<3) + (1<<2) + (nb_read_access<<1) + (1)
            message_bytes = [0,0,0,0];
            for i in range(0,4):
                message_bytes[i] = (message>>(i*8))&0xff
                success = self.onu_wrack(13*4+i,message_bytes[i],onu_addr)
                if (not success):
                    self.logger.warn('Error OLT2ONU_WRACK')					
                    return [-1]

        i2c_done = 0
        i2c_error = 0
        i2c_drop_req = 0		
        while(i2c_done==0 and i2c_error==0 and i2c_drop_req==0):
            i2c_stat=0		
            for i in range(0,4):
                [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(14*4+i,onu_addr)
                if (not success):
                    self.logger.warn('Error OLT2ONU_RD in onu_i2c_read')
                    return [-1]
                i2c_stat = i2c_stat + (reg_value<<(i*8))
            i2c_done     = (i2c_stat & 0x00000001) >> 0
            i2c_error    = (i2c_stat & 0x00000002) >> 1
            i2c_drop_req = (i2c_stat & 0x00000004) >> 2
            i2c_data     = (i2c_stat & 0x0007FFF8) >> 3	

        if(i2c_error):
            self.logger.warn('IIC read error')				
            return [-1]
        elif(i2c_drop_req):
            self.logger.warn('IIC drop request error')				
            return [-1]		
        else:
            if(nb_read_access > 0):		
                data_return_list = []        		
                for i in range(0,(nb_read_access+1)):
                    data  = (i2c_data >> i*8) & 0xFF
                    data_return_list.append(data)
                return data_return_list
            else:
                return [i2c_data&0xFF]

    ### DRP CTRL (begin) ###
    def onu_drp_write(self, addr, value, onu_addr):
        rdy=0	
        while(rdy==0):
            drpen=1
            drpwe=1
            success = self.onu_wrack(17*4+3,0,onu_addr)
            if (not success):
                self.logger.warn('Error OLT2ONU_WRACK')					
                return 0
					
            message = (1<<31) + (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + 0
            message_bytes = [0,0,0,0];
            for i in range(0,4):
                message_bytes[i] = (message>>(i*8))&0xff
                success = self.onu_wrack(17*4+i,message_bytes[i],onu_addr)
                if (not success):
                    self.logger.warn('Error OLT2ONU_WRACK')					
                    return 0
            stat=0
            for i in range(0,4):
                [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(17*4+i,onu_addr)
                if (not success):
                    self.logger.warn('Error OLT2ONU_RD in onu_drp_write')
                    return 0				
                self.read_onu_msg()
                stat = stat + (reg_value<<(i*8))

            value =  (stat&0x0000ffff) >> 0;
            addr  =  (stat&0x01ff0000) >> 16;
            rdy   =  (stat&0x04000000) >> 26;

        return rdy

    def onu_drp_read(self, addr, onu_addr):
        rdy=0
        while(rdy==0):
            drpen=1
            drpwe=0
            success = self.onu_wrack(17*4+3,0,onu_addr)
            if (not success):
                self.logger.warn('Error OLT2ONU_WRACK')					
                return 0			
            message = (1<<31) + (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + 0
            message_bytes = [0,0,0,0];
            for i in range(0,4):
                message_bytes[i] = (message>>(i*8))&0xff
                success = self.onu_wrack(17*4+i,message_bytes[i],onu_addr)
                if (not success):
                    self.logger.warn('Error OLT2ONU_WRACK')					
                    return 0
            stat=0
            for i in range(0,4):
                [success, ope, reg_addr, reg_value, onu_addr] =self.onu_rd(17*4+i,onu_addr)
                if (not success):
                    self.logger.warn('Error OLT2ONU_WRACK')					
                    return 0
                stat = stat + (reg_value<<(i*8))

            value =  (stat&0x0000ffff) >> 0;
            addr  =  (stat&0x01ff0000) >> 16;	
            rdy   =  (stat&0x04000000) >> 26;
        return [addr, value, rdy]

    ### Phy Error Monitoring ###
    def onu_phymon_clear(self, onu_addr):
        [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(5*4+0,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_RD in onu_phymon_clear')
            return 0
        stat = reg_value

        stat = stat | 1
        success = self.onu_wrack(5*4,stat,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_WRACK')					
            return 0

        stat = stat ^ 1
        success = self.onu_wrack(5*4,stat,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_WRACK')					
            return 0

        return 1

    # It returns a list of seven elements
    # [success, seen_mgt_tx_nready, seen_mgt_rx_nready, seen_mgt_txpll_nlock, seen_mgt_rxpll_nlock, seen_rx_nlocked, seen_phase_ngood]
    def onu_phymon_read_seen(self, onu_addr):
        [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(5*4+1,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_RD in onu_phymon_read_seen')
            return [0, 0, 0, 0, 0, 0, 0]
        stat = reg_value
        seen_mgt_tx_nready      = (stat>>0)&0x1
        seen_mgt_rx_nready      = (stat>>1)&0x1
        seen_mgt_txpll_nlock    = (stat>>2)&0x1
        seen_mgt_rxpll_nlock    = (stat>>3)&0x1
        seen_rx_nlocked         = (stat>>4)&0x1
        seen_phase_ngood        = (stat>>5)&0x1				
        # Return value
        mon_return = [1, seen_mgt_tx_nready, seen_mgt_rx_nready, seen_mgt_txpll_nlock, seen_mgt_rxpll_nlock, seen_rx_nlocked, seen_phase_ngood]
        return mon_return
		
    ### Error Monitor on-the-fly based on FEC correction capability and SC-CRC-7 error detection ###
    def onu_errmon_clear(self, onu_addr):   # onu_addr: 1 or 4
        [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(3*4+0,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_RD in onu_errmon_clear(%d)'%onu_addr)
            return 0
        stat = reg_value

        stat = stat | 1
        success = self.onu_wrack(3*4,stat,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_WRACK in onu_errmon_clear(%d)'%onu_addr)
            return 0

        stat = stat ^ 1
        success = self.onu_wrack(3*4,stat,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_WRACK in onu_errmon_clear(%d)'%onu_addr)
            return 0

        return 1

    # It returns a list of four elements
    # [success, single_error_corrected_fec, double_error_corrected_fec, slow_control_crc_err_detect]
    def onu_errmon_read(self, onu_addr):
        # Latch
        [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(3*4+0,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_RD in onu_errmon_read(%d)'%onu_addr)
            return [0, 0, 0, 0]
        stat = reg_value

        stat = stat | (1<<1)
        success = self.onu_wrack(3*4,stat,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_WRACK')					
            return [0, 0, 0, 0]

        stat = stat ^ (1<<1)
        success = self.onu_wrack(3*4,stat,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_WRACK')					
            return [0, 0, 0, 0]

        # Read status
        mon_return = [1]
        for i in range(1,4):
            [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(3*4+i,onu_addr)
            if (not success):
                self.logger.warn('Error OLT2ONU_RD in onu_errmon_read.2')
                mon_return[0] = 0
            mon_return.append(reg_value)

        return mon_return

    # It returns a list of four elements
    # [success, seen_single_error_corrected, seen_double_error_corrected, seen_crc_err_detect]
    def onu_errmon_read_seen_error(self, onu_addr):
        [success, ope, reg_addr, reg_value, onu_addr] = self.onu_rd(4*4+0,onu_addr)
        if (not success):
            self.logger.warn('Error OLT2ONU_RD in onu_errmon_read_seen_error')
            return [0, 0, 0, 0]
        stat = reg_value
        serror_fec    = (stat>>0)&0x1
        derror_fec    = (stat>>1)&0x1
        sc_crc_errdet = (stat>>2)&0x1		
        # Return value
        mon_return = [1, serror_fec, derror_fec, sc_crc_errdet]
        return mon_return

    # All the methods to get diagnostic from SFP return a list that can have up to two elements
    # The first element indicates if the operation was successful and the other element is the diagnostic asked for
	# The SFP module is supposed to be internally calibrated for this method, otherwise a review on
    # the calibration formulas shall be made
    # SFP related methods
    def onu_get_sfp_vendor(self, onu_addr):
        slv_addr = 0x50	
        reg_bank = []
        for reg_addr in range(20,36,2):
            reg_value = self.onu_i2c_read(slv_addr, reg_addr,2, onu_addr) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]			
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)
        vendor_name = ''.join(chr(i) for i in reg_bank)
        return [1, string.strip(vendor_name)]

    def onu_get_sfp_part_number(self, onu_addr):
        slv_addr = 0x50		
        reg_bank = []
        for reg_addr in range(40,56,2):
            reg_value = self.onu_i2c_read(slv_addr, reg_addr,2, onu_addr) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]			
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)
        vendor_part_number = ''.join(chr(i) for i in reg_bank)
        return [1, string.strip(vendor_part_number)]

    def onu_get_sfp_serial_number(self, onu_addr):
        slv_addr = 0x50		
        reg_bank = []
        for reg_addr in range(68,84,2):
            reg_value = self.onu_i2c_read(slv_addr, reg_addr,2, onu_addr) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)
        vendor_serial_number = ''.join(chr(i) for i in reg_bank)
        return [1, string.strip(vendor_serial_number)]
	
    def onu_get_sfp_temp(self, onu_addr):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 96
        reg_value  = self.onu_i2c_read(slv_addr, reg_addr,2, onu_addr) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        temp    = (reg_bank[0]<<8) + reg_bank[1]

        # TEMPERATURE:
        # 1) Internally measured transceiver temperature. Represented as a 16 bit signed twos complement value in
        # increments of 1/256 degrees Celsius
        if(temp&0x8000):
            sign = -1
            temp_conv = (temp ^ 0xFFFF) + 1
        else:
            sign = 1
            temp_conv = temp
        temp_conv = sign*temp_conv*(1.0/256.0) # UNIT: degrees Celsius

        return [1, temp_conv]

    def onu_get_sfp_vcc(self, onu_addr):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 98
        reg_value  = self.onu_i2c_read(slv_addr, reg_addr,2, onu_addr) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        vcc     = (reg_bank[0]<<8) + reg_bank[1]

        # SUPPLY VOLTAGE:
        # 2) Internally measured transceiver supply voltage. Represented as a 16 bit unsigned integer with the
        # voltage defined as the full 16 bit value (0-65535) with LSB equal to 100 uVolt
        vcc_conv = (vcc*100e-6) # UNIT: V

        return [1, vcc_conv]

    def onu_get_sfp_tx_bias(self, onu_addr):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 100
        reg_value  = self.onu_i2c_read(slv_addr, reg_addr,2, onu_addr) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        tx_bias = (reg_bank[0]<<8) + reg_bank[1]	

		# TX BIAS CURRENT:
        # 3) Measured TX bias current in uA. Represented as a 16 bit unsigned integer with the current defined as the
        # full 16 bit value (0-65535) with LSB equal to 2 uA
        tx_bias_conv = (tx_bias*2e-6)*1e3 # UNIT: mA

        return [1, tx_bias_conv]
		
    def onu_get_sfp_tx_pwr(self, onu_addr):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 102
        reg_value  = self.onu_i2c_read(slv_addr, reg_addr,2, onu_addr) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        tx_pwr  = (reg_bank[0]<<8) + reg_bank[1]

		# TX OUTPUT POWER:
        # 4) Measured TX output power in mW. Represented as a 16 bit unsigned integer with the power defined as
        # the full 16 bit value (0-65535) with LSB equal to 0.1 uW
        tx_pwr_conv = 10*math.log10((tx_pwr * 0.1 * 1e-6)/1e-3) # UNIT: dBm

        return [1, tx_pwr_conv]

    def onu_get_sfp_rx_pwr(self, onu_addr):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 104
        reg_value  = self.onu_i2c_read(slv_addr, reg_addr,2, onu_addr) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        rx_pwr  = (reg_bank[0]<<8) + reg_bank[1]

		# RX RECEIVED OPTICAL POWER:
        # 5) Measured RX received optical power in mW. Value can represent either average received power or OMA
        # depending upon how bit 3 of byte 92 (A0h) is set. Represented as a 16 bit unsigned integer with the
        # power defined as the full 16 bit value (0-65535) with LSB equal to 0.1 uW
        if(rx_pwr>0):
            rx_pwr_conv = 10*math.log10((rx_pwr * 0.1 * 1e-6)/1e-3) # UNIT: dBm
        else:
            rx_pwr_conv = -1*float('inf')

        return [1, rx_pwr_conv]

    ### Detect ONU ###
    # Detect if ONU is present
    # The detection is based in a write w/ acknowledgment to ONU
    # OLT sends a random message to an IDLE byte (not used by any ONU important parameter) of ONU with an acknowledment request
    # Checks if OLT has received the same message it has sent
    # The test is successful if OLT has received a certain consecutive number of correct acknowledgements
    # onu_addr          : ONU address to verify presence in the up line
    # nb_consec_success : minimum number of consecutive successful write w/ acknowledment operation to say ONU is present
    # idle_byte_addr    : byte address used for ONU detection
    def detect_onu(self, onu_addr, nb_consec_success, idle_byte_addr):
        olt2onu_success = 1
        index_access = 1
        while(olt2onu_success == 1 and index_access <= nb_consec_success):
            # Generate a random byte value
            byte_value = randint(0, 255)		
            olt2onu_success = self.onu_wrack(idle_byte_addr, byte_value, onu_addr)
            self.logger.debug('Access number:' + str(index_access))
            index_access += 1

        return olt2onu_success

    ### Measure position of ONU ###
    # Measure position of a given ONU
    # The test is successful if OLT has received a certain consecutive number of positions varying from a maximum of range_pos_accept
    # nb_consec_success : minimum number of consecutive correct positions
    # range_pos_accept  : range accepted for position variation (unit: 0.104ns)
    # This method returns the position with the [fine_offset (0.104ns), coarse_offset (8.33ns)] convention
    # [0,0] when it does not converge
    def measpos_onu(self, nb_consec_success, range_pos_accept):
        measpos_success = 1
        index_access = 1
        pos_meas = 0
        return_pos = [0, 0]
        pos_meas_vector = self.calib_read_pos()
        pos_meas = pos_meas_vector[0] + (self.c_COARSE_ROUNDTRIP_UNIT / self.c_FINE_ROUNDTRIP_UNIT) * pos_meas_vector[1]
        pos_meas_previous = pos_meas
        self.logger.debug('Measurement number: 1')
        self.logger.debug('Measured position :' + str(pos_meas))
        while(measpos_success == 1 and index_access < nb_consec_success):
            pos_meas_vector = self.calib_read_pos()
            pos_meas = pos_meas_vector[0] + (self.c_COARSE_ROUNDTRIP_UNIT / self.c_FINE_ROUNDTRIP_UNIT) * pos_meas_vector[1]
            self.logger.debug('Measurement number:' + str(index_access + 1))
            self.logger.debug('Measured position (raw) :' + str(pos_meas_vector))
            self.logger.debug('Measured position (0.104ns) :' + str(pos_meas))

           # Check if the position measured is within acceptable limit of
           # previous
            if(abs(pos_meas - pos_meas_previous) > range_pos_accept):
                measpos_success = 0

            pos_meas_previous = pos_meas
            index_access += 1

        if measpos_success:
            return_pos = pos_meas_vector

        return return_pos

    ### OLT SFP RX RESET POSITIONING ###
    # Find position for which sfp_rx_reset is positioned in the end of the burst
    # The SFP_RX_RESET positioning is based on a fast-BERT
    # onu1_addr     : onu1 used for BERT
    # onu2_addr     : onu2 used for BERT (can be 0 if only one ONU is being analysed)
    # consec_nvalid : number of consecutive non-valid positions to detect rising_edge
    # consec_valid  : number of consecutive valid positions to detect rising_edge
    # margin_pos    : margin to the sfp_rx_reset position (more description below)
    # This methods returns:
    # position in the range 0-29 when it works
    # -1 when no position is working
    # -2 when more than one position is found
    def find_sfp_reset_pos_bert(self, onu1_addr, onu2_addr, consec_nvalid, consec_valid, mod_reset_pos, margin_pos):
        # step1: Find pos_valid vector
                # pos   |0|1|2|.............|26|27|28|29|
                #        ______              ___________
                # valid        \____________/
        pos_valid = []

        mod_reset_pos = int(math.ceil(mod_reset_pos))

        for rst_pos in range(0, mod_reset_pos):
            self.sfp_reset(0, 0, mod_reset_pos)
            self.sfp_reset(1, rst_pos, mod_reset_pos)
            self.bert_config(onu1_addr, onu2_addr)
            self.bert_clear()
            bert_status = [0, 0, 0, 0, 0, 0]

            # while BERT is not finished
            while((not bert_status[0]) or (not bert_status[3])):
                # bert_status returns: (onu1_finished,onu1_locked,onu1_errors,onu2_finished,onu2_locked,onu2_errors)
                bert_status = self.bert_read()
            # print('nb_errors: ' + str(bert_status[2]) + ' - - - locked:' + str(bert_status[1]))
            # pos_valid_onu1 = 1 if(bert_status[1]) else 0
            pos_valid_onu1 = 1 if(bert_status[1]) else 0			
            pos_valid_onu2 = 1 if(bert_status[4] or onu2_addr == 0) else 0
            pos_valid.append(pos_valid_onu1 & pos_valid_onu2)

            self.logger.debug('POSITION:' + str(rst_pos) + ' / VALID:' + str(pos_valid[rst_pos]))

        # step2: Find position for rising_edge of pos_valid vector (e.g. 26)
        # Based on a "circular convolution" between the pos_valid vector and mask_pos vector (e.g. [0001111] consec_nvalid=3, consec_valid=4)
        # pos   |0|1|2|.............|26|27|28|29|
        #        ______              ___________
        # valid        \____________/
        # OBS  : rising edge is considered when 'consec_nvalid' 0's and 'consec_valid' 1's are found in the vector
        # OBS2 : margin is given to the position through the 'margin_pos'
        # parameter (e.g. position given will be 26+margin_pos)
        mask_pos = ([0] * consec_nvalid) + ([1] * consec_valid)
        found_pos = 0
        ideal_pos = -1
        for i in range(0, len(pos_valid)):
            diff = 0
            for j in range(0, len(mask_pos)):
                index_rst = (i + j) % len(pos_valid)
                diff += pos_valid[index_rst] ^ mask_pos[j]

            if(diff == 0):
                ideal_pos = (i + consec_nvalid + margin_pos) % len(pos_valid)
                self.logger.info('Ideal position found at:' + str(ideal_pos))  # DEBUG
                if found_pos:
                    ideal_pos = -2
                    self.logger.warn('More than a single position found for sfp_rx_reset')
                found_pos = 1

        return ideal_pos

    # Find position for which sfp_rx_reset is positioned inside the burst
    def find_sfp_reset_pos_sd(self, mod_reset_pos, margin_pos=2):
        mod_reset_pos = int(math.ceil(mod_reset_pos))
        # Find sd_fal_to_ris_max vector
        vec = [0] * mod_reset_pos
        for i in range(0,len(vec)):
            self.sfp_reset(1,i, len(vec))
            self.sfp_sd_clear()
            #time.sleep(0.001)			
            meas_res = self.get_sfp_sd_stat()
            vec[i] = meas_res[1]
            self.logger.debug('POSITION:' + str(i) + ' / SD_RIS_TO_FALL:' + str(vec[i]))			
        #print(vec)
        # Find differential curve
        diff = [0] * len(vec)
        for i in range(0,len(vec)):
            diff[i] = ((vec[(i+3)%len(vec)]+vec[(i+2)%len(vec)]+vec[(i+1)%len(vec)]) - (vec[(i)%len(vec)]+vec[(i-1)%len(vec)]+vec[(i-2)%len(vec)]))/3

        #print(diff)

        # Find maxima
        m = max(vec)		

        pos_m = [i for i, j in enumerate(vec) if j==m]		
        # print(pos_m)	
        pos_m = pos_m[0]

        # Find first point from maxima where the differential becomes smaller than one
        k = 0
        found = 0
        while(k<len(diff) and found==0):
            i = (k+pos_m) % len(diff)
            if(abs(diff[i])<1.0 and vec[i]<=8):
                found = 1
                pos_rst0 = i
            k = k + 1

        if (found==0):
            self.logger.warn('No position found for SFP_RX_RESET')
            return -1

        # Position given to reset is pos_rst0 + margin_pos
        pos_rst = (pos_rst0 + margin_pos) % len(diff)
        self.logger.info('Ideal position found at:' + str(pos_rst))  # DEBUG
        return pos_rst

    ### Find SFP SD EN MASK POSITIONING ###
    def find_sfp_sd_en_mask_pos(self):
        window_pos = 0
        self.sfp_sd_en_mask()
        valid_window = 1
        found_one_valid_window = 0		
        while(valid_window==1 and window_pos<=self.c_MAX_SFP_SD_EN_MASK_DELAY):
            self.sfp_sd_en_mask_delay(window_pos)
            finished = 0
            time0 = time.time()
            deltaT = 0
            self.sfp_sd_comma_corr_clear()			
            while(finished==0 and deltaT<=self.c_MAX_TIME_SFP_SD_COMMA_CORR_STAT):
                [finished, comma_corr_stat] = self.get_sfp_sd_comma_corr_stat()
                deltaT = time.time() - time0

            if((finished==0 or comma_corr_stat <= self.c_MIN_COMMA_CORR_STAT_VALID) and found_one_valid_window==1):
                valid_window=0
            else:
                window_pos = window_pos + 1
                found_one_valid_window = 1

        if(not(valid_window) and not(found_one_valid_window)):
            self.logger.warn('Could not find any valid SFP SD EN MASK')		
            return -1
        elif(not(valid_window) and found_one_valid_window):
            final_window_pos = window_pos-self.c_MARGIN_SFP_SD_EN_MASK
            if(final_window_pos>=0):
                return final_window_pos			
            else:
                self.logger.warn('!!! No much margin for SFP SD EN MASK - review delay scheme!!!!')
                return 0				
        else:
            self.logger.warn('All positions for SFP SD EN MASK seem to be valid, this should be reviewed')
            return -1

    ### ONU SFP ENABLE DELAY POSITIONING ###
    # Find position for sfp_en_delay
    # The SFP_ENABLE_DELAY positioning is based on a fast-BERT
    # onu_addr      : onu which the position is to be calibrated
    # margin_pos    : margin to the sfp_en_delay position
    # This method returns the first valid position + a margin given
    # -1 when no position is found
    def find_onu_sfp_delay(self, onu_addr, margin_pos):
        found_pos = 0
        sfp_en_delay = 0
        ideal_pos = -1

        while((not found_pos) and sfp_en_delay <= (self.c_MAX_SFP_EN_DELAY - margin_pos)):			
            self.onu_sfp_en_delay(onu_addr, sfp_en_delay)
            self.bert_config(onu_addr, 0)
            self.bert_clear()
            bert_status = [0, 0, 0, 0, 0, 0]
            # time.sleep(1)

            # while BERT is not finished
            while(not bert_status[0]):
                # bert_status returns
                # (onu1_finished,onu1_locked,onu1_errors,onu2_finished,onu2_locked,onu2_errors)
                bert_status = self.bert_read()
            #print(bert_status)
            pos_valid_onu1 = 1 if(bert_status[1]) else 0
            self.logger.debug('POSITION:' + str(sfp_en_delay) + ' / VALID:' + str(pos_valid_onu1))

            if pos_valid_onu1:
                found_pos = 1
                ideal_pos = sfp_en_delay + margin_pos
                self.onu_sfp_en_delay(onu_addr, ideal_pos)
                self.logger.info('Ideal position found at:' + str(ideal_pos))
            sfp_en_delay += 1
        if found_pos == 0:
            self.logger.warn('No position found for ONU_SFP_EN_DELAY')

        return ideal_pos

    ### Calibration calculation ###
    # Calculate ONU offset in TDM
    # onu_ref_pos      : reference position
    # onu_meas_pos     : measured position
    # onu_ideal_offset : ideal offset if there were no latency difference between ONUs
    # mod_onu_offset   : modulo of onu local timer
    # This method returns
    def calc_offset(self, onu_ref_pos, onu_meas_pos, onu_ideal_offset, mod_onu_offset):
        # Calculate roundtrip in fine_roundtrip_unit
        diff_pos = (onu_ref_pos[1] * self.c_COARSE_ROUNDTRIP_UNIT + onu_ref_pos[0] * self.c_FINE_ROUNDTRIP_UNIT) - (onu_meas_pos[1] * self.c_COARSE_ROUNDTRIP_UNIT + onu_meas_pos[0] * self.c_FINE_ROUNDTRIP_UNIT)
        # Calculate roundtrip in fine_offset_unit
        diff_pos = diff_pos // (self.c_FINE_OFFSET_UNIT // self.c_FINE_ROUNDTRIP_UNIT)

        fine_offset = int(diff_pos % (self.c_COARSE_OFFSET_UNIT // self.c_FINE_OFFSET_UNIT))

        coarse_offset = int((onu_ideal_offset + ((diff_pos // (self.c_COARSE_OFFSET_UNIT // self.c_FINE_OFFSET_UNIT)) * (self.c_BURST_LENGTH_ONURXCLK // self.c_BURST_LENGTH_ONUTXCLK))) % mod_onu_offset)
        return [fine_offset, coarse_offset]

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
    ### Full roundtrip measurement procedure ###
    # list_onu_dict      : list of ONU with system previously calibrated (full_calibration)
    # full_nlight_calib  : 1 if full calibration, 0 if light calibration
	# Detects ONUs present in the system (only detect ONUs that were previously detected if light calibration) and measure roundtrip for each
    def full_roundtrip_meas(self, list_onu_dict, full_nlight_calib):

        # Init
        number_onu_detected = 0
        self.tdm_init(64)
        self.calib_config(1, 1, 1)
        self.sfp_reset(0, 0)

        onu_dict_list = list_onu_dict

        # Detect ONUs that have coarse_offset != -1 (they were at least correctly calibrated once)
        # ONU being detected is put in calibration mode (others are disabled)
        # detection is based on a write with acknowledgment for all the ONUs expected to be in the system	
        if(full_nlight_calib) : gen_onu_detected = (onu for onu in onu_dict_list)
        else : gen_onu_detected = (onu for onu in onu_dict_list if onu['coarse_offset'] != -1)  # Only try to detect ONUs which have been previously initialized

        for onu_detected in gen_onu_detected:
            try_detect = 1
            index_trial = 1
            self.logger.info('OLT trying to detect ONU ADDR ' + str(onu_detected['onu_addr']) + '... ')
            self.calib_config(0, 1, onu_detected['onu_addr'])
            # try to detect ONU a certain number of times
            while(try_detect and index_trial <= self.c_MAX_TRIALS_ONU_DETECT):
                self.logger.debug('Trial number:' + str(index_trial))
                self.onu_mode_disable(self.c_BROADCAST_ADDR)   # -> msg2onu
                self.onu_mode_calib(onu_detected['onu_addr'])  # -> msg2onu	
                time.sleep(0.1)
                onu_was_detected = self.detect_onu(onu_detected['onu_addr'], self.c_CONSECUTIVE_SUCCESS_DETECT, self.c_IDLE_BYTE_ADDR)   # -> msg2onu
                if onu_was_detected:
                    self.logger.info('ONU ' + str(onu_detected['onu_addr']) + ' detected in trial number ' + str(index_trial))
                    try_detect = 0
                    number_onu_detected += 1
                    onu_dict_list[onu_dict_list.index(onu_detected)]['detected']=1
                    try_measpos = 1
                    index_measpos = 1
                    # try to measure ONU position
                    while(try_measpos and index_measpos <= self.c_MAX_TRIALS_ONU_MEAS_POS):
                        onu_meas_pos = self.measpos_onu(self.c_CONSECUTIVE_SUCCESS_MEASPOS, self.c_RANGE_ACCEPT_MEASPOS)					
                        if (onu_meas_pos != [0, 0]):
                            try_measpos = 0
                            onu_dict_list[onu_dict_list.index(onu_detected)]['fine_roundtrip']   = onu_meas_pos[0]
                            onu_dict_list[onu_dict_list.index(onu_detected)]['coarse_roundtrip'] = onu_meas_pos[1]							
                            break
                        self.onu_mode_calib(onu_detected['onu_addr'])
                        index_measpos += 1
                    if try_measpos:
                        self.logger.warn('Position measured in calibration for ONU ' + str(onu_detected['onu_addr']) + ' does not converge')
                    else:					
                        # Measure power
                        i2c_transaction = self.get_sfp_rx_pwr()											
                        i2c_success = i2c_transaction[0]						
                        if(i2c_success):
                            rx_pwr = i2c_transaction[1]  						
                            onu_dict_list[onu_dict_list.index(onu_detected)]['rx_pwr']=rx_pwr
                            self.logger.info('Measured received power from ONU ' + str(onu_detected['onu_addr']) + ': ' + str(rx_pwr) + ' dBm')
                            # return - DEBUGGING							
                        else:
                            self.logger.warn('Could not measure received power from ONU' + str(onu_detected['onu_addr']))

                index_trial += 1

            if not onu_was_detected:
                self.logger.warn('ONU ' + str(onu_detected['onu_addr']) + ' NOT detected')
                onu_dict_list[onu_dict_list.index(onu_detected)]['detected'] = 0

        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        for onu_detected in gen_onu_detected: onu_detected['nb_onu_system'] = number_onu_detected

        self.calib_config(0,0,0)

        return onu_dict_list

    ### Network update offset procedure ###
    # list_onu_dict      : list of ONU with system previously calibrated (full_calibration)
    # full_nlight_calib  : 1 if full calibration, 0 if light calibration
	# Calculates 'fine_offset' and 'coarse_offset' field for each ONU in list_onu_dict
    def network_update_offset(self, list_onu_dict, full_nlight_calib):
        onu_dict_list = list_onu_dict
        # Take as reference ONU the position of the first detected ONU and check number of ONUs in the system
        # if(full_nlight_calib) : gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        # else : gen_onu_detected = (onu for onu in onu_dict_list if onu['coarse_offset'] != -1)
        # onu_detected = next(gen_onu_detected)
        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        onu_detected = next(gen_onu_detected)
		
        if(full_nlight_calib):
            onu_ref_pos = [onu_detected['fine_roundtrip'], onu_detected['coarse_roundtrip']]
            ideal_offset = 0
        else :
            onu_ref_pos = [onu_detected['fine_roundtrip'] + onu_detected['fine_offset'] * self.c_FINE_OFFSET_UNIT, onu_detected['coarse_roundtrip']]
            ideal_offset = onu_detected['coarse_offset'] % self.c_BURST_LENGTH_ONURXCLK

        number_onu_detected = onu_detected['nb_onu_system']
        mod_onu_offset = self.c_BURST_LENGTH_ONURXCLK * number_onu_detected

        self.logger.info("Starting Offset calculation ...")

        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        for onu_detected in gen_onu_detected:
            offset = self.calc_offset(onu_ref_pos, [onu_detected['fine_roundtrip'], onu_detected['coarse_roundtrip']], ideal_offset, mod_onu_offset)
            onu_dict_list[onu_dict_list.index(onu_detected)]['coarse_offset'] = offset[1]
            onu_dict_list[onu_dict_list.index(onu_detected)]['fine_offset'] = offset[0]
            ideal_offset += self.c_BURST_LENGTH_ONURXCLK
            self.logger.debug("Offset calculation for ONU " + str(onu_detected['onu_addr']))
            self.logger.debug("Coarse Offset (ONU " + str(onu_detected['onu_addr']) + "): " + str(offset[1]))
            self.logger.debug("Fine   Offset (ONU " + str(onu_detected['onu_addr']) + "): " + str(offset[0]))

        return onu_dict_list

    ### Initialize network procedure ###
    # list_onu_dict      : list of ONU with system previously calibrated (full_calibration)
	# writes to the ONUs the fields 'coarse_offset', 'fine_offset', 'nb_onu_system', 'sfp_en_delay'
    def network_init(self, list_onu_dict):
        onu_dict_list = list_onu_dict
        self.onu_mode_disable(self.c_BROADCAST_ADDR)
        # For each detected ONU
        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        for onu_detected in gen_onu_detected:
           number_onu_detected = onu_detected['nb_onu_system']
           self.onu_coarse_offset(onu_detected['onu_addr'], onu_dict_list[onu_dict_list.index(onu_detected)]['coarse_offset'])
           self.onu_fine_offset(onu_detected['onu_addr'], onu_dict_list[onu_dict_list.index(onu_detected)]['fine_offset'])
           self.onu_heartbeat(onu_detected['onu_addr'], onu_dict_list[onu_dict_list.index(onu_detected)]['nb_onu_system'])
           self.onu_sfp_en_delay(onu_detected['onu_addr'], onu_dict_list[onu_dict_list.index(onu_detected)]['sfp_en_delay'])
           self.onu_mode_operational(onu_detected['onu_addr'])

        self.tdm_init(number_onu_detected)

        return 1

    ### Check system health procedure ###
    # OBS: THIS METHOD BLOCKS USER UPSTREAM TRANSMISSION
    # list_onu_dict      : list of ONU with system previously calibrated (full_calibration)
	# This method returns the list of ONU with the 'prbs7_errors' / 'prbs7_locked' fields updated
    def network_health_check(self, list_onu_dict):
        onu_dict_list = list_onu_dict
        # ONUs detected set as prbs burst mode
        self.logger.info("Starting PRBS analysis ...")
        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        for onu_detected in gen_onu_detected:
            self.onu_mode_prbsburst(onu_detected['onu_addr'])

        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        for onu_detected in gen_onu_detected:
            onu1 = onu_detected
            onu2 = next(gen_onu_detected,None)
            onu1_addr = onu1['onu_addr']
            if (onu2==None):
                onu2_addr = 0
            else:
                onu2_addr = onu2['onu_addr']
            self.bert_config(onu1_addr, onu2_addr)
            self.bert_clear()
            bert_status = [0, 0, 0, 0, 0, 0]
            # while BERT is not finished
            while((not bert_status[0]) or (not bert_status[3])):
                bert_status = self.bert_read() # (onu1_finished,onu1_locked,onu1_errors,onu2_finished,onu2_locked,onu2_errors)

            onu_dict_list[onu_dict_list.index(onu1)]['prbs7_errors'] = bert_status[2]
            onu_dict_list[onu_dict_list.index(onu1)]['prbs7_locked'] = bert_status[1]
            self.logger.info("ONU " + str(onu1_addr) + " - ERRORS:" + str(bert_status[2]) + " - PRBS_LOCKED:" + str(bert_status[1]))
            if (onu2!=None):
                onu_dict_list[onu_dict_list.index(onu2)]['prbs7_errors'] = bert_status[5]
                onu_dict_list[onu_dict_list.index(onu2)]['prbs7_locked'] = bert_status[4]
                self.logger.info("ONU " + str(onu2_addr) + " - ERRORS:" + str(bert_status[5]) + " - PRBS_LOCKED:" + str(bert_status[4]))

        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        for onu_detected in gen_onu_detected:
            self.onu_mode_operational(onu_detected['onu_addr'])

        return onu_dict_list

    ### Check ONU presence procedure ###
    # list_onu_dict      : list of ONU with system previously calibrated (full_calibration)
	# This method returns the list of ONU with the 'detected' field updated
    def network_presence_check(self, list_onu_dict):
        onu_dict_list = list_onu_dict

        self.logger.info('Starting network presence check ...')

        gen_onu_detected = (onu for onu in onu_dict_list if onu['coarse_offset'] != -1)  # Only try to detect ONUs which have been previously initialized

        for onu_detected in gen_onu_detected:
            try_detect = 1
            index_trial = 1
            self.logger.info('OLT trying to detect ONU ADDR ' + str(onu_detected['onu_addr']) + '... ')
            #self.calib_config(0, 1, onu_detected['onu_addr'])
            # try to detect ONU a certain number of times
            while(try_detect and index_trial <= self.c_MAX_TRIALS_ONU_DETECT):
                self.logger.debug('Trial number:' + str(index_trial))
                onu_was_detected = self.detect_onu(onu_detected['onu_addr'], self.c_CONSECUTIVE_SUCCESS_DETECT, self.c_IDLE_BYTE_ADDR)
                if onu_was_detected:
                    self.logger.info('ONU ' + str(onu_detected['onu_addr']) + ' detected' + ' in trial number ' + str(index_trial))
                    try_detect = 0
                index_trial += 1

            if not onu_was_detected:
                self.logger.warn('ONU ' + str(onu_detected['onu_addr']) + ' NOT detected')
                onu_dict_list[onu_dict_list.index(onu_detected)]['detected'] = 0

        return onu_dict_list

    ### Full Calibration Procedure ###
    # Run calibration
    # list_onu_addr      : list containing the ONU ADDRESSES present in the system - ONU address has to be unique per ONU (ex.: [1,2,3,4,5,6,7,8])
    # This method returns a list of dictionary containing the calibration information relative to each ONU in the system
    def full_calibration_run(self, list_onu_addr):

        time_start = time.time()

        # SFP SD EN MASK is deactivated during calibration run		
        self.sfp_sd_dis_mask()

        # Check if all elements in the list are different and in the range 1-254	
        for x in list_onu_addr:
            if (list_onu_addr.count(x) > 1) :
                self.logger.warn("Addresses must be different, finishing calibration")
                return
            elif (x >254 or x < 1) :
                self.logger.warn("Addresses must be in the range 1-254")
                return

        # OLT configurations
        # Optimal offset cancellation		
        self.drp_write(55,0x0000)
        # Get OLT part number
        # The technique for full calibration depends on the OLT module		
        i2c_transaction = self.get_sfp_part_number()
        if i2c_transaction==None: return None
        i2c_success = i2c_transaction[0]		
        if(not i2c_success):
            self.logger.warn('COULD NOT IDENTIFY OLT SFP PART NUMBER')
            sfp_part_number= "?"
        else:
            sfp_part_number = i2c_transaction[1]
            self.logger.info('OLT PART NUMBER:' + sfp_part_number)
        self.c_OLT_SFP_PART_NUMBER = sfp_part_number
        self.c_OLT_SFP_PART_NUMBER='SOGX'
        # Init
        onu_dict_list = []

        # Init ONU dictionary list
        for i in range(0,len(list_onu_addr)):
            onu_dict_list.append({})
            onu_dict_list[i]['onu_addr']         = list_onu_addr[i]
            onu_dict_list[i]['detected']         = -1
            onu_dict_list[i]['rx_pwr']           = -1			
            onu_dict_list[i]['nb_onu_system']    = -1
            onu_dict_list[i]['coarse_roundtrip'] = -1
            onu_dict_list[i]['fine_roundtrip']   = -1
            onu_dict_list[i]['coarse_offset']    = -1
            onu_dict_list[i]['fine_offset']      = -1
            onu_dict_list[i]['sfp_en_delay']     = -1
            onu_dict_list[i]['prbs7_errors']     = -1
            onu_dict_list[i]['prbs7_locked']     = -1

        # Detect ONUs and calculate roundtrip for each
        onu_dict_list = self.full_roundtrip_meas(onu_dict_list, 1)

        # Take as reference ONU the position of the first detected ONU and check number of ONUs in the system
        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        onu_detected = next(gen_onu_detected,None)

        if (onu_detected == None):
            self.logger.info("No ONU was detected, calibration routine will be finished")
            return onu_dict_list

        onu_ref_pos = [onu_detected['fine_roundtrip'], onu_detected['coarse_roundtrip']]

        test_number_onu = 1.4
        # Find OLT SFP rx reset position
        # Configure the system as if 1.4 ONUs are present (125+50) ns, only first detected ONU is used for the positioning
        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        onu_detected = next(gen_onu_detected,{'onu_addr':0})
        ideal_offset = self.c_BURST_LENGTH_ONURXCLK
        mod_onu_offset = int(math.ceil(self.c_BURST_LENGTH_ONURXCLK * test_number_onu))
        self.tdm_init(test_number_onu)
        self.onu_heartbeat(self.c_BROADCAST_ADDR, test_number_onu)

        # Disable all ONUs except first detected
        self.onu_mode_disable(self.c_BROADCAST_ADDR)
        self.onu_mode_prbscont(onu_detected['onu_addr'])

        # Give offset to ONU
        offset = self.calc_offset(onu_ref_pos, [onu_detected['fine_roundtrip'], onu_detected['coarse_roundtrip']], ideal_offset, mod_onu_offset)
        self.onu_coarse_offset(onu_detected['onu_addr'], offset[1])
        self.onu_fine_offset(onu_detected['onu_addr'], offset[0])

        # Find SFP RX reset position
        if(self.c_OLT_SFP_PART_NUMBER[0:4]=='SOGX'):
            self.logger.info('Started first SFP RX reset positioning ...')
            sfp_rx_reset = self.find_sfp_reset_pos_sd(test_number_onu*self.c_BURST_LENGTH_OLTTXCLK, self.c_MARGIN_SFP_RX_RESET_GFOT)		
        else:
            self.logger.info('Started SFP RX reset positioning ...')
            sfp_rx_reset = self.find_sfp_reset_pos_bert(onu_detected['onu_addr'], 0, self.c_MASK_NVALID_SFP_RX_RESET, self.c_MASK_VALID_SFP_RX_RESET, test_number_onu*self.c_BURST_LENGTH_OLTTXCLK, self.c_MARGIN_SFP_RX_RESET)

        if (sfp_rx_reset == -1):
            self.logger.warn("No position found for SFP RX reset, calibration routine will be finished")
            return onu_dict_list

        # Enable SFP RX RESET	
        self.sfp_reset(1, sfp_rx_reset,test_number_onu*self.c_BURST_LENGTH_OLTTXCLK)

        # Find ONU SFP enable delay
        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        for onu_detected in gen_onu_detected:
            # disable all ONUs expect the one being calibrated
            self.onu_mode_disable(self.c_BROADCAST_ADDR)
            self.onu_mode_prbsburst(onu_detected['onu_addr'])
            offset = self.calc_offset(onu_ref_pos, [onu_detected['fine_roundtrip'], onu_detected['coarse_roundtrip']], ideal_offset, mod_onu_offset)
            self.onu_coarse_offset(onu_detected['onu_addr'], offset[1])
            self.onu_fine_offset(onu_detected['onu_addr'], offset[0])

            # Find SFP EN delay for ONU
            self.logger.info('Started SFP enable delay positioning for ONU ' + str(onu_detected['onu_addr']))	
            sfp_delay = self.find_onu_sfp_delay(onu_detected['onu_addr'], self.c_MARGIN_SFP_EN_DELAY)
            onu_dict_list[onu_dict_list.index(onu_detected)]['sfp_en_delay'] = sfp_delay

		# Update Network Offset
        onu_dict_list = self.network_update_offset(onu_dict_list, 1)

        # Network init
        self.network_init(onu_dict_list)
        self.onu_mode_prbsburst(self.c_BROADCAST_ADDR)

        # Find SFP RX reset position
        self.logger.info('Started SFP RX reset positioning ...')		
        if(self.c_OLT_SFP_PART_NUMBER[0:4]=='SOGX'):
            sfp_rx_reset = self.find_sfp_reset_pos_sd(self.c_BURST_LENGTH_OLTTXCLK, self.c_MARGIN_SFP_RX_RESET_GFOT)
        else:
            gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
            onu_detected1 = next(gen_onu_detected,None)
            onu_detected2 = next(gen_onu_detected,None)
            if(onu_detected2==None) : sfp_rx_reset = self.find_sfp_reset_pos_bert(onu_detected1['onu_addr'], 0, self.c_MASK_NVALID_SFP_RX_RESET, self.c_MASK_VALID_SFP_RX_RESET, self.c_BURST_LENGTH_OLTTXCLK, self.c_MARGIN_SFP_RX_RESET)
            else : sfp_rx_reset = self.find_sfp_reset_pos_bert(onu_detected1['onu_addr'], onu_detected2['onu_addr'], self.c_MASK_NVALID_SFP_RX_RESET, self.c_MASK_VALID_SFP_RX_RESET, self.c_BURST_LENGTH_OLTTXCLK, self.c_MARGIN_SFP_RX_RESET)
        if (sfp_rx_reset == -1):
            self.logger.warn("No position found for SFP RX reset, calibration routine will be finished")
            return onu_dict_list			
        self.sfp_reset(1, sfp_rx_reset)

        # Find SFP_SD_BLIND_WINDOW
        self.logger.info('Started SFP SD EN MASK positioning ...')		
        sfp_sd_en_mask_pos = self.find_sfp_sd_en_mask_pos()
        if (sfp_sd_en_mask_pos == -1):
            self.logger.warn("No position found for SFP EN MASK, mask will not be activated")
            self.sfp_sd_dis_mask()
        else:
            self.logger.info('SFP SD EN MASK positioned at ' + str(sfp_sd_en_mask_pos))		
            self.sfp_sd_en_mask()		
            self.sfp_sd_en_mask_delay(sfp_sd_en_mask_pos)

        # Network Health Check
        onu_dict_list = self.network_health_check(onu_dict_list)

        self.logger.info("Full calibration done")

        time_stop = time.time()
        self.logger.info('Time elapsed: ' + str(time_stop-time_start) + ' s')

        return onu_dict_list

    ### Light Calibration Procedure ###
    # list_onu_dict      : list of ONU with system previously calibrated (full_calibration)
    # This method returns a list of dictionary containing the calibration information relative to each ONU in the system
    def light_calibration_run(self, list_onu_dict):

        time_start = time.time()

        # Save sfp reset state
        sfp_reset_stat = self.get_sfp_reset()

        # Detect ONUs which have been previously calibrated and calculate roundtrip for each
        onu_dict_list = self.full_roundtrip_meas(list_onu_dict, 0)

        # Check if at least one ONU was detected
        gen_onu_detected = (onu for onu in onu_dict_list if onu['detected'] == 1)
        onu_detected = next(gen_onu_detected,None)

        if (onu_detected == None):
            self.logger.info("No ONU was detected, calibration routine will be finished")
            return onu_dict_list

        # Update Network Offset
        onu_dict_list = self.network_update_offset(onu_dict_list, 0)

        # Set SFP RX RESET
        self.sfp_reset(1, sfp_reset_stat[1])

        # Network init
        self.network_init(onu_dict_list)

        # Network Health Check
        onu_dict_list = self.network_health_check(onu_dict_list)

        self.logger.info("Light calibration done")

        time_stop = time.time()
        self.logger.info('Time elapsed: ' + str(time_stop-time_start) + ' s')

        return onu_dict_list

    ### Clear olt monitoring ###
    # Clear all sticky status and error correction/detection counters	
    def clear_olt_monitor_status(self):
        self.errmon_clear()
        self.phymon_clear()
        return 1

    ### Full OLT Monitoring Procedure ###	
    def full_olt_monitor_run(self):
        olt_dict = {}	
        olt_dict['time_stamp']           = time.time()
        olt_dict['sfp_temp']             = float('nan')
        olt_dict['sfp_vcc']              = float('nan')
        olt_dict['sfp_tx_bias']          = float('nan')
        olt_dict['sfp_tx_pwr']           = float('nan')
        olt_dict['sfp_rx_pwr']           = float('nan')
        olt_dict['stk_mgt_tx_nready']    = float('nan')
        olt_dict['stk_mgt_rx_nready']    = float('nan')
        olt_dict['stk_mgt_pll_nlock']    = float('nan')
        olt_dict['stk_missing_burst']    = float('nan')

        # SFP Monitor
        sfp_temp_conv = self.get_sfp_temp()	
        if(sfp_temp_conv[0] != 0) : olt_dict['sfp_temp'] = sfp_temp_conv[1]

        sfp_vcc_conv = self.get_sfp_vcc()	
        if(sfp_vcc_conv[0] != 0) : olt_dict['sfp_vcc'] = sfp_vcc_conv[1]

        sfp_tx_bias_conv = self.get_sfp_tx_bias()
        if(sfp_tx_bias_conv[0] != 0) : olt_dict['sfp_tx_bias'] = sfp_tx_bias_conv[1]

        sfp_tx_pwr_conv = self.get_sfp_tx_pwr()
        if(sfp_tx_pwr_conv[0] != 0) : olt_dict['sfp_tx_pwr'] = sfp_tx_pwr_conv[1]

        sfp_rx_pwr_conv = self.get_sfp_rx_pwr()
        if(sfp_rx_pwr_conv[0] != 0) : olt_dict['sfp_rx_pwr'] = sfp_rx_pwr_conv[1]

        # Phy Sticky Status
        [olt_seen_missing_burst, olt_seen_mgt_tx_nready, olt_seen_mgt_rx_nready, olt_seen_mgt_pll_nlock] = self.phymon_read_seen()
        olt_dict['stk_mgt_tx_nready']   = olt_seen_mgt_tx_nready
        olt_dict['stk_mgt_rx_nready']   = olt_seen_mgt_rx_nready
        olt_dict['stk_mgt_pll_nlock']   = olt_seen_mgt_pll_nlock
        olt_dict['stk_missing_burst']   = olt_seen_missing_burst

        return olt_dict

    ### Clear onu monitoring ###
    # Clear all sticky status and error correction/detection counters	
    def clear_onu_monitor_status(self,onu_addr):
        onu_errmon_comm_success = self.onu_errmon_clear(onu_addr)
        onu_phymon_comm_success = self.onu_phymon_clear(onu_addr)
        return (onu_errmon_comm_success & onu_phymon_comm_success)

    ### Full ONU Monitoring Procedure ###	
    def full_onu_monitor_run(self,onu_addr):
        onu_dict = {}	
        onu_dict['onu_addr']             = onu_addr
        onu_dict['time_stamp']           = time.time()
        onu_dict['sfp_temp']             = float('nan')
        onu_dict['sfp_vcc']              = float('nan')
        onu_dict['sfp_tx_bias']          = float('nan')
        onu_dict['sfp_tx_pwr']           = float('nan')
        onu_dict['sfp_rx_pwr']           = float('nan')
        onu_dict['stk_mgt_tx_nready']    = float('nan')
        onu_dict['stk_mgt_rx_nready']    = float('nan')
        onu_dict['stk_mgt_txpll_nlock']  = float('nan')
        onu_dict['stk_mgt_rxpll_nlock']  = float('nan')
        onu_dict['stk_rx_nlock']         = float('nan')
        onu_dict['stk_phase_ngood']      = float('nan')
        onu_dict['stk_dn_fec_serr_corr'] = float('nan')
        onu_dict['stk_dn_fec_derr_corr'] = float('nan')
        onu_dict['stk_dn_sc_crc_errdet'] = float('nan')
        onu_dict['dn_fec_serr_corr_cnt'] = float('nan')
        onu_dict['dn_fec_derr_corr_cnt'] = float('nan')
        onu_dict['dn_sc_crc_errdet_cnt'] = float('nan')
        onu_dict['stk_up_8b10b_errdet']  = float('nan')
        onu_dict['burst_pos']            = float('nan')

        # Config for RT measurement
        self.calib_config(0,0,onu_addr)
        [fine, coarse] = self.measpos_onu(3,3)

        # SFP Monitor
        sfp_temp_conv = self.onu_get_sfp_temp(onu_addr)	
        if(sfp_temp_conv[0] != 0) : onu_dict['sfp_temp'] = sfp_temp_conv[1]

        sfp_vcc_conv = self.onu_get_sfp_vcc(onu_addr)	
        if(sfp_vcc_conv[0] != 0) : onu_dict['sfp_vcc'] = sfp_vcc_conv[1]

        sfp_tx_bias_conv = self.onu_get_sfp_tx_bias(onu_addr)
        if(sfp_tx_bias_conv[0] != 0) : onu_dict['sfp_tx_bias'] = sfp_tx_bias_conv[1]

        sfp_tx_pwr_conv = self.onu_get_sfp_tx_pwr(onu_addr)
        if(sfp_tx_pwr_conv[0] != 0) : onu_dict['sfp_tx_pwr'] = sfp_tx_pwr_conv[1]

        sfp_rx_pwr_conv = self.onu_get_sfp_rx_pwr(onu_addr)
        if(sfp_rx_pwr_conv[0] != 0) : onu_dict['sfp_rx_pwr'] = sfp_rx_pwr_conv[1]

        # Phy Sticky Status
        [comm_success, onu_seen_mgt_tx_nready, onu_seen_mgt_rx_nready, onu_seen_mgt_txpll_nlock, onu_seen_mgt_rxpll_nlock, onu_seen_rx_nlocked, onu_seen_phase_ngood] = self.onu_phymon_read_seen(onu_addr)
        if(comm_success) : 
            onu_dict['stk_mgt_tx_nready']   = onu_seen_mgt_tx_nready
            onu_dict['stk_mgt_rx_nready']   = onu_seen_mgt_rx_nready
            onu_dict['stk_mgt_txpll_nlock'] = onu_seen_mgt_txpll_nlock
            onu_dict['stk_mgt_rxpll_nlock'] = onu_seen_mgt_rxpll_nlock
            onu_dict['stk_rx_nlock']        = onu_seen_rx_nlocked
            onu_dict['stk_phase_ngood']     = onu_seen_phase_ngood

        # Downstream Error Monitoring
        [comm_success, onu_seen_serror_fec, onu_seen_derror_fec, onu_seen_sc_crc_errdet] = self.onu_errmon_read_seen_error(onu_addr)
        if(comm_success) : 
            onu_dict['stk_dn_fec_serr_corr'] = onu_seen_serror_fec
            onu_dict['stk_dn_fec_derr_corr'] = onu_seen_derror_fec
            onu_dict['stk_dn_sc_crc_errdet'] = onu_seen_sc_crc_errdet

        [comm_success, onu_serror_fec_cntr, onu_derror_fec_cntr, onu_sc_crc_errdet_cntr] = self.onu_errmon_read(onu_addr)
        if(comm_success) :
            onu_dict['dn_fec_serr_corr_cnt'] = onu_serror_fec_cntr
            onu_dict['dn_fec_derr_corr_cnt'] = onu_derror_fec_cntr
            onu_dict['dn_sc_crc_errdet_cnt'] = onu_sc_crc_errdet_cntr

        # Upstream Error Monitoring
        up_error_list = self.errmon_read_seen_error()
        if(up_error_list[0]):        
            if(onu_addr in up_error_list[1:]) : onu_dict['stk_up_8b10b_errdet'] = 1
            else : onu_dict['stk_up_8b10b_errdet'] = 0
        else: onu_dict['stk_up_8b10b_errdet'] = 0

        [fine, coarse] = self.measpos_onu(3,3)
        if(fine!=0 or coarse!=0):
            onu_dict['burst_pos'] = coarse * self.c_COARSE_ROUNDTRIP_UNIT + fine * self.c_FINE_ROUNDTRIP_UNIT

        return onu_dict
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
    # Save OLT configuration in .csv file
    def save_olt_config_csv(self, name_file):
        with open(name_file + '.csv', 'w') as csvfile:
            fieldnames = ['olt_reg_addr', 'olt_reg_value']
            fieldnames_formatted = []

            list_olt_regbank = []

            # format 20 spaces for each value
            for i in range(0, len(fieldnames)):
                fieldnames_formatted.append('%20s' % fieldnames[i])

            for i in range(0,self.c_NUMBER_OLT_REGS):
                list_olt_regbank.append({})
                list_olt_regbank[-1][fieldnames_formatted[0]] = '%20d' % i
                list_olt_regbank[-1][fieldnames_formatted[1]] = '%20d' % self.read_ctrl_reg(i)

            writer = csv.DictWriter(
                csvfile,
                fieldnames=fieldnames_formatted,
                delimiter=',',
                lineterminator='\n')
            writer.writeheader()

            for olt_dict in list_olt_regbank:
                writer.writerow(olt_dict)

        return 1

    # Load OLT configuration from .csv file and write to OLT
    def load_olt_config_csv(self, name_file):

        list_olt_regbank = []

        with open(name_file + '.csv') as csvfile:
            fieldnames = ['olt_reg_addr', 'olt_reg_value']

            reader = csv.DictReader(csvfile)
            for row in reader:
                list_olt_regbank.append({})
                for i in range(0,len(fieldnames)):
                    list_olt_regbank[-1][fieldnames[i]] = int(row['%20s' % fieldnames[i]])

        for i in range(0,len(list_olt_regbank)):
            self.write_ctrl_reg(list_olt_regbank[i][fieldnames[0]], list_olt_regbank[i][fieldnames[1]])

        return 1

    # Save ONU configuration to .csv file
    @staticmethod			
    def save_onu_config_csv(list_onu_dict, name_file):
        with open(name_file + '.csv', 'w') as csvfile:	
            fieldnames = ['onu_addr', 'detected','rx_pwr', 'nb_onu_system', 'coarse_roundtrip', 'fine_roundtrip', 'coarse_offset', 'fine_offset', 'sfp_en_delay', 'prbs7_errors', 'prbs7_locked']
            fieldnames_formatted = []

            list_onu_dict_formatted = []
            for j in range(0,len(list_onu_dict)) : list_onu_dict_formatted.append({})

    		# format 20 spaces for each value
            for i in range(0, len(fieldnames)):
                fieldnames_formatted.append('%20s' % fieldnames[i])
                for j in range(0,len(list_onu_dict)):
                    if(fieldnames[i]=='rx_pwr'): list_onu_dict_formatted[j][fieldnames_formatted[i]] = '%20.4f' % list_onu_dict[j][fieldnames[i]]
                    else                       : list_onu_dict_formatted[j][fieldnames_formatted[i]] = '%20d' % list_onu_dict[j][fieldnames[i]]
            writer = csv.DictWriter(
                csvfile,
                fieldnames=fieldnames_formatted,
                delimiter=',',
                lineterminator='\n')
            writer.writeheader()

            for onu_dict in list_onu_dict_formatted:
                writer.writerow(onu_dict)

        return 1

    # Read ONU configuration from .csv file and returns a list of dictionaries (which can be used to run a light calibration)
    # This method returns a list of ONU dictionaries read from the .csv file
    @staticmethod    
    def read_onu_config_csv(name_file):

        list_onu_dict = []

        with open(name_file + '.csv') as csvfile:
            fieldnames = ['onu_addr', 'detected','rx_pwr', 'nb_onu_system', 'coarse_roundtrip', 'fine_roundtrip', 'coarse_offset', 'fine_offset', 'sfp_en_delay', 'prbs7_errors', 'prbs7_locked']

            reader = csv.DictReader(csvfile)
            for row in reader:
                list_onu_dict.append({})
                for i in range(0,len(fieldnames)):
                    if(fieldnames[i]=='rx_pwr') : list_onu_dict[-1][fieldnames[i]] = float(row['%20s' % fieldnames[i]])				
                    else                        : list_onu_dict[-1][fieldnames[i]] = int(row['%20s' % fieldnames[i]])

        return list_onu_dict

    # Save OLT monitor dict to .csv file
    @staticmethod			
    def save_olt_monitor_csv(list_olt_dict, name_file):
        file_exists = os.path.isfile(name_file + '.csv')	
        with open(name_file + '.csv', 'a') as csvfile:
            fieldnames = ['time_stamp', 'sfp_temp', 'sfp_vcc', 'sfp_tx_bias', 'sfp_tx_pwr', 'sfp_rx_pwr', 'stk_mgt_tx_nready', 'stk_mgt_rx_nready', 'stk_mgt_pll_nlock', 'stk_missing_burst']
            fieldnames_formatted = []

            list_olt_dict_formatted = {}

    		# format 20 spaces for each value
            for i in range(0, len(fieldnames)):
                fieldnames_formatted.append('%20s' % fieldnames[i])
                if(type(list_olt_dict[fieldnames[i]]) is float) : list_olt_dict_formatted[fieldnames_formatted[i]] = '%20.4f' % list_olt_dict[fieldnames[i]]
                elif(type(list_olt_dict[fieldnames[i]]) is int) : list_olt_dict_formatted[fieldnames_formatted[i]] = '%20d' % list_olt_dict[fieldnames[i]]
                else : list_olt_dict_formatted[fieldnames_formatted[i]] = '%20s' % list_olt_dict[fieldnames[i]]

            writer = csv.DictWriter(
                csvfile,
                fieldnames=fieldnames_formatted,
                delimiter=',',
                lineterminator='\n')

            if (not file_exists) : writer.writeheader()

            writer.writerow(list_olt_dict_formatted)

        return 1

    # Read OLT monitor from .csv file and returns a list of dictionaries
    # This method returns a list of OLT dictionaries read from the .csv file
    @staticmethod	
    def read_olt_monitor_csv(name_file):

        list_olt_dict = []

        with open(name_file + '.csv') as csvfile:
            fieldnames = ['time_stamp', 'sfp_temp', 'sfp_vcc', 'sfp_tx_bias', 'sfp_tx_pwr', 'sfp_rx_pwr', 'stk_mgt_tx_nready', 'stk_mgt_rx_nready', 'stk_mgt_pll_nlock', 'stk_missing_burst']

            reader = csv.DictReader(csvfile)
            for row in reader:
                list_olt_dict.append({})
                for i in range(0,len(fieldnames)):
                    list_olt_dict[-1][fieldnames[i]] = float(row['%20s' % fieldnames[i]])

        return list_olt_dict

    # Save ONU monitor dict to .csv file
    @staticmethod			
    def save_onu_monitor_csv(list_onu_dict, name_file):
        file_exists = os.path.isfile(name_file + '.csv')	
        with open(name_file + '.csv', 'a') as csvfile:
            fieldnames = ['onu_addr', 'time_stamp', 'sfp_temp', 'sfp_vcc', 'sfp_tx_bias', 'sfp_tx_pwr', 'sfp_rx_pwr', 'stk_mgt_tx_nready', 'stk_mgt_rx_nready', 'stk_mgt_txpll_nlock',
                          'stk_mgt_rxpll_nlock', 'stk_rx_nlock', 'stk_phase_ngood', 'stk_dn_fec_serr_corr', 'stk_dn_fec_derr_corr', 'stk_dn_sc_crc_errdet', 'dn_fec_serr_corr_cnt',
                          'dn_fec_derr_corr_cnt', 'dn_sc_crc_errdet_cnt', 'stk_up_8b10b_errdet', 'burst_pos']
            fieldnames_formatted = []

            list_onu_dict_formatted = {}

    		# format 20 spaces for each value
            for i in range(0, len(fieldnames)):
                fieldnames_formatted.append('%20s' % fieldnames[i])
                if(type(list_onu_dict[fieldnames[i]]) is float) : list_onu_dict_formatted[fieldnames_formatted[i]] = '%20.4f' % list_onu_dict[fieldnames[i]]
                elif(type(list_onu_dict[fieldnames[i]]) is int) : list_onu_dict_formatted[fieldnames_formatted[i]] = '%20d' % list_onu_dict[fieldnames[i]]
                else : list_onu_dict_formatted[fieldnames_formatted[i]] = '%20s' % list_onu_dict[fieldnames[i]]
            writer = csv.DictWriter(
                csvfile,
                fieldnames=fieldnames_formatted,
                delimiter=',',
                lineterminator='\n')

            if (not file_exists) : writer.writeheader()

            writer.writerow(list_onu_dict_formatted)
			
        return 1

    # Read ONU monitor from .csv file and returns a list of dictionaries
    # This method returns a list of ONU dictionaries read from the .csv file
    @staticmethod    
    def read_onu_monitor_csv(name_file):

        list_onu_dict = []

        with open(name_file + '.csv') as csvfile:
            fieldnames = ['onu_addr', 'time_stamp', 'sfp_temp', 'sfp_vcc', 'sfp_tx_bias', 'sfp_tx_pwr', 'sfp_rx_pwr', 'stk_mgt_tx_nready', 'stk_mgt_rx_nready', 'stk_mgt_txpll_nlock',
                          'stk_mgt_rxpll_nlock', 'stk_rx_nlock', 'stk_phase_ngood', 'stk_dn_fec_serr_corr', 'stk_dn_fec_derr_corr', 'stk_dn_sc_crc_errdet', 'dn_fec_serr_corr_cnt',
                          'dn_fec_derr_corr_cnt', 'dn_sc_crc_errdet_cnt', 'stk_up_8b10b_errdet', 'burst_pos']

            reader = csv.DictReader(csvfile)
            for row in reader:
                list_onu_dict.append({})
                for i in range(0,len(fieldnames)):
                    list_onu_dict[-1][fieldnames[i]] = float(row['%20s' % fieldnames[i]])

        return list_onu_dict


