#!/usr/bin/python3

import os
import socket
import logging
import datetime
import time
import csv
import math
from random import randint

from . driver_comm import DriverComm
from ttcpon_board_config.pon_fmc_i2c_program import get_pll_reg

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3


class OnuCore(DriverComm):

    def __init__(self, ip='192.168.1.11', port=9001, onun=None):
        self.onun= onun
        DriverComm.__init__(self, ip, port)
        #self.logger.warn("OnuCore...")
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

    ### I2C CTRL (begin) ###
    def i2c_write(self, slv_addr, reg_addr, reg_value):
        if self.conver!="ublaze":
          #print "onu_%d.i2c_write: not implemented"%self.onun
          #return [-1]
          pass #print "DBG onu_%d.i2c_write 0x%x 0x%x val:0x%x"%(self.onun,slv_addr, reg_addr, reg_value)
        slv_addr_format  = slv_addr & 0x7F
        reg_addr_format  = reg_addr & 0xFF
        reg_value_format = reg_value & 0xFF
        if((slv_addr != slv_addr_format) | (reg_addr != reg_addr_format) | (reg_value != reg_value_format)):
            self.logger.warn('Error parameters formatting for IIC write command')		
            return -1
        else:
            message =(0<<31)+(reg_value<<18) + (reg_addr<<10) + (slv_addr<<3) + (0<<1) + (0<<1) + (1)			
            self.write_ctrl_reg(13, message)
            message =(1<<31)+(reg_value<<18) + (reg_addr<<10) + (slv_addr<<3) + (0<<1) + (0<<1) + (1)
            self.write_ctrl_reg(13, message)

        i2c_done = 0
        i2c_error = 0
        i2c_drop_req = 0
        while(i2c_done==0 and i2c_error==0 and i2c_drop_req==0):
            i2c_stat     = self.read_ctrl_reg(14)
            i2c_done     = (i2c_stat & 0x00000001) >> 0
            i2c_error    = (i2c_stat & 0x00000002) >> 1
            i2c_drop_req = (i2c_stat & 0x00000004) >> 2

        if(i2c_error):
            self.logger.warn('IIC write error')				
            return 0
        elif(i2c_drop_req):
            self.logger.warn('IIC drop request error')				
            return 0			
        else:
            return 1

    # Read IIC (nb_read_access can be 1 or 2 for double bursty read)
    def i2c_read(self, slv_addr, reg_addr, nb_read_access=1):
        if self.conver!="ublaze":
          #print "onu_%d.i2c_read: not implemented"%self.onun
          #return [-1]
          pass #print "DBG onu_%d.i2c_read  0x%x 0x%x bytes:0x%x"%(self.onun,slv_addr, reg_addr, nb_read_access)
        slv_addr_format  = slv_addr & 0x7F;
        reg_addr_format  = reg_addr & 0xFF;
        nb_read_access   = nb_read_access - 1
        nb_read_access_format = nb_read_access & 0x1

        if((slv_addr != slv_addr_format) | (reg_addr != reg_addr_format) | (nb_read_access_format != nb_read_access)):
            self.logger.warn('Error parameters formatting for IIC read command')		
            return [-1]
        else:
            message = (0<<31)+(0<<18) + (reg_addr<<10) + (slv_addr<<3) + (1<<2) + (nb_read_access<<1) + (1)	
            self.write_ctrl_reg(13, message)		
            message = (1<<31)+(0<<18) + (reg_addr<<10) + (slv_addr<<3) + (1<<2) + (nb_read_access<<1) + (1)	
            self.write_ctrl_reg(13, message)

        i2c_done = 0
        i2c_error = 0
        i2c_drop_req = 0 ; i2c_hangs= 0; rattempts= 0;
        while(i2c_done==0 and i2c_error==0 and i2c_drop_req==0):
            i2c_stat  = self.read_ctrl_reg(14)
            i2c_done     = (i2c_stat & 0x00000001) >> 0
            i2c_error    = (i2c_stat & 0x00000002) >> 1
            i2c_drop_req = (i2c_stat & 0x00000004) >> 2
            i2c_data     = (i2c_stat & 0x0007FFF8) >> 3			
            rattempts= rattempts +1
            if (i2c_stat & 0x7)==0:
              pass #print "i2c_read:%d: 0x%x"%(rattempts, i2c_stat)
            if rattempts>10:
              i2c_hangs=1 ; break
        if(i2c_error):
            self.logger.warn('IIC read error')				
            returnrc= [-1]
        elif(i2c_hangs):
            self.logger.warn('IIC read hangs, stat: 0x%x'%i2c_stat)				
            returnrc= [-1]
        elif(i2c_drop_req):
            self.logger.warn('IIC drop request error')
            returnrc= [-1]			
        else:
            if(nb_read_access > 0):		
                data_return_list = []        		
                for i in range(0,(nb_read_access+1)):
                    data  = (i2c_data >> i*8) & 0xFF
                    data_return_list.append(data)
                returnrc= data_return_list
            else:
                returnrc= [i2c_data&0xFF]
        #print "DBG2 onu_%d.i2c_read  0x%x 0x%x bytes:0x%x"%(self.onun,slv_addr, reg_addr, nb_read_access),returnrc
        return returnrc
    def get_pll_reg(self, reg_addr):
        if self.conver=="ublaze":
          reg_value= get_pll_reg(self,reg_addr)
        else:
          reg_value= self.ctrlcon.plldev.rreg(reg_addr>>8, reg_addr&0xff)
        print "0x%x: 0x%x"%(reg_addr, reg_value)

    ### DRP CTRL (begin) ###
    def drp_write(self, addr, value):
        rdy=0	
        while(rdy==0):
            drpen=1
            drpwe=1
            value = (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + ((value&0x0000ffff)<<0)
            self.write_ctrl_reg(17,value)
            value = (1<<31) + (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + ((value&0x0000ffff)<<0)
            self.write_ctrl_reg(17,value)

            stat = self.read_ctrl_reg(17)

            value =  (stat&0x0000ffff) >> 0;
            addr  =  (stat&0x01ff0000) >> 16;
            rdy   =  (stat&0x04000000) >> 26;

        return rdy

    def drp_read(self, addr):
        rdy=0	
        while(rdy==0):
            drpen=1
            drpwe=0
            value = (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + 0
            self.write_ctrl_reg(17,value)			
            value = (1<<31) + (drpen<<26) + (drpwe<<25) + ((addr&0x000001ff)<<16) + 0
            self.write_ctrl_reg(17,value)

            stat = self.read_ctrl_reg(17)

            value =  (stat&0x0000ffff) >> 0;
            addr  =  (stat&0x01ff0000) >> 16;	
            rdy   =  (stat&0x04000000) >> 26;
        return [addr, value, rdy]

   ### Phy Error Monitoring ###
    def phymon_clear(self):
        stat = self.read_ctrl_reg(5)
        stat = stat | 1
        self.write_ctrl_reg(5, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(5, stat)
        return 1

    # It returns a list of six elements
    # [seen_mgt_tx_nready, seen_mgt_rx_nready, seen_mgt_txpll_nlock, seen_mgt_rxpll_nlock, seen_rx_nlocked, seen_phase_ngood]
    def phymon_read_seen(self):
        stat = self.read_ctrl_reg(5)
        seen_mgt_tx_nready      = (stat>>8)&0x1
        seen_mgt_rx_nready      = (stat>>9)&0x1
        seen_mgt_txpll_nlock    = (stat>>10)&0x1
        seen_mgt_rxpll_nlock    = (stat>>11)&0x1
        seen_rx_nlocked         = (stat>>12)&0x1
        seen_phase_ngood        = (stat>>13)&0x1				
        # Return value
        mon_return = [seen_mgt_tx_nready, seen_mgt_rx_nready, seen_mgt_txpll_nlock, seen_mgt_rxpll_nlock, seen_rx_nlocked, seen_phase_ngood]
        return mon_return

    ###### Error Monitor on-the-fly based on FEC correction capability and SC-CRC-7 error detection ######
    def errmon_clear(self):
        stat = self.read_ctrl_reg(3)
        stat = stat | 1
        self.write_ctrl_reg(3, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(3, stat)
        return 1

    # It returns a list of three elements
    # [single_error_corrected_fec, double_error_corrected_fec, slow_control_crc_err_detect]
    def errmon_read(self):
        # Latch
        stat = self.read_ctrl_reg(3)
        stat = stat | (1 << 1)
        self.write_ctrl_reg(3, stat)
        stat = stat ^ (1 << 1)
        self.write_ctrl_reg(3, stat)
        # Read status
        stat = self.read_ctrl_reg(3)
        serror_fec    = (stat>>8)&0xff
        derror_fec    = (stat>>16)&0xff
        sc_crc_errdet = (stat>>24)&0xff		
        # Return value
        mon_return = [serror_fec, derror_fec, sc_crc_errdet]
        return mon_return

    # It returns a list of three elements
    # [seen_single_error_corrected, seen_double_error_corrected, seen_crc_err_detect]
    def errmon_read_seen_error(self):
        # Latch
        stat = self.read_ctrl_reg(4)
        serror_fec    = (stat>>0)&0x1
        derror_fec    = (stat>>1)&0x1
        sc_crc_errdet = (stat>>2)&0x1		
        # Return value
        mon_return = [serror_fec, derror_fec, sc_crc_errdet]
        return mon_return

	# ONU Interrupt
    # Interrupt mask is a list of the first N elements of the interrupt mask vector (N<=31 always)
    #interrupt_async <= (interrupt_mask(0) and mgt_seen_tx_nready_mngrsync_meta)    or
    #                   (interrupt_mask(1) and mgt_seen_rx_nready_mngrsync_meta)    or
    #                   (interrupt_mask(2) and mgt_seen_txpll_nlock_mngrsync_meta)  or
    #                   (interrupt_mask(3) and mgt_seen_rxpll_nlock_mngrsync_meta)  or
    #                   (interrupt_mask(4) and seen_rx_nlocked_mngrsync_meta)       or
    #                   (interrupt_mask(5) and seen_phase_ngood_mngrsync_meta)      or
    #                   (interrupt_mask(6) and rx_seen_serror_correc)               or					  
    #                   (interrupt_mask(7) and rx_seen_derror_correc)               or	
    #                   (interrupt_mask(8) and rx_seen_crc_errdetect);
    # Ex. set an interrupt with rx_seen_derror_correc or rx_seen_crc_errdetect or seen_rx_nlocked:
    # onu.set_interrupt_mask([0 0 0 0 1 0 0 1 1])
    # If any value is not given to the interrupt, it assumes 0
    def set_interrupt_mask(self, interrupt_mask):
        message = 0
        for i in range(0,len(interrupt_mask)):
            message = message | (interrupt_mask[i]<<i)
        self.write_ctrl_reg(6, message)
        return 1

    # it returns a list with 31 elements
    def get_interrupt_mask(self):
        interrupt_mask = [0] * 31
        stat = self.read_ctrl_reg(6)
        for i in range(0,31):
            interrupt_mask[i] = ((stat>>i)&0x1)
        return interrupt_mask

    # it returns 0=interrupt not activated/1=interrupt activated
    def get_interrupt(self):
        interrupt = 0
        stat = self.read_ctrl_reg(6)
        interrupt = (stat>>31)&0x1
        return interrupt

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3

    # SFP related methods
    def get_sfp_vendor(self):
        slv_addr = 0x50	
        reg_bank = []
        for reg_addr in range(20,36,2):
            reg_value = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]			
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)
        vendor_name = ''.join(chr(i) for i in reg_bank)
        return [1, vendor_name]

    def get_sfp_part_number(self):
        slv_addr = 0x50		
        reg_bank = []
        for reg_addr in range(40,56,2):
            reg_value = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]			
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)
        vendor_part_number = ''.join(chr(i) for i in reg_bank)
        return [1, vendor_part_number]

    def get_sfp_serial_number(self):
        slv_addr = 0x50		
        reg_bank = []
        for reg_addr in range(68,84,2):
            reg_value = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
            reg_value1 = reg_value[0]
            if(reg_value1==-1):
                self.logger.warn('Failed to read from SFP')
                return [0]
            reg_value2 = reg_value[1]
            reg_bank.append(reg_value1)
            reg_bank.append(reg_value2)
        vendor_serial_number = ''.join(chr(i) for i in reg_bank)
        return [1, vendor_serial_number]

    def get_sfp_temp(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 96
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        temp    = (reg_bank[0]<<8) + reg_bank[1]

        # TEMPERATURE:
        # 1) Internally measured transceiver temperature. Represented as a 16 bit signed twos complement value in
        # increments of 1/256 degrees Celsius
        if(temp&0x8000):
            sign = -1
            temp_conv = (temp ^ 0xFFFF) + 1
        else:
            sign = 1
            temp_conv = temp
        temp_conv = sign*temp_conv*(1.0/256.0) # UNIT: degrees Celsius

        return [1, temp_conv]

    def get_sfp_vcc(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 98
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        vcc     = (reg_bank[0]<<8) + reg_bank[1]

        # SUPPLY VOLTAGE:
        # 2) Internally measured transceiver supply voltage. Represented as a 16 bit unsigned integer with the
        # voltage defined as the full 16 bit value (0-65535) with LSB equal to 100 uVolt
        vcc_conv = (vcc*100e-6) # UNIT: V

        return [1, vcc_conv]

    def get_sfp_tx_bias(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 100
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        tx_bias = (reg_bank[0]<<8) + reg_bank[1]	

		# TX BIAS CURRENT:
        # 3) Measured TX bias current in uA. Represented as a 16 bit unsigned integer with the current defined as the
        # full 16 bit value (0-65535) with LSB equal to 2 uA
        tx_bias_conv = (tx_bias*2e-6)*1e3 # UNIT: mA

        return [1, tx_bias_conv]
		
    def get_sfp_tx_pwr(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 102
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        tx_pwr  = (reg_bank[0]<<8) + reg_bank[1]

		# TX OUTPUT POWER:
        # 4) Measured TX output power in mW. Represented as a 16 bit unsigned integer with the power defined as
        # the full 16 bit value (0-65535) with LSB equal to 0.1 uW
        tx_pwr_conv = 10*math.log10((tx_pwr * 0.1 * 1e-6)/1e-3) # UNIT: dBm

        return [1, tx_pwr_conv]

    def get_sfp_rx_pwr(self):
        slv_addr  = 0x51
        # Read digital diagnostic
        reg_bank = []
        reg_addr = 104
        reg_value  = self.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            self.logger.warn('Failed to read from ONU - SFP')
            return [0]
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

        rx_pwr  = (reg_bank[0]<<8) + reg_bank[1]

		# RX RECEIVED OPTICAL POWER:
        # 5) Measured RX received optical power in mW. Value can represent either average received power or OMA
        # depending upon how bit 3 of byte 92 (A0h) is set. Represented as a 16 bit unsigned integer with the
        # power defined as the full 16 bit value (0-65535) with LSB equal to 0.1 uW
        if(rx_pwr>0):
            rx_pwr_conv = 10*math.log10((rx_pwr * 0.1 * 1e-6)/1e-3) # UNIT: dBm
        else:
            rx_pwr_conv = -1*float('inf')

        return [1, rx_pwr_conv]
