#!/usr/bin/python3

import csv
import datetime
import math
import time
import logging
from random import randint

logger_pon = logging.getLogger('ttc_pon')

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
def rx_eye_scan_masksetup(drp_master):
    # ES_QUAL_MASK is all 1's for statistical eye 
    for i in range(0x44,0x49):
        drp_master.drp_write(i,2**16-1)

    drp_master.drp_write(0x49,0)
    drp_master.drp_write(0x4A,0)
    drp_master.drp_write(0x4B,0xFF00)
    drp_master.drp_write(0x4C,2**16-1)
    drp_master.drp_write(0x4D,2**16-1)

def rx_eye_scan_prescale(drp_master, prescale_factor):
    stat = drp_master.drp_read(0x3C)[1]
    stat = stat & 0xFFE0
    stat = stat | (prescale_factor & 0x001F)		
    return drp_master.drp_write(0x3C,stat)

#!/usr/bin/python3
def rx_eye_scan_enable(drp_master):
    stat = drp_master.drp_read(0x3C)[1]
    stat_new = stat | 0x0300
    if(stat_new!=stat):
        return drp_master.drp_write(0x3C,stat_new)
    else:
        return 0

#!/usr/bin/python3
def rx_eye_scan_run(drp_master):
    stat = drp_master.drp_read(0x3C)[1]
    stat_new = stat | (0x0400)
    return drp_master.drp_write(0x3C,stat_new)

#!/usr/bin/python3
def rx_eye_scan_stop(drp_master):
    stat = drp_master.drp_read(0x3C)[1]
    stat_new = stat & (0x03FF)
    return drp_master.drp_write(0x3C,stat_new)

def rx_eye_scan_horz_offset(drp_master, horz_offset):
    stat = drp_master.drp_read(0x4F)[1]
    horz_offset_cp2 = horz_offset		
    if(horz_offset<0):
        horz_offset_cp2 = -1*horz_offset_cp2
        horz_offset_cp2 = horz_offset_cp2^0xFFF + 1
    stat_new = (stat&0x000F | horz_offset_cp2<<4)
    return drp_master.drp_write(0x4F, stat_new)

def rx_eye_scan_vert_offset(drp_master, vert_offset, step_volt):
    # step_volt can be:
    # 0 : 1.5mV/div
    # 1 : 1.8mV/div
    # 2 : 2.2mV/div
    # 3 : 2.8mV/div
    stat = drp_master.drp_read(0x97)[1]
    vert_offset_cp1 = vert_offset		
    if(vert_offset<0):
        vert_offset_cp1 = -1*vert_offset_cp1		
        vert_offset_cp1 = vert_offset_cp1 | 0x100
    
    vert_offset_cp1 = vert_offset_cp1 << 2
    vert_offset_cp1 = vert_offset_cp1 |(step_volt&0x0003)
    stat_new = (stat&0xF800 | vert_offset_cp1)
    return drp_master.drp_write(0x97, stat_new)

#!/usr/bin/python3
def rx_eye_scan_read_fsm(drp_master):
    stat = drp_master.drp_read(0x153)[1]
    stat_new = (stat & (0x000F))
    return stat_new

def rx_eye_scan_read_errors(drp_master):
    stat = drp_master.drp_read(0x151)[1]
    return stat

def rx_eye_scan_read_samples(drp_master):
    stat = drp_master.drp_read(0x152)[1]
    return stat

def rx_eye_scan(drp_master, eyescan_file_name, save_to_file=1):
    c_STATE_WAIT   = 0b0001
    c_STATE_RESET  = 0b0010
    c_STATE_COUN   = 0b0110
    c_STATE_END    = 0b0101
    c_STATE_ARMED  = 0b1010
    c_STATE_READ   = 0b1001

    # Hard parameters (do not change)
    c_GTH_INT_DATAWIDTH = 40 #datapath width
    c_PRESCALE_VALUE    = 5
    c_OFFSET_VOLTAGE    = 3 # 2.8mV/div
    ber_matrix = []

    rx_eye_scan_masksetup(drp_master)
    rx_eye_scan_prescale(drp_master, c_PRESCALE_VALUE)		
    rx_eye_scan_stop(drp_master)		

    stat = rx_eye_scan_enable(drp_master)

    for vert_offset in range(127,-127,-4):
        ber_matrix.append([])		
        rx_eye_scan_vert_offset(drp_master, vert_offset, c_OFFSET_VOLTAGE)		
        line_print = ''
        for horz_offset in range(-32,32):				
            rx_eye_scan_horz_offset(drp_master, horz_offset)
            rx_eye_scan_run(drp_master)
            stat = 0
			
            # wait scan is done (sample acc saturated)
            while((stat&0x0001)==0):
                stat = rx_eye_scan_read_fsm(drp_master)

            errs = rx_eye_scan_read_errors(drp_master)
            bits = rx_eye_scan_read_samples(drp_master)
            errs_scaled = errs				
            bits_scaled = bits*c_GTH_INT_DATAWIDTH*(2**(1+c_PRESCALE_VALUE))
            ber = (1.0*errs_scaled/bits_scaled) if (errs_scaled>0) else (1./bits_scaled)
            ber_matrix[-1].append(ber)
            rx_eye_scan_stop(drp_master)
            #print(ber)
            if(errs_scaled==0): append_char=' '
            else : append_char = '*'			
            line_print = line_print + append_char
        logger_pon.info(line_print)

    # save eyescan to file
    if(save_to_file):		
        with open(eyescan_file_name + '.csv', 'w') as csvfile:
            fieldnames = []
            for i in range(0, len(ber_matrix[-1])):
                fieldnames += ['hor' + str(i)]

            for i in range(0, len(fieldnames)):
                # format 20 spaces for each value
                fieldnames[i] = '%20s' % fieldnames[i]

            writer = csv.DictWriter(
                csvfile,
                fieldnames=fieldnames,
                delimiter=',',
                lineterminator='\n')
            writer.writeheader()

            dict_write = {}
            for j in range(0, len(ber_matrix)):
                for i in range(0, len(ber_matrix[-1])):
                    dict_write[fieldnames[i]] = "%20.15f" % ber_matrix[j][i]

                writer.writerow(dict_write)
        logger_pon.info('Finished Saving eye scan to file ' + eyescan_file_name + '.csv')
