#!/usr/bin/python3

import time
import atbwrap,symbreg
#!/usr/bin/python3
from ttcpon_exdsg import (OltMlinkExdsg, OnuExdsg, OnuMlinkExdsg)
#!/usr/bin/python3
from driver_comm import (DriverComm)

#!/usr/bin/python3
I2C_FMC_SWITCH_SLAVE_ADDRESS   = 0x74
I2C_FMC_SEL_SFP                = 0x20

oltc_regs_fiv010= (
("HB_PERIOD_25",         0,0xffff,0,"RW","LSB=25ns. Heartbeat period for TDM arbitration, set to Number_ONU's x 125ns/25ns."),
("HB_PERIOD_8_3",        1,0xfff,0,"RW","LSB=8.33ns. Heartbeat period for calibration / bbert, set to Number_ONU's x 125ns/8.33ns"), 
("SLOW_CTRL_MSG_ONU",    2,0xffffff,0,"RW","ONU manager is not necessary. OLT can write to the ONU register bank.The frame format:(OPE-4b / ONU_REGISTER_BYTE_ADDR - 9b / ONU_REGISTER_VALUE - 8b / ONU_ADDRESS - 8b) where OPE can be: READ=0001, WRITE=1111, WRITE WITH ACKNOWLEDGMENT=1110"), 
("CAL_ONU_ADDR",         3,0xff,0,"RW","Address of ONU which is in calibration mode"),
("CAL_ON",               3,1,8,"RW","Keep 1 during calibration"),
("CAL_INIT",             3, 1, 9,"RW","At the begginnning of calibration, keep to '1' for at least 2 x HEARTBEAT_PERIOD"),
("ONU_POSITION_LATCH",   3,1,10,"RW","Take measurement of ONU position during calibration"), 
("BERT_ONU_A_ADDR",      4, 0xff,0,"RW",),
("BERT_ONU_B_ADDR",      4, 0xff,8,"RW",),
("BERT_CLEAR",           4, 1, 16,"RW",),
("BERT_LATCH",           4, 1, 17,"RW",),
("BURST_LENGTH-1",       5,0x3f,0,"RW","LSB=4.16ns. Set to 29 for CBM125"),
("SFP_RESET_RX_POS",     5,0x3f,6,"RW","OLT Burst Mode Rx reset position (mod 30)"),
("SFP_RESET_RX_ENABLE",  5,1,12,"RW","enable OLT Burst Mode Rx reset"),
("SLOW_CTRL_FROM_ONU",   6,0xffffff,0,"R","Received message from ONU (response: WR_ACK, READ, RCVD_ERROR=10000)"),
("RONDTRIP_FINE_DELAY",  7,0x7f,0,"R","LSB=0.104ns. Measured fine position ONU"),
("RONDTRIP_COARSDELAY",  7,0xffff,7,"R","LSB=8.33ns. Measured coarse position ONU"),
("ERROR_SUM_ONU_A",      8,0xffff,0,"R",),
("ERROR_SUM_ONU_B",      8,0xffff,16,"R",),
("BERT_LOCKED_ONU_A",    9,1,0,"R",),
("BERT_LOCKED_ONU_B",    9,1,1,"R",),
("BERT_FINISHED_ONU_A",  9,1,2,"R",),
("BERT_FINISHED_ONU_B",  9,1,3,"R",),
("MGT_TX_POLARITY",      10,1,0,"RW",),
("MGT_RX_POLARITY",      10,1,1,"RW",),
("PHY_MONITOR_CLEAR",    10,1,2,"RW","Clear PHY sticky bits monitor"),
("PHY_MGT_SEEN_TXNRDY",  10,1,16,"R","Sticky bit"),
("PHY_MGT_SEEN_RXNRDY",  10,1,17,"R","Sticky bit"),
("PHY_MGT_SEEN_PLLNLOCK",10,1,18,"R","Sticky bit"),
("SD_STAT_CLEAR",        11,1,0,"RW","Clear signal detect failing/raising edge to statistics"), 
("SD_STAT_LATCH",        11,1,1,"RW","Latch signal detect failing/raising edge to statistics"), 
("SD_COMMA_CORR_CLEAR",  11,1,2,"RW","Clear signal detect x comma correlation statistics"), 
("SD_COMMA_CORR_LATCH",  11,1,3,"RW","Latch signal detect x comma correlation statistics"), 
("SD_MASK_OSRX",         11,1,4,"RW","Enable OSRX masking with signal detect"), 
("SD_MASK_OSRX_DELAY",   11,0xf,5,"RW","Delay to signal detect for OSRX masking (unit:TX_CLK_PERIOD)"), 
("SD_STAT",              12,0x3fffffff,0,"R","Signal detect falling/rising edge counter statistics used for calibration in order to find optimal reset position and implement missing burst functions"), 
("I2C_MASTER_CTL",       13,0xffffffff,0,"RW",), 
("I2C_MASTER_STAT",      14,0xffffffff,0,"R",), 
("8B10B_ERRDET_CLEAR",   15,1,    0,"RW","Clear err. detect monitor"),
("8B10B_ERRDET_LATCH",   15,1,    1,"R","Latch err. detect monitor"),
("8B10B_ERRDET_PAGESEL", 15,0x1f, 2,"RW","Page sel. for 8b10b error det. seen. It chooses which ONUs will be in the 8B10B_ERRDET_SEEN_ONU reg. 0:ONU7-0, 1:ONU15-8, ..."),
("8B10B_ERRDET_CNTR_ADD",15,0xff, 7,"RW","ONU addr for 8b10b err detect counter"),
("8B10B_ERRDET_SEEN_ANY",15,1,    15,"R","Sticky bit if any 8x10b error detect seen"),
("8B10B_ERRDET_SEEN_ONU",15,0xff, 16,"R","Sticky bit if 8x10b error detect seen for ONU addresses selected by reg 8B10B_ERRDET_PAGESEL"),
("8B10B_ERRDET_CNTR_ONU",15,0xff, 24,"R","Counter for 8b10b err detection of a given ONU (selected by 8B10B_ERRDET_CNTR_ADD)"),
("INTERRUPT_CTRL",16,0x7fffffff,0,"RW","Interrupt mask control"), 
("INTERRUPT",16,1,31,"R","Interrupt value"), 
("DRPACCESS",17,0x7ffffff,0,"R",) 
)
oltu_regs_fiv010= (
("OLT_CORE_RESET",                0x0,1,0,"RW","0->1: RESET OLT CORE"),
("CLEAR_STICKY_BITS",             0x0,1,1,"RW","CLEAR STICKY BITS OF REGISTER 0x3"),
("RX_BBERT_CLEAR",                0x1,1,0,"RW","CLEAR BBERT"),
("RX_BBERT_LATCH",                0x1,1,1,"RW","LATCH RESULTS BBERT"),
("RX_BBERT_HEARTBEAT_PERIOD",     0x1,0xffff,2,"RW","NUMBER_ONUS*(125ns/8.33ns)"),
("SEL_TX_PATTERN",                0x2,3,0,"RW","00=PRBS7 / 01=PRBS23 / 10=CONSTANT / 11=BYTE COUNTER"),
("MUX_TX_DATA_FROM_ONU_RX",       0x2,1, 2,"RW","0=SEND DATA FROM LOCAL DATA GEN (SELECTED WITH BITS 1-0).  1=SEND DATA THAT IS RECEIVED FROM ONU IN THE FIRST PON LAYER"),
("EN_TX_MATCH_FLAG",              0x2,1,3,"RW", "Enable TX match flag for latency measurements"),
("OLT_MGT_PLL_LOCK",              0x3,1,0,"R","STATUS"),
("OLT_MGT_TX_READY",              0x3,1,1,"R","STATUS"),
("OLT_MGT_RX_READY",              0x3,1,2,"R","STATUS"),
("PHASE_GOOD_ONU2OLT",            0x3,1,3,"R","STATUS - PHASE BETWEEN onu_clk_rxusr240_i and olt_clk_txusr240_i is in a good position. Recommended to perform a phase scanning with the sticky bit"),
("OLT_MGT_PLL_LOCK_STICKY",       0x3,1,16,"R","STATUS - STICKY BIT"),
("OLT_MGT_TX_READY_STICKY",       0x3,1,17,"R","STATUS - STICKY BIT"),
("OLT_MGT_RX_READY_STICKY",       0x3,1,18,"R","STATUS - STICKY BIT"),
("PHASE_GOOD_ONU2OLT_STICKY",     0x3,1,19,"R","STATUS - STICKY BIT"),
("BBERT_ONU_BITSUM(31:0)",        0x4,0xffffffff,0,"R","LINK QUALITY EVALUATION"),
("BBERT_ONU_BITSUM(47:32)",       0x5,0xffff,0,"R","LINK QUALITY EVALUATION"),
("BBERT_ONU_ERRORSUM(31:0)",      0x6,0xffffffff,0,"R","LINK QUALITY EVALUATION"),
("BBERT_ONU_ERRORSUM(47:32)",     0x7,0xffff,0,"R","LINK QUALITY EVALUATION"),
("BBERT_ONU_PACKETLOSSSUM(31:0)", 0x8,0xffffffff, 0,"R","LINK QUALITY EVALUATION"),
("BBERT_ONU_PACKETLOSSSUM(47:32)",0x9,0xffff,0,"R","LINK QUALITY EVALUATION"),
("BBERT_SELECT_CHANNEL(3:0)",     0xA,0xf,0,"RW","LINK QUALITY EVALUATION - This revision contains only a single BBERT per OLT user logic. Checked address is in this reg, 0 corresponds to ONU1, 1 to ONU2 and so on ... This reg has to be set BEFORE the BBERT is cleared")
)
onuc_regs_fiv010= (
("ONU_MODE",                   0x0, 0xff, 0,      "RW","ONU - MODES: ONU doesn't transmit any info (avoid ONUs overlap). 233 to be written here to allow user operation."),
("HEARTBEAT_PERIOD-1",         0x1, 0xffff, 0,    "RW","LSB=4.16ns. Heartbeat period, set to Number_ONU's*125/4.16-1"),
("COARSE_OFFSET_TDM",          0x1, 0xffff, 16,   "RW","Coarse offset in TDM. For a symmetrical network (all fiber lengths are the same), set this value to 1+ONU_ADDR*125/4.16. For asymmetric network calculated by calibration."),
("SFP_ENABLE_DELAY",           0x2, 0xf, 0,       "RW","Adjust this value such the burst is in the middle of the window (will be implemented automatically -> calib)"),
("FINE_OFFSET_TDM",            0x2, 0x1f,8,       "RW","Fine offset in TDM (LSB=0.416ns). For a symmetrical network, set to 0"),
("CLEAR_FEC_CRC_ERROR",        0x3, 1, 0,         "RW","Clear FEC and CRC error counters/sticky bits statistics (wr 1+0)"),
("LATCH_FEC_CRC_ERROR",        0x3, 1, 1,         "RW","Latch FEC and CRC error counters/sticky bits statistics"),
("SLOW_CONTROL_FRAMING_LOCKED",0x3, 1, 7,         "R", "Indicates if slow control framing is locked"),
("FEC_SINGLE_ERROR_CNTR",      0x3, 0xff, 8,       "R", "FEC single error counter"),  # 3 8bits counters used only during fullcal
("FEC_DOUBLE_ERROR_CNTR",      0x3, 0xff,16,       "R", "FEC double error counter"),
("CRC_ERROR_CNTR",             0x3, 0xff,24,       "R", "CRC error detected counter (increments every time an error is detected in the slow control frame)"),
("FEC_SINGLE_ERROR_SEEN",      0x4, 1, 0,         "R", "FEC single error counter seen"),
("FEC_DOUBLE_ERROR_SEEN",      0x4, 1, 1,         "R", "FEC double error counter seen"),
("CRC_ERROR_DETECT_SEEN",      0x4, 1, 2,         "R", "CRC error detected seen"),
("PHY_MONITOR_CLEAR",          0x5, 1, 0,         "RW","Clear PHY sticky bits monitor (wr 1+0)"),
("PHY_MGT_SEEN_TX_NREADY",     0x5, 1, 8,         "R", "Sticky bit corresponding to MGT TX not ready"),
("PHY_MGT_SEEN_RX_NREADY",     0x5, 1, 9,         "R", "Sticky bit corresponding to MGT RX not ready"),
("PHY_MGT_SEEN_TXPLL_NLOCK",   0x5, 1, 10,        "R", "Sticky bit corresponding to MGT TX PLL not locked"),
("PHY_MGT_SEEN_RXPLL_NLOCK",   0x5, 1, 11,        "R", "Sticky bit corresponding to MGT TX PLL not locked"),
("PHY_SEEN_RX_NLOCK",          0x5, 1, 12,        "R", "Sticky bit corresponding to downstream framing not locked"),
("PHY_SEEN_PHASE_NGOOD",       0x5, 1, 13,        "R", "Sticky bit corresponding to a phase not good for the RX->TX CDC"),
("INTERUPT_MASK",              0x6, 0x3fffffff, 0,"RW","Interrupt mask"),
("INTERRUPT",                  0x6, 1, 31,        "R", "Interrupt"),
("I2C_MASTER_CONTROL",         0x0D,0xffffffff, 0,"RW","I2C master control"),
("I2C_MASTER_STAT",            0x0E,0xffffffff, 0,"R", "I2C master stat"),
("OLT_COMM_DISABLE (DEBUG!)",  0xF, 1, 0,         "RW","manager only. Disable OLT slow control communication with the ONU (only controllable by local manager)"),
("DRP ACCESS",                 0x11,0x7ffffff, 0, "R", "Only available for Xilinx devices")
)
onuu_regs_fiv010= (
("ONU_CORE_RESET",            0x0, 1, 0,   "RW",    "RESET ONU CORE"),
("ONU_ADDRESS",               0x0, 0xff,1, "RW",    "ONU ADDRESS"),
("RX_BERT_CLEAR",             0x1, 1, 0,   "RW",    "CLEAR RESULTS BERT"),
("RX_BERT_LATCH",             0x1, 1, 1,   "RW",    "LATCH RESULTS BERT"),
("SEL_TX_PATTERN",            0x2, 0x3,0,  "RW",    "00=PRBS7 / 01=PRBS23 / 10=CONSTANT / 11=BYTE COUNTER"),
("TX_INTERFACING_MODE",       0x2, 1, 2,   "RW",    "TX INTERFACING"),
("TX_INJECT_ERROR",           0x2, 1, 3,   "RW",    "KEEP TO '1' FOR AR LEAST ONE HB PERIOD TO INSERT A SINGLE ERROR in TX data"),
("ONU_RX40_LOCKED",           0x3, 1, 0,   "R",     "STATUS"),
("ONU_PHASE_GOOD",            0x3, 1, 1,   "R",     "STATUS"),
("ONU_RX_LOCKED",             0x3, 1, 2,   "R",     "STATUS"),
("ONU_OPERATIONAL",           0x3, 1, 3,   "R",     "STATUS"),
("ONU_MGT_TX_READY",          0x3, 1, 4,   "R",     "STATUS"),
("ONU_MGT_RX_READY",          0x3, 1, 5,   "R",     "STATUS"),
("ONU_MGT_TXPLL_LOCK",        0x3, 1, 6,   "R",     "STATUS"),
("ONU_MGT_RXPLL_LOCK",        0x3, 1, 7,   "R",     "STATUS"),
("BERT_BITSUM(31:0)",         0x4, 0xffffffff,0,"R","LINK QUALITY EVALUATION"),
("BERT_BITSUM(47:32)",        0x5, 0xffff,0,"R",    "LINK QUALITY EVALUATION"),
("BERT_ERRORSUM(31:0)",       0x6, 0xffffffff,0,"R","LINK QUALITY EVALUATION"),
("BERT_ERRORSUM(47:32)",      0x7, 0xffff,0,"R",    "LINK QUALITY EVALUATION"),
("FEC_SINGLE_ERRORSUM(31:0)", 0x8, 0xffffffff,0,"R","LINK QUALITY EVALUATION"),
("FEC_SINGLE_ERRORSUM(47:32)",0x9, 0xffff,0,"R",    "LINK QUALITY EVALUATION"),
("FEC_DOUBLE_ERRORSUM(31:0)", 0xA, 0xffffffff,0,"R","LINK QUALITY EVALUATION"),
("FEC_DOUBLE_ERRORSUM(47:32)",0xB, 0xffff,0,"R",    "LINK QUALITY EVALUATION")
)
def printTable(oo, table, coru, poss=30,only1flags=True):
    """
    coru: 
    """
    ralast= -1
    for ir in range(len(table)):
      ra= table[ir][1]   # rel. words
      if ralast!= ra: 
        ralast= ra
        msg="%2d:"%ra
        valueword= oo.read_ctrl_reg(ra | coru)
      else: msg="  :"
      bitm= table[ir][2]
      value= (valueword>>table[ir][3]) & bitm
      if only1flags and ((bitm>0xffff) or (value==0)): continue
      fmt= " %%%d.%ds >%%2d 0x%%8x:0x%%8x"%(poss,poss)
      msg= msg + fmt%(table[ir][0], 
        table[ir][3], bitm, value)
      print msg
class OltMlinkExdsgAlice(OltMlinkExdsg):
  def __init__(self, ip='192.168.1.2', port=9001, lane=0, mode="s"): #, nonus=0):
    OltMlinkExdsg.__init__(self, ip, port, lane)
    # self.nonus= nonus ; self.bbert= (nonus*15)<<2   seems not used 
    self.setolt(lane+1)
    self.gsmode= mode   # 's' or 'g'
    # set MUX_TX_DATA_FROM_ONU_RX bit in olt_u:
    val= self.read_ctrl_reg(self.Asel_tx_pattern)
    if mode=='g':
      val= val | 0x4
    else:
      val= val & 0xfffffffb
    self.write_ctrl_reg(self.Asel_tx_pattern, val)
  def setolt(self, nolt):   # 1..9
    # define muxes (different for kc/atb and oltX/onu):
    #print "setolt:",nolt, self.conver
    if self.conver=="ipbus":
      if nolt==1: self.mux=(0x74,0, 0x76,1<<4)
      elif nolt==2: self.mux=(0x74,0, 0x76,1<<5)
      elif nolt==3: self.mux=(0x74,0, 0x76,1<<6)
      elif nolt==4: self.mux=(0x74,0, 0x76,1<<7)
      elif nolt==5: self.mux=(0x76,0, 0x74,1<<6)
      elif nolt==6: self.mux=(0x76,0, 0x74,1<<7)
      elif nolt==7: self.mux=(0x74,0, 0x76,1<<2)
      elif nolt==8: self.mux=(0x74,0, 0x76,1<<3)
      elif nolt==9: self.mux=(0x76,0, 0x74,1<<2)
      else:
        print "Bad OLT%d, using OLT1"%nolt
        nolt=1; self.mux=(0x74,0, 0x76,1<<4)
    elif self.conver=="ublaze":
      if nolt!=1:
        print "Bad OLT%d, using OLT1 on ublaze board"%nolt
        nolt=1 
      self.mux=(I2C_FMC_SWITCH_SLAVE_ADDRESS,I2C_FMC_SEL_SFP)   # FMC switch
    else:
      print "setolt: Bad OLT%d, only ulaze or ipbus boards supported"%nolt
      return
    # mozno toto netreba: self.lane_offset= 128*(nolt-1)
    uroffset= self.addr_offset #+ (nolt-1)*0x80
    self.Arx_bbert= 1 + uroffset
    self.Amux_select= 16 + uroffset
    self.Asel_tx_pattern= 2 + uroffset
    self.Astatus= 3 + uroffset
    self.Abbert_sel_chan= 10 + uroffset
    DriverComm.setolt(self, nolt)
  def setmux(self, ipbusaccess=False) :  
    rc= True
    if self.conver=="ublaze":
      print "OLT setmux: set fmc mux on kcu/kc board (ublaze conenction): nothing done."
      return rc
    if ipbusaccess==True:
      self.ctrlcon.board.Write("i2cmainttc.mux", 11)
      # when on ipbus, setmux called from i2c. methods for given device
      print "i2cmainttc.mux:11(ipbus)"
      return rc
    else:
      try:
        self.ctrlcon.board.Write("i2cmainttc.mux", self.multiolt)
      except:
        rc= False
        return rc
    if self.mux: 
      print "olt setmux: olt%d ..."%self.multiolt
      for ix in range(0,len(self.mux),2):
        ma= self.mux[ix] ; val= self.mux[ix+1]
        # like in ttcpon_board_config/pon_fmc_i2c_program.py 
        rc= self.i2c_write(ma, val, val)   # set mux
        if rc!= 1:
          print "olt%d setmux 0x%x:0x%x failed..."%(self.multiolt, ma, val)
          rc= False
          return rc
        else:
          print "         mux 0x%x:0x%x ok."%(ma, val)
    else:
      print "olt setmux: no mux ?, seems internal error"
      rc= False
    return rc
  def set_remotemux(self, onuaddr, rmmuxes):
    """using olt_core.onu_i2c_write to set remote mux: 
Assuming i2cttpon.mux is set already (can be set ONLY through ipbus access!)
Cases:
1. remote ONU/atb
  (0x74,0, 0x76,1<<4) # fixed, only 1 ONU (SFP2) on each alice board
2. remote OLT/atb -not implemented
3. remote ONU/kc[u] or OLT/kc[u] on TTCPON FMC card
   (0x74,0x20) ublaze FMC (olt or onu)
4. remote CRU: do nothing (no mux)
"""
    #print "set_remotemux: onuaddr:%d (%s)..."%(onuaddr, rmmuxes)
    for ix in range(0,len(rmmuxes),2):
      ma= rmmuxes[ix] ; val= rmmuxes[ix+1]
      # like in ttcpon_board_config/pon_fmc_i2c_program.py 
      rc= self.onu_i2c_write(ma, val, val, onuaddr)
      if rc!= 1:
        print "Error in set_remotemux: olt%d setmux 0x%x:0x%x failed..."%(self.multiolt, ma, val)
        return False
      else:
        pass #print "             : olt%d remotemux 0x%x:0x%x ok..."%(self.multiolt, ma, val)
    return True  
  def oltreset(self):
    print "OLT%d reset..."%self.getoltn()
    self.exd_core_reset()
    #self.write_ctrl_reg(self.addr_offset,1)
    #time.sleep(0.1)
    #self.write_ctrl_reg(self.addr_offset,0)
  def pllreset(self):
    print "OLT pllreset..."
    if self.ctrlcon.plldev:
      self.ctrlcon.plldev.hwreset()
      time.sleep(1)
      self.ctrlcon.plldev.checkgs(self.gsmode,load=True, checkbefore=False)
  def setseltx(self, seltx):
    #Aseltx= 2 | self.uroffset
    val= self.read_ctrl_reg(self.Asel_tx_pattern)
    val= (val & 0xfffffffc) | seltx
    self.write_ctrl_reg(self.Asel_tx_pattern, val)
  def bertClear(self, onuaddr1=None, Nonus=None):
    if (onuaddr1==None) or (onuaddr1==0):
      print "onuN has to be >1"
      return
    if Nonus!=None: 
      self.exd_bert_set_heartbeat(Nonus)
    else:
      print "bertClear: exd_bert_set_heartbeat(Nonus) not called (Nonus:None)"
    # channel has to be set before bert_clear:
    #self.write_ctrl_reg(self.Abbert_sel_chan, onuaddr1-1)
    # let's use ttcpon original one:
    self.exd_bert_select_channel(onuaddr1-1)
    self.exd_bert_clear()
    #print "bertClear:"+self.name+" for onu:", onun
    #print "bertClear:"+self.name+" for ONU_%d"%onuaddr1
  def bertLatch(self):
    self.exd_bert_latch()
  def cfg_downstream(self):
    #self.write_ctrl_reg(self.Aolt_core_reset, 1)
    #time.sleep(1)
    #self.write_ctrl_reg(self.Aolt_core_reset, 0)
    #time.sleep(1)
    self.exd_core_reset()
    self.write_ctrl_reg(self.Amux_select, 0)
    self.write_ctrl_reg(self.Asel_tx_pattern, 0)   # PRBS7
  def getUpCounts(self):
    self.bertLatch()
    Abs= 4 | self.addr_offset + (self.getoltn()-1)*0x80
    rc= []
    for ix in range(3):
      bs2= self.read_ctrl_reg(Abs+2*ix+1) & 0xffff
      bs1= self.read_ctrl_reg(Abs+2*ix) 
      rc.append(bs1); rc.append(bs2)
    return rc   # sum32,16 errsum32,16 pcklost32,16
  def getr(self, cu, addr, desc=True):
    if cu=='u':
      #coru= self.addr_offset + (self.getoltn()-1)*0x80
      coru= self.addr_offset
    elif cu=='c':
      #coru= (self.getoltn()-1)*0x80
      coru= 0
    else:
      return None
    value= self.read_ctrl_reg(addr | coru)
    if desc:
      print "getr %c 0x%x | 0x%x: 0x%x"%(cu, addr, coru, value)
      self.prtDesc(cu, addr, value)
    return value
  def getoltstatus(self):
    return self.read_ctrl_reg(self.Astatus)
  def putr(self, cu, addr, value, desc=True):
    if cu=='u':
      coru= self.addr_offset + (self.getoltn()-1)*0x80
    elif cu=='c':
      coru= (self.getoltn()-1)*0x80
    else:
      return None
    self.write_ctrl_reg(addr | coru, value)
    if desc:
      print "putr %c %d: 0x%x"%(cu, addr, value)
      self.prtDesc(cu, addr, value)
    return value
  def prtDesc(self, cu, addr, val):
    if cu=='c':
      rd= oltc_regs_fiv010
    else:
      rd= oltu_regs_fiv010
    # "name" addr, mask, bit_from_to_the_left,"RW","description"
    #  0     1     2     3                    4     5
    for ix in range(len(rd)):
      if rd[ix][1]==addr:
        desc= ""
        if len(rd[ix])>5: desc= rd[ix][5]
        print "%30s: 0x%x %s"%(rd[ix][0], ((val>>rd[ix][3]) & rd[ix][2]), desc)

  def printRegs(self, wregs="cu", only1flags=True):
    if self.getoltn()==0:
      print "mulitolt:0"
      return
    fiinfo= "fi > v4 (multiolt), using OLT%d."%self.getoltn()
    oltc_regs= oltc_regs_fiv010
    oltu_regs= oltu_regs_fiv010
    if wregs.find('c')>=0:
      print "OLT CORE:", fiinfo
      #self.printTable(oltc_regs, (self.getoltn()-1)*0x80, 21)
      printTable(self, oltc_regs, (self.getoltn()-1)*0x80, 21, only1flags)
    if wregs.find('u')>=0:
      print "OLT USER:"
      #self.printTable(oltu_regs, self.uroffset, 30)
      printTable(self, oltu_regs, self.addr_offset + (self.getoltn()-1)*0x80, 30, only1flags)
      if not only1flags:
        cs= self.getUpCounts()
        print "BITSUM: %d %d   ERRSUM:%d %d   PACKETLOSS:%d %d"%(cs[0], cs[1], cs[2], cs[3], cs[4], cs[5])

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
    old="""
class OnuMlinkExdsgAlice(OnuExdsgAlice):
  def __init__(self, ip='192.168.1.2', port=9001, onun=None, mononu= False):
    OnuExdsgAlice.__init__(self, ip, port,onun)
    self.monionu= mononu
    #self.lane_offset =OnuMlinkExdsg.lane_offset   # 0x500=128*10
    self.lane_offset = 0x500   # =128*10
    #self.addr_offset = OnuMlinkExdsg.addr_offset   # 0x4000= 16384
    self.addr_offset = 0x4000   # = 16384
    print "OnuMlinkExdsgAlice lane addr offsets:",self.lane_offset,self.addr_offset
  def write_ctrl_reg(self, reg_addr, reg_value):
    return OnuExdsg.write_ctrl_reg(self, reg_addr + self.lane_offset, reg_value)

  def read_ctrl_reg(self, reg_addr):
    return OnuExdsg.read_ctrl_reg(self, reg_addr + self.lane_offset)
"""

class Gbt:
  AVAGO=0x100  # bit9(0x200) on: force Data header bit to 1 for both receivers
  VTRX=0x2100  # 
  def __init__(self, bn='atb_xx', gbtn=1, dest="vtrx"):   # vtrx sfp sfpbsr
    self.board= atbwrap.OpenATBoard(bn) 
    if gbtn==0:
      self.slave="gbt"
    else:
      self.slave="gbt%d"%gbtn
    self.dest= dest
    #self.slave="gbt"   # v5 has just 1 slave "gbt"
    if self.board.nodExists(self.slave, True):
      #self.init()
      self.reg_status= symbreg.Register("status",
        TxAlignedComp=0,
        TxAlignedBank=1,
        MgtReady=2,
        rxBitSlips=(3,8),
        FrameReady=11,
        ErrDetected=12,
        RxDataF=13,
        GbtRxReady=17,
        GbtBankTxReady=18)
      self.reg_txc= symbreg.Register("txctrl",
        TxShift=(0,8),
        TxEnc= 8,
        Txls=9,        # 0:idle header pattern, 1: data hp
        ClkMuxSel=12,
        GbtTxPol=13)   # 0: normal, 1: Inverted (vtrx)
      self.reg_rxc= symbreg.Register("rxctrl",
        rxEnc=0, 
        rxBitSlipRst=4,
        Loopback=(5,3),
        GbtRxPol=8)    # 0: normal, 1: Inverted (vtrx)
    #print "%s -> %s"%(self.slave, dest)
    rc= self.board.Read(self.slave+".txctrl")
    self.exprxctrl= 0x1
    if self.dest=="vtrx":
      if rc != self.VTRX:
        print "%s -> vtrx: setting .txctrl register to: 0x%x"%(self.slave, self.VTRX)
        self.board.Write(self.slave+".txctrl",self.VTRX)  
      else:
        print "%s -> vtrx: .txctrl register ok: 0x%x"%(self.slave, self.VTRX)
    elif self.dest=="sfp" or self.dest=="sfpbsr":
      if rc != self.AVAGO:
        print "%s -> %s: setting .txctrl register to: 0x%x"%(self.slave, self.dest, 
          self.AVAGO)
        self.board.Write(self.slave+".txctrl",self.AVAGO)  
      else:
        print "%s -> %s: .txctrl register ok: 0x%x"%(self.slave, self.dest, self.AVAGO)
      if self.dest=="sfpbsr":
        self.exprxctrl= 0x11
      else:
        self.exprxctrl= 0x1
      rc= self.board.Read(self.slave+".rxctrl")
      if rc== self.exprxctrl:
        print "%s -> %s: .rxctrl register ok: 0x%x"%(self.slave, self.dest, 
          self.exprxctrl)
      else:
        print "%s -> %s: setting .rxctrl register to: 0x%x"%(self.slave, self.dest, 
          self.exprxctrl)
        self.board.Write(self.slave+".rxctrl",self.exprxctrl)  
    else:
      print "%s -> %s ???"%(self.slave, self.dest)
    #self.showstatus()
  def init(self, loopback=0):
    dest= self.dest
    # config tx for avago/vldb: 0x100
    # another option: cofig tx for vtrx/vldb: 0x2100
    if dest=="sfp":
      receiver= self.AVAGO
      deste="sfp, downstream+upstream"
    elif dest=="vtrx":
      receiver=self.VTRX
      deste="vtrx, downstream only"
    else:
      print "Error: %s init: bad destination: "%self.slave, dest
      return
    print "%s: sending to %s."%(self.slave,deste)
    self.board.Write(self.slave+".txctrl",receiver)  
    rc= self.board.Read(self.slave+".txctrl")
    if rc != receiver:
      print "%s Init error txctrl: 0x%x instead 0x%x"%(self.slave, rc, receiver)
    # loopback: 0: default (no loopback)
    # r3doc/GBT_in_ATB_v5.pdf
    #https://www.xilinx.com/support/documentation/user_guides/ug576-ultrascale-gth-transceivers.pdf (Pag 86)
    # 
    self.board.Write(self.slave+".rxctrl",(self.exprxctrl|(loopback<<5)))  
    rc= self.board.Read(self.slave+".rxctrl")
    if rc != self.exprxctrl:
      print "Warning: %s Init rxctrl: 0x%x instead 0x%x"%(self.slave, rc, self.exprxctrl)
    self.reset()
    rc= self.board.Read(self.slave+".status")
    if dest=="sfp":
      if rc != 0x60807:
        print "%s Init error (status): 0x%x instead 0x60807"%(self.slave,rc)
      else:
        print "%s Init (SFP) ok"%self.slave
    elif dest=="vtrx":
      if (rc&0x41003) != 0x41003:
        print "%s Init error (status): 0x%x (0x41003 bits expected)"%(self.slave,rc)
      else:
        print "%s Init (VTRX) ok"%self.slave
    else:
      print "%s Init Error: "%self.slave,dest
  def reset(self, bit="cpu"):
    if bit=="cpu": 
      msk=0x1
    elif bit=="rx": 
      msk=0x4
    elif bit=="tx": 
      msk=0x2
    else:
      print "Error: %s, only cpu rx or tx resets available", bit
      return
    # write only, read no sense, rr= self.board.Read(self.slave+".resets")
    rw1=  msk
    self.board.Write(self.slave+".resets",rw1)  # ->1
    rw2= 0
    self.board.Write(self.slave+".resets",rw2)  # ->0
    print "gbt reset 0x%x 0x%x"%(rw1,rw2)
  # todo: check + repair gbt bidirectional connection:
  # if status not stable:
  #   olt reset    -on both sides!
  #   if status not ok (0x60807), but stable:
  #     gbt N reset rx
  #   elif status not stable:
  #     gbt N reset cpu
  def settxrx(self, gv):   # "sfp" or "vtrx"
    if gv=="sfp":
      self.board.Write(self.slave+".txctrl",self.AVAGO)  
    elif gv=="vtrx":
      self.board.Write(self.slave+".txctrl",self.VTRX)  
    elif gv=="lb0":
      rx= self.board.Read(self.slave+".rxctrl")
      rx= (rx & 0xffffff1f ) # no loopback
      self.board.Write(self.slave+".rxctrl", rx)
    elif gv=="lb1":
      rx= self.board.Read(self.slave+".rxctrl")
      rx= (rx & 0xffffff1f ) | 0x20  # internal loopback 1
      self.board.Write(self.slave+".rxctrl", rx)
    elif gv=="bsr0":   # rxBitSlipRst off
      rx= self.board.Read(self.slave+".rxctrl")
      rx= (rx & 0xffffffef )
      self.board.Write(self.slave+".rxctrl", rx)
    elif gv=="bsr1":   # rxBitSlipRst on
      rx= self.board.Read(self.slave+".rxctrl")
      rx= (rx | 0x10 )
      self.board.Write(self.slave+".rxctrl", rx)
    else:
      print "fp or vtrx expected..."
      return
    tx= self.board.Read(self.slave+".txctrl")
    self.reg_txc.prt(tx)
  def status(self):
    st= self.board.Read(self.slave+".status")
    tx= self.board.Read(self.slave+".txctrl")
    rx= self.board.Read(self.slave+".rxctrl")
    return st,tx,rx
  def showbitdefs(self, bits=None):
    if bits==None:
      self.reg_status.prtme()
      self.reg_txc.prtme()
      self.reg_rxc.prtme()
    else:
      self.reg_status.prt(bits)
      self.reg_txc.prt(bits)
      self.reg_rxc.prt(bits)
  def showstatus(self):
    st,tx,rx= self.status()
    #self.reg_status.prtme()
    self.reg_status.prt(st, self.slave)
    self.reg_txc.prt(tx)
    self.reg_rxc.prt(rx)

