#!/usr/bin/python3

import time
import ttcpon_alice
from . olt_core import OltCore
from . onu_core import OnuCore

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3


class OltExdsg(OltCore):

    def __init__(self, ip='192.168.1.2', port=9001):
        OltCore.__init__(self, ip, port)
        self.addr_offset = 1024

    def exd_core_reset(self):
        stat = self.read_ctrl_reg(self.addr_offset + 0)
        stat = stat | 1
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        return 1

    def exd_bert_set_heartbeat(self, number_onu):
        self.write_ctrl_reg(self.addr_offset + 1, (number_onu * 15) << 2)
        return 1

    def exd_bert_clear(self):
        stat = self.read_ctrl_reg(self.addr_offset + 1)
        stat = stat | 1
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        return 1

    def exd_bert_latch(self):
        stat = self.read_ctrl_reg(self.addr_offset + 1)
        stat = stat | (1 << 1)
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        stat = stat ^ (1 << 1)
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        return 1

    def exd_bert_select_channel(self, channel):
        #print "exd_bert_select_channel (rev010):", self.addr_offset,channel
        if channel>15:
          print "Error: exd_bert_select_channel: %d > 15, only low 4 bits valid!"%channel
        self.write_ctrl_reg(self.addr_offset + 10, channel)
        return 1

    def exd_read_error_sum(self):
        msb = self.read_ctrl_reg(self.addr_offset + 7)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 6)
        return (lsb + msb)

    def exd_read_bit_sum(self):
        msb = self.read_ctrl_reg(self.addr_offset + 5)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 4)
        return (lsb + msb)

    def exd_read_ploss_sum(self):
        msb = self.read_ctrl_reg(self.addr_offset + 9)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 8)
        return (lsb + msb)

    def exd_tx_cntr(self):
        self.write_ctrl_reg(self.addr_offset + 2, 3)

    def exd_tx_cte(self):
        self.write_ctrl_reg(self.addr_offset + 2, 2)
        return 1

    def exd_tx_prbs23(self):
        self.write_ctrl_reg(self.addr_offset + 2, 1)		
        return 1

    def exd_tx_prbs7(self):
        self.write_ctrl_reg(self.addr_offset + 2, 0)		
        return 1

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3


class OnuExdsg(OnuCore):

    def __init__(self, ip='192.168.1.11', port=9001, onun= None):
        OnuCore.__init__(self, ip, port, onun)
        self.addr_offset = 1024

    def exd_core_reset(self):
        stat = self.read_ctrl_reg(self.addr_offset + 0)
        stat = stat | 1
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        return 1

    def exd_read_onu_status(self, print_enable=0):
        stat = self.read_ctrl_reg(self.addr_offset + 3)
        rx40_locked    = (stat & 0x00000001) >> 0
        phase_good     = (stat & 0x00000002) >> 1	
        rx_locked      = (stat & 0x00000004) >> 2
        operational    = (stat & 0x00000008) >> 3
        mgt_tx_ready   = (stat & 0x00000010) >> 4
        mgt_rx_ready   = (stat & 0x00000020) >> 5	
        mgt_txpll_lock = (stat & 0x00000040) >> 6
        mgt_rxpll_lock = (stat & 0x00000080) >> 7

        if(print_enable):
            print("ONU_STATUS:")
            print("rx40_locked:"    + str(rx40_locked))
            print("phase_good:"     + str(phase_good)) 
            print("rx_locked:"      + str(rx_locked))
            print("operational:"    + str(operational)) 
            print("mgt_tx_ready:"   + str(mgt_tx_ready))
            print("mgt_rx_ready:"   + str(mgt_rx_ready))
            print("mgt_txpll_lock:" + str(mgt_txpll_lock))
            print("mgt_rxpll_lock:" + str(mgt_rxpll_lock))

        return [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock]

    def exd_set_addr(self, addr_value):
        self.write_ctrl_reg(self.addr_offset + 0, addr_value << 1)
        return 1

    def exd_get_addr(self):
        stat = self.read_ctrl_reg(self.addr_offset + 0)
        addr = (stat>>1)&0xFF
        return addr

    def exd_bert_clear(self):
        stat = self.read_ctrl_reg(self.addr_offset + 1)
        stat = stat | 1
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        stat = stat ^ 1
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        return 1

    def exd_bert_latch(self):
        stat = self.read_ctrl_reg(self.addr_offset + 1)
        stat = stat | (1 << 1)
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        stat = stat ^ (1 << 1)
        self.write_ctrl_reg(self.addr_offset + 1, stat)
        return 1

    def exd_read_derror_correc_sum(self):
        msb = self.read_ctrl_reg(self.addr_offset + 11)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 10)
        return (lsb + msb)

    def exd_read_serror_correc_sum(self):
        msb = self.read_ctrl_reg(self.addr_offset + 9)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 8)
        return (lsb + msb)

    def exd_read_error_sum(self):
        msb = self.read_ctrl_reg(self.addr_offset + 7)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 6)
        return (lsb + msb)

    def exd_read_bit_sum(self):
        msb = self.read_ctrl_reg(self.addr_offset + 5)
        msb = msb << 32
        lsb = self.read_ctrl_reg(self.addr_offset + 4)
        return (lsb + msb)

    def exd_tx_inject_error(self):
        stat = self.read_ctrl_reg(self.addr_offset + 2)
        stat = stat | (1 << 3)
        self.write_ctrl_reg(self.addr_offset + 2, stat)
        time.sleep(1)
        stat = stat ^ (1 << 3)
        self.write_ctrl_reg(self.addr_offset + 2, stat)
        return 1

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3


class OnuFMCExdsg(OnuExdsg):

    def __init__(self, ip='192.168.1.11', port=9001):
        OnuExdsg.__init__(self, ip, port)
        self.fmc_offset = 2048

    def write_ctrl_reg(self, reg_addr, reg_value):
        return OnuExdsg.write_ctrl_reg(self, reg_addr + self.fmc_offset, reg_value)

    def read_ctrl_reg(self, reg_addr):
        return OnuExdsg.read_ctrl_reg(self, reg_addr + self.fmc_offset)

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3


class OltMlinkExdsg(OltExdsg):
    def __init__(self, ip='192.168.1.2', port=9001, lane=0):
        OltExdsg.__init__(self, ip, port)
        self.lane_offset = 128*lane
        self.addr_offset = 16384

    def write_ctrl_reg(self, reg_addr, reg_value):
        return OltExdsg.write_ctrl_reg(self, reg_addr + self.lane_offset, reg_value)

    def read_ctrl_reg(self, reg_addr):
        return OltExdsg.read_ctrl_reg(self, reg_addr + self.lane_offset)

    def exd_clear_sticky(self):
        stat = self.read_ctrl_reg(self.addr_offset + 0)
        stat = stat | 2
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        stat = stat ^ 2
        self.write_ctrl_reg(self.addr_offset + 0, stat)
        return 1

    def exd_read_olt_status_sticky(self):
        stat = self.read_ctrl_reg(self.addr_offset + 3)
        mgt_pll_lock       = (stat>>16)&1
        mgt_tx_ready       = (stat>>17)&1
        mgt_rx_ready       = (stat>>18)&1
        phase_good_onu2olt = (stat>>19)&1

        return [mgt_pll_lock, mgt_tx_ready, mgt_rx_ready, phase_good_onu2olt]

    def exd_tx_cascaded_forward_data(self):
        stat = self.read_ctrl_reg(self.addr_offset + 2)
        stat = stat | 4
        self.write_ctrl_reg(self.addr_offset + 2, stat)
        return 1

class OnuExdsgAlice(OnuExdsg):
  def __init__(self, ip='192.168.1.11', port=9001, onun= None):
    OnuExdsg.__init__(self, ip, port, onun)
    #print "OnuExdsgAlice ip:%s %s self.addr_offset: 0x%x"%\
    #  (ip, self.conver, self.addr_offset)
    if self.conver=="ublaze":
      self.mux=(I2C_FMC_SWITCH_SLAVE_ADDRESS,I2C_FMC_SEL_SFP)   # FMC switch
    elif self.conver=="ipbus":
      self.mux= (0x76,0, 0x74,1<<3) # fixed, only 1 ONU (SFP2) on each alice board
    else:
      self.mux= None    # CRU
    #self.Asel_tx_pattern= 2 + self.addr_offset
    self.notvalid= False
  def getura(self, addr):
    return(addr+self.addr_offset)
    #def whoami(self):
    #if (self.onun==3) or (self.onun==6) or (self.onun==7):
    #  return "cru"
    #elif self.onun>=10:
    #  return "atb"
    #else:
    #  return "kc"
    return self.conver
  def onureset(self):
    if self.conver=="cru":
      print "ONU_%d reset not done (cru)"%self.onun
    else:
      print "ONU_%d reset..."%self.onun
      self.exd_core_reset()
      # also ok self.write_ctrl_reg(self.addr_offset,1)
  def pllreset(self):
    if self.conver=="cru":
      print "ONU_%d pllreset not done (cru)"%self.onun
    elif self.monionu:
      print "ONU_%d pllreset not done (monitoring onu)"%self.onun
    else:
      print "ONU_%d pllreset..."%self.onun
      if self.ctrlcon.plldev:
        self.ctrlcon.plldev.hwreset()
        time.sleep(1)
        self.ctrlcon.plldev.load()
  def onuseta(self):
    # set ONU number
    self.write_ctrl_reg(self.addr_offset, self.onun*2)
  def getonustatus(self):
    Aonu_rx40_locked= self.getura(3)
    val= self.read_ctrl_reg(Aonu_rx40_locked)
    return val
  def onuinit_delme(self):   # not used -> see onu_init in ttcpon_board_config.init_onu
    print "onuinit: %d"%self.onun
    onu_addr= self.onun
    # Program FMC devices
    #program_clkmux(onu, 0x37)
    #program_si570(onu, 1)
    #program_si5344(onu, 1)
    #
    # Wait some time to PLL output to be stable
    time.sleep(0.6)
    self.exd_core_reset()
    time.sleep(1.0)
    #
    # Program ONU upstream TDM address
    self.exd_set_addr(onu_addr)
    #
    # Check ONU status before doing phase alignment - ONU must be locked
    status_onu = self.exd_read_onu_status()
    [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
    timeout_cntr = 0
    while((not (mgt_rx_ready and mgt_rxpll_lock and rx40_locked and rx_locked and mgt_txpll_lock and mgt_tx_ready)) and timeout_cntr< c_MAX_TIMEOUT_MGT_READY):
        status_onu = self.exd_read_onu_status()
        [rx40_locked, phase_good, rx_locked, operational, mgt_tx_ready, mgt_rx_ready, mgt_txpll_lock, mgt_rxpll_lock] = status_onu
        timeout_cntr = timeout_cntr + 1
        time.sleep(0.1)
        if(timeout_cntr >= c_MAX_TIMEOUT_MGT_READY):
            logger_pon.warn("ONU is not getting ready after timeout, check ONU status!")		
            return 0
    #
    for i in range(0,1):	
        logger_pon.info("Phase adjustment nb. "+str(i))
        #stat = dbg_automatic_phase_good(onu,i)
        stat = self.automatic_phase_good(onu)	
        #onu.exd_core_reset()
        #time.sleep(1.0)
    #
    #rogram_leds(onu, 0x00)
    #monitor_sfp(onu,1)
    #
    return stat
  def seti2cmainttcmux(self, val):
    if self.conver == "ipbus":
      #was= self.ctrlcon.board.Read("i2cmainttc.mux")
      #self.ctrlcon.board.Write("i2cmainttc.mux", val)
      #  better: i2c.I2Ctpcore.seti2cttcmux
      self.ctrlcon.board.i2cg.seti2cttcmux(val)
      # val: 1..11 corresponding to olt1..9,onu,ipbus
      #print "seti2cmainttcmux: %d -> %d on onuaddr:%d"%(was,val,self.onun)
    else:
      print "Warning seti2cmainttcmux: not ipbus connection:", self.conver
  def setmux(self, stdout=False):  
    if self.conver=="ipbus":
      self.seti2cmainttcmux("onu")   # 10
    if self.mux: 
      for ix in range(0,len(self.mux),2):
        ma= self.mux[ix] ; val= self.mux[ix+1]
        # like in ttcpon_board_config/pon_fmc_i2c_program.py 
        if stdout: print "onu setmux 0x%x: 0x%x"%(ma, val)
        rc= self.i2c_write(ma, val, val)   # set mux
        if rc!= 1:
          print "setmux 0x%x:0x%x failed..."%(ma,val)
    #self.ctrlcon.board.Write("i2cmainttc.mux", 10)
  def getmux(self):
    """ test (not working in ttcpon v010: i2c implementation does not allow read)"""
    self.seti2cmainttcmux("onu")
    if self.mux:
      for ix in range(0,len(self.mux),2):
        ma= self.mux[ix] ; vale= self.mux[ix+1]
        val= self.i2c_read(ma, ma, 1)
        print "getmux 0x%x:  expected:0x%x got:"%(ma, vale), val
  def setseltx(self, seltx):
    #Aseltx= 2 | self.uroffset
    #print "setseltx:    0x%x correct:0x%x"%(self.Asel_tx_pattern, 2 + self.addr_offset)
    Asel_tx_pattern= self.getura(2)
    val= self.read_ctrl_reg(Asel_tx_pattern)
    val= (val & 0xfffffffc) | seltx
    self.write_ctrl_reg(Asel_tx_pattern, val)
  def clrDownCounts(self):
    if self.conver == 'ipbus':
      print "ONU_%d bert/errmon clear..."%self.onun
      self.exd_bert_clear()   # user regs
      self.errmon_clear()    # control regs
    else:
      print "ONU_%d bert/errmon clear NOT DONE (cru)"%self.onun
  def bertLatch(self):
    #print "ONU%d BERT latch..."%self.onun
    self.exd_bert_latch()
  def getDownCounts(self):
    if self.conver=="cru":   # not availbale for CRU-ONU
      print "getDownCounts: not availbale for CRU-ONU"
      rc=[0]*8
    else:
      self.bertLatch()
      Abs= 4 | self.addr_offset ; rc= []
      for ix in range(4):
        bs2= self.read_ctrl_reg(Abs+2*ix+1) & 0xffff
        bs1= self.read_ctrl_reg(Abs+2*ix) 
        rc.append(bs1); rc.append(bs2)
      #print "OnuCore.getDownCounts:", bs1, bs2
      onuC= self.errmon_read()   #fec_sing8 fec_doub8 sc_crc8
      rc= rc + onuC 
    return rc   # sum32,16 err32,16 fec_sing32,16 fec_doub32,16
  def excluded(self):
    return self.notvalid
  def exclude(self):
    self.notvalid= True
    if self.conver != "ipbus":
      print "Did you set Core reg0 (ONU_MODE) to 0?"
    else:
      onumode= self.read_ctrl_reg(0)
      self.write_ctrl_reg(0, 0)
      print "Core reg0 ONU_MODE set to 0 (excluded), was 0x%x"%(onumode)
  def include(self):
    self.notvalid= False
    if self.conver != "ipbus":
      print "Did you set Core reg0 (ONU_MODE) to 0xe9?"
    else:
      onumode= self.read_ctrl_reg(0)
      self.write_ctrl_reg(0, 0xe9)
      print "Core reg0 ONU_MODE set to 0xe9 (included), was 0x%x"%(onumode)
  def printRegs(self, wregs="cu", only1flags=True):
    if self.conver == "cru":
      print "ONU not accessible (cru)"
      return
    onuc_regs= onuc_regs_fiv010
    onuu_regs= onuu_regs_fiv010
    if wregs.find('c')>=0:
      print "ONU_%d CORE:"%self.onun
      ttcpon_alice.printTable(self, onuc_regs, 0, 27, only1flags)
    if wregs.find('u')>=0:
      print "ONU_%d USER:"%self.onun
      #self.printTable(oltu_regs, self.uroffset, 30)
      ttcpon_alice.printTable(self, onuu_regs, self.addr_offset, 26, only1flags)
      if not only1flags:
        cs= self.getDownCounts()
        print "BITSUM: %d %d   ERRSUM:%d %d   FEC_S:%d %d   FEC_D:%d %d"%(cs[0], cs[1], cs[2], cs[3], cs[4], cs[5], cs[6], cs[7])
  def cfg_upstream(self):
    if self.conver=="cru":
      print "ONU_%d cfg_upstream not done (cru)"%self.onun
    else:
      self.onureset()
      self.setseltx(0)   # 0:PRBS7
      # ONU status
      #print("ONU_%d MODE 0x%x:"%(self.onun,self.read_ctrl_reg(0)))
      #print("HEARTBEAT_PERIOD-1: 0x%x"%self.read_ctrl_reg(1)) 
      #print "SEL_TX_PATTERN:0x%x"%self.read_ctrl_reg(self.getura(2))

  def printRegs_v4(self):
    if self.conver == "cru":
      print "ONU not accessible (cru)"
      return
    mode= self.read_ctrl_reg(0)
    hb= self.read_ctrl_reg(1)
    print "ONU%d CORE:MODE[7..0]:%d  HB_PERIOD[31-16]:%d  HB_PERIOD[15-0]:%d"%\
      (self.onun, mode, hb>>16, hb&0xffff)
    sfp= self.read_ctrl_reg(2)
    if self.conver == "ublaze":
      print "     SFP_ENA_DEL:%d  FINE__OFFSET_TDM:%d"%\
        (sfp&0xf, (sfp>>8)&0x1f)
    elif self.conver == "ipbus":   # fy v4 (1 reg. more)
      drpaccess= self.read_ctrl_reg(0xe)
      print "     SFP_ENA_DEL:%d  FINE__OFFSET_TDM:%d DRPACCESS:0x%x"%\
        (sfp&0xf, (sfp>>8)&0x1f, drpaccess&0x7ffffff)
    else:
      print "INTERROR: self.conver:", self.conver
    print "USER:"
    add= (self.read_ctrl_reg(0+self.addr_offset)>>1)&0xff
    cl= self.read_ctrl_reg(1+self.addr_offset)
    print "ADD:%d  BERT_CLEAR 1[0]:%d  BERT_LATCH 1[1]:%d"%\
      (add, cl&1, (cl>>1)&1)
    tx= self.read_ctrl_reg(2+self.addr_offset)
    txp= tx&0x3; txps=("PRBS7","PRBS23","CONST","BYTECNT")
    print "SEL_TX_PAT 2[1-0]:%d=%s TX_INT_MODE 2[2]:%d  TX_INJECT_ERR 2[3]:%d"%\
      (txp,txps[txp], (tx>>2)&1, (tx>>3)&1)
    mg= self.read_ctrl_reg(3+self.addr_offset)&0xff
    ur3bits=("RX40LOCKED","PHASEGOOD","RXLOCKED","OPER","TXRDY","RXRDY",
      "TXPLLLOCK", "RXPLLLOCK")
    mgs=""
    for ix in range(7,-1,-1):
      if mg & (1<<ix): mgs= mgs+" "+ur3bits[ix]
    print "3[7-0]:0x%x=%s"%(mg, mgs)
    bs1,bs2,es1,es2= self.getDownCounts()
    print "BITSUM: %d %d   ERRSUM:%d %d"%(bs1,bs2,es1,es2)

#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3
#!/usr/bin/python3


class OnuMlinkExdsg(OnuExdsgAlice):

    def __init__(self, ip='192.168.1.2', port=9001, onun=None):
        OnuExdsgAlice.__init__(self, ip, port,onun)
        self.lane_offset = 128*10
        self.addr_offset = 16384

    def write_ctrl_reg(self, reg_addr, reg_value):
        return OnuExdsg.write_ctrl_reg(self, reg_addr + self.lane_offset, reg_value)

    def read_ctrl_reg(self, reg_addr):
        return OnuExdsg.read_ctrl_reg(self, reg_addr + self.lane_offset)

class OnuMlinkExdsgAlice(OnuMlinkExdsg):
  def __init__(self, ip='192.168.1.2', port=9001, onun=None, mononu= False):
    OnuMlinkExdsg.__init__(self, ip, port,onun)
    self.monionu= mononu


